Package: marginaleffects
Title: Predictions, Comparisons, Slopes, Marginal Means, and Hypothesis
        Tests
Version: 0.32.0
Authors@R: 
    c(person(given = "Vincent",
             family = "Arel-Bundock",
             role = c("aut", "cre", "cph"),
             email = "vincent.arel-bundock@umontreal.ca",
             comment = c(ORCID = "0000-0003-2042-7063")),
      person(given = "Noah",
             family = "Greifer",
             role = "ctb",
             email = "noah.greifer@gmail.com",
             comment = c(ORCID = "0000-0003-3067-7154")),
      person("Etienne", "Bacher",
             email = "etienne.bacher@protonmail.com",
             role = "ctb",
             comment = c(ORCID = "0000-0002-9271-5075")),
      person(given = "Grant",
             family = "McDermott",
             role = c("ctb"),
             comment = c(ORCID = "0000-0001-7883-8573")),
      person(given = "Andrew",
             family = "Heiss",
             role = c("ctb"),
             comment = c(ORCID = "0000-0002-3948-3914"))
             )
Description: Compute and plot predictions, slopes, marginal means, and comparisons (contrasts, risk ratios, odds, etc.) for over 100 classes of statistical and machine learning models in R. Conduct linear and non-linear hypothesis tests, or equivalence tests. Calculate uncertainty estimates using the delta method, bootstrapping, or simulation-based inference. Details can be found in Arel-Bundock, Greifer, and Heiss (2024) <doi:10.18637/jss.v111.i09>.
License: GPL (>= 3)
Copyright: inst/COPYRIGHTS
Encoding: UTF-8
URL: https://marginaleffects.com/
BugReports: https://github.com/vincentarelbundock/marginaleffects/issues
RoxygenNote: 7.3.3
Depends: R (>= 3.6.0)
Imports: backports, checkmate, data.table (>= 1.16.4), generics,
        Formula, insight (>= 1.3.0), methods, rlang
Suggests: AER, Amelia, afex, aod, arrow, bayestestR, bench, betareg,
        BH, bife, biglm, blme, boot, brglm2, brms, brmsmargins, broom,
        car, carData, causaldata, clarify, cjoint, cobalt, collapse,
        conflicted, countrycode, covr, crch, DALEXtra, DCchoice,
        dbarts, DirichletReg, distributional, dfidx, dplyr, emmeans,
        equivalence, estimatr, fixest, flexsurv, fmeffects, fontquiver,
        future, future.apply, fwb, gam, gamlss, gamlss.dist, geepack,
        ggdist, ggokabeito, ggplot2, ggrepel, glmmTMB, glmtoolbox,
        glmx, haven, here, itsadug, ivreg, kableExtra, lme4, lmerTest,
        logistf, magrittr, margins, MatchIt, MASS, mclogit, MCMCglmm,
        mhurdle, missRanger, mgcv, mice, miceadds, mlogit, mlr3verse,
        modelbased, modelsummary, multcomp, mvgam, mvtnorm,
        nanoparquet, nlme, nnet, numDeriv, nycflights13, optmatch,
        ordbetareg, ordinal, parameters, parsnip, partykit, patchwork,
        pkgdown, phylolm, pbapply, plm, polspline, posterior, probably,
        pscl, purrr, quantreg, quantregForest, Rchoice, REndo,
        rcmdcheck, Rdatasets, remotes, reticulate, rmarkdown, rms,
        robust, robustbase, robustlmm, rsample, rstanarm, rstantools,
        rstpm2, rstudioapi, rsvg, sampleSelection, sandwich, scam,
        spelling, speedglm, splines, survey, survival, svglite,
        systemfit, systemfonts, tibble, tictoc, tidymodels, tidyr,
        tidyverse, tinysnapshot (>= 0.0.7), tinytable (>= 0.6.1),
        tinytest, titanic, truncreg, tsModel, withr, workflows, yaml,
        xgboost, altdoc, knitr, quarto
Collate: 'add_hypothesis.R' 'add_rowid.R' 'attributes.R' 'autodiff.R'
        'backtransform.R' 'broom.R' 'by.R' 'class.R' 'comparisons.R'
        'components.R' 'construct_call.R' 'datagrid.R' 'equivalence.R'
        'get_ci.R' 'get_coef.R' 'get_comparisons.R'
        'get_comparisons_data.R' 'get_comparisons_data_character.R'
        'get_comparisons_data_factor.R'
        'get_comparisons_data_logical.R'
        'get_comparisons_data_numeric.R' 'get_dataset.R'
        'get_degrees_of_freedom.R' 'get_draws.R' 'get_group_names.R'
        'get_hypotheses.R' 'get_jacobian.R' 'get_labels.R'
        'get_model_matrix.R' 'get_modeldata.R' 'get_predict.R'
        'get_predictions.R' 'get_se_delta.R' 'get_vcov.R'
        'github_issue.R' 'hush.R' 'hypotheses.R' 'hypotheses_joint.R'
        'hypothesis.R' 'hypothesis_formula.R' 'hypothesis_function.R'
        'hypothesis_matrix.R' 'hypothesis_string.R' 'imputation.R'
        'inferences.R' 'inferences_boot.R' 'inferences_conformal.R'
        'inferences_conformal_cv.R' 'inferences_conformal_full.R'
        'inferences_conformal_quantile.R'
        'inferences_conformal_split.R' 'inferences_fwb.R'
        'inferences_rsample.R' 'inferences_simulation.R'
        'matrix_apply.R' 'mean_or_mode.R' 'methods.R' 'set_coef.R'
        'methods_AER.R' 'sanity_model.R' 'methods_DirichletReg.R'
        'methods_MASS.R' 'methods_MCMCglmm.R' 'methods_Rchoice.R'
        'methods_WeightIt.R' 'methods_aaa.R' 'methods_afex.R'
        'methods_aod.R' 'methods_betareg.R' 'methods_bife.R'
        'methods_biglm.R' 'methods_nnet.R' 'methods_brglm2.R'
        'methods_brms.R' 'methods_crch.R' 'methods_dataframe.R'
        'methods_dbarts.R' 'methods_fixest.R' 'methods_flexsurv.R'
        'methods_gamlss.R' 'methods_glmmTMB.R' 'methods_glmtoolbox.R'
        'methods_glmx.R' 'methods_ivreg.R' 'methods_lme4.R'
        'methods_mclogit.R' 'methods_mgcv.R' 'methods_mhurdle.R'
        'methods_mice.R' 'methods_mlm.R' 'methods_mlogit.R'
        'methods_mlr3.R' 'methods_nlme.R' 'methods_ordinal.R'
        'methods_plm.R' 'methods_pscl.R' 'methods_quantreg.R'
        'methods_rms.R' 'methods_robustlmm.R' 'methods_rstanarm.R'
        'methods_rstpm2.R' 'methods_sampleSelection.R' 'methods_scam.R'
        'methods_survey.R' 'methods_survival.R' 'methods_systemfit.R'
        'methods_tidymodels.R' 'methods_tobit1.R' 'modelarchive.R'
        'multcomp.R' 'myTryCatch.R' 'package.R' 'pad.R' 'plot.R'
        'plot_build.R' 'plot_comparisons.R' 'plot_predictions.R'
        'plot_slopes.R' 'predictions.R' 'print.R' 'prune.R' 'refit.R'
        'sanitize_comparison.R' 'sanitize_condition.R'
        'sanitize_conf_level.R' 'sanitize_hypothesis_formula.R'
        'sanitize_interaction.R' 'sanitize_newdata.R'
        'sanitize_numderiv.R' 'sanitize_reserved.R' 'sanitize_type.R'
        'sanitize_variables.R' 'sanitize_vcov.R' 'sanity.R'
        'sanity_by.R' 'sanity_dots.R' 'sanity_inferences.R'
        'sanity_multcomp.R' 'settings.R' 'slopes.R' 'sort.R'
        'tinytest.R' 'type_dictionary.R' 'unpack_matrix_cols.R'
        'utils.R' 'zzz.R'
Language: en-US
VignetteBuilder: knitr, quarto
NeedsCompilation: no
Packaged: 2026-02-14 02:05:39 UTC; vincent
Author: Vincent Arel-Bundock [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-2042-7063>),
  Noah Greifer [ctb] (ORCID: <https://orcid.org/0000-0003-3067-7154>),
  Etienne Bacher [ctb] (ORCID: <https://orcid.org/0000-0002-9271-5075>),
  Grant McDermott [ctb] (ORCID: <https://orcid.org/0000-0001-7883-8573>),
  Andrew Heiss [ctb] (ORCID: <https://orcid.org/0000-0002-3948-3914>)
Maintainer: Vincent Arel-Bundock <vincent.arel-bundock@umontreal.ca>
Repository: CRAN
Date/Publication: 2026-02-14 06:40:08 UTC
Built: R 4.6.0; ; 2026-04-26 09:59:28 UTC; unix

Package: marginaleffects
Title: Predictions, Comparisons, Slopes, Marginal Means, and Hypothesis
        Tests
Version: 0.32.0
Authors@R: 
    c(person(given = "Vincent",
             family = "Arel-Bundock",
             role = c("aut", "cre", "cph"),
             email = "vincent.arel-bundock@umontreal.ca",
             comment = c(ORCID = "0000-0003-2042-7063")),
      person(given = "Noah",
             family = "Greifer",
             role = "ctb",
             email = "noah.greifer@gmail.com",
             comment = c(ORCID = "0000-0003-3067-7154")),
      person("Etienne", "Bacher",
             email = "etienne.bacher@protonmail.com",
             role = "ctb",
             comment = c(ORCID = "0000-0002-9271-5075")),
      person(given = "Grant",
             family = "McDermott",
             role = c("ctb"),
             comment = c(ORCID = "0000-0001-7883-8573")),
      person(given = "Andrew",
             family = "Heiss",
             role = c("ctb"),
             comment = c(ORCID = "0000-0002-3948-3914"))
             )
Description: Compute and plot predictions, slopes, marginal means, and comparisons (contrasts, risk ratios, odds, etc.) for over 100 classes of statistical and machine learning models in R. Conduct linear and non-linear hypothesis tests, or equivalence tests. Calculate uncertainty estimates using the delta method, bootstrapping, or simulation-based inference. Details can be found in Arel-Bundock, Greifer, and Heiss (2024) <doi:10.18637/jss.v111.i09>.
License: GPL (>= 3)
Copyright: inst/COPYRIGHTS
Encoding: UTF-8
URL: https://marginaleffects.com/
BugReports: https://github.com/vincentarelbundock/marginaleffects/issues
RoxygenNote: 7.3.3
Depends: R (>= 3.6.0)
Imports: backports, checkmate, data.table (>= 1.16.4), generics,
        Formula, insight (>= 1.3.0), methods, rlang
Suggests: AER, Amelia, afex, aod, arrow, bayestestR, bench, betareg,
        BH, bife, biglm, blme, boot, brglm2, brms, brmsmargins, broom,
        car, carData, causaldata, clarify, cjoint, cobalt, collapse,
        conflicted, countrycode, covr, crch, DALEXtra, DCchoice,
        dbarts, DirichletReg, distributional, dfidx, dplyr, emmeans,
        equivalence, estimatr, fixest, flexsurv, fmeffects, fontquiver,
        future, future.apply, fwb, gam, gamlss, gamlss.dist, geepack,
        ggdist, ggokabeito, ggplot2, ggrepel, glmmTMB, glmtoolbox,
        glmx, haven, here, itsadug, ivreg, kableExtra, lme4, lmerTest,
        logistf, magrittr, margins, MatchIt, MASS, mclogit, MCMCglmm,
        mhurdle, missRanger, mgcv, mice, miceadds, mlogit, mlr3verse,
        modelbased, modelsummary, multcomp, mvgam, mvtnorm,
        nanoparquet, nlme, nnet, numDeriv, nycflights13, optmatch,
        ordbetareg, ordinal, parameters, parsnip, partykit, patchwork,
        pkgdown, phylolm, pbapply, plm, polspline, posterior, probably,
        pscl, purrr, quantreg, quantregForest, Rchoice, REndo,
        rcmdcheck, Rdatasets, remotes, reticulate, rmarkdown, rms,
        robust, robustbase, robustlmm, rsample, rstanarm, rstantools,
        rstpm2, rstudioapi, rsvg, sampleSelection, sandwich, scam,
        spelling, speedglm, splines, survey, survival, svglite,
        systemfit, systemfonts, tibble, tictoc, tidymodels, tidyr,
        tidyverse, tinysnapshot (>= 0.0.7), tinytable (>= 0.6.1),
        tinytest, titanic, truncreg, tsModel, withr, workflows, yaml,
        xgboost, altdoc, knitr, quarto
Collate: 'add_hypothesis.R' 'add_rowid.R' 'attributes.R' 'autodiff.R'
        'backtransform.R' 'broom.R' 'by.R' 'class.R' 'comparisons.R'
        'components.R' 'construct_call.R' 'datagrid.R' 'equivalence.R'
        'get_ci.R' 'get_coef.R' 'get_comparisons.R'
        'get_comparisons_data.R' 'get_comparisons_data_character.R'
        'get_comparisons_data_factor.R'
        'get_comparisons_data_logical.R'
        'get_comparisons_data_numeric.R' 'get_dataset.R'
        'get_degrees_of_freedom.R' 'get_draws.R' 'get_group_names.R'
        'get_hypotheses.R' 'get_jacobian.R' 'get_labels.R'
        'get_model_matrix.R' 'get_modeldata.R' 'get_predict.R'
        'get_predictions.R' 'get_se_delta.R' 'get_vcov.R'
        'github_issue.R' 'hush.R' 'hypotheses.R' 'hypotheses_joint.R'
        'hypothesis.R' 'hypothesis_formula.R' 'hypothesis_function.R'
        'hypothesis_matrix.R' 'hypothesis_string.R' 'imputation.R'
        'inferences.R' 'inferences_boot.R' 'inferences_conformal.R'
        'inferences_conformal_cv.R' 'inferences_conformal_full.R'
        'inferences_conformal_quantile.R'
        'inferences_conformal_split.R' 'inferences_fwb.R'
        'inferences_rsample.R' 'inferences_simulation.R'
        'matrix_apply.R' 'mean_or_mode.R' 'methods.R' 'set_coef.R'
        'methods_AER.R' 'sanity_model.R' 'methods_DirichletReg.R'
        'methods_MASS.R' 'methods_MCMCglmm.R' 'methods_Rchoice.R'
        'methods_WeightIt.R' 'methods_aaa.R' 'methods_afex.R'
        'methods_aod.R' 'methods_betareg.R' 'methods_bife.R'
        'methods_biglm.R' 'methods_nnet.R' 'methods_brglm2.R'
        'methods_brms.R' 'methods_crch.R' 'methods_dataframe.R'
        'methods_dbarts.R' 'methods_fixest.R' 'methods_flexsurv.R'
        'methods_gamlss.R' 'methods_glmmTMB.R' 'methods_glmtoolbox.R'
        'methods_glmx.R' 'methods_ivreg.R' 'methods_lme4.R'
        'methods_mclogit.R' 'methods_mgcv.R' 'methods_mhurdle.R'
        'methods_mice.R' 'methods_mlm.R' 'methods_mlogit.R'
        'methods_mlr3.R' 'methods_nlme.R' 'methods_ordinal.R'
        'methods_plm.R' 'methods_pscl.R' 'methods_quantreg.R'
        'methods_rms.R' 'methods_robustlmm.R' 'methods_rstanarm.R'
        'methods_rstpm2.R' 'methods_sampleSelection.R' 'methods_scam.R'
        'methods_survey.R' 'methods_survival.R' 'methods_systemfit.R'
        'methods_tidymodels.R' 'methods_tobit1.R' 'modelarchive.R'
        'multcomp.R' 'myTryCatch.R' 'package.R' 'pad.R' 'plot.R'
        'plot_build.R' 'plot_comparisons.R' 'plot_predictions.R'
        'plot_slopes.R' 'predictions.R' 'print.R' 'prune.R' 'refit.R'
        'sanitize_comparison.R' 'sanitize_condition.R'
        'sanitize_conf_level.R' 'sanitize_hypothesis_formula.R'
        'sanitize_interaction.R' 'sanitize_newdata.R'
        'sanitize_numderiv.R' 'sanitize_reserved.R' 'sanitize_type.R'
        'sanitize_variables.R' 'sanitize_vcov.R' 'sanity.R'
        'sanity_by.R' 'sanity_dots.R' 'sanity_inferences.R'
        'sanity_multcomp.R' 'settings.R' 'slopes.R' 'sort.R'
        'tinytest.R' 'type_dictionary.R' 'unpack_matrix_cols.R'
        'utils.R' 'zzz.R'
Language: en-US
VignetteBuilder: knitr, quarto
NeedsCompilation: no
Packaged: 2026-02-14 02:05:39 UTC; vincent
Author: Vincent Arel-Bundock [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-2042-7063>),
  Noah Greifer [ctb] (ORCID: <https://orcid.org/0000-0003-3067-7154>),
  Etienne Bacher [ctb] (ORCID: <https://orcid.org/0000-0002-9271-5075>),
  Grant McDermott [ctb] (ORCID: <https://orcid.org/0000-0001-7883-8573>),
  Andrew Heiss [ctb] (ORCID: <https://orcid.org/0000-0002-3948-3914>)
Maintainer: Vincent Arel-Bundock <vincent.arel-bundock@umontreal.ca>
Repository: CRAN
Date/Publication: 2026-02-14 06:40:08 UTC
Built: R 4.6.0; ; 2026-04-25 16:07:16 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(coef,comparisons)
S3method(coef,hypotheses)
S3method(coef,predictions)
S3method(coef,slopes)
S3method(components,comparisons)
S3method(components,hypotheses)
S3method(components,marginaleffects)
S3method(components,predictions)
S3method(components,slopes)
S3method(df.residual,comparisons)
S3method(df.residual,predictions)
S3method(df.residual,slopes)
S3method(get_autodiff_args,glm)
S3method(get_autodiff_args,ivreg)
S3method(get_autodiff_args,lm)
S3method(get_coef,DirichletRegModel)
S3method(get_coef,afex_aov)
S3method(get_coef,betareg)
S3method(get_coef,bracl)
S3method(get_coef,brmsfit)
S3method(get_coef,brmultinom)
S3method(get_coef,comparisons)
S3method(get_coef,data.frame)
S3method(get_coef,default)
S3method(get_coef,gam)
S3method(get_coef,gamlss)
S3method(get_coef,glmgee)
S3method(get_coef,glmmTMB)
S3method(get_coef,lmerMod)
S3method(get_coef,lmerModLmerTest)
S3method(get_coef,mblogit)
S3method(get_coef,merMod)
S3method(get_coef,mlm)
S3method(get_coef,multinom)
S3method(get_coef,multinom_weightit)
S3method(get_coef,nls)
S3method(get_coef,polr)
S3method(get_coef,predictions)
S3method(get_coef,scam)
S3method(get_coef,selection)
S3method(get_coef,slopes)
S3method(get_coef,svyolr)
S3method(get_coef,systemfit)
S3method(get_coef,workflow)
S3method(get_group_names,bracl)
S3method(get_group_names,brmsfit)
S3method(get_group_names,clm)
S3method(get_group_names,default)
S3method(get_group_names,hurdle)
S3method(get_group_names,mblogit)
S3method(get_group_names,mlm)
S3method(get_group_names,multinom)
S3method(get_group_names,polr)
S3method(get_group_names,svyolr)
S3method(get_mean_or_mode,character)
S3method(get_mean_or_mode,data.frame)
S3method(get_mean_or_mode,default)
S3method(get_mean_or_mode,factor)
S3method(get_mean_or_mode,logical)
S3method(get_model_matrix,default)
S3method(get_predict,DirichletRegModel)
S3method(get_predict,Learner)
S3method(get_predict,MCMCglmm)
S3method(get_predict,afex_aov)
S3method(get_predict,aft)
S3method(get_predict,bart)
S3method(get_predict,betareg)
S3method(get_predict,bife)
S3method(get_predict,biglm)
S3method(get_predict,brmsfit)
S3method(get_predict,brmultinom)
S3method(get_predict,clm)
S3method(get_predict,coxph)
S3method(get_predict,crch)
S3method(get_predict,default)
S3method(get_predict,fixest)
S3method(get_predict,flexsurvreg)
S3method(get_predict,gamlss)
S3method(get_predict,glimML)
S3method(get_predict,glm)
S3method(get_predict,glmgee)
S3method(get_predict,glmmPQL)
S3method(get_predict,glmmTMB)
S3method(get_predict,gsm)
S3method(get_predict,hxlr)
S3method(get_predict,lda)
S3method(get_predict,lm)
S3method(get_predict,lmerMod)
S3method(get_predict,lmerModLmerTest)
S3method(get_predict,lrm)
S3method(get_predict,mblogit)
S3method(get_predict,merMod)
S3method(get_predict,mhurdle)
S3method(get_predict,mlogit)
S3method(get_predict,model_fit)
S3method(get_predict,multinom)
S3method(get_predict,ols)
S3method(get_predict,orm)
S3method(get_predict,polr)
S3method(get_predict,pstpm2)
S3method(get_predict,rlmerMod)
S3method(get_predict,rms)
S3method(get_predict,rq)
S3method(get_predict,stanreg)
S3method(get_predict,stpm2)
S3method(get_predict,svyglm)
S3method(get_predict,svyolr)
S3method(get_predict,systemfit)
S3method(get_predict,tobit1)
S3method(get_predict,workflow)
S3method(get_vcov,Learner)
S3method(get_vcov,MCMCglmm)
S3method(get_vcov,afex_aov)
S3method(get_vcov,aft)
S3method(get_vcov,bart)
S3method(get_vcov,biglm)
S3method(get_vcov,brmsfit)
S3method(get_vcov,comparisons)
S3method(get_vcov,default)
S3method(get_vcov,gamlss)
S3method(get_vcov,glimML)
S3method(get_vcov,glmmTMB)
S3method(get_vcov,gsm)
S3method(get_vcov,hypotheses)
S3method(get_vcov,lda)
S3method(get_vcov,mhurdle)
S3method(get_vcov,model_fit)
S3method(get_vcov,orm)
S3method(get_vcov,predictions)
S3method(get_vcov,pstpm2)
S3method(get_vcov,scam)
S3method(get_vcov,slopes)
S3method(get_vcov,stpm2)
S3method(get_vcov,systemfit)
S3method(get_vcov,workflow)
S3method(glance,comparisons)
S3method(glance,hypotheses)
S3method(glance,marginaleffects_mids)
S3method(glance,predictions)
S3method(glance,slopes)
S3method(knitr::knit_print, marginaleffects)
S3method(knitr::knit_print,comparisons)
S3method(knitr::knit_print,hypotheses)
S3method(knitr::knit_print,predictions)
S3method(knitr::knit_print,slopes)
S3method(plot,comparisons)
S3method(plot,predictions)
S3method(plot,slopes)
S3method(print,comparisons)
S3method(print,hypotheses)
S3method(print,marginaleffects)
S3method(print,predictions)
S3method(print,slopes)
S3method(prune,comparisons)
S3method(prune,hypotheses)
S3method(prune,marginaleffects)
S3method(prune,predictions)
S3method(prune,slopes)
S3method(refit,comparisons)
S3method(refit,hypotheses)
S3method(refit,marginaleffects)
S3method(refit,predictions)
S3method(refit,slopes)
S3method(sanitize_model_specific,DirichletRegModel)
S3method(sanitize_model_specific,Learner)
S3method(sanitize_model_specific,bart)
S3method(sanitize_model_specific,brmsfit)
S3method(sanitize_model_specific,gamlss)
S3method(sanitize_model_specific,lrm)
S3method(sanitize_model_specific,mblogit)
S3method(sanitize_model_specific,merMod)
S3method(sanitize_model_specific,ols)
S3method(sanitize_model_specific,orm)
S3method(sanitize_model_specific,svyglm)
S3method(sanitize_model_specific,svyolr)
S3method(set_coef,DirichletRegModel)
S3method(set_coef,afex_aov)
S3method(set_coef,aft)
S3method(set_coef,betareg)
S3method(set_coef,crch)
S3method(set_coef,data.frame)
S3method(set_coef,default)
S3method(set_coef,flexsurvreg)
S3method(set_coef,gamlss)
S3method(set_coef,glimML)
S3method(set_coef,glm)
S3method(set_coef,glmgee)
S3method(set_coef,glmmPQL)
S3method(set_coef,glmmTMB)
S3method(set_coef,glmx)
S3method(set_coef,gsm)
S3method(set_coef,hetprob)
S3method(set_coef,hurdle)
S3method(set_coef,hxlr)
S3method(set_coef,ivpml)
S3method(set_coef,lm)
S3method(set_coef,lme)
S3method(set_coef,lmerMod)
S3method(set_coef,lmerModLmerTest)
S3method(set_coef,merMod)
S3method(set_coef,mlm)
S3method(set_coef,model_fit)
S3method(set_coef,multinom)
S3method(set_coef,nls)
S3method(set_coef,polr)
S3method(set_coef,pstpm2)
S3method(set_coef,rlmerMod)
S3method(set_coef,scam)
S3method(set_coef,selection)
S3method(set_coef,stpm2)
S3method(set_coef,survreg)
S3method(set_coef,svyolr)
S3method(set_coef,systemfit)
S3method(set_coef,tobit)
S3method(set_coef,workflow)
S3method(set_coef,zeroinfl)
S3method(tidy,comparisons)
S3method(tidy,hypotheses)
S3method(tidy,marginaleffects_mids)
S3method(tidy,predictions)
S3method(tidy,slopes)
S3method(vcov,comparisons)
S3method(vcov,hypotheses)
S3method(vcov,predictions)
S3method(vcov,slopes)
export(autodiff)
export(avg_comparisons)
export(avg_predictions)
export(avg_slopes)
export(comparisons)
export(components)
export(datagrid)
export(expect_comparisons)
export(expect_hypotheses)
export(expect_margins)
export(expect_predictions)
export(expect_slopes)
export(get_coef)
export(get_dataset)
export(get_draws)
export(get_group_names)
export(get_model_matrix)
export(get_predict)
export(get_vcov)
export(glance)
export(hypotheses)
export(inferences)
export(plot_comparisons)
export(plot_predictions)
export(plot_slopes)
export(posterior_draws)
export(predictions)
export(prune)
export(refit)
export(set_coef)
export(slopes)
export(tidy)
import(data.table)
importFrom(generics,components)
importFrom(generics,glance)
importFrom(generics,prune)
importFrom(generics,tidy)

The `marginaleffects` package for `R` and `Python` offers a single point
of entry to easily interpret the results of [over 100 classes of
models,](https://marginaleffects.com/bonus/supported_models.html) using
a simple and consistent user interface.

This package comes with a free full-length online book, with extensive
tutorials: <https://marginaleffects.com>

The package’s benefits include:

- *Powerful:* It can compute and plot predictions; comparisons
  (contrasts, risk ratios, etc.); slopes; and conduct hypothesis and
  equivalence tests for over 100 different classes of models in `R`.
- *Simple:* All functions share a simple and unified interface.
- *Documented*: Each function is thoroughly documented with abundant
  examples. The Marginal Effects Zoo website includes 20,000+ words of
  vignettes and case studies.
- *Efficient:* [Some
  operations](https://marginaleffects.com/bonus/performance.html) can be
  up to 1000 times faster and use 30 times less memory than with the
  `margins` package.  
- *Valid:* When possible, [numerical results are
  checked](https://marginaleffects.com/bonus/supported_models.html)
  against alternative software like `Stata` or other `R` packages.
- *Thin:* The `R` package requires relatively few dependencies.
- *Standards-compliant:* `marginaleffects` follows “tidy” principles and
  returns simple data frames that work with all standard `R` functions.
  The outputs are easy to program with and feed to other packages like
  [`ggplot2`](https://marginaleffects.com/bonus/plot.html) or
  [`modelsummary`.](https://marginaleffects.com/bonus/tables.html)
- *Extensible:* Adding support for new models is very easy, often
  requiring less than 10 lines of new code. Please submit [feature
  requests on
  Github.](https://github.com/vincentarelbundock/marginaleffects/issues)
- *Active development*: Bugs are fixed promptly.

To cite marginaleffects in publications use:

Arel-Bundock V, Greifer N, Heiss A (2024). “How to Interpret Statistical
Models Using marginaleffects for R and Python.” *Journal of Statistical
Software*, *111*(9), 1-32. doi:10.18637/jss.v111.i09
<https://doi.org/10.18637/jss.v111.i09>.

A BibTeX entry for LaTeX users is

@Article{, title = {How to Interpret Statistical Models Using
{marginaleffects} for {R} and {Python}}, author = {Vincent Arel-Bundock
and Noah Greifer and Andrew Heiss}, journal = {Journal of Statistical
Software}, year = {2024}, volume = {111}, number = {9}, pages = {1–32},
doi = {10.18637/jss.v111.i09}, }

#' check type sanity
#'
#' @param model model object
#' @param type character vector
#' @noRd
sanitize_type <- function(model, type, by = FALSE, calling_function = "raw") {
    # mlr3
    if (inherits(model, "Learner")) {
        if (is.null(type)) type <- "response"
        valid <- setdiff(model$predict_types, "se")
        checkmate::assert_choice(type, choices = valid, null.ok = TRUE)
        return(type)
    }

    checkmate::assert_character(type, len = 1, null.ok = TRUE)
    
    if (inherits(model, "model_fit")) {
        cl <- "model_fit"
    } else {
        cl <- class(model)[1]
    }

    # if (!cl %in% type_dictionary$class) {
    #     cl <- "other"
    # }

    dict <- type_dictionary
    # raw is often invoked by `get_predict()`, which is required for {clarify} and others.
    # we only allow invlink(link) in predictions() and marginal_means(), which are handled by {marginaleffects}

    # invlink(link) only supported by predictions()
    if (!isTRUE(calling_function %in% "predictions")) {
        dict <- dict[dict$type != "invlink(link)", , drop = FALSE]
    }

    # invlink(link) only default for predictions() if by=FALSE
    if (!isTRUE(type == "invlink(link)") && !isFALSE(by)) {
        dict <- dict[dict$type != "invlink(link)", , drop = FALSE]
    }

    # fixest: invlink(link) only supported for glm model
    if (inherits(model, "fixest")) {
        if (!isTRUE(hush(model[["method_type"]]) %in% c("feglm"))) {
            dict <- dict[dict$type != "invlink(link)", , drop = FALSE]
        }
    }

    dict <- dict[dict$class == cl, , drop = FALSE]
    if (nrow(dict) > 0) {
        checkmate::assert_choice(type, choices = dict$type, null.ok = TRUE)
        if (is.null(type)) {
            type <- dict$type[1]
        }
    }

    if (is.null(type)) type <- "response"

    return(type)
}

#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.multinom <- function(model, coefs, ...) {
    # internally, coefficients are held in the `wts` vector, with 0s
    # interspersed. When transforming that vector to a matrix, we see that the
    # first row and first column are all zeros. 
    # NOTE: must use `newdata` in predict otherwise returns stored object.
    b_original <- get_coef(model)
    model$wts[match(b_original, model$wts)] <- coefs
    return(model)
}


#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.multinom <- function(model, ...) {
    out <- insight::get_parameters(model, ...)
    out <- stats::setNames(
        out$Estimate,
        sprintf("%s:%s", out$Response, out$Parameter))
    return(out)
}


#' @include get_group_names.R
#' @rdname get_group_names
#' @export
get_group_names.multinom <- function(model, ...) {
    resp <- insight::get_response(model)
    if (is.factor(resp)) {
        out <- levels(resp)
    } else {
        out <- unique(resp)
    }
    return(out[2:length(out)])
}


#' @include get_predict.R
#' @rdname get_predict
#' @export
get_predict.multinom <- function(model,
                                 newdata = insight::get_data(model),
                                 type = "probs",
                                 ...) {

    type <- sanitize_type(model, type, calling_function = "predictions")

    is_latent <- is_mclogit <- is_nnet <- FALSE
    if (isTRUE(type == "latent") && inherits(model, c("mblogit", "mclogit"))) {
        is_latent <- TRUE
        is_mclogit <- TRUE
        type <- "link"

    } else if (isTRUE(type == "latent") && inherits(model, "multinom")) {
        is_latent <- TRUE
        is_nnet <- TRUE
        type <- "probs"
    } 

    # needed because `predict.multinom` uses `data` rather than `newdata`
    pred <- stats::predict(model,
                           newdata = newdata,
                           type = type,
                           ...)

    # atomic vector means there is only one row in `newdata`
    # two levels DV returns a vector 
    if (isTRUE(checkmate::check_atomic_vector(pred))) {
        y_original <- sort(unique(insight::get_response(model)))
        two_levels <- length(y_original) == 2
        if (isTRUE(two_levels)) {
            pred <- matrix(pred)
            colnames(pred) <- as.character(y_original[2])
        } else {
            pred <- matrix(pred, nrow = 1, dimnames = list(NULL, names(pred)))
        }

    }

    if (is_latent && is_mclogit) {
        missing_level <- as.character(unique(insight::get_response(model)))
        missing_level <- setdiff(missing_level, colnames(pred))
        if (length(missing_level) == 1) {
            pred <- cbind(0, pred)
            colnames(pred)[1] <- missing_level
            pred <- pred - rowMeans(pred)
        } else {
            insight::format_error("Unable to compute predictions on the latent scale.")
        }

    } else if (is_latent && is_nnet) {
        inverse_softMax <- function(mu) {
            log_mu <- log(mu)
            return(sweep(log_mu, 1, STATS = rowMeans(log_mu), FUN = "-"))
        }
        pred <- inverse_softMax(pred)
    }


    # matrix with outcome levels as columns
    out <- data.frame(
        group = rep(colnames(pred), each = nrow(pred)),
        estimate = c(pred))
    out$group <- group_to_factor(out$group, model)

    # usually when `newdata` is supplied by `comparisons`
    if ("rowid" %in% colnames(newdata)) {
        out$rowid <- rep(newdata$rowid, times = ncol(pred))
    } else {
        out$rowid <- rep(seq_len(nrow(pred)), times = ncol(pred))
    }

    return(out)
}

supported_engine <- function(x) {
    insight::check_if_installed("parsnip")
    tmp <- parsnip::extract_fit_engine(x)
    flag <- inherits(try(sanitize_model(tmp), silent = TRUE), "try-error")
    return(!flag)
}


#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.model_fit <- function(model, coefs, ...) {

  if (!"fit" %in% names(model)) {
    return(model)
  }

  model$fit <- set_coef(model$fit, coefs, ...)

  return(model)
}


#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.workflow <- function(model, coefs, ...) {
    if ("fit" %in% names(model) && "fit" %in% names(model$fit)) {
        model$fit$fit <- set_coef(model$fit$fit, coefs, ...)
    }
    return(model)
}


#' @include get_predict.R
#' @rdname get_predict
#' @keywords internal
#' @export
get_predict.model_fit <- function(model, newdata, type = NULL, ...) {
    out <- stats::predict(model, new_data = newdata, type = type)

    if (type == "numeric") {
        v <- intersect(c(".pred", ".pred_res"), colnames(out))[1]
        out <- data.frame(rowid = seq_len(nrow(out)), estimate = out[[v]])

    } else if (type == "class") {
        out <- data.frame(rowid = seq_len(nrow(out)), estimate = out[[".pred_class"]])

    } else if (type == "prob") {
        colnames(out) <- substr(colnames(out), 7, nchar(colnames(out)))
        out$rowid <- seq_len(nrow(out))
        out <- data.table::melt(
            out,
            id.vars = "rowid",
            variable.name = "group",
            value.name = "estimate")
    }

    return(out)
}


#' @include get_predict.R
#' @rdname get_predict
#' @keywords internal
#' @export
get_predict.workflow <- get_predict.model_fit


#' @include get_vcov.R
#' @rdname get_vcov
#' @keywords internal
#' @export
get_vcov.model_fit <- function(model, type = NULL, ...) {
    if (isTRUE(type == "class")) {
        return(FALSE)
    }
    if (isTRUE(supported_engine(model))) {
        tmp <- parsnip::extract_fit_engine(model)
        out <- get_vcov(tmp)
    } else {
        out <- FALSE
    }
    return(out)
}


#' @include get_vcov.R
#' @rdname get_vcov
#' @keywords internal
#' @export
get_vcov.workflow <- get_vcov.model_fit


#' @include get_coef.R
#' @rdname get_coef
#' @keywords internal
#' @export
get_coef.workflow <- function(model, ...) {
    if (isTRUE(supported_engine(model))) {
        tmp <- parsnip::extract_fit_engine(model)
        out <- get_coef(tmp)
    } else {
        out <- NULL
    }
    return(out)
}

sanity_by <- function(by, newdata) {
    checkmate::assert(
        checkmate::check_flag(by),
        checkmate::check_data_frame(by, min.cols = 1, min.rows = 1),
        checkmate::check_character(by, min.len = 1),
        checkmate::check_null(by))

    known <- c("by", "group", "term", "rowid", "rowidcf", "contrast", colnames(newdata))

    if (isTRUE(by == "group") && "group" %in% colnames(newdata)) {
        msg <- 'The "group" variable name is forbidden to avoid conflicts with the column names of the outputs produced by the `marginaleffects` package. Please rename your variable of change the value of the `by` argument.'
        insight::format_error(msg)
    }

    if (isTRUE(checkmate::check_flag(by))) {
        flag <- FALSE
    } else if (inherits(by, "data.frame")) {
        flag <- !all(colnames(by) %in% known) || !"by" %in% colnames(by)
    } else {
        flag <- !all(by %in% known | grepl("^contrast_", by))
    }

    if (flag) {
        bycols <- paste(setdiff(colnames(newdata), c("rowid", "rowidcf", "term", "group")), collapse = ", ")
        msg <- c(
            "The `by` argument must be either:", "",
            sprintf("1. Character vector in which each element is part of: %s", bycols),
            "",
            sprintf("2. A data frame with a `by` column of labels, and in which all other columns are elements of: %s", bycols),
            "",
            "It can sometimes be useful to supply a data frame explicitly to the `newdata` argument in order to be able to group by different columns."
        )
        stop(insight::format_message(msg), call. = FALSE)
    }
}

hypothesis_string <- function(x, hypothesis) {
    out_list <- draws_list <- list()
    lab <- attr(hypothesis, "label")
    tmp <- expand_wildcard(hypothesis, nrow(x), lab)
    hypothesis <- tmp[[1]]
    labs <- tmp[[2]]
    for (i in seq_along(hypothesis)) {
        out_list[[i]] <- eval_string_hypothesis(x, hypothesis[i], labs[i])
        draws_list[[i]] <- attr(out_list[[i]], "posterior_draws")
    }
    out <- do.call(rbind, out_list)
    attr(out, "posterior_draws") <- do.call(rbind, draws_list)
    attr(out, "label") <- if (!is.null(attr(labs, "names"))) {
        attr(labs, "names")
    } else {
        labs
    }
    return(out)
}


eval_string_hypothesis <- function(x, hypothesis, lab) {
    # row indices: `hypotheses` includes them, but `term` does not
    if (isTRUE(grepl("\\bb\\d+\\b", hypothesis)) && !any(grepl("\\bb\\d+\\b", x[["term"]]))) {
        msg <- "
It is essential to check the order of estimates when specifying hypothesis tests using positional indices like b1, b2, etc. The indices of estimates can change depending on the order of rows in the original dataset, user-supplied arguments, model-fitting package, and version of `marginaleffects`.

It is also good practice to use assertions that ensure the order of estimates is consistent across different runs of the same code. Example:

```r
mod <- lm(mpg ~ am * carb, data = mtcars)

# assertion for safety
p <- avg_predictions(mod, by = 'carb')
stopifnot(p$carb[1] != 1 || p$carb[2] != 2)

# hypothesis test
avg_predictions(mod, by = 'carb', hypothesis = 'b1 - b2 = 0')
```

Disable this warning with: `options(marginaleffects_safe = FALSE)`
"
        if (isTRUE(getOption("marginaleffects_safe", default = TRUE))) {
            warn_once(msg, "hypothesis_positional_indices_are_dangerous")
        }


        bmax <- regmatches(lab, gregexpr("\\bb\\d+\\b", lab))[[1]]
        bmax <- tryCatch(max(as.numeric(gsub("b", "", bmax))), error = function(e) 0)
        if (bmax > nrow(x)) {
            msg <- "%s cannot be used in `hypothesis` because the call produced just %s estimate(s). Try executing the exact same command without the `hypothesis` argument to see which estimates are available for hypothesis testing."
            msg <- sprintf(msg, paste0("b", bmax), nrow(x))
            insight::format_error(msg)
        }
        for (i in seq_len(nrow(x))) {
            tmp <- paste0("marginaleffects__", i)
            hypothesis <- gsub(paste0("b", i), tmp, hypothesis)
        }
        rowlabels <- paste0("marginaleffects__", seq_len(nrow(x)))

        # term names
    } else {
        if (!"term" %in% colnames(x) || anyDuplicated(x$term) > 0) {
            msg <- c(
                'To use term names in a `hypothesis` string, the same function call without `hypothesis` argument must produce a `term` column with unique row identifiers. You can use `b1`, `b2`, etc. indices instead of term names in the `hypotheses` string Ex: "b1 + b2 = 0" Alternatively, you can use the `newdata`, `variables`, or `by` arguments:',
                "",
                "mod <- lm(mpg ~ am * vs + cyl, data = mtcars)",
                'comparisons(mod, newdata = "mean", hypothesis = "b1 = b2")',
                'comparisons(mod, newdata = "mean", hypothesis = "am = vs")',
                'comparisons(mod, variables = "am", by = "cyl", hypothesis = ~pairwise)')
            insight::format_error(msg)
        }
        rowlabels <- x$term
    }

    eval_string_function <- function(vec, hypothesis, rowlabels) {
        envir <- parent.frame()
        void <- sapply(
            seq_along(vec), function(i) {
                assign(rowlabels[i], vec[i], envir = envir)
            })
        out <- eval(parse(text = hypothesis), envir = envir)
        return(out)
    }

    if (!is.null(attr(lab, "names"))) {
        lab <- attr(lab, "names")
    } else {
        lab <- gsub("\\s+", "", lab)
    }

    draws <- attr(x, "posterior_draws")
    if (!is.null(draws)) {
        insight::check_if_installed("collapse", minimum_version = "1.9.0")
        tmp <- apply(
            draws,
            MARGIN = 2,
            FUN = eval_string_function,
            hypothesis = hypothesis,
            rowlabels = rowlabels)
        draws <- matrix(tmp, ncol = ncol(draws))
        out <- data.table(
            hypothesis = lab,
            tmp = collapse::dapply(draws, MARGIN = 1, FUN = collapse::fmedian))
    } else {
        out <- eval_string_function(
            x[["estimate"]],
            hypothesis = hypothesis,
            rowlabels = rowlabels)
        out <- data.table(
            hypothesis = lab,
            tmp = out)
    }

    setnames(out, old = "tmp", new = "estimate")
    attr(out, "posterior_draws") <- draws
    return(out)
}


expand_wildcard <- function(hyp, bmax, lab) {
    # Find all occurrences of b*
    bstar_indices <- gregexpr("b\\*", hyp)[[1]]
    if (bstar_indices[1] == -1) {
        return(list(hyp, lab))
    }
    bstar_count <- length(bstar_indices)
    if (bstar_count > 1) {
        insight::format_error("Error: More than one 'b*' substring found.")
    }

    # Replace b* with b1, b2, b3, ..., bmax
    labs <- character(bmax)
    result <- character(bmax)
    for (i in 1:bmax) {
        result[i] <- sub("b\\*", paste0("b", i), hyp)
        labs[i] <- sub("b\\*", paste0("b", i), lab)
    }

    return(list(result, labs))
}

bootstrap_simulation <- function(model, INF_FUN, vcov, ...) {
    insight::check_if_installed("MASS")

    dots <- list(...)

    B <- get_coef(model)
    V <- get_vcov(model, vcov = vcov)
    R <- attr(model, "inferences_R")
    coefmat <- MASS::mvrnorm(R, mu = B, Sigma = V)

    # avoid recursion
    args <- dots
    args[["vcov"]] <- FALSE
    attr(model, "inferences_method") <- NULL

    delta <- do.call(INF_FUN, c(list(model), args))

    inner_fun <- function(i = NULL) {
        args <- dots
        mod_tmp <- set_coef(model, coefmat[i, ])
        attr(mod_tmp, "inferences_method") <- NULL
        args[["model"]] <- mod_tmp
        args[["vcov"]] <- FALSE
        do.call(INF_FUN, args)$estimate
    }

    draws <- lapply(seq_len(nrow(coefmat)), inner_fun)
    draws <- do.call("cbind", draws)

    out <- get_ci(delta, draws = draws, conf_level = args$conf_level)

    # do not use simulation mean as point estimate
    # https://doi.org/10.1017/psrm.2023.8
    out$estimate <- delta$estimate

    attr(out, "posterior_draws") <- draws

    return(out)
}

# Combined function that sanitizes and computes degrees of freedom from mfx object
# Stores the result in mfx@df and returns the numeric df for backward compatibility
add_degrees_of_freedom <- function(mfx, df = Inf, by = NULL, hypothesis = NULL, vcov = NULL) {
    model <- mfx@model
    newdata <- mfx@newdata

    # K-W changes both the vcov and the df
    # Satterthwaite changes the df but not the vcov
    if (isTRUE(checkmate::check_choice(vcov, c("satterthwaite", "kenward-roger")))) {
        if (
            !isFALSE(by) ||
                !isTRUE(checkmate::check_number(hypothesis, null.ok = TRUE))
        ) {
            msg <- sprintf(
                "Satterthwaite and Kenward-Roger adjustments are not supported in this command with the `by` or `hypothesis` arguments, or with the `avg_` function prefix. One common strategy is to use the smallest unit-level degree of freedom. This can be obtained from the `df` column generated by calling the same function without `by` or `hypothesis` argument."
            )
            stop(msg, call. = FALSE)
        }
        df <- vcov
    }

    # Sanitize df input
    checkmate::assert(
        checkmate::check_true(df),
        checkmate::check_number(df, lower = 1),
        checkmate::check_numeric(df, len = nrow(newdata)),
        checkmate::check_choice(df, c("residual", "satterthwaite", "kenward-roger"))
    )

    if (isTRUE(checkmate::check_choice(df, c("satterthwaite", "kenward-roger")))) {
        checkmate::assert_data_frame(newdata)
        # predict.lmerTest requires the DV
        if (inherits(model, "lmerMod")) {
            dv <- mfx@variable_names_response
            if (!dv %in% colnames(newdata)) {
                newdata[[dv]] <- mean(insight::get_response(model))
            }
        }

        tmp <- try(
            insight::get_df(x = model, data = newdata, type = df, df_per_obs = TRUE),
            silent = TRUE
        )
        if (inherits(tmp, "try-error")) {
            msg <- "Unable to extract degrees of freedom of type `%s` for model of class `%s`."
            stop_sprintf(msg, df, class(model)[1])
        }
        df <- tmp
    } else if (isTRUE(checkmate::check_choice(df, "residual"))) {
        tmp <- try(
            insight::get_df(x = model, data = newdata, type = df, df_per_obs = TRUE),
            silent = TRUE
        )
        if (inherits(tmp, "try-error")) {
            msg <- "Unable to extract degrees of freedom of type `%s` for model of class `%s`."
            stop_sprintf(msg, df, class(model)[1])
        }
        df <- tmp
    } else if (
        isTRUE(checkmate::check_number(df, lower = Inf)) ||
            isFALSE(df) ||
            isTRUE(checkmate::check_null(df))
    ) {
        mfx@df <- Inf
        return(mfx)
    } else if (isTRUE(checkmate::check_numeric(df))) {
        # pass
    } else if (isTRUE(df)) {
        df <- try(insight::get_df(x = model), silent = TRUE)
    } else {
        stop("Invalid arguments for `get_df`.")
    }

    if (isTRUE(checkmate::check_data_frame(newdata))) {
        checkmate::assert(
            checkmate::check_numeric(df, len = 1),
            checkmate::check_numeric(df, len = nrow(newdata))
        )
    }

    # Store the computed degrees of freedom in the mfx object
    mfx@df <- df
    return(mfx)
}

#' internal function to build the type dictionary
#'
#' @noRd
type_dictionary_build <- function() {
text <-
'class,type
bam,response
bam,link
bart,ev
bart,ppd
betareg,response
betareg,link
betareg,precision
betareg,quantile
betareg,variance
bife,response
bife,link
bracl,probs
brglmFit,response
brglmFit,link
brmsfit,response
brmsfit,link
brmsfit,prediction
brmsfit,average
brmultinom,probs
brmultinom,class
clm,prob
clm,cum.prob
clm,linear.predictor
clogit,expected
clogit,lp
clogit,risk
clogit,survival
coxph,survival
coxph,expected
coxph,lp
coxph,risk
coxph_weightit,survival
coxph_weightit,expected
coxph_weightit,lp
coxph_weightit,risk
crch,response
crch,location
crch,scale
crch,density
hetprob,pr
hetprob,xb
hxlr,location
hxlr,cumprob
hxlr,scale
hxlr,density
ivpml,pr
ivpml,xb
flexsurvreg,survival,
flexsurvreg,response,
flexsurvreg,mean,
flexsurvreg,link,
flexsurvreg,lp,
flexsurvreg,linear,
flexsurvreg,rmst,
flexsurvreg,hazard,
flexsurvreg,cumhaz,
fixest,invlink(link)
fixest,response
fixest,link
hurdle,response
hurdle,prob
hurdle,count
hurdle,zero
iv_robust,response
lm,response
gam,response
gam,link
Gam,invlink(link)
Gam,response
Gam,link
geeglm,response
geeglm,link
Gls,lp
glimML,response
glimML,link
glm,invlink(link)
glm,response
glm,link
glmerMod,response
glmerMod,link
glmgee,response
glmrob,response
glmrob,link
glmmTMB,response
glmmTMB,link
glmmTMB,conditional
glmmTMB,zprob
glmmTMB,zlink
glmmTMB,disp
glmmPQL,response
glmmPQL,link
glmx,response
glm_weightit,invlink(link)
glm_weightit,probs
glm_weightit,response
glm_weightit,lp
glm_weightit,link
ivreg,response
lmerMod,response
lmerModLmerTest,response
lmrob,response
lm_robust,response
lrm,fitted
lrm,lp
lrm,mean
mblogit,response
mblogit,latent
mblogit,link
mclogit,response
mclogit,latent
mclogit,link
MCMCglmm,response
model_fit,numeric
model_fit,prob
model_fit,class
workflow,numeric
workflow,prob
workflow,class
multinom,probs
multinom,latent
multinom_weightit,probs
multinom_weightit,response
multinom_weightit,mean
mhurdle,E
mhurdle,Ep
mhurdle,p
mlogit,response
mvgam,response
mvgam,link
mvgam,expected
mvgam,detection
mvgam,latent_N
negbin,invlink(link)
negbin,response
negbin,link
ols,lp
"oohbchoice", "probability",
"oohbchoice", "utility",
orm,fitted
orm,mean
orm,lp
ordinal_weightit,probs
ordinal_weightit,response
ordinal_weightit,link
ordinal_weightit,lp
ordinal_weightit,mean
polr,probs
rendo.base,response
rendo.base,link
rlm,response
selection,response
selection,link
selection,unconditional
speedlm,response
speedglm,response
speedglm,link
stanreg,response
stanreg,link
survreg,response
survreg,link
survreg,quantile
svyglm,response
svyglm,link
svyolr,probs
tobit,response
tobit1,expvalue
tobit1,linpred
tobit1,prob
zeroinfl,response
zeroinfl,prob
zeroinfl,count
zeroinfl,zero'
out <- utils::read.csv(
    text = text,
    colClasses = c("character", "character"))
for (i in 1:2) {
    out[[i]] <- trimws(out[[i]])
}
return(out)
}


#' type dictionary
#'
#' insight::get_predict accepts a `predict` argument
#' stats::predict accepts a `type` argument
#' this dictionary converts
#' @noRd
type_dictionary <- type_dictionary_build()

#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.glmx <- function(model, coefs, ...) {
    out <- model
    out$coefficients$glm <- coefs[names(out$coefficients$glm)]
    out$coefficients$extra <- coefs[names(out$coefficients$extra)]
    return(out)
}

#' Pad newdata with factor levels and merge with original data
#'
#' `model.matrix` breaks when `newdata` includes a factor
#' variable, but not all levels are present in the data. This is bad for us
#' because we often want to get predictions with one (or few) rows, where some
#' factor levels are inevitably missing.
#' @param model Model object to check for mlogit class
#' @param newdata Data frame to pad
#' @keywords internal
#' @noRd
pad <- function(model, newdata) {
    checkmate::assert_data_frame(newdata)

    # pad factors: `model.matrix` breaks when factor levels are missing
    # support `newdata` and assume no padding the `idx` column is necessary for
    # `get_predict` but it breaks binding, so we can't remove it in
    # sanity_newdata and we can't rbind it with padding
    # pad factors: `model.matrix` breaks when factor levels are missing
    if (inherits(model, "mlogit")) {
        return(newdata)
    }

    # store variables with missing factors
    vault <- list()
    for (v in colnames(newdata)) {
        if (is.factor(newdata[[v]])) {
            if (!all(levels(newdata[[v]]) %in% newdata[[v]])) {
                vault[[v]] <- factor(levels(newdata[[v]]), levels = levels(newdata[[v]]))
            }
        }
    }

    # create padding
    if (length(vault) > 0) {
        # HACK: Some models use a character variable with many levels (e.g.,
        # mixed-effects groups). This creates a massive padding dataset, and making
        # predictions can become very expensive.
        if (isTRUE(sum(lengths(vault)) > 100)) {
            return(newdata)
        }

        padding <- utils::head(newdata, 1)
        data.table::setDT(padding)
        for (v in names(vault)) {
            padding[[v]] <- NULL
        }
        gr <- do.call(data.table::CJ, vault)
        padding <- cjdt(list(padding, gr))
        to_keep <- colnames(newdata)
        padding[, ..to_keep]
        setcolorder(padding, to_keep)
        data.table::setDF(padding)

        padding$rowid <- -1 * padding$rowid

        # merge padding with original newdata
        return(rbindlist(list(padding, newdata)))
    } else {
        return(newdata)
    }
}

# unpad
unpad <- function(out, draws) {
    if (!"rowid" %in% colnames(out)) {
        return(out)
    }
    idx <- out$rowid > 0
    out <- out[idx, , drop = FALSE]
    if (!is.null(draws)) {
        draws <- draws[idx, , drop = FALSE]
    }
    return(list(out = out, draws = draws))
}

#' @importFrom generics prune
#' @export
generics::prune


#' Prune marginaleffects objects to reduce memory usage
#'
#' Remove large attributes from marginaleffects objects to reduce memory usage.
#' Warning: This will disable many useful post-processing features of `marginaleffects`
#' @param tree A marginaleffects object (predictions, comparisons, slopes, or hypotheses)
#' @param component A character string indicating which component to prune: "all" or "modeldata".
#' @param ... Unused
#' @return A pruned marginaleffects object
#' @details ...
#' @export
prune.marginaleffects <- function(tree, component = "all", ...) {
    checkmate::assert_choice(component, c("all", "modeldata"))

    mfx <- components(tree, "all")

    if (component == "all") {
        if (!is.null(mfx)) {
            mfx@model <- NULL
            mfx@newdata <- NULL
            mfx@modeldata <- NULL
            mfx@call <- NULL
            mfx@jacobian <- matrix()
        }
        essential_attrs <- c("names", "row.names", "class", "mfx")
        for (nm in setdiff(names(attributes(tree)), essential_attrs)) {
            attr(tree, nm) <- NULL
        }
        attr(tree, "lean") <- TRUE
    } else if (component == "modeldata") {
        # Do not prune modeldata for dbarts, mlr3, or tidymodels models
        if (inherits(mfx@model, c("bart", "Learner", "model_fit", "workflow"))) {
            attr(tree, "marginaleffects") <- mfx
            return(tree)
        }

        fml <- hush(insight::find_formula(mfx@model))
        fml <- unlist(lapply(fml, all.vars))

        keepers <- c(
            fml,
            names(mfx@variables),
            mfx@variable_names_datagrid,
            mfx@variable_names_response,
            mfx@variable_names_wts,
            mfx@variable_names_by,
            mfx@variable_names_by_hypothesis
        )

        # fixest-specific syntax: i.groupid and ~weights
        # before name lookup in modeldata
        if (inherits(mfx@model, "fixest")) {
            keepers <- gsub("^~|^i\\.", "", keepers)
        }

        # preserve order
        keepers <- unique(intersect(names(mfx@modeldata), unique(keepers)))

        if (inherits(mfx@modeldata, "data.table")) {
            mfx@modeldata <- mfx@modeldata[, ..keepers, drop = FALSE]
        } else {
            mfx@modeldata <- mfx@modeldata[, keepers, drop = FALSE]
        }
    }

    attr(tree, "marginaleffects") <- mfx
    return(tree)
}

#' @export
prune.predictions <- prune.marginaleffects

#' @export
prune.hypotheses <- prune.marginaleffects

#' @export
prune.slopes <- prune.marginaleffects

#' @export
prune.comparisons <- prune.marginaleffects

backtransform <- function(x, transform) {

    # transform can be a function or a named list of length 1 with a function, but could be NULL
    if (!is.function(transform)) {
        if (is.null(transform[[1]])) {
            return(x)
        } else {
            transform <- transform[[1]]
        }
    }
   

    checkmate::assert_data_frame(x)
    checkmate::assert_function(transform)
    cols <- intersect(colnames(x), c("estimate", "conf.low", "conf.high"))
    draws <- attr(x, "posterior_draws")

    if (!is.null(draws)) {
        dim_pre <- dim(draws)
        draws <- transform(draws)
        dim_post <- dim(draws)
        if (!identical(dim_pre, dim_post)) {
            insight::format_error("The `transform` function must return an object of the same class as its input: a matrix input must return a matrix of the same size, and a vector input must return a vector of the same length.")
        }
    }

    for (col in cols) {
        x[[col]] <- transform(x[[col]])
    }

    # Issue #1204: Some inverse link functions swap the order of low and high
    if (all(c("conf.low", "conf.high") %in% colnames(x))) {
        if (all(x$conf.high < x$conf.low)) {
            lo <- x[["conf.low"]]
            hi <- x[["conf.high"]]
            x[["conf.low"]] <- hi
            x[["conf.high"]] <- lo
        }
    }

    for (col in c("std.error", "statistic")) {
        x[[col]] <- NULL
    }

    attr(x, "posterior_draws") <- draws

    return(x)
}

#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.glm <- function(model, coefs, ...) {
    model[["coefficients"]] <- sub_named_vector(
        model[["coefficients"]], coefs
    )
    ## But, there's an edge case!! When `predict(model, se.fit = TRUE)` is called without `newdata`, `predict.lm()` isn't called.
    ## Instead `model$linear.predictors` is returned directly if `type = "link"` and
    ## `model$fitted.values` is returned directly if `type = "response"`.
    ## `marginal_effects()` for "glm" is always called with `newdata`, so we won't hit this.
    model
}


#' @rdname set_coef
#' @export
set_coef.lm <- function(model, coefs, ...) {
    model[["coefficients"]] <- sub_named_vector(
        model[["coefficients"]], coefs
    )
    model
}


#' @rdname get_predict
#' @export
get_predict.lm <- function(model, newdata = insight::get_data(model), type = "response", ...) {
    MM <- attr(newdata, "marginaleffects_model_matrix")
    beta <- get_coef(model)
    if (!isTRUE(checkmate::check_matrix(MM)) || ncol(MM) != length(beta)) {
        out <- get_predict.default(model = model, newdata = newdata, type = type, ...)
        return(out)
    }
    p <- model$rank
    p1 <- seq_len(p)
    piv <- if (p) qr(model)$pivot[p1]
    if (!all(seq_len(ncol(MM)) %in% piv)) {
        MM <- MM[, piv, drop = FALSE]
        beta <- beta[piv]
    }
    if (getOption("marginaleffects_linalg", default = "RcppEigen") == "RcppEigen") {
        pred <- eigenMatMult(MM, beta)
    } else {
        pred <- drop(MM %*% beta)
    }
    
    # `pred` is a secret argument which re-uses the default get_predict to format a vector a data frame using correct `rowid`
    out <- get_predict.default(model = model, newdata = newdata, type = type, pred = pred, ...)
    return(out)
}


#' @rdname get_predict
#' @export
get_predict.glm <- function(model, newdata = insight::get_data(model), type = "response", ...) {
    out <- NULL
    MM <- attr(newdata, "marginaleffects_model_matrix")
    if (isTRUE(checkmate::check_matrix(MM))) {
        if (isTRUE(checkmate::check_choice(type, c("link")))) {
            out <- get_predict.lm(model = model, newdata = newdata, ...)
        } else if (isTRUE(checkmate::check_choice(type, "response")) || is.null(type)) {
            out <- get_predict.lm(model = model, newdata = newdata, ...)
            out$estimate <- stats::family(model)$linkinv(out$estimate)
        }
    }
    
    if (is.null(out)) {
        out <- get_predict.default(model = model, newdata = newdata, type = type, ...)
    }
    
    return(out)
}



#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.nls <- function(model, coefs, ...) {
    out <- model
    out$m$setPars(coefs)
    return(out)
}


#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.nls <- function(model, ...) {
    model$m$getPars()
}

#' Download and Read Datasets from `marginaleffects` or Rdatasets
#'
#' @description
#' Downloads a dataset from the `marginaleffects` or the Rdatasets archives, and return it as a data frame. Opens the documentation as an HTML page. Search available datasets.
#'
#' [https://vincentarelbundock.github.io/Rdatasets/](https://vincentarelbundock.github.io/Rdatasets/)
#'
#' @param dataset String. Name of the dataset to download.
#'   - `marginaleffects` archive: `affairs`, `airbnb`, `ces_demographics`, `ces_survey`, `immigration`, `lottery`, `military`, `thornton`, `factorial_01`, `interaction_01`, `interaction_02`, `interaction_03`, `interaction_04`, `polynomial_01`, `polynomial_02`
#'   - Rdatasets archive: The name of a dataset listed on the Rdatasets index. See the website or the `search` argument.
#' @param package String. Package name that originally published the data.
#' @param docs Logical. If TRUE open the documentation using `getOption("viewer")` or the Rstudio viewer.
#' @param search Regular expression. Download the dataset index from Rdatasets; search the "Package", "Item", and "Title" columns; and return the matching rows.
#' @return A data frame containing the dataset.
#' @examplesIf FALSE
#' dat <- get_dataset("Titanic", "Stat2Data")
#' head(dat)
#'
#' get_dataset(search = "(?i)titanic")
#'
#' # View documentation in the browser
#' get_dataset("Titanic", "Stat2Data", docs = TRUE)
#' @export
get_dataset <- function(
    dataset = "thornton",
    package = NULL,
    docs = FALSE,
    search = NULL) {
    checkmate::assert_string(dataset)
    checkmate::assert_string(package, null.ok = TRUE)
    checkmate::assert_flag(docs)
    checkmate::assert_string(search, null.ok = TRUE)
    insight::check_if_installed("Rdatasets")

    if (!is.null(search)) {
        return(get_dataset_search(search))
    }

    # Define data dictionary for marginaleffects datasets
    data_dict <- c(
        "affairs" = "https://marginaleffects.com/data/affairs.parquet",
        "airbnb" = "https://marginaleffects.com/data/airbnb.parquet",
        "cameras" = "https://marginaleffects.com/data/cameras.parquet",
        "ces_demographics" = "https://marginaleffects.com/data/ces_demographics.parquet",
        "ces_survey" = "https://marginaleffects.com/data/ces_survey.parquet",
        "immigration" = "https://marginaleffects.com/data/immigration.parquet",
        "lottery" = "https://marginaleffects.com/data/lottery.parquet",
        "military" = "https://marginaleffects.com/data/military.parquet",
        "thornton" = "https://marginaleffects.com/data/thornton.parquet",
        "factorial_01" = "https://marginaleffects.com/data/factorial_01.parquet",
        "interaction_01" = "https://marginaleffects.com/data/interaction_01.parquet",
        "interaction_02" = "https://marginaleffects.com/data/interaction_02.parquet",
        "interaction_03" = "https://marginaleffects.com/data/interaction_03.parquet",
        "interaction_04" = "https://marginaleffects.com/data/interaction_04.parquet",
        "polynomial_01" = "https://marginaleffects.com/data/polynomial_01.parquet",
        "polynomial_02" = "https://marginaleffects.com/data/polynomial_02.parquet"
    )

    # Check if it's a marginaleffects dataset first
    if (is.null(package) && dataset %in% names(data_dict)) {
        package <- "marginaleffects"
    }

    # marginaleffects-specific datasets
    if (identical(package, "marginaleffects")) {
        if (docs) {
            get_dataset_docs(dataset, data_dict)
            return(invisible(NULL))
        } else {
            rdpath <- getOption("marginaleffects_website_path", default = NULL)
            if (is.null(rdpath)) {
                out <- get_dataset_data(dataset, data_dict)
            } else {
                insight::check_if_installed("nanoparquet")
                out <- nanoparquet::read_parquet(file.path(
                    rdpath, "data", paste0(dataset, ".parquet")))
            }
            out <- as.data.frame(out, check.names = FALSE)
            return(out)
        }
        # Rdatasets
    } else if (is.null(package)) {
        if (docs) {
            Rdatasets::rddocs(dataset)
            return(invisible(NULL))
        } else {
            return(Rdatasets::rddata(dataset))
        }
    } else {
        if (docs) {
            Rdatasets::rddocs(dataset, package)
            return(invisible(NULL))
        } else {
            rdpath <- getOption("marginaleffects_rdataset_path", default = NULL)
            if (is.null(rdpath)) {
                return(Rdatasets::rddata(dataset, package))
            }

            insight::check_if_installed("nanoparquet")
            out <- nanoparquet::read_parquet(file.path(
                rdpath, "parquet", package,
                paste0(dataset, ".parquet")))
            out <- as.data.frame(out, check.names = FALSE)
            return(out)
        }
    }
}

get_dataset_search <- function(search) {
    # Use Rdatasets package to search all fields
    idx <- Rdatasets::rdsearch(search)
    # Convert column names to match the original format for compatibility
    if (nrow(idx) > 0 && "Dataset" %in% names(idx)) {
        names(idx)[names(idx) == "Dataset"] <- "Item"
    }
    return(idx)
}

get_dataset_data <- function(dataset, data_dict) {
    insight::check_if_installed("nanoparquet")
    checkmate::assert_choice(dataset, names(data_dict))

    data_url <- data_dict[dataset]
    temp_file <- tempfile(fileext = ".parquet")
    utils::download.file(data_url, temp_file, mode = "wb", quiet = TRUE)
    data <- nanoparquet::read_parquet(temp_file)
    return(data)
}

get_dataset_docs <- function(dataset, data_dict) {
    checkmate::assert_choice(dataset, names(data_dict))

    data_url <- data_dict[dataset]
    if (grepl("factorial|interaction|polynomial", dataset)) {
        documentation <- "https://marginaleffects.com/data/model_to_meaning_simulated_data.html"
    } else if (grepl("^ces", dataset)) {
        documentation <- "https://marginaleffects.com/data/ces.html"
    } else {
        documentation <- sub("parquet$", "html", data_url)
    }

    temp_doc <- tempfile(fileext = ".html")
    utils::download.file(documentation, temp_doc, mode = "w", quiet = TRUE)

    if (requireNamespace("rstudioapi")) {
        if (isTRUE(rstudioapi::isAvailable())) {
            rstudioapi::viewer(temp_doc)
            return()
        }
    }

    msg <- "Please choose a default browser with a command like: `options(browser = 'firefox')`"
    if (identical(getOption("browser"), "")) stop(msg, call. = FALSE)

    viewer <- getOption("viewer", utils::browseURL)
    if (!is.function(viewer)) stop(msg, call. = FALSE)

    viewer(temp_doc)
}

# Shared utilities for conformal prediction methods

get_conformal_score <- function(x, score, mfx = NULL) {
    if (is.null(mfx)) {
        mfx <- attr(x, "marginaleffects")
    }
    model <- mfx@model
    response_name <- insight::find_response(model)
    response <- x[[response_name]]
    if (!is.numeric(response) && score != "softmax") {
        stop_sprintf(
            'The response must be numeric. Did you want to use `conformal_score="softmax"`?'
        )
    }
    if (score == "residual_abs") {
        out <- abs(response - x$estimate)
    } else if (score == "residual_sq") {
        out <- (response - x$estimate)^2
        attr(out, "residual_abs") <- abs(response - x$estimate)
    } else if (score == "softmax") {
        response <- x[[insight::find_response(model)]]
        if (is.numeric(response) && is_binary(response)) {
            # See p.4 of Angelopoulos, Anastasios N., and Stephen Bates. 2022. "A
            # Gentle Introduction to Conformal Prediction and Distribution-Free
            # Uncertainty Quantification." arXiv.
            # https://doi.org/10.48550/arXiv.2107.07511.
            # 1 minus the softmax output of the true class
            out <- ifelse(response == 1, 1 - x$estimate, x$estimate)
        } else if ("group" %in% colnames(x)) {
            # HACK: is this fragile? I think `group` should always be character.
            idx <- as.character(response) == as.character(x$group)
            out <- 1 - x$estimate[idx]
        } else {
            stop_sprintf("Failed to compute the conformity score.")
        }
    }
    return(out)
}


get_conformal_bounds <- function(x, score, conf_level, mfx = NULL) {
    if (is.null(mfx)) {
        mfx <- attr(x, "marginaleffects")
    }
    model <- mfx@model
    response_name <- insight::find_response(model)
    response <- x[[response_name]]
    resid <- attr(score, "residual_abs")
    if (is.null(resid)) {
        resid <- score
    }
    d <- min(resid[score > stats::quantile(score, probs = conf_level)])
    if ("group" %in% colnames(x)) {
        q <- stats::quantile(
            score,
            probs = (length(score) + 1) * conf_level / length(score)
        )
        out <- x[x$estimate > (1 - q), ]
        data.table::setDT(out)
        out <- out[,
            .(pred.set = list(unique(group))),
            by = c("rowid", response_name)
        ]
        setorder(out, rowid)
        data.table::setDF(out)
        class(out) <- c("predictions", class(out))
        attr(out, "variables_datagrid") <- response_name
        return(out)
    } else {
        # continuous outcome: conformity half-width
        x$pred.low <- x$estimate - d
        x$pred.high <- x$estimate + d
    }
    return(x)
}

#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.afex_aov <- function(model, coefs, ...) {
    if (isTRUE(checkmate::check_matrix(model$lm$coefficients))) {
        mat <- matrix(coefs, ncol = ncol(model$lm$coefficients))
        dimnames(mat) <- dimnames(model$lm$coefficients)
        model$lm$coefficients <- mat
    } else {
        model$lm$coefficients <- coefs
    }
    return(model)
}


#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.afex_aov <- function(model, ...) {
    b <- insight::get_parameters(model)
    b <- stats::setNames(b$Estimate, paste(b$Parameter, b$Response, sep = ":"))
    return(b)
}


#' @include get_vcov.R
#' @rdname get_vcov
#' @export
get_vcov.afex_aov <- function(model, vcov = NULL, ...) {
    if (!is.null(vcov) && !is.logical(vcov)) {
        stop("The `vcov` argument is not supported for models of this class.")
    }
    insight::get_varcov(model)
}


#' @include get_predict.R
#' @rdname get_predict
#' @export
get_predict.afex_aov <- function(model, newdata = NULL, ...) {
    out <- stats::predict(model, newdata = newdata)
    out <- data.frame(estimate = out)
    if (isTRUE("rowid" %in% colnames(newdata))) {
        out[["rowid"]] <- newdata[["rowid"]]
    } else {
        out[["rowid"]] <- seq_len(nrow(out))
    }
    return(out)
}

conformal_split <- function(x,
                            data_test,
                            data_calib,
                            score,
                            conf_level,
                            mfx = NULL,
                            ...) {
    # calibration
    # use original model---fitted on the training set---to make predictions in the calibration set
    p_calib <- refit(x, newdata = data_calib, vcov = FALSE)
    score <- get_conformal_score(p_calib, score = score, mfx = mfx)

    # test
    # use original model to make predictions in the test set
    p_test <- refit(x, newdata = data_test, vcov = FALSE)

    # bounds
    out <- get_conformal_bounds(p_test,
        score = score,
        conf_level = conf_level,
        mfx = mfx)

    return(out)
}

#' @include get_predict.R
#' @rdname get_predict
#' @keywords internal
#' @export
get_predict.Learner <- function(model, newdata, type = NULL, ...) {
    if (!is.null(type) && !type %in% model$predict_types) {
        msg <- sprintf("Valid `type` values: %s", toString(model$predict_types))
        stop_sprintf(msg)
    }
    out <- drop(stats::predict(model, newdata = newdata, predict_type = type))
    out <- data.table(estimate = out)
    out <- add_rowid(out, newdata)
    return(out)
}


#' @include get_vcov.R
#' @rdname get_vcov
#' @export
get_vcov.Learner <- function(model, ...) {
    return(FALSE)
}


#' @include sanity_model.R
#' @rdname sanitize_model_specific
#' @export
sanitize_model_specific.Learner <- function(model, calling_function, ...) {
    if (calling_function == "hypotheses") {
        msg <- "`marginaleffects` does not support hypothesis tests for models of this class."
        stop_sprintf(msg)
    }
    return(model)
}

# TODO: heckit standard errors are not available because `vcov` is block diagonal with NAs

#' @rdname get_coef
#' @export
get_coef.selection <- function(model, ...) {
    if ("coefficients" %in% names(model)) {
        out <- model$coefficients
    } else if ("estimate" %in% names(model)) {
        out <- model$estimate
    } else {
        stop("Model does not have coefficients or estimates.")
    }
    return(out)
}


#' @rdname set_coef
#' @export
set_coef.selection <- function(model, coefs, ...) {
    # sampleSelection::selection
    if ("coefficients" %in% names(model)) {
        model[["coefficients"]] <- coefs
    } else if ("estimate" %in% names(model)) {
        model[["estimate"]] <- coefs
    } else {
        stop("Model does not have coefficients or estimates.")
    }
    return(model)
}

hypothesis_formula_list <- list(
    reference = list(
        ratio = list(
            comparison = function(x) (x / x[1])[2:length(x)],
            label = function(x) sprintf("(%s) / (%s)", x, x[1])[2:length(x)]
        ),
        difference = list(
            comparison = function(x) (x - x[1])[2:length(x)],
            label = function(x) sprintf("(%s) - (%s)", x, x[1])[2:length(x)]
        )
    ),
    sequential = list(
        ratio = list(
            comparison = function(x) (x / data.table::shift(x))[2:length(x)],
            label = function(x) sprintf("(%s) / (%s)", x, data.table::shift(x))[2:length(x)]
        ),
        difference = list(
            comparison = function(x) (x - data.table::shift(x))[2:length(x)],
            label = function(x) sprintf("(%s) - (%s)", x, data.table::shift(x))[2:length(x)]
        )
    ),
    pairwise = list(
        ratio = list(
            comparison = function(x) {
                safe_mode <- getOption("marginaleffects_safe", default = TRUE)
                if (length(x) > 25 && isTRUE(safe_mode)) {
                    msg <- "This command will generate many estimates. Set `options(marginaleffects_safe=FALSE)` to circumvent this guardrail."
                    stop(msg, call. = FALSE)
                }
                out <- outer(x, x, "/")
                diag(out) <- NA
                out[upper.tri(out)] <- NA # Set lower triangle to NA
                out <- as.vector(out)
                out <- out[!is.na(out)] # Keep only non-NA values
                out
            },
            label = function(x) {
                x <- sprintf("(%s)", x)
                out <- outer(x, x, paste, sep = " / ")
                out[upper.tri(out)] <- NA # Set lower triangle to NA
                diag(out) <- NA
                out <- as.vector(out)
                out[!is.na(out)]
            }),
        difference = list(
            comparison = function(x) {
                safe_mode <- getOption("marginaleffects_safe", default = TRUE)
                if (length(x) > 25 && isTRUE(safe_mode)) {
                    msg <- "This command will generate many estimates. Set `options(marginaleffects_safe=FALSE)` to circumvent this guardrail."
                    stop(msg, call. = FALSE)
                }
                out <- outer(x, x, "-")
                diag(out) <- NA
                out[upper.tri(out)] <- NA # Set lower triangle to NA
                out <- as.vector(out)
                out <- out[!is.na(out)] # Keep only non-NA values
                out
            },
            label = function(x) {
                x <- sprintf("(%s)", x)
                out <- outer(x, x, paste, sep = " - ")
                out[upper.tri(out)] <- NA # Set lower triangle to NA
                diag(out) <- NA
                out <- as.vector(out)
                out[!is.na(out)]
            })
    ),
    revpairwise = list(
        ratio = list(
            comparison = function(x) {
                out <- outer(x, x, "/")
                diag(out) <- NA
                out[lower.tri(out)] <- NA # Set lower triangle to NA
                out <- as.vector(out)
                out <- out[!is.na(out)] # Keep only non-NA values
                safe_mode <- getOption("marginaleffects_safe", default = TRUE)
                if (length(out) > 25 && isTRUE(safe_mode)) {
                    msg <- "This command will generate many estimates. Set `options(marginaleffects_safe=FALSE)` to circumvent this guardrail."
                    stop(msg, call. = FALSE)
                }
                out
            },
            label = function(x) {
                x <- sprintf("(%s)", x)
                out <- outer(x, x, paste, sep = " / ")
                out[lower.tri(out)] <- NA # Set lower triangle to NA
                diag(out) <- NA
                out <- as.vector(out)
                out[!is.na(out)]
            }),
        difference = list(
            comparison = function(x) {
                out <- outer(x, x, "-")
                diag(out) <- NA
                out[lower.tri(out)] <- NA # Set lower triangle to NA
                out <- as.vector(out)
                out <- out[!is.na(out)] # Keep only non-NA values
                safe_mode <- getOption("marginaleffects_safe", default = TRUE)
                if (length(out) > 25 && isTRUE(safe_mode)) {
                    msg <- "This command will generate many estimates. Set `options(marginaleffects_safe=FALSE)` to circumvent this guardrail."
                    stop(msg, call. = FALSE)
                }
                out
            },
            label = function(x) {
                x <- sprintf("(%s)", x)
                out <- outer(x, x, paste, sep = " - ")
                out[lower.tri(out)] <- NA # Set lower triangle to NA
                diag(out) <- NA
                out <- as.vector(out)
                out[!is.na(out)]
            })
    ),
    trt_vs_ctrl = list(
        ratio = list(
            comparison = function(x) mean(x[2:length(x)] / x[1]),
            label = function(x) "Mean(Trt) / Ctrl"
        ),
        difference = list(
            comparison = function(x) mean(x[2:length(x)] - x[1]),
            label = function(x) "Mean(Trt) - Ctrl"
        )
    ),
    meandev = list(
        ratio = list(
            comparison = function(x) x / mean(x),
            label = function(x) sprintf("(%s) / %s", x, "Mean")
        ),
        difference = list(
            comparison = function(x) x - mean(x),
            label = function(x) sprintf("(%s) - %s", x, "Mean")
        )
    ),
    meanotherdev = list(
        ratio = list(
            comparison = function(x) {
                s <- sum(x)
                m_other <- (s - x) / (length(x) - 1)
                x / m_other
            },
            label = function(x) sprintf("(%s) / %s", x, "Mean (other)")
        ),
        difference = list(
            comparison = function(x) {
                s <- sum(x)
                m_other <- (s - x) / (length(x) - 1)
                x - m_other
            },
            label = function(x) sprintf("(%s) - %s", x, "Mean (other)")
        )
    ),
    poly = list(
        dotproduct = list(
            comparison = function(x) {
                nx <- length(x)
                w <- stats::contr.poly(nx)
                w <- w[, 1:min(5, ncol(w))]
                as.vector(crossprod(w, matrix(x)))
            },
            label = function(x) {
                c("Linear", "Quadratic", "Cubic", "Quartic", "Quintic")[1:min(5, (length(x) - 1))]
            })
    ),
    helmert = list(
        dotproduct = list(
            comparison = function(x) {
                nx <- length(x)
                w <- stats::contr.helmert(nx)
                out <- as.vector(x %*% w)
                names(out) <- paste("Helmert", seq_along(out))
                return(out)
            },
            label = function(x) x
        )
    )
)


#' Internal function
#'
#' @noRd
#' @keywords internal
hypothesis_formula <- function(x, hypothesis, newdata, by) {
    insight::check_if_installed("collapse")
    # default values
    draws <- attr(x, "posterior_draws")

    if (inherits(x, "data.frame")) {
        data.table::setDT(x)
    }
    if (inherits(newdata, "data.frame")) {
        data.table::setDT(newdata)
        if (nrow(newdata) != nrow(x)) {
            newdata <- NULL
        }
    }


    form <- sanitize_hypothesis_formula(hypothesis)

    group <- form$group

    if (isTRUE(checkmate::check_character(by))) {
        bycols <- setdiff(by, group)
    } else {
        bycols <- by
    }
    labels <- get_labels(x, by = bycols, hypothesis_by = group)

    if (isTRUE(form$lhs == "arbitrary_function")) {
        fun_comparison <- sprintf("function(x) %s", form$rhs)
        fun_label <- sprintf("function(x) suppressWarnings(names(%s))", form$rhs)
        fun_comparison <- eval(parse(text = fun_comparison))
        fun_label <- eval(parse(text = fun_label))
    } else {
        fun_label <- hypothesis_formula_list[[form$rhs]][[form$lhs]]$label
        fun_comparison <- hypothesis_formula_list[[form$rhs]][[form$lhs]]$comparison
    }

    args <- list(matrix(x$estimate), FUN = fun_comparison)

    if (is.null(labels)) {
        labels <- paste("Row", seq_len(nrow(x)))
    }

    if (!is.null(group)) {
        if (any(!group %in% c(colnames(x), colnames(newdata)))) {
            msg <- "Some `~ | groupid` variables were not found in `newdata`."
            stop(msg, call. = FALSE)
        }
        col_x <- intersect(group, colnames(x))
        col_newdata <- intersect(group, colnames(newdata))
        groupval <- list()
        if (length(col_x) > 0) {
            groupval <- c(groupval, list(x[, ..col_x, drop = FALSE]))
        } else if (length(col_newdata) > 0) {
            groupval <- c(groupval, list(newdata[, ..col_newdata, drop = FALSE]))
        }
        groupval <- do.call(cbind, Filter(is.data.frame, groupval))
    } else {
        groupval <- NULL
    }

    combined <- list(x[, "estimate", drop = FALSE], groupval)
    combined <- Filter(function(x) inherits(x, "data.frame"), combined)
    combined <- do.call(cbind, combined)
    data.table::setDT(combined)

    if (is.null(groupval)) {
        estimates <- combined[, lapply(.SD, fun_comparison)]
    } else {
        estimates <- combined[, lapply(.SD, fun_comparison), keyby = groupval]
    }

    lab <- function(x) suppressWarnings(names(fun_comparison(x)))
    lab <- tryCatch(combined[, lapply(.SD, lab), keyby = groupval], error = function(e) NULL)

    if (inherits(lab, "data.frame") && nrow(lab) == nrow(estimates)) {
        data.table::setnames(lab, old = "estimate", "hypothesis")
        cols <- setdiff(colnames(lab), colnames(estimates))
        estimates <- cbind(lab[, ..cols], estimates)
    }

    if (!is.null(labels) && !inherits(lab, "data.frame") || nrow(lab) == 0) {
        combined[, estimate := labels]
        labels <- tryCatch(combined[, lapply(.SD, fun_label), keyby = groupval],
            error = function(e) NULL)
        if (inherits(labels, "data.frame") && nrow(labels) == nrow(estimates)) {
            data.table::setnames(labels, old = "estimate", "hypothesis")
            estimates <- cbind(labels, estimates)
        }
    }

    out <- estimates

    # Sometimes we get duplicated `term` columns
    # drop all instances after the first
    idx <- grep("^term$", colnames(out))
    if (length(idx) > 1) {
        idx <- idx[2:length(idx)]
        out <- out[, -..idx]
    }

    if (!is.null(draws)) {
        draws <- matrix_apply_column(draws, FUN = fun_comparison, by = groupval)
        if ("hypothesis" %in% colnames(out)) {
            row.names(draws) <- out$hypothesis
        }
    }


    attr(out, "posterior_draws") <- draws
    attr(out, "hypothesis_function_by") <- form$group

    return(out)
}

#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.survreg <- function(model, coefs, ...) {
    # Reverse engineering insight::get_get_parameters.survreg(),
    # which uses summary.survreg()

    nvar0 <- length(model$coefficients)
    nvar <- nrow(model$var)
    if (nvar > nvar0) {
        model[["coefficients"]][] <- coefs[-nvar0]
        model[["scale"]][] <- exp(coefs[nvar0])
    } else {
        model$coefficients[] <- coefs
    }

    model
}

#' @rdname get_predict
#' @export
get_predict.coxph <- function(
    model,
    newdata = insight::get_data(model),
    type = "lp",
    ...) {
    out <- stats::predict(model, newdata = newdata, type = type, ...)
    out <- data.table(estimate = out)
    out <- add_rowid(out, newdata)
    return(out)
}


#' @rdname sanitize_model_specific
sanitize_model_specific.coxph <- function(model, vcov, ...) {
    insight::check_if_installed("survival")
    flag1 <- !isFALSE(vcov)
    flag2 <- !isTRUE(checkmate::check_choice(vcov, choices = c("rsample", "boot", "fwb")))
    flag3 <- isTRUE(getOption("marginaleffects_safe", default = TRUE))
    if (flag1 && flag2 && flag3) {
        msg <- 'The default delta method standard errors for `coxph` models only take into account uncertainty in the regression coefficients. Standard errors may be too small. Use the `inferences()` function or set `vcov` to "rsample", "boot"  or "fwb" to compute confidence intervals by bootstrapping. Set `vcov` to `FALSE` or `options(marginaleffects_safe=FALSE)` to silence this warning.'
        warning(msg, call. = FALSE)
    }
    return(model)
}

get_contrast_data_numeric <- function(model,
                                      newdata,
                                      variable,
                                      modeldata,
                                      ...) {
    
    h <- variable[["eps"]]

    s <- m <- NA
    if (is.numeric(modeldata[[variable$name]])) {
        s <- stats::sd(modeldata[[variable$name]], na.rm = TRUE)
        m <- mean(modeldata[[variable$name]], na.rm = TRUE)
    }
    x <- newdata[[variable$name]]
    xmd <- modeldata[[variable$name]]

    make_label <- function(lab, val) {
        if (identical(lab, "custom")) return(lab)
        args <- append(list(lab), as.list(val))
        out <- tryCatch(
            do.call("sprintf", args),
            error = function(e) lab)
        return(out)
    }

    # slope
    # by default variable$value, so we need to check this first
    slopes <- c(
        "dY/dX",
        "eY/eX",
        "eY/dX",
        "dY/eX",
        "mean(dY/dX)",
        "mean(eY/eX)",
        "mean(eY/dX)",
        "mean(dY/eX)")

    # manual high
    if (isTRUE(checkmate::check_data_frame(variable$value))) {
        if (all(c("low", "high") %in% colnames(variable$value))) {
            low <- variable$value$low
            high <- variable$value$high
        } else {
            low <- variable$value[[1]]
            high <- variable$value[[2]]
        }
        lab <- "manual"

    } else if (isTRUE(variable$label %in% slopes)) {
        low <- x - h / 2
        high <- x + h / 2
        lab <- variable$label

    } else if (identical(variable$label, "exp(dY/dX)")) {
        low <- x - h / 2
        high <- x + h / 2
        lab <- "exp(dY/dX)"

    # contrast_label is designed for categorical predictors
    # numeric contrasts first
    } else if (isTRUE(checkmate::check_numeric(variable$value, len = 1))) {

        direction <- getOption("marginaleffects_contrast_direction", default = "forward")
        if (isTRUE(direction == "center")) {
            low <- x - variable$value / 2
            high <- x + variable$value / 2
        } else if (isTRUE(direction == "backward")) {
            low <- x - variable$value
            high <- x
        } else {
            low <- x
            high <- x + variable$value
        }

        # wrap in parentheses, unless mean() because there are already parentheses
        # important to display ratios of x+1, etc.
        # label should not be `(mpg+1) - mpg` because that is misleading for centered contrast
        if (!isTRUE(grepl("mean", variable$label))) {
            lab <- sprintf("+%s", variable$value)
        } else {
            lab <- sprintf("mean(+%s)", variable$value)
        }

    } else if (isTRUE(checkmate::check_numeric(variable$value, len = 2))) {
        variable$value <- sort(variable$value)
        low <- variable$value[1]
        high <- variable$value[2]
        gap <- diff(variable$value)
        lab <- make_label(variable$label, rev(variable$value))

    # character contrasts
    } else if (identical(variable$value, "sd")) {
        low <- m - s / 2
        high <- m + s / 2
        lab <- c("x + sd/2", "x - sd/2")
        if (!isTRUE(grepl("mean", variable$label))) {
            lab <- sprintf("(%s)", lab)
        }
        lab <- make_label(variable$label, lab)

    } else if (identical(variable$value, "2sd")) {
        low <- m - s
        high <- m + s
        lab <- c("x + sd", "x - sd")
        if (!isTRUE(grepl("mean", variable$label))) {
            lab <- sprintf("(%s)", lab)
        }
        lab <- make_label(variable$label, lab)

    } else if (identical(variable$value, "iqr")) {
        low <- stats::quantile(xmd, probs = 0.25, na.rm = TRUE)
        high <- stats::quantile(xmd, probs = 0.75, na.rm = TRUE)
        lab <- make_label(variable$label, c("Q3", "Q1"))

    } else if (identical(variable$value, "minmax")) {
        low <- min(xmd, na.rm = TRUE)
        high <- max(xmd, na.rm = TRUE)
        lab <- make_label(variable$label, c("Max", "Min"))

    } else if (isTRUE(checkmate::check_function(variable$value))) {
        tmp <- variable$value(x)
        low <- tmp[, 1]
        high <- tmp[, 2]
        lab <- "custom"
    }

    lo <- hi <- newdata
    lo[[variable$name]] <- low
    hi[[variable$name]] <- high
    out <- list(rowid = seq_len(nrow(newdata)),
                lo = lo,
                hi = hi,
                original = newdata,
                ter = rep(variable$name, nrow(newdata)),
                lab = rep(lab, nrow(newdata)))
    return(out)
}

# Helper functions for add_variables()

get_predictors <- function(variables, mfx) {
    model <- mfx@model
    modeldata <- mfx@modeldata
    newdata <- mfx@newdata
    calling_function <- mfx@calling_function

    stop_zero <- function(predictors) {
        if (length(predictors) == 0) {
            msg <- "There is no valid predictor variable. Please change the `variables` argument or supply an alternative data frame to the `newdata` argument."
            stop_sprintf(msg)
        }
    }

    # variables -> predictors (named list)
    if (isTRUE(checkmate::check_list(variables, names = "unique"))) {
        predictors <- variables
    } else if (isTRUE(checkmate::check_character(variables))) {
        predictors <- stats::setNames(
            rep(list(NULL), length(variables)),
            variables)
    } else {
        # we don't want group ids in multi-level models and other weirdness
        predictors <- insight::find_predictors(mfx@model, component = "all")

        # mhurdle names variables weirdly
        if (inherits(model, "mhurdle")) {
            predictors <- unlist(predictors)
            predictors <- list(conditional = predictors)
        }

        known <- c("fixed", "conditional", "zero_inflated", "scale", "nonlinear")
        if (any(known %in% names(predictors))) {
            predictors <- predictors[known]
        }
        predictors <- unlist(predictors)
        predictors <- setdiff(predictors, mfx@variable_names_response)
        predictors <- stats::setNames(
            rep(list(NULL), length(predictors)),
            predictors)
    }
    stop_zero(predictors)

    # warn if missing from `newdata`
    miss <- setdiff(names(predictors), colnames(newdata))
    if (length(miss) > 0) {
        msg <- sprintf("Variable(s) not found in `newdata`: %s.", toString(miss))
        warn_sprintf(msg)
        predictors <- predictors[names(predictors) %in% colnames(newdata)]
    }
    stop_zero(predictors)

    # warn if matrix column
    mat <- sapply(names(predictors), function(v) {
        check_variable_class(mfx, variable = v, compare = "matrix")
    })
    if (any(mat)) {
        warn_sprintf("Matrix columns are not supported: %s. Use the `variables` argument to specify valid predictors, or use a function like `drop()` to convert your matrix columns into vectors.", toString(names(mat)[mat]))
        predictors <- predictors[setdiff(names(predictors), names(mat)[mat])]
    }
    stop_zero(predictors)

    # sometimes `insight` returns interaction component as if
    # it were a constituent variable
    idx <- grepl(":", names(predictors))
    predictors <- predictors[!idx]
    stop_zero(predictors)

    return(predictors)
}

sanitize_predictor_container <- function(predictors) {
    # character -> list
    if (isTRUE(checkmate::check_character(predictors))) {
        predictors <- stats::setNames(
            rep(list(NULL), length(predictors)),
            predictors
        )
    }
    predictors
}

add_prediction_functions <- function(predictors, modeldata) {
    # functions to values
    # only for predictions; get_comparisons_data_numeric handles this for comparisons()
    # do this before NULL-to-defaults so we can fill it in with default in case of failure
    for (v in names(predictors)) {
        if (is.function(predictors[[v]])) {
            tmp <- hush(predictors[[v]](modeldata[[v]]))
            if (is.null(tmp)) {
                msg <- sprintf(
                    "The `%s` function produced invalid output when applied to the dataset used to fit the model.",
                    v
                )
                warn_sprintf(msg)
            }
            predictors[[v]] <- hush(predictors[[v]](modeldata[[v]]))
        }
    }
    predictors
}

add_numeric_shortcuts <- function(predictors, mfx) {
    # string shortcuts for predictions only
    modeldata <- mfx@modeldata
    for (v in names(predictors)) {
        if (check_variable_class(mfx, v, "numeric") || check_variable_class(mfx, v, "integer")) {
            if (identical(predictors[[v]], "iqr")) {
                predictors[[v]] <- stats::quantile(
                    modeldata[[v]],
                    probs = c(0.25, 0.75),
                    na.rm = TRUE
                )
            } else if (identical(predictors[[v]], "minmax")) {
                predictors[[v]] <- c(
                    min(modeldata[[v]], na.rm = TRUE),
                    max(modeldata[[v]], na.rm = TRUE)
                )
            } else if (identical(predictors[[v]], "sd")) {
                s <- stats::sd(modeldata[[v]], na.rm = TRUE)
                m <- mean(modeldata[[v]], na.rm = TRUE)
                predictors[[v]] <- c(m - s / 2, m + s / 2)
            } else if (identical(predictors[[v]], "2sd")) {
                s <- stats::sd(modeldata[[v]], na.rm = TRUE)
                m <- mean(modeldata[[v]], na.rm = TRUE)
                predictors[[v]] <- c(m - s, m + s)
            } else if (identical(predictors[[v]], "threenum")) {
                s <- stats::sd(modeldata[[v]], na.rm = TRUE)
                m <- mean(modeldata[[v]], na.rm = TRUE)
                predictors[[v]] <- c(m - s, m, m + s)
            } else if (identical(predictors[[v]], "fivenum")) {
                predictors[[v]] <- stats::fivenum(modeldata[[v]])
            } else if (is.character(predictors[[v]])) {
                msg <- sprintf(
                    '%s is a numeric variable. The summary shortcuts supported by the variables argument are: "iqr", "minmax", "sd", "2sd", "threenum", "fivenum".',
                    v
                )
                stop_sprintf(msg)
            }
        }
    }
    predictors
}

add_default_values <- function(predictors, mfx) {
    calling_function <- mfx@calling_function

    # NULL to defaults
    for (v in names(predictors)) {
        if (is.null(predictors[[v]])) {
            if (check_variable_class(mfx, v, "binary")) {
                predictors[[v]] <- 0:1
            } else if (check_variable_class(mfx, v, "numeric") || check_variable_class(mfx, v, "integer")) {
                if (calling_function == "comparisons") {
                    predictors[[v]] <- 1
                } else if (calling_function == "predictions") {
                    v_unique <- unique(mfx@modeldata[[v]])
                    if (length(v_unique) < 6) {
                        predictors[[v]] <- v_unique
                    } else {
                        predictors[[v]] <- stats::fivenum(mfx@modeldata[[v]])
                    }
                }
            } else {
                if (calling_function == "comparisons") {
                    predictors[[v]] <- "reference"
                } else if (calling_function == "predictions") {
                    # TODO: warning when this is too large. Here or elsewhere?
                    predictors[[v]] <- unique(mfx@modeldata[[v]])
                }
            }
        }
    }
    predictors
}

sanitize_predictor_specs <- function(predictors, mfx) {
    modeldata <- mfx@modeldata
    newdata <- mfx@newdata
    calling_function <- mfx@calling_function

    # shortcuts and validity
    for (v in names(predictors)) {
        if (
            isTRUE(checkmate::check_data_frame(
                predictors[[v]],
                nrows = nrow(newdata)
            ))
        ) {
            # do nothing, but don't take the other validity check branches
        } else if (check_variable_class(mfx, v, "binary")) {
            if (
                !isTRUE(checkmate::check_numeric(predictors[[v]])) ||
                    !is_binary(predictors[[v]])
            ) {
                msg <- sprintf(
                    "The `%s` variable is binary. The corresponding entry in the `variables` argument must be 0 or 1.",
                    v
                )
                stop_sprintf(msg)
            }
            # get_comparisons_data requires both levels
            if (calling_function == "comparisons") {
                if (length(predictors[[v]]) != 2) {
                    msg <- sprintf(
                        "The `%s` variable is binary. The corresponding entry in the `variables` argument must be a vector of length 2.",
                        v
                    )
                    stop_sprintf(msg)
                }
            }
        } else if (check_variable_class(mfx, v, "numeric") || check_variable_class(mfx, v, "integer")) {
            if (calling_function == "comparisons") {
                # For comparisons(), the string shortcuts are processed in contrast_data_* functions because we need fancy labels.
                # Eventually it would be nice to consolidate, but that's a lot of work.
                valid_str <- c("iqr", "minmax", "sd", "2sd")
                flag <- isTRUE(checkmate::check_numeric(
                    predictors[[v]],
                    min.len = 1,
                    max.len = 2
                )) ||
                    isTRUE(checkmate::check_choice(
                        predictors[[v]],
                        choices = valid_str
                    )) ||
                    isTRUE(checkmate::check_function(predictors[[v]]))
                if (!isTRUE(flag)) {
                    msg <- "The %s element of the `variables` argument is invalid."
                    msg <- sprintf(msg, v)
                    stop_sprintf(msg)
                }
            } else if (calling_function == "predictions") {
                # string shortcuts
                if (identical(predictors[[v]], "iqr")) {
                    predictors[[v]] <- stats::quantile(
                        modeldata[[v]],
                        probs = c(0.25, 0.75),
                        na.rm = TRUE
                    )
                } else if (identical(predictors[[v]], "minmax")) {
                    predictors[[v]] <- c(
                        min(modeldata[[v]], na.rm = TRUE),
                        max(modeldata[[v]], na.rm = TRUE)
                    )
                } else if (identical(predictors[[v]], "sd")) {
                    s <- stats::sd(modeldata[[v]], na.rm = TRUE)
                    m <- mean(modeldata[[v]], na.rm = TRUE)
                    predictors[[v]] <- c(m - s / 2, m + s / 2)
                } else if (identical(predictors[[v]], "2sd")) {
                    s <- stats::sd(modeldata[[v]], na.rm = TRUE)
                    m <- mean(modeldata[[v]], na.rm = TRUE)
                    predictors[[v]] <- c(m - s, m + s)
                } else if (identical(predictors[[v]], "threenum")) {
                    s <- stats::sd(modeldata[[v]], na.rm = TRUE)
                    m <- mean(modeldata[[v]], na.rm = TRUE)
                    predictors[[v]] <- c(m - s, m, m + s)
                } else if (identical(predictors[[v]], "fivenum")) {
                    predictors[[v]] <- stats::fivenum
                } else if (is.character(predictors[[v]])) {
                    msg <- sprintf(
                        '%s is a numeric variable. The summary shortcuts supported by the variables argument are: "iqr", "minmax", "sd", "2sd", "threenum", "fivenum".',
                        v
                    )
                    stop_sprintf(msg)
                }
            }
        } else {
            if (calling_function == "comparisons") {
                valid <- c(
                    "reference",
                    "sequential",
                    "pairwise",
                    "all",
                    "revpairwise",
                    "revsequential",
                    "revreference"
                )
                # minmax needs an actual factor in the original data to guarantee correct order of levels.
                if (is.factor(modeldata[[v]])) {
                    valid <- c(valid, "minmax")
                }
                flag1 <- checkmate::check_choice(predictors[[v]], choices = valid)
                flag2 <- checkmate::check_vector(predictors[[v]], len = 2)
                flag3 <- checkmate::check_data_frame(
                    predictors[[v]],
                    nrows = nrow(newdata),
                    ncols = 2
                )
                flag4 <- checkmate::check_function(predictors[[v]])
                flag5 <- checkmate::check_data_frame(predictors[[v]])
                if (
                    !isTRUE(flag1) &&
                        !isTRUE(flag2) &&
                        !isTRUE(flag3) &&
                        !isTRUE(flag4) &&
                        !isTRUE(flag5)
                ) {
                    msg <- "The %s element of the `variables` argument must be a vector of length 2 or one of: %s"
                    msg <- sprintf(msg, v, toString(valid))
                    stop_sprintf(msg)
                }
            } else if (calling_function == "predictions") {
                if (is.character(predictors[[v]]) || is.factor(predictors[[v]])) {
                    if (
                        !all(
                            as.character(predictors[[v]]) %in% as.character(modeldata[[v]])
                        )
                    ) {
                        invalid <- intersect(
                            as.character(predictors[[v]]),
                            c(
                                "pairwise",
                                "reference",
                                "sequential",
                                "revpairwise",
                                "revreference",
                                "revsequential"
                            )
                        )
                        if (length(invalid) > 0) {
                            msg <- "These values are only supported by the `variables` argument in the `comparisons()` function: %s"
                            msg <- sprintf(msg, toString(invalid))
                        } else {
                            msg <- "Some elements of the `variables` argument are not in their original data. Check this variable: %s"
                            msg <- sprintf(msg, v)
                        }
                        stop_sprintf(msg)
                    }
                }
            }
        }
    }
    predictors
}

get_comparison_functions <- function(mfx) {
    # goals:
    # allow multiple function types: slopes() uses both difference and dydx
    # when comparison is defined, use that if it works or turn back to defaults
    # predictors list elements: name, value, function, label
    comparison <- mfx@comparison
    by <- mfx@by

    if (is.null(comparison)) {
        fun_numeric <- fun_categorical <- comparison_function_dict[["difference"]]
        lab_numeric <- lab_categorical <- comparison_label_dict[["difference"]]
    } else if (is.function(comparison)) {
        fun_numeric <- fun_categorical <- comparison
        lab_numeric <- lab_categorical <- "custom"
    } else if (is.character(comparison)) {
        # switch to the avg version when there is a `by` function
        if (
            (isTRUE(checkmate::check_character(by)) || isTRUE(by)) &&
                !isTRUE(grepl("avg$", comparison))
        ) {
            comparison <- paste0(comparison, "avg")
        }

        # weights if user requests `avg` or automatically switched
        if (
            isTRUE(grepl("avg$", comparison)) &&
                "marginaleffects_wts_internal" %in% colnames(mfx@newdata)
        ) {
            comparison <- paste0(comparison, "wts")
        }

        fun_numeric <- fun_categorical <- comparison_function_dict[[comparison]]
        lab_numeric <- lab_categorical <- comparison_label_dict[[comparison]]
        if (isTRUE(grepl("dydxavgwts|eyexavgwts|dyexavgwts|eydxavgwts", comparison))) {
            fun_categorical <- comparison_function_dict[["differenceavgwts"]]
            lab_categorical <- comparison_label_dict[["differenceavgwts"]]
        } else if (isTRUE(grepl("dydxavg|eyexavg|dyexavg|eydxavg", comparison))) {
            fun_categorical <- comparison_function_dict[["differenceavg"]]
            lab_categorical <- comparison_label_dict[["differenceavg"]]
        } else if (isTRUE(grepl("dydx$|eyex$|dyex$|eydx$", comparison))) {
            fun_categorical <- comparison_function_dict[["difference"]]
            lab_categorical <- comparison_label_dict[["difference"]]
        }
    }

    list(
        fun_numeric = fun_numeric,
        fun_categorical = fun_categorical,
        lab_numeric = lab_numeric,
        lab_categorical = lab_categorical,
        comparison = comparison
    )
}

add_functions_and_labels <- function(predictors, comparison_config, mfx) {
    modeldata <- mfx@modeldata

    for (v in names(predictors)) {
        if (
            (check_variable_class(mfx, v, "numeric") || check_variable_class(mfx, v, "integer")) &&
                !check_variable_class(mfx, v, "binary")
        ) {
            fun <- comparison_config$fun_numeric
            lab <- comparison_config$lab_numeric
        } else {
            fun <- comparison_config$fun_categorical
            lab <- comparison_config$lab_categorical
        }
        predictors[[v]] <- list(
            "name" = v,
            "function" = fun,
            "label" = lab,
            "value" = predictors[[v]],
            "comparison" = comparison_config$comparison
        )
    }
    predictors
}

add_epsilon_values <- function(predictors, mfx, eps) {
    modeldata <- mfx@modeldata

    # epsilon for finite difference
    for (v in names(predictors)) {
        if (!is.null(eps)) {
            predictors[[v]][["eps"]] <- eps
        } else if (is.numeric(modeldata[[v]])) {
            predictors[[v]][["eps"]] <- 1e-4 *
                diff(range(modeldata[[v]], na.rm = TRUE, finite = TRUE))
            # 1-row grid has 0 range
            if (predictors[[v]][["eps"]] == 0) {
                predictors[[v]][["eps"]] <- 1e-4
            }
        } else {
            predictors[[v]]["eps"] <- list(NULL)
        }
    }
    predictors
}

sanitize_internal_variables <- function(predictors, mfx, model, calling_function) {
    # can't take the slope of an outcome, except in weird brms models (issue #1006)
    if (!inherits(model, "brmsfit") || !isTRUE(length(model$formula$forms) > 1)) {
        dv <- mfx@variable_names_response
        # sometimes insight doesn't work
        if (length(dv) > 0) {
            predictors <- predictors[setdiff(names(predictors), dv)]
        }
    }
    if (length(predictors) == 0) {
        stop_sprintf(
            "There is no valid predictor variable. Please make sure your model includes predictors and use the `variables` argument."
        )
    }

    # sort variables alphabetically
    predictors <- predictors[sort(names(predictors))]

    # internal variables are not predictors
    predictors <- predictors[
        !names(predictors) %in% c("marginaleffects_wts_internal", "rowid_dedup")
    ]
    if (length(predictors) == 0 && calling_function == "comparisons") {
        stop_sprintf("No valid predictor variable.")
    }

    predictors
}

get_weight_variables <- function(mfx) {
    w <- mfx@variable_names_wts
    w <- intersect(w, colnames(mfx@newdata))
    w
}

# input: character vector or named list
# output: named list of lists where each element represents a variable with: name, value, function, label
add_variables <- function(
    variables,
    mfx) {
    # Input validation
    checkmate::assert(
        checkmate::check_null(variables),
        checkmate::check_character(variables, min.len = 1, names = "unnamed"),
        checkmate::check_list(variables, min.len = 1, names = "unique"),
        combine = "or"
    )

    # Extract and normalize predictors
    predictors <- get_predictors(variables, mfx)
    predictors <- sanitize_predictor_container(predictors)

    # Handle predictions-specific processing
    if (mfx@calling_function == "predictions") {
        modeldata <- mfx@modeldata
        predictors <- add_prediction_functions(predictors, modeldata)
        predictors <- add_numeric_shortcuts(predictors, mfx)
    }

    # Set defaults and validate
    modeldata <- mfx@modeldata
    predictors <- add_default_values(predictors, mfx)
    predictors <- sanitize_predictor_specs(predictors, mfx)

    # Configure comparison functions and labels
    comparison_config <- get_comparison_functions(mfx)
    predictors <- add_functions_and_labels(predictors, comparison_config, mfx)

    # Add epsilon values
    predictors <- add_epsilon_values(predictors, mfx, mfx@eps)

    # Final sanitization
    predictors <- sanitize_internal_variables(predictors, mfx, mfx@model, mfx@calling_function)

    # output
    mfx@variables <- predictors
    return(mfx)
}


detect_variable_class <- function(modeldata, model = NULL) {
    if (is.null(modeldata)) {
        return(character())
    }

    # this can be costly on large datasets, when only a portion of
    # variables are used in the model
    variables <- NULL
    if (!is.null(model)) {
        variables <- tryCatch(
            unlist(insight::find_variables(model, flatten = TRUE, verbose = FALSE), use.names = FALSE),
            error = function(e) NULL
        )
        # Always include all columns that are actually present in modeldata
        # This ensures that variables used in 'by' arguments get properly classified
        if (!is.null(variables)) {
            variables <- union(variables, colnames(modeldata))
        }
    }
    if (is.null(variables)) variables <- colnames(modeldata)

    out <- modeldata

    cl <- NULL
    for (col in variables) {
        if (is.matrix(out[[col]])) {
            cl[col] <- "matrix"
        } else if (is.logical(out[[col]])) {
            cl[col] <- "logical"
        } else if (is.character(out[[col]])) {
            cl[col] <- "character"
        } else if (is.factor(out[[col]])) {
            cl[col] <- "factor"
        } else if (inherits(out[[col]], "Surv")) {
            # is numeric but breaks the %in% 0:1 check
            cl[col] <- "other"
        } else if (isTRUE(checkmate::check_integerish(out[[col]]))) {
            if (is_binary(out[[col]])) {
                cl[col] <- "binary"
            } else {
                cl[col] <- "integer"
            }
        } else if (is.numeric(out[[col]])) {
            cl[col] <- "numeric"
        } else {
            cl[col] <- "other"
        }
    }

    if (is.null(model)) {
        return(cl)
    }

    te <- hush(insight::find_terms(model, flatten = TRUE))

    # in-formula factor
    regex <- "^(^as\\.factor|^factor)\\((.*)\\)"
    idx <- gsub(
        regex,
        "\\2",
        Filter(function(x) grepl(regex, x), te)
    )
    cl[names(cl) %in% idx] <- "factor"

    # in-formula categoricals
    regex <- "^(^mo|^strata)\\((.*)\\)"
    idx <- gsub(
        regex,
        "\\2",
        Filter(function(x) grepl(regex, x), te)
    )
    cl[names(cl) %in% idx] <- "strata"

    # in-formula numeric
    regex <- "^numeric\\((.*)\\)$|^as.numeric\\((.*)\\)$"
    idx <- gsub(
        regex,
        "\\1",
        Filter(function(x) grepl(regex, x), te)
    )
    cl[names(cl) %in% idx] <- "numeric"

    # in-formula logical
    regex <- "^logical\\((.*)\\)$|^as.logical\\((.*)\\)$"
    idx <- gsub(
        regex,
        "\\1",
        Filter(function(x) grepl(regex, x), te)
    )
    cl[names(cl) %in% idx] <- "logical"

    # in-formula: fixest::i()
    fi <- NULL
    idx <- grepl("^i\\(.*\\)$", te)
    if (sum(idx) > 0) {
        arg1 <- lapply(te[idx], function(x) hush(as.character(str2lang(x)[[2]])))
        arg2 <- lapply(te[idx], function(x) hush(as.character(str2lang(x)[[3]])))
        arg1 <- unlist(arg1, recursive = TRUE)
        arg2 <- unlist(arg2, recursive = TRUE)
        arg2 <- gsub("^i\\.", "", arg2[grepl("^i\\.", arg2)])
        fi <- unique(c(arg1, arg2))
    }
    for (f in fi) {
        cl[f] <- "cluster"
    }

    # random effects groups
    re <- hush(insight::find_random(model, flatten = TRUE))
    cl[names(cl) %in% re] <- "cluster"

    return(cl)
}


check_variable_class <- function(newdata, variable = NULL, compare = NULL) {
    if (inherits(newdata, "marginaleffects_internal")) {
        cl <- newdata@variable_class
    } else if ("marginaleffects_variable_class" %in% names(attributes(newdata))) {
        cl <- attributes(newdata)$marginaleffects_variable_class
    } else if (isTRUE(checkmate::check_character(newdata))) {
        cl <- newdata
    } else {
        cl <- character()
    }

    if (is.null(compare) && is.null(variable)) {
        out <- cl
    } else if (is.null(compare)) {
        # Handle vector of variables
        out <- cl[variable]
        out[!names(out) %in% names(cl)] <- NA
    } else if (is.null(variable)) {
        if (isTRUE(compare == "categorical")) {
            out <- cl[
                cl %in%
                    c("factor", "character", "logical", "strata", "cluster", "binary")
            ]
        } else {
            out <- cl[cl %in% compare]
        }
    } else {
        # Handle vector of variables: use any() to check if ANY variable matches
        vars_in_cl <- variable %in% names(cl)
        if (any(vars_in_cl)) {
            if (isTRUE(compare == "categorical")) {
                out <- any(cl[variable[vars_in_cl]] %in%
                    c("factor", "character", "logical", "strata", "cluster", "binary"))
            } else {
                out <- any(cl[variable[vars_in_cl]] %in% compare)
            }
        } else {
            out <- FALSE
        }
    }

    return(out)
}

# Global variable to avoid R CMD check note
utils::globalVariables("mAD")

eval_fun_with_numpy_arrays <- function(FUN, ...) {
    dots <- list(...)
    # Handle special cases for JAX indexing
    for (i in seq_along(dots)) {
        if (names(dots)[i] %in% c("num_groups", "link_type", "family_type", "comparison_type")) {
            # Keep num_groups, link_type, family_type as is (integer scalars)
        } else if (names(dots)[i] == "groups") {
            # Convert groups to integer array explicitly
            dots[[i]] <- reticulate::np_array(dots[[i]], dtype = "int32")
        } else {
            # Convert to numpy array
            dots[[i]] <- reticulate::np_array(dots[[i]])
        }
    }
    # Call Python function - use reticulate's $ operator which handles Python callables
    # Build the call dynamically
    result <- rlang::exec(FUN, !!!dots)
    return(result)
}


autodiff_warning <- function(feature) {
    msg <- "Automatic differentiation with JAX does not support %s. Reverting to finite difference."
    warning(sprintf(msg, feature), call. = FALSE)
}


#' @keywords internal
#' @noRd
get_autodiff_args <- function(model, mfx) {
    UseMethod("get_autodiff_args")
}


#' @rdname get_autodiff_args
#' @keywords internal
#' @noRd
get_autodiff_args.default <- function(model, mfx) {
    return(NULL)
}



jax_align_group_J <- function(jac_fun, mfx, original, estimates, X, X_hi, X_lo) {
    if (isTRUE(grepl("_byG", jac_fun))) {
        # Case 1: comparisons with by=character uses pre-assigned group IDs
        if (!is.null(original) && "marginaleffects_group_id" %in% colnames(original)) {
            groups <- original$marginaleffects_group_id - 1L  # Convert to 0-indexed
            num_groups <- max(groups) + 1L
        } else {
            # Case 2: predictions with by=character, or comparisons with by=TRUE
            # Need to create groups on the fly

            # Determine grouping columns
            if (is.character(mfx@by)) {
                # predictions with by=character: use user-specified columns from newdata
                bycols <- mfx@by
                if (inherits(mfx@newdata, "data.table")) {
                    groups_data <- mfx@newdata[, ..bycols]
                } else {
                    groups_data <- mfx@newdata[, bycols, drop = FALSE]
                }
            } else if (!is.null(original)) {
                # comparisons with by=TRUE: use term/contrast from original
                bycols <- intersect(c("term", "contrast"), colnames(original))
                groups_data <- subset(original, select = bycols)
            } else {
                bycols <- NULL
            }

            if (length(bycols) > 0) {
                # Sort data by group columns to ensure consistent ordering
                idx <- do.call(order, groups_data)
                groups_data <- groups_data[idx, , drop = FALSE]
                if (!is.null(X)) X <- X[idx, , drop = FALSE]
                if (!is.null(X_hi)) X_hi <- X_hi[idx, , drop = FALSE]
                if (!is.null(X_lo)) X_lo <- X_lo[idx, , drop = FALSE]
                # Create group IDs
                groups_combined <- apply(groups_data, 1, function(x) paste0(x, collapse = "_"))
                groups <- as.integer(factor(groups_combined, levels = unique(groups_combined))) - 1L
                num_groups <- max(groups) + 1L
            } else {
                groups <- num_groups <- NULL
            }
        }
    } else {
        groups <- num_groups <- NULL
    }

    return(list(groups = groups, num_groups = num_groups, X = X, X_hi = X_hi, X_lo = X_lo))
}



#' Add model matrix attribute to a data frame
#' @keywords internal
#' @noRd
add_model_matrix_attribute_data <- function(mfx, data) {
    # Temporarily set as newdata to get model matrix
    original_newdata <- mfx@newdata
    mfx@newdata <- data
    data_with_mm <- add_model_matrix_attribute(mfx)
    return(data_with_mm)
}


#' Sanitize mfx object for autodiff
#'
#' Check common requirements for autodiff support and return NULL if unsupported
#' @param mfx marginaleffects object
#' @return mfx object if all checks pass, NULL otherwise
#' @keywords internal
#' @noRd
sanitize_autodiff_mfx <- function(mfx) {
    # Check model support via get_autodiff_args
    autodiff_args <- get_autodiff_args(mfx@model, mfx)
    if (is.null(autodiff_args)) {
        return(NULL)
    }

    # Check for unsupported wts argument
    if (!is.null(mfx@wts)) {
        autodiff_warning("the `wts` argument")
        return(NULL)
    }

    # Check for unsupported hypothesis argument
    if (!is.null(mfx@hypothesis)) {
        autodiff_warning("the `hypothesis` argument")
        return(NULL)
    }

    # Check for unsupported by argument values
    if (!isTRUE(mfx@by) && !isFALSE(mfx@by) && !is.character(mfx@by)) {
        autodiff_warning("values of `by` other than TRUE, FALSE, or a character vector")
        return(NULL)
    }

    # All checks passed, return mfx
    return(mfx)
}


#' Compute predictions with JAX autodiff (estimates, SE, jacobian)
#' @return List with estimate, std.error, jacobian, or NULL if unsupported
#' @keywords internal
#' @noRd
jax_predictions <- function(mfx, vcov_matrix, ...) {
    # Validate mfx object for autodiff support
    mfx <- sanitize_autodiff_mfx(mfx)
    if (is.null(mfx)) return(NULL)

    # Get autodiff args (already validated in sanitize_autodiff_mfx)
    autodiff_args <- get_autodiff_args(mfx@model, mfx)

    mAD <- settings_get("mAD")

    # Get model matrix
    X <- attr(mfx@newdata, "marginaleffects_model_matrix")
    if (is.null(X)) return(NULL)

    # Get coefficients
    coefs <- get_coef(mfx@model, ...)

    # Check for NA coefficients (e.g., from aliased terms)
    if (anyNA(coefs)) {
        autodiff_warning("models with NA coefficients (aliased terms)")
        return(NULL)
    }

    # Determine aggregation function
    if (isFALSE(mfx@by)) {
        fun_name <- "predictions"
        groups <- NULL
        num_groups <- NULL
    } else if (isTRUE(mfx@by)) {
        fun_name <- "predictions_byT"
        groups <- NULL
        num_groups <- NULL
    } else if (is.character(mfx@by)) {
        fun_name <- "predictions_byG"
        # Prepare group indices
        group_result <- jax_align_group_J("jacobian_byG", mfx, NULL, NULL, X, NULL, NULL)
        groups <- group_result$groups
        num_groups <- group_result$num_groups
        X <- group_result$X

        # If groups couldn't be created, fall back to finite differences
        if (is.null(groups) || is.null(num_groups)) {
            return(NULL)
        }
    }

    # Select autodiff function
    # e.g., mAD$linear$predictions is a module containing predictions(), predictions_byT(), predictions_byG()
    # The base module name (predictions or comparisons)
    base_module_name <- if (grepl("predictions", fun_name)) "predictions" else "comparisons"
    module <- mAD[[autodiff_args$model_type]][[base_module_name]]
    FUN <- module[[fun_name]]  # e.g., module$predictions or module$predictions_byT

    # Build arguments (without FUN)
    args <- list(
        beta = coefs,
        X = X,
        vcov = vcov_matrix,
        groups = groups,
        num_groups = num_groups,
        family_type = autodiff_args$family_type,
        link_type = autodiff_args$link_type
    )
    args <- Filter(function(x) !is.null(x), args)

    # Call Python function using eval_fun_with_numpy_arrays
    result <- do.call(eval_fun_with_numpy_arrays, c(list(FUN = FUN), args))

    # Convert to R objects
    J <- as.matrix(result[["jacobian"]])

    # Ensure jacobian is (n_predictions x n_coefs), transpose if needed
    # Only transpose if we have (n_coefs x 1) instead of (1 x n_coefs)
    if (nrow(J) == length(coefs) && ncol(J) == 1) {
        J <- t(J)
    }

    out <- list(
        estimate = as.vector(result[["estimate"]]),
        std.error = as.vector(result[["std_error"]]),
        jacobian = J
    )

    # Add column names to jacobian
    if (!is.null(names(coefs)) && ncol(out$jacobian) == length(coefs)) {
        colnames(out$jacobian) <- names(coefs)
    }

    if (isTRUE(getOption("marginaleffects_autodiff_message", default = FALSE))) {
        message("\nJAX is fast!")
    }

    return(out)
}


#' Compute comparisons with JAX autodiff (estimates, SE, jacobian)
#' @return List with estimate, std.error, jacobian, or NULL if unsupported
#' @keywords internal
#' @noRd
jax_comparisons <- function(mfx, vcov_matrix, hi, lo, original, ...) {
    # Validate mfx object for autodiff support
    mfx <- sanitize_autodiff_mfx(mfx)
    if (is.null(mfx)) return(NULL)

    # Get autodiff args (already validated in sanitize_autodiff_mfx)
    autodiff_args <- get_autodiff_args(mfx@model, mfx)

    # Check comparison-specific requirements
    if (!is.character(mfx@comparison) || !mfx@comparison %in% c("difference", "ratio")) {
        comp_str <- if (is.character(mfx@comparison)) mfx@comparison else "custom function"
        autodiff_warning(sprintf("`comparison='%s'` (only 'difference' and 'ratio' supported)", comp_str))
        return(NULL)
    }

    # Ratio comparisons with by=TRUE compute ratio-then-average instead of average-then-ratio
    if (isTRUE(mfx@by) && mfx@comparison == "ratio") {
        autodiff_warning("`comparison='ratio'` with `by=TRUE` (averaging order differs from finite differences)")
        return(NULL)
    }

    mAD <- settings_get("mAD")

    # Get model matrices
    X_hi <- attr(hi, "marginaleffects_model_matrix")
    X_lo <- attr(lo, "marginaleffects_model_matrix")
    if (is.null(X_hi) || is.null(X_lo)) return(NULL)

    # Get coefficients
    coefs <- get_coef(mfx@model, ...)

    # Check for NA coefficients (e.g., from aliased terms)
    if (anyNA(coefs)) {
        autodiff_warning("models with NA coefficients (aliased terms)")
        return(NULL)
    }

    # Map comparison type
    comparison_type <- switch(mfx@comparison,
        difference = mAD$comparisons$ComparisonType$DIFFERENCE,
        ratio = mAD$comparisons$ComparisonType$RATIO
    )

    # Determine aggregation function
    if (isFALSE(mfx@by)) {
        fun_name <- "comparisons"
        groups <- NULL
        num_groups <- NULL
    } else {
        # Both by=TRUE and by=character use comparisons_byG with grouping
        # by=TRUE aggregates by term/contrast, by=character adds user-specified variables
        fun_name <- "comparisons_byG"
        # Prepare group indices (includes contrast, term, and optionally user variables)
        group_result <- jax_align_group_J("jacobian_byG", mfx, original, NULL, NULL, X_hi, X_lo)
        groups <- group_result$groups
        num_groups <- group_result$num_groups
        X_hi <- group_result$X_hi
        X_lo <- group_result$X_lo

        # If groups couldn't be created, fall back to finite differences
        if (is.null(groups) || is.null(num_groups)) {
            return(NULL)
        }
    }

    # Select autodiff function
    # e.g., mAD$linear$comparisons is a module containing comparisons(), comparisons_byT(), comparisons_byG()
    # The base module name (predictions or comparisons)
    base_module_name <- if (grepl("predictions", fun_name)) "predictions" else "comparisons"
    module <- mAD[[autodiff_args$model_type]][[base_module_name]]
    FUN <- module[[fun_name]]  # e.g., module$comparisons or module$comparisons_byT

    # Build arguments (without FUN)
    args <- list(
        beta = coefs,
        X_hi = X_hi,
        X_lo = X_lo,
        vcov = vcov_matrix,
        comparison_type = comparison_type,
        groups = groups,
        num_groups = num_groups,
        family_type = autodiff_args$family_type,
        link_type = autodiff_args$link_type
    )
    args <- Filter(function(x) !is.null(x), args)

    # Call Python function using eval_fun_with_numpy_arrays
    result <- do.call(eval_fun_with_numpy_arrays, c(list(FUN = FUN), args))

    # Convert to R
    J <- as.matrix(result[["jacobian"]])

    # Ensure jacobian is (n_comparisons x n_coefs), transpose if needed
    # Only transpose if we have (n_coefs x 1) instead of (1 x n_coefs)
    if (nrow(J) == length(coefs) && ncol(J) == 1) {
        J <- t(J)
    }

    out <- list(
        estimate = as.vector(result[["estimate"]]),
        std.error = as.vector(result[["std_error"]]),
        jacobian = J
    )

    if (!is.null(names(coefs)) && ncol(out$jacobian) == length(coefs)) {
        colnames(out$jacobian) <- names(coefs)
    }

    if (isTRUE(getOption("marginaleffects_autodiff_message", default = FALSE))) {
        message("\nJAX is fast!")
    }

    return(out)
}


#' EXPERIMENTAL -- Enable Automatic Differentiation with JAX
#'
#' This function enables or disables automatic differentiation using the JAX
#' package in Python, which can considerably speed up and increase the accuracy
#' of standard errors when a model includes many parameters.
#'
#' @param autodiff Logical flag. If `TRUE`, enables automatic differentiation
#'   with JAX. If `FALSE`, disables automatic differentiation and
#'   reverts to finite difference methods. If `NULL` (default), the function
#'   simply returns the current autodiff setting without changing it.
#' @param install Logical flag. If `TRUE`, installs the `marginaleffects`
#'   Python package via `reticulate::py_install()`. Default is `FALSE`. This is
#'   only necessary if you are self-managing a Python installation.
#'
#' @details
#'
#' Automatic differentiation needs to be enabled once per session.
#'
#' When `autodiff = TRUE`, this function:
#' - Imports the `marginaleffects.autodiff` Python module via [reticulate::import()]
#' - Sets the internal jacobian function to use JAX-based automatic differentiation
#' - Provides faster and more accurate gradient computation for supported models
#' - Falls back on the default finite difference method for unsupported models and calls.
#'
#' Currently supports:
#' - Model types: `lm`, `glm`, `ols`
#' - Functions: [predictions()] and [comparisons()], along with `avg_` and `plot_` variants.
#' - `type`: "response" or "link"
#' - `by`: `TRUE`, `FALSE`, or character vector.
#' - `comparison`: "difference" and "ratio"
#'
#' For unsupported models or options, the function automatically falls back to
#' the default finite difference method.
#'
#' # Python Configuration
#'
#' By default, no manual configuration of Python should be necessary. On most
#' machines, unless you have explicitly configured `reticulate`, reticulate
#' defaults to an automatically managed ephemeral virtual environment with all
#' Python requirements declared via `reticulate::py_require()`.
#'
#' If you prefer to use a manually managed Python installation, you can direct
#' `reticulate` and specify which Python executable or environment to use.
#' `reticulate` selects a Python installation using its [Order of
#' Discovery](https://rstudio.github.io/reticulate/articles/versions.html#order-of-discovery).
#' As a convenience `autodiff(install=TRUE)` will install the `marginaleffects` Python
#' package in a self-managed virtual environment.
#'
#' To specify an alternate Python version:
#' ```r
#' library(reticulate)
#' use_python("/usr/local/bin/python")
#' ```
#'
#' To use a virtual environment:
#' ```r
#' use_virtualenv("myenv")
#' ```
#'
#' These configuration commands should be called before calling `autodiff()`.
#'
#' @return When `autodiff` is `NULL`, returns `TRUE` if autodiff is enabled and
#'   `FALSE` otherwise. Otherwise called for side effects of enabling or
#'   disabling automatic differentiation or installing the Python package.
#'
#' @examples
#' \dontrun{
#' # Install the Python package (only needed once)
#' autodiff(install = TRUE)
#'
#' # Enable automatic differentiation
#' autodiff(TRUE)
#'
#' # Fit a model and compute marginal effects
#' mod <- glm(am ~ hp + wt, data = mtcars, family = binomial)
#' avg_comparisons(mod) # Will use JAX for faster computation
#'
#' # Disable automatic differentiation
#' autodiff(FALSE)
#' }
#'
#' @export
autodiff <- function(autodiff = NULL, install = FALSE) {
    checkmate::assert_flag(autodiff, null.ok = TRUE)
    checkmate::assert_flag(install)
    if (isTRUE(install) || isTRUE(autodiff)) {
        insight::check_if_installed("reticulate")
    }
    if (isTRUE(install)) {
        reticulate::py_install("marginaleffects")
    }
    if (is.null(autodiff)) {
        state <- isTRUE(settings_get("autodiff"))
        if (state) {
            cat("Autodiff is enabled.\n")
        } else {
            cat("Autodiff is disabled.\n")
        }
        return(invisible(state))
    }
    if (isFALSE(autodiff)) {
        settings_set("autodiff", FALSE)
    } else if (isTRUE(autodiff)) {
        mAD <- reticulate::import("marginaleffects.autodiff", delay_load = FALSE)
        settings_set("mAD", mAD)
        settings_set("autodiff", TRUE)
    }
}

bootstrap_fwb <- function(model, INF_FUN, ...) {

  # attached by `inferences()`
  conf_type <- attr(model, "inferences_conf_type")
  checkmate::assert_choice(conf_type, choices = c("perc", "norm", "basic", "bc", "bca"))

  # bootstrap using the original data and call
  modeldata <- get_modeldata(model, additional_variables = FALSE)

  # evaluate the {marginaleffects} call to get output without inferences()
  # use ... because arguments are not the same for different {marginaleffects} functions
  dots <- list(...)
  dots[["vcov"]] <- FALSE

  # avoid recursion
  attr(model, "inferences_method") <- NULL
  out <- do.call(INF_FUN, c(list(model), dots))

  # default confidence level may be implicit in original call, but we need numeric
  if (is.null(dots[["conf_level"]])) {
    conf_level <- 0.95
  } else {
    conf_level <- dots[["conf_level"]]
  }

  bootfun <- function(data, w) {
    # If model has weights, multiply them by random weights
    if (!is.null(w0 <- stats::weights(model))) w <- w * w0

    # Update the model's call and evaluate
    modboot <- stats::update(model, weights = w, evaluate = TRUE)

    # {marginaleffects} function needs to incorporate weights if
    # averaging. May be a problem if other weights supplied to
    # `wts` argument.
    dots[["wts"]] <- w
    args <- c(list(modboot), dots)
    out <- do.call(INF_FUN, args)$estimate
    return(out)
  }
  args <- list("data" = modeldata, "statistic" = bootfun)
  args <- c(args, attr(model, "inferences_dots"))
  args <- args[unique(names(args))]

  B <- do.call(fwb::fwb, args)

  # print.boot prints an ugly nested call
  B$call <- match.call()

  # Extract SEs and CIs
  fwb_summary <- summary(B, conf = conf_level,
                         ci.type = conf_type)

  out$std.error <- fwb_summary[, "Std. Error"]
  out$conf.low <- fwb_summary[, 3]
  out$conf.high <- fwb_summary[, 4]

  attr(out, "inferences") <- B
  attr(out, "posterior_draws") <- t(B$t)
  return(out)
}

github_issue <- function() {
    msg <- insight::format_string(
        "Please report this error along with a minimal reproducible example using
    publicly available data on Github:

    https://github.com/vincentarelbundock/marginaleffects/issues
    "
    )
    return(msg)
}

#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.glm <- function(model, coefs, ...) {
    model[["coefficients"]] <- sub_named_vector(
        model[["coefficients"]],
        coefs
    )
    model
}


#' @rdname set_coef
#' @export
set_coef.lm <- function(model, coefs, ...) {
    model[["coefficients"]] <- sub_named_vector(
        model[["coefficients"]],
        coefs
    )
    model
}


#' @rdname get_predict
#' @export
get_predict.lm <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    ...) {
    MM <- attr(newdata, "marginaleffects_model_matrix")
    beta <- get_coef(model)
    if (!isTRUE(checkmate::check_matrix(MM)) || ncol(MM) != length(beta)) {
        out <- get_predict.default(
            model = model,
            newdata = newdata,
            type = type,
            ...
        )
        return(out)
    }
    p <- model$rank
    p1 <- seq_len(p)
    piv <- if (p) qr(model)$pivot[p1]
    if (!all(seq_len(ncol(MM)) %in% piv)) {
        MM <- MM[, piv, drop = FALSE]
        beta <- beta[piv]
    }
    pred <- drop(MM %*% beta)

    # `pred` is a secret argument which re-uses the default get_predict to format a vector a data frame using correct `rowid`
    out <- get_predict.default(
        model = model,
        newdata = newdata,
        type = type,
        pred = pred,
        ...
    )
    return(out)
}


#' @rdname get_predict
#' @export
get_predict.glm <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    ...) {
    out <- NULL
    MM <- attr(newdata, "marginaleffects_model_matrix")
    if (isTRUE(checkmate::check_matrix(MM))) {
        if (isTRUE(checkmate::check_choice(type, c("link")))) {
            out <- get_predict.lm(model = model, newdata = newdata, ...)
        } else if (isTRUE(checkmate::check_choice(type, "response")) || is.null(type)) {
            out <- get_predict.lm(model = model, newdata = newdata, ...)
            out$estimate <- stats::family(model)$linkinv(out$estimate)
        }
    }

    if (is.null(out)) {
        out <- get_predict.default(
            model = model,
            newdata = newdata,
            type = type,
            ...
        )
    }

    return(out)
}


#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.nls <- function(model, coefs, ...) {
    out <- model
    out$m$setPars(coefs)
    return(out)
}


#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.nls <- function(model, ...) {
    model$m$getPars()
}


#' @keywords internal
#' @export
get_autodiff_args.lm <- function(model, mfx) {
    # no inheritance! Important to avoid breaking other models
    if (!class(model)[1] %in% c("lm", "ivreg")) {
        return(NULL)
    }

    if (!is.null(model$offset)) {
        autodiff_warning("models with offsets")
        return(NULL)
    }

    if ("weights" %in% names(model$call)) {
        autodiff_warning("models with weights")
        return(NULL)
    }

    # Check type support
    if (!mfx@type %in% c("response", "link", "invlink(link)")) {
        autodiff_warning(sprintf("`type='%s'`", mfx@type))
        return(NULL)
    }

    # If all checks pass, return supported arguments
    out <- list(model_type = "linear")
    return(out)
}


#' @keywords internal
#' @export
get_autodiff_args.glm <- function(model, mfx) {
    # no inheritance! Important to avoid breaking other models
    if (!class(model)[1] == "glm") {
        return(NULL)
    }

    if (!is.null(model$offset)) {
        autodiff_warning("models with offsets")
        return(NULL)
    }

    if ("weights" %in% names(model$call)) {
        autodiff_warning("models with weights")
        return(NULL)
    }

    # Check type support
    if (!mfx@type %in% c("response", "link", "invlink(link)")) {
        autodiff_warning(sprintf("`type='%s'`", mfx@type))
        return(NULL)
    }

    # Check comparison type for comparisons function
    if (mfx@calling_function == "comparisons" && (!is.character(mfx@comparison) || !mfx@comparison %in% c("difference", "ratio"))) {
        comp_str <- if (is.character(mfx@comparison)) mfx@comparison else "custom function"
        autodiff_warning(sprintf("`comparison='%s'` (only 'difference' and 'ratio' supported)", comp_str))
        return(NULL)
    }

    # For GLM, check if family/link combination is supported
    family_name <- model$family$family
    link_name <- model$family$link

    # Supported family types
    supported_families <- c("gaussian", "binomial", "poisson", "Gamma")
    if (!family_name %in% supported_families) {
        autodiff_warning("unsupported GLM family/link combinations")
        return(NULL)
    }

    # Supported link types
    supported_links <- c("identity", "log", "logit", "probit", "inverse", "sqrt", "cloglog")
    if (!link_name %in% supported_links) {
        autodiff_warning("unsupported GLM family/link combinations")
        return(NULL)
    }

    # For link/invlink(link) type with GLM, use linear model approach
    model_type <- if (mfx@type %in% c("link", "invlink(link)")) "linear" else "glm"

    # Convert to Python enum values for GLM models
    family_type <- NULL
    link_type <- NULL
    if (model_type == "glm") {
        # Import Family and Link enums from Python
        mAD <- settings_get("mAD")
        Family <- mAD$glm$families$Family
        Link <- mAD$glm$families$Link

        # Family types using Python enum
        family_type <- switch(family_name,
            "gaussian" = Family$GAUSSIAN,
            "binomial" = Family$BINOMIAL,
            "poisson" = Family$POISSON,
            "Gamma" = Family$GAMMA,
            NULL
        )

        # Link types using Python enum
        link_type <- switch(link_name,
            "identity" = Link$IDENTITY,
            "log" = Link$LOG,
            "logit" = Link$LOGIT,
            "probit" = Link$PROBIT,
            "inverse" = Link$INVERSE,
            "sqrt" = Link$SQRT,
            "cloglog" = Link$CLOGLOG,
            NULL
        )
    }

    # If all checks pass, return supported arguments
    out <- list(
        model_type = model_type,
        family_type = family_type,
        link_type = link_type

    )
    return(out)
}

add_rowid <- function(p, newdata) {
    if (!"rowid" %in% colnames(newdata)) {
        return(p)
    }
    if (nrow(p) == nrow(newdata)) {
        p$rowid <- newdata$rowid
    } else {
        p$rowid <- rep(newdata$rowid, nrow(p) / nrow(newdata))
    }
    return(p)
}

# License: https://creativecommons.org/licenses/by-sa/3.0/
# Source: https://stackoverflow.com/a/24569739/342331
myTryCatch <- function(expr) {
    warn <- err <- NULL
    value <- withCallingHandlers(
        tryCatch(expr, error = function(e) {
            err <<- e
            NULL
        }),
        warning = function(w) {
            warn <<- w
            invokeRestart("muffleWarning")
        }
    )
    list(value = value, warning = warn, error = err)
}

#' @rdname get_predict
#' @export
get_predict.fixest <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    mfx = NULL,
    ...) {
    insight::check_if_installed("fixest")

    if (is.null(type)) {
        calling_function <- if (!is.null(mfx)) mfx@calling_function else "predictions"
        type <- sanitize_type(
            model = model,
            type = type,
            calling_function = calling_function
        )
    }

    dots <- list(...)

    # some predict methods raise warnings on unused arguments
    unused <- c(
        "normalize_dydx",
        "step_size",
        "numDeriv_method",
        "conf.int",
        "internal_call"
    )
    dots <- dots[setdiff(names(dots), unused)]

    # fixest is super slow when using do call because of some `deparse()` call
    # issue #531: we don't want to waste time computing intervals or risk having
    # them as leftover columns in contrast computations
    pred <- try(
        stats::predict(
            object = model,
            newdata = newdata,
            type = type
        ),
        silent = TRUE
    )

    if (inherits(pred, "try-error")) {
        return(pred)
    }

    out <- data.table(estimate = as.numeric(pred))
    out <- add_rowid(out, newdata)

    return(out)
}


#' @rdname sanitize_model_specific
sanitize_model_specific.fixest <- function(model, vcov = TRUE, calling_function = "predictions", ...) {
    # issue #1487 is only a problem for standard errors
    if (isFALSE(vcov)) {
        return(model)
    }
    if (!isTRUE(getOption("marginaleffects_safe", default = TRUE))) {
        return(model)
    }

    msg <- paste(
        "For this model type, `marginaleffects` cannot take into account the uncertainty in fixed-effects parameters.",
        "This is especially consequential for predictions and their uncertainty intervals.",
        "Set `vcov=FALSE` to compute estimates without standard errors."
    )

    if (!is.null(model[["fixef_vars"]])) {
        # issue #1487: fixed-effects always matter for predictions
        if (identical(calling_function, "predictions")) {
            stop_sprintf(msg)
        }

        # issue #1487: fixed-effects matter for slopes and contrasts, except for linear models
        if (!is.null(model$family)) {
            if (identical(calling_function, "hypotheses")) {
                if (isTRUE(getOption("marginaleffects_safe", default = TRUE))) {
                    warn_sprintf(msg)
                }
            } else {
                stop_sprintf(msg)
            }
        }
    }

    return(model)
}

get_jacobian <- function(func, x, numderiv) {
    numDeriv_options <- getOption("marginaleffects_numDeriv", default = NULL)
    if (is.null(numDeriv_options)) {
        method <- numderiv[[1]]
        numderiv[[1]] <- NULL
        numderiv[["func"]] <- func
        numderiv[["x"]] <- x
        if (identical(method, "richardson")) {
            df <- do.call(get_jacobian_richardson, numderiv)
        } else if (identical(method, "fdforward")) {
            df <- do.call(get_jacobian_fdforward, numderiv)
        } else if (identical(method, "fdcenter")) {
            df <- do.call(get_jacobian_fdcenter, numderiv)
        }
    } else {
        insight::check_if_installed("numDeriv")
        numDeriv_options[["func"]] <- func
        numDeriv_options[["x"]] <- x
        ndFUN <- get("jacobian", asNamespace("numDeriv"))
        df <- do.call(ndFUN, numDeriv_options)
    }
    return(df)
}


get_jacobian_fdforward <- function(func, x, eps = NULL) {
    # old version. probably not optimal. Keep for posterity.
    # h <- max(1e-8, 1e-4 * min(abs(x), na.rm = TRUE))
    baseline <- func(x)

    # we pre-chunk because future does not cache datasets in nodes, which means we 
    # must pass every worker the full data and model for every future.
    inner_loop <- function(chunk, ...) {
        out <- list()
        for (i in chunk) {
            if (is.null(eps)) {
                h <- max(abs(x[i]) * sqrt(.Machine$double.eps), 1e-10)
            } else {
                h <- eps
            }
            dx <- x
            dx[i] <- dx[i] + h
            out <- c(out, list((func(dx) - baseline) / h))
        }
        out <- do.call(cbind, out)
        return(out)
    }

    if (isTRUE(getOption("marginaleffects_parallel", default = FALSE))) {
        insight::check_if_installed("future.apply")
        insight::check_if_installed("future")
        insight::check_if_installed("parallel")
        chunks <- parallel::splitIndices(length(x), future::nbrOfWorkers())
        df <- future.apply::future_lapply(
            chunks,
            inner_loop, 
            future.seed = TRUE)
        df <- do.call("cbind", df)
    } else {
        chunks <- seq_along(x)
        df <- lapply(chunks, inner_loop)
        df <- do.call("cbind", df)
    }

    return(df)
}


get_jacobian_fdcenter <- function(func, x, eps = NULL) {
    baseline <- func(x)
    # we pre-chunk because future does not cache datasets in nodes, which means we 
    # must pass every worker the full data and model for every future.
    inner_loop <- function(chunk, ...) {
        out <- list()
        for (i in chunk) {
            if (is.null(eps)) {
                h <- max(abs(x[i]) * sqrt(.Machine$double.eps), 1e-10)
            } else {
                h <- eps
            }
            dx_hi <- dx_lo <- x
            dx_hi[i] <- dx_hi[i] + h / 2
            dx_lo[i] <- dx_lo[i] - h / 2
            out <- c(out, list((func(dx_hi) - func(dx_lo)) / h))
        }
        out <- do.call(cbind, out)
        return(out)
    }

    if (isTRUE(getOption("marginaleffects_parallel", default = FALSE))) {
        insight::check_if_installed("future.apply")
        insight::check_if_installed("future")
        insight::check_if_installed("parallel")
        chunks <- parallel::splitIndices(length(x), future::nbrOfWorkers())
        df <- future.apply::future_lapply(
            chunks,
            inner_loop, 
            future.seed = TRUE)
        df <- do.call("cbind", df)
    } else {
        chunks <- seq_along(x)
        df <- lapply(chunks, inner_loop)
        df <- do.call("cbind", df)
    }
    return(df)
}


# Code adapted from the `numDeriv` package by Paul Gilbert and Ravi Varadhan
# GPL-3: https://cran.r-project.org/package=numDeriv
get_jacobian_richardson <- function(
    func,
    x,
    eps = 1e-4,
    d = 1e-4,
    zero_tol = sqrt(.Machine$double.eps / 7e-7),
    side = NULL,
    r = 4,
    v = 2,
    ...) {

    n <- length(x)
    f <- func(x, ...)

    a <- array(NA, c(length(f), r, n))
    h <- abs(d * x) + eps * (abs(x) < zero_tol)
    pna <- (side == 1) & !is.na(side) # double these on plus side
    mna <- (side == -1) & !is.na(side) # double these on minus side
    for (k in 1:r) { # successively reduce h
        ph <- mh <- h
        ph[pna] <- 2 * ph[pna]
        ph[mna] <- 0
        mh[mna] <- 2 * mh[mna]
        mh[pna] <- 0
        for (i in 1:n) {
            a[, k, i] <- (func(x + ph * (i == seq(n)), ...) -
                func(x - mh * (i == seq(n)), ...)) / (2 * h[i])
            # if((k != 1)) a[,(abs(a[,(k-1),i]) < 1e-20)] <- 0 #some func are unstable near zero
        }
        h <- h / v # Reduced h by 1/v.
    }
    for (m in 1:(r - 1)) {
        a <- (a[, 2:(r + 1 - m), , drop = FALSE] * (4^m) - a[, 1:(r - m), , drop = FALSE]) / (4^m - 1)
    }
    # drop second dim of a, which is now 1 (but not other dim's even if they are 1
    return(array(a, dim(a)[c(1, 3)]))
}

#' @rdname set_coef
#' @export
set_coef.scam <- function(model, coefs, ...) {
    # in basic model classes coefficients are named vector
    model[["coefficients.t"]][names(coefs)] <- coefs
    model
}


#' @rdname get_coef
#' @export
get_coef.scam <- function(model, ...) {
    model$coefficients.t
}


#' @rdname get_vcov
#' @export
get_vcov.scam <- function(model,
                          vcov = NULL,
                          ...) {

    if (isTRUE(checkmate::check_matrix(vcov))) {
        return(vcov)
    }

    # email from developer Natalya Pya
    # "one of the elements of the returned 'scam' object is 'Vp.t' which is an
    # estimated covariance matrix for the reparametrized parameters,
    # 'model$coefficients.t'."

    if (!is.null(vcov) && !is.logical(vcov)) {
        stop("The `vcov` argument is not supported for models of class `scam`.", .call = FALSE)
    }

    V <- model$Vp.t
    b <- model$coefficients.t
    if (length(b) != nrow(V)) {
        stop("The size of the variance-covariance matrix does not match the length of the coefficients vector.", call. = FALSE)
    }
    colnames(V) <- row.names(V) <- names(b)
    return(V)
}

inferences_boot <- function(x, R = 1000, conf_level = 0.95, conf_type = "perc", estimator = NULL, data_train = NULL, mfx = NULL, ...) {
    insight::check_if_installed("boot")

    out <- x

    if (!is.null(estimator)) {
        bootfun <- function(data, indices) {
            d <- data[indices, , drop = FALSE]
            boot_mfx <- estimator(d)
            return(boot_mfx$estimate)
        }
    } else {
        bootfun <- function(data, indices) {
            d <- data[indices, , drop = FALSE]
            boot_mfx <- refit(x, data = d, vcov = FALSE)
            return(boot_mfx$estimate)
        }
    }

    args <- list("data" = data_train, "statistic" = bootfun, R = R)
    args <- c(args, list(...))
    B <- do.call(boot::boot, args)

    # print.boot prints an ugly nested call
    t <- matrix(B$t, nrow = nrow(B$t))
    op <- cbind(
        colMeans(t, na.rm = TRUE),
        sqrt(apply(t, 2L, function(t.st) stats::var(t.st[!is.na(t.st)])))
    )

    # extract from weird boot.ci() list (inspired from `broom::tidy.broom` under MIT)
    ci_list <- suppressWarnings(lapply(
        seq_along(B$t0),
        boot::boot.ci,
        boot.out = B,
        conf = conf_level,
        type = conf_type
    )) # extreme order statistics
    pos <- pmatch(conf_type, names(ci_list[[1]]))
    if (conf_type == "norm") {
        cols <- 2:3
    } else {
        cols <- 4:5
    }
    ci <- lapply(ci_list, function(x) x[[pos]])
    ci <- do.call("rbind", ci)[, cols]

    # add CI to original {marginaleffects} call
    if (is.matrix(ci)) {
        out$conf.low <- ci[, 1]
        out$conf.high <- ci[, 2]
    } else {
        out$conf.low <- ci[1]
        out$conf.high <- ci[2]
    }

    cols <- setdiff(names(out), c("p.value", "std.error", "statistic", "s.value", "df"))
    out <- out[, cols, drop = FALSE]

    mfx@draws <- t(B$t)
    mfx@inferences <- B
    attr(out, "marginaleffects") <- mfx

    return(out)
}

#' Plot Conditional or Marginal Predictions
#'
#' @description
#' Plot predictions on the y-axis against values of one or more predictors (x-axis, colors/shapes, and facets).
#'
#' The `by` argument is used to plot marginal predictions, that is, predictions made on the original data, but averaged by subgroups. This is analogous to using the `by` argument in the `predictions()` function.
#'
#' The `condition` argument is used to plot conditional predictions, that is, predictions made on a user-specified grid. This is analogous to using the `newdata` argument and `datagrid()` function in a `predictions()` call. All variables whose values are not specified explicitly are treated as usual by `datagrid()`, that is, they are held at their mean or mode (or rounded mean for integers). This includes grouping variables in mixed-effects models, so analysts who fit such models may want to specify the groups of interest using the `condition` argument, or supply model-specific arguments to compute population-level estimates. See details below.
#'
#' See the "Plots" vignette and website for tutorials and information on how to customize plots:
#'
#' * https://marginaleffects.com/bonus/plot.html
#' * https://marginaleffects.com
#'
#' @param condition Conditional predictions.
#' + Character vector (max length 4): Names of the predictors to display.
#' + Named list (max length 4): List names correspond to predictors. List elements can be:
#'   - Numeric vector
#'   - Function which returns a numeric vector or a set of unique categorical values
#'   - Shortcut strings for common reference values: "minmax", "quartile", "threenum"
#' + 1: x-axis. 2: color/shape. 3: facet (wrap if no fourth variable, otherwise cols of grid). 4: facet (rows of grid).
#' + Numeric variables in positions 2 and 3 are summarized by Tukey's five numbers `?stats::fivenum`
#' @param by Marginal predictions
#' + Character vector (max length 3): Names of the categorical predictors to marginalize across.
#' + 1: x-axis. 2: color. 3: facets.
#' @param newdata When `newdata` is `NULL`, the grid is determined by the `condition` argument. When `newdata` is not `NULL`, the argument behaves in the same way as in the `predictions()` function. Note that the `condition` argument builds its own grid, so the `newdata` argument is ignored if the `condition` argument is supplied.
#' @param points Number between 0 and 1 which controls the transparency of raw data points. 0 (default) does not display any points. Warning: The points displayed are raw data, so the resulting plot is not a "partial residual plot."
#' @param draw `TRUE` returns a `ggplot2` plot. `FALSE` returns a `data.frame` of the underlying data.
#' @inheritParams plot_slopes
#' @inheritParams predictions
#' @template model_specific_arguments
#' @template type
#' @return A `ggplot2` object or data frame (if `draw=FALSE`)
#' @export
#' @examplesIf interactive() || isTRUE(Sys.getenv("R_DOC_BUILD") == "true")
#' mod <- lm(mpg ~ hp + wt, data = mtcars)
#' plot_predictions(mod, condition = "wt")
#'
#' mod <- lm(mpg ~ hp * wt * am, data = mtcars)
#' plot_predictions(mod, condition = c("hp", "wt"))
#'
#' plot_predictions(mod, condition = list("hp", wt = "threenum"))
#'
#' plot_predictions(mod, condition = list("hp", wt = range))
#'
#' # marginal predictions
#' mod <- lm(mpg ~ hp * am, data = mtcars)
#' plot_predictions(mod, by = "am")
#'
#' # marginal predictions on a counterfactual grid
#' plot_predictions(mod,
#'     by = "am",
#'     newdata = datagrid(am = 0:1, grid_type = "counterfactual")
#' )
#'
plot_predictions <- function(
  model,
  condition = NULL,
  by = NULL,
  newdata = NULL,
  type = NULL,
  vcov = NULL,
  conf_level = 0.95,
  wts = FALSE,
  transform = NULL,
  points = 0,
  rug = FALSE,
  gray = getOption("marginaleffects_plot_gray", default = FALSE),
  draw = TRUE,
  ...
) {
    checkmate::assert_number(points, lower = 0, upper = 1)

    if (inherits(model, c("mira", "amest"))) {
        msg <- "This function does not support multiple imputation. Call `predictions()` or `avg_predictions()` instead. These functions return easy to plot data frames."
        stop_sprintf(msg)
    }


    # init
    call <- construct_call(model, "comparisons")
    model <- sanitize_model(model, call = call, newdata = newdata, wts = wts, vcov = vcov, by = by, ...)
    mfx <- new_marginaleffects_internal(
        call = call,
        model = model
    )

    if (inherits(mfx@model, "mira") && is.null(newdata)) {
        msg <- "Please supply a data frame to the `newdata` argument explicitly."
        stop_sprintf(msg)
    }

    # order of the first few paragraphs is important
    scall <- rlang::enquo(newdata)
    newdata <- sanitize_newdata_call(scall, newdata, mfx = mfx, by = by)
    if (!isFALSE(wts) && is.null(by)) {
        stop_sprintf("The `wts` argument requires a `by` argument.")
    }
    checkmate::assert_character(by, null.ok = TRUE)

    # sanity check
    checkmate::assert_character(
        by,
        null.ok = TRUE,
        max.len = 4,
        min.len = 1,
        names = "unnamed"
    )
    if ((!is.null(condition) && !is.null(by)) || (is.null(condition) && is.null(by))) {
        msg <- "One of the `condition` and `by` arguments must be supplied, but not both."
        stop_sprintf(msg)
    }

    # mlr3 and tidymodels
    if (is.null(mfx@modeldata) || nrow(mfx@modeldata) == 0) {
        mfx@modeldata <- newdata
    }

    # conditional
    if (!is.null(condition)) {
        condition <- sanitize_condition(
            mfx = mfx,
            condition = condition,
            variables = NULL
        )
        v_x <- condition$condition1
        v_color <- condition$condition2
        v_facet_1 <- condition$condition3
        v_facet_2 <- condition$condition4
        datplot <- predictions(
            mfx@model,
            newdata = condition$newdata,
            type = type,
            vcov = vcov,
            conf_level = conf_level,
            transform = transform,
            modeldata = mfx@modeldata,
            wts = wts,
            ...
        )
    }

    # marginal
    if (!isFALSE(by) && !is.null(by)) {
        # switched from NULL above
        condition <- NULL

        newdata <- sanitize_newdata(
            mfx = mfx,
            newdata = newdata,
            by = by,
            wts = wts
        )

        # tidymodels & mlr3
        if (is.null(mfx@modeldata)) {
            mfx@modeldata <- newdata
        }

        datplot <- predictions(
            mfx@model,
            by = by,
            type = type,
            vcov = vcov,
            conf_level = conf_level,
            wts = wts,
            transform = transform,
            newdata = newdata,
            modeldata = mfx@modeldata,
            ...
        )
        v_x <- by[[1]]
        v_color <- hush(by[[2]])
        v_facet_1 <- hush(by[[3]])
        v_facet_2 <- hush(by[[4]])
    }

    datplot <- plot_preprocess(
        datplot,
        v_x = v_x,
        v_color = v_color,
        v_facet_1 = v_facet_1,
        v_facet_2 = v_facet_2,
        condition = condition,
        mfx = mfx
    )

    dv <- mfx@variable_names_response
    dv_label <- dv
    if (!is.null(mfx@modeldata) && length(dv) == 1 && dv %in% colnames(mfx@modeldata)) {
        dv_label <- attr(mfx@modeldata[[dv]], "label")
    } else {
        dv_label <- "Prediction"
    }
    attr(datplot$estimate, "label") <- dv_label

    # return immediately if the user doesn't want a plot
    if (isFALSE(draw)) {
        out <- as.data.frame(datplot)
        return(out)
    }

    # ggplot2
    insight::check_if_installed("ggplot2", minimum_version = "4.0.0")
    p <- plot_build(
        datplot,
        v_x = v_x,
        v_color = v_color,
        v_facet_1 = v_facet_1,
        v_facet_2 = v_facet_2,
        points = points,
        dv = dv,
        rug = rug,
        gray = gray,
        mfx = mfx
    )

    return(p)
}

#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.crch <- function(model, coefs, ...) {
    # coefs are split between location coefs (which can be length 0) and scale coefs
    # (which must be length > 0 and always start with "(scale)_" due to get_parameters(),
    # to match with get_varcov(., component = "all") output). In crch object, these
    # are stored as two elements in a list, with scale coefs lacking the "(scale)_"
    # prefix, so we remove it.
    location_coefs <- coefs[!startsWith(names(coefs), "(scale)_")]
    scale_coefs <- coefs[startsWith(names(coefs), "(scale)_")]
    names(scale_coefs) <- sub("(scale)_", "", names(scale_coefs), fixed = TRUE)

    if (length(location_coefs) > 0) {
        model[["coefficients"]]$location[names(location_coefs)] <- location_coefs
    }
    model[["coefficients"]]$scale[names(scale_coefs)] <- scale_coefs

    model
}


#' @include get_predict.R
#' @rdname get_predict
#' @export
get_predict.crch <- function(model, newdata = NULL, type = "location", ...) {
    pred <- stats::predict(model, newdata = newdata, type = type)
    sanity_predict_vector(
        pred = pred,
        model = model,
        newdata = newdata,
        type = type
    )
    sanity_predict_numeric(
        pred = pred,
        model = model,
        newdata = newdata,
        type = type
    )
    out <- data.table(estimate = pred)
    out <- add_rowid(out, newdata)
    return(out)
}


#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.hxlr <- function(model, coefs, ...) {
    # in crch::hxlr, coefficients are held in:
    # model$coefficients$intercept
    # model$coefficients$location
    # model$coefficients$scale
    #
    # note: there are no prefixes in coef() output, so coefs may have same name
    out <- model

    idx_int <- length(model$coefficients$intercept)
    idx_loc <- length(model$coefficients$location)

    out$coefficients$intercept[] <- coefs[1:idx_int]
    out$coefficients$location[] <- coefs[idx_int + 1:idx_loc]
    out$coefficients$scale[] <- coefs[(idx_int + idx_loc + 1):length(coefs)]

    return(out)
}


#' @export
#' @noRd
get_predict.hxlr <- get_predict.crch

# unexported functions for use in tests

modelarchive_model <- function(name) {
    tmp <- tempfile()
    url <- paste0("https://raw.github.com/vincentarelbundock/modelarchive/main/data/", name, ".rds")
    try(utils::download.file(url, tmp, quiet = TRUE), silent = TRUE)
    out <- try(readRDS(tmp), silent = TRUE)
    return(out)
}

modelarchive_data <- function(name) {
    dat <- sprintf(
        "https://raw.githubusercontent.com/vincentarelbundock/modelarchive/main/data-raw/%s.csv",
        name)
    out <- utils::read.csv(dat)
    return(out)
}

.onAttach <- function(lib, pkg) {
    msg <- c(
        "Please cite the software developers who make your work possible.",
        'One package:             citation("package_name")',
        "All project packages:    softbib::softbib()"
    )
    msg <- paste(msg, collapse = "\n")

    # once every 24 hours
    last_time <- settings_get("startup_message_time")
    if (inherits(last_time, "POSIXct")) {
        flag_time <- difftime(Sys.time(), last_time, units = "sec") >= 24 * 60 * 60
    } else {
        flag_time <- TRUE
    }

    flag_option <- isTRUE(getOption("marginaleffects_startup_message", TRUE))

    if (interactive() && flag_time && flag_option) {
        packageStartupMessage(msg)
        settings_set("startup_message_time", Sys.time(), persistent = TRUE)
    }

    invisible()
}

.onLoad <- function(lib, pkg) {
    backports::import(pkg)
    if (isNamespaceLoaded("reticulate")) {
        py_require_marginaleffects()
    } else {
        setHook(packageEvent("reticulate", "onLoad"), py_require_marginaleffects)
    }
}

py_require_marginaleffects <- function(...) {
    reticulate::py_require("marginaleffects")
    reticulate::py_require("jax")
}

#' @rdname get_predict
#' @export
get_predict.clm <- function(model,
                            newdata = insight::get_data(model),
                            type = "prob",
                            ...) {

    # `predict.clm()` only makes predictions for the observed response group of
    # each observation in `newdata`. When we remove the response from
    # `newdata`, `predict.clm()` makes predictions for all levels, which is
    # what we want.
    resp <- insight::find_response(model)

    # otherwise `predict.clm` does not see some columns (mystery)
    data.table::setDF(newdata)

    newdata <- newdata[, setdiff(colnames(newdata), resp), drop = FALSE]

    pred <- stats::predict(model, newdata = newdata, type = type)
    
    contenders <- c("fit", "eta1", "eta2", "cprob1", "cprob2")
    tmp <- NULL
    for (con in contenders) {
        if (is.null(tmp) && con %in% names(pred)) {
            tmp <- pred[[con]]
        }
    }
    pred <- tmp

    out <- data.frame(
        group = rep(colnames(pred), each = nrow(pred)),
        estimate = c(pred))
    out$group <- group_to_factor(out$group, model)

    # often an internal call
    if ("rowid" %in% colnames(newdata)) {
        out$rowid <- rep(newdata$rowid, times = ncol(pred))
    } else {
        out$rowid <- rep(seq_len(nrow(pred)), times = ncol(pred))
    }

    return(out)
}

#' @include get_group_names.R
#' @rdname get_group_names
#' @export
get_group_names.clm <- get_group_names.polr


#' @include sanity_model.R
#' @rdname sanitize_model_specific
#' @keywords internal
sanitize_model_specific.clm <- function(model, ...) {

    # Corner case: The `predict.clm` method does not make predictions when the
    # response was transformed to a factor in the formula AND the response is
    # missing from `newdata`.
    lhs <- names(attr(stats::terms(model), "dataClasses"))[1]
    if (isTRUE(grepl("^factor\\(", lhs))) {
        stop("The response variable should not be transformed to a factor in the formula. Please convert the variable to factor before fitting your model.",
             call. = FALSE)
    }
    return(model)
}

#' @include get_predict.R
#' @rdname get_predict
#' @keywords internal
#' @export
get_predict.bife <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    ...) {
    pred <- stats::predict(model, X_new = newdata, type = type, ...)

    sanity_predict_vector(
        pred = pred,
        model = model,
        newdata = newdata,
        type = type
    )
    sanity_predict_numeric(
        pred = pred,
        model = model,
        newdata = newdata,
        type = type
    )

    out <- data.table(estimate = pred)
    out <- add_rowid(out, newdata)

    return(out)
}

mean_i <- function(x) as.integer(round(mean(x, na.rm = TRUE)))
mean_na <- function(x) mean(x, na.rm = TRUE)
unique_s <- function(x) sort(unique(x))


#' Data grids
#'
#' @description
#' Generate a data grid of user-specified values for use in the `newdata` argument of the `predictions()`, `comparisons()`, and `slopes()` functions. This is useful to define where in the predictor space we want to evaluate the quantities of interest. Ex: the predicted outcome or slope for a 37 year old college graduate.
#'
#' @param ... named arguments with vectors of values or functions for user-specified variables.
#' + Functions are applied to the variable in the `model` dataset or `newdata`, and must return a vector of the appropriate type.
#' + Character vectors are automatically transformed to factors if necessary.
#' + The output will include all combinations of these variables (see Examples below.)
#' @param model Model object
#' @param newdata data.frame (one and only one of the `model` and `newdata` arguments can be used.)
#' @param by character vector with grouping variables within which `FUN_*` functions are applied to create "sub-grids" with unspecified variables.
#' @param response Logical should the response variable be included in the grid, even if it is not specified explicitly.
#' @param FUN a function to be applied to all variables in the grid. This is useful when you want to apply the same function to all variables, such as `mean` or `median`. If you specify `FUN`, it will override the `grid_type` defaults, but not other `FUN_*` arguments below.
#' @param FUN_character the function to be applied to character variables.
#' @param FUN_factor the function to be applied to factor variables. This only applies if the variable in the original data is a factor. For variables converted to factor in a model-fitting formula, for example, `FUN_character` is used.
#' @param FUN_logical the function to be applied to logical variables.
#' @param FUN_integer the function to be applied to integer-ish variables (including columns without decimal places).
#' @param FUN_binary the function to be applied to binary variables.
#' @param FUN_numeric the function to be applied to numeric variables.
#' @param FUN_other the function to be applied to other variable types.
#' @param grid_type character. Determines the functions to apply to each variable. The defaults can be overridden by defining individual variables explicitly in `...`, or by supplying a function to one of the `FUN_*` arguments.
#'   * "mean_or_mode": Character, factor, logical, and binary variables are set to their modes. Numeric, integer, and other variables are set to their means.
#'   * "balanced": Each unique level of character, factor, logical, and binary variables are preserved. Numeric, integer, and other variables are set to their means. Warning: When there are many variables and many levels per variable, a balanced grid can be very large. In those cases, it is better to use `grid_type="mean_or_mode"` and to specify the unique levels of a subset of named variables explicitly.
#'   * "dataframe": Similar to "mean_or_mode" but creates a data frame by binding columns element-wise rather than taking the cross-product. All explicitly specified vectors must have the same length (or length 1), and the result has as many rows as the longest vector. This differs from other grid types which use `expand.grid()` or `data.table::CJ()` to create all combinations.
#'   * "counterfactual": the entire dataset is duplicated for each combination of the variable values specified in `...`. Variables not explicitly supplied to `datagrid()` are set to their observed values in the original dataset.
#' @details
#' If `datagrid` is used in a `predictions()`, `comparisons()`, or `slopes()` call as the
#' `newdata` argument, the model is automatically inserted in the `model` argument of `datagrid()`
#' call, and users do not need to specify either the `model` or `newdata` arguments. The same behavior will occur when the value supplied to `newdata=` is a function call which starts with "datagrid". This is intended to allow users to create convenience shortcuts like:
#'
#' **Warning about hierarchical grouping variables:** When using the default `grid_type = "mean_or_mode"` with hierarchical models (such as mixed models with nested grouping factors), `datagrid()` may create invalid combinations of grouping variables. For example, if you have students nested within schools, or countries nested within regions, the modal values of each grouping variable may not correspond to valid nested relationships in the data. This can cause prediction errors. To avoid this issue, explicitly specify valid combinations of hierarchical grouping variables in the `datagrid()` call, or use `grid_type = "counterfactual"` to preserve the original data structure.
#'
#' \preformatted{
#' mod <- lm(mpg ~ am + vs + factor(cyl) + hp, mtcars)
#' datagrid_bal <- function(...) datagrid(..., grid_type = "balanced")
#' predictions(model, newdata = datagrid_bal(cyl = 4))
#' }
#'
#'
#' If users supply a model, the data used to fit that model is retrieved using
#' the `insight::get_data` function.
#' @return
#' A `data.frame` in which each row corresponds to one combination of the named
#' predictors supplied by the user via the `...` dots. Variables which are not
#' explicitly defined are held at their mean or mode.
#' @export
#' @examples
#' # The output only has 2 rows, and all the variables except `hp` are at their
#' # mean or mode.
#' datagrid(newdata = mtcars, hp = c(100, 110))
#'
#' # We get the same result by feeding a model instead of a data.frame
#' mod <- lm(mpg ~ hp, mtcars)
#' datagrid(model = mod, hp = c(100, 110))
#'
#' # Use in `marginaleffects` to compute "Typical Marginal Effects". When used
#' # in `slopes()` or `predictions()` we do not need to specify the
#' # `model` or `newdata` arguments.
#' slopes(mod, newdata = datagrid(hp = c(100, 110)))
#'
#' # datagrid accepts functions
#' datagrid(hp = range, cyl = unique, newdata = mtcars)
#' comparisons(mod, newdata = datagrid(hp = fivenum))
#'
#' # The full dataset is duplicated with each observation given counterfactual
#' # values of 100 and 110 for the `hp` variable. The original `mtcars` includes
#' # 32 rows, so the resulting dataset includes 64 rows.
#' dg <- datagrid(newdata = mtcars, hp = c(100, 110), grid_type = "counterfactual")
#' nrow(dg)
#'
#' # We get the same result by feeding a model instead of a data.frame
#' mod <- lm(mpg ~ hp, mtcars)
#' dg <- datagrid(model = mod, hp = c(100, 110), grid_type = "counterfactual")
#' nrow(dg)
#'
#' # Use `by` to hold variables at group-specific values
#' mod2 <- lm(mpg ~ hp + cyl, mtcars)
#' datagrid(model = mod2, hp = mean, by = "cyl")
#'
#' # Use `FUN` to apply function to all variables
#' datagrid(model = mod2, FUN = median)
#'
#' # Use `grid_type="dataframe"` for column-wise binding instead of cross-product
#' datagrid(model = mod2, hp = c(100, 200), cyl = c(4, 6), grid_type = "dataframe")
datagrid <- function(
    ...,
    model = NULL,
    newdata = NULL,
    by = NULL,
    grid_type = "mean_or_mode",
    response = FALSE,
    FUN = NULL,
    FUN_character = NULL,
    FUN_factor = NULL,
    FUN_logical = NULL,
    FUN_numeric = NULL,
    FUN_integer = NULL,
    FUN_binary = NULL,
    FUN_other = NULL) {
    checkmate::assert_choice(
        grid_type,
        choices = c("mean_or_mode", "balanced", "counterfactual", "dataframe")
    )
    checkmate::assert_function(FUN, null.ok = TRUE)
    checkmate::assert_function(FUN_character, null.ok = TRUE)
    checkmate::assert_function(FUN_factor, null.ok = TRUE)
    checkmate::assert_function(FUN_logical, null.ok = TRUE)
    checkmate::assert_function(FUN_binary, null.ok = TRUE)
    checkmate::assert_function(FUN_integer, null.ok = TRUE)
    checkmate::assert_function(FUN_numeric, null.ok = TRUE)
    checkmate::assert_function(FUN_other, null.ok = TRUE)
    checkmate::assert_character(by, null.ok = TRUE)
    checkmate::assert_data_frame(newdata, null.ok = TRUE)
    checkmate::assert_flag(response)

    if (is.null(newdata) && is.null(model)) {
        stop_sprintf("One of `newdata` and `model` must not be `NULL`.")
    }

    # internal model settings are useful if available
    if ("marginaleffects_internal" %in% ...names()) {
        mfx <- ...get("marginaleffects_internal")
    } else {
        mfx <- new_marginaleffects_internal(
            model = model,
            call = call("predictions"))
    }

    # priority: newdata. fallback: modeldata
    if (is.null(newdata)) {
        newdata <- mfx@modeldata
    }

    # sanity checks
    if (!is.null(by)) {
        checkmate::assert_true(all(by %in% colnames(newdata)))
    }

    # prune data, otherwise it can be very expensive to summarize all columns
    # hack because prune works on final marginaleffects objects
    if (!is.null(mfx@model)) {
        variable_names <- c(
            hush(insight::find_variables(mfx@model, flatten = TRUE)),
            hush(insight::find_weights(mfx@model, flatten = TRUE)),
            "marginaleffects_wts_internal",
            by)
        variable_names <- intersect(sort(unique(variable_names)), colnames(newdata))
        if (length(variable_names) > 0) {
            newdata <- subset(newdata, select = variable_names)
        }
    } else {
        variable_names <- colnames(newdata)
    }

    variable_class <- detect_variable_class(newdata, model = model)

    if (is.null(by)) {
        idx <- data.frame()
    } else {
        idx <- newdata[, by, drop = FALSE]
    }
    if (ncol(idx) == 0) {
        newdata_split <- list(newdata)
    } else {
        newdata_split <- split(newdata, idx, drop = TRUE)
    }

    modeldata_for_levels <- if (NROW(mfx@modeldata) > 0) mfx@modeldata else NULL

    values_split <- lapply(newdata_split, function(x) {
        datagrid_newdata_to_list(
            ...,
            newdata = x,
            model = model,
            modeldata = modeldata_for_levels,
            variable_class = variable_class,
            grid_type = grid_type,
            FUN = FUN,
            FUN_character = FUN_character,
            FUN_factor = FUN_factor,
            FUN_logical = FUN_logical,
            FUN_numeric = FUN_numeric,
            FUN_integer = FUN_integer,
            FUN_binary = FUN_binary,
            FUN_other = FUN_other
        )
    })

    if (grid_type %in% c("balanced", "mean_or_mode")) {
        out_split <- lapply(seq_along(values_split), function(i) {
            x <- c(values_split[[i]][[1]], values_split[[i]][[2]], use.names = TRUE)
            len <- setdiff(unique(lengths(x)), 1)
            do.call(data.table::CJ, x)
        })
    } else if (grid_type == "dataframe") {
        out_split <- lapply(seq_along(values_split), function(i) {
            x <- c(values_split[[i]][[1]], values_split[[i]][[2]], use.names = TRUE)
            len <- setdiff(unique(lengths(x)), 1)
            if (length(len) > 1) {
                stop_sprintf('With `grid_type="dataframe"`, the length of each vector must be 1 or be the same for every variable.')
            }
            as.data.frame(x, check.names = FALSE)
        })
    } else if (grid_type == "counterfactual") {
        out_split <- lapply(seq_along(values_split), function(i) {
            explicit <- do.call(data.table::CJ, values_split[[i]]$explicit)
            implicit <- newdata_split[[i]]
            data.table::setDF(implicit)
            data.table::setDF(explicit)
            implicit <- implicit[, setdiff(names(implicit), names(explicit)),
                drop = FALSE]
            implicit <- cbind(data.frame(rowidcf = seq_len(nrow(implicit))), implicit)
            merge(implicit, explicit, all = NULL, sort = FALSE)
        })
    } else {
        stop_sprintf("Unknown grid_type '%s'.", grid_type)
    }

    out <- data.table::rbindlist(out_split)

    out <- data.frame(out, check.names = FALSE)

    if (!"rowid" %in% colnames(out) && nrow(out) > 0) {
        out <- cbind(data.frame(rowid = seq_len(nrow(out))), out)
    }

    attr(out, "variable_names_datagrid") <- names(list(...))


    return(out)
}


datagrid_newdata_to_list <- function(
    ...,
    newdata,
    model,
    modeldata = NULL,
    variable_class,
    grid_type,
    FUN = NULL,
    FUN_character = NULL,
    FUN_factor = NULL,
    FUN_logical = NULL,
    FUN_numeric = NULL,
    FUN_integer = NULL,
    FUN_binary = NULL,
    FUN_other = NULL) {
    FUN_balanced <- list(
        ".FUN_character" = unique_s,
        ".FUN_factor" = unique_s,
        ".FUN_logical" = unique_s,
        ".FUN_binary" = unique_s,
        ".FUN_numeric" = mean_na,
        ".FUN_integer" = mean_i,
        ".FUN_other" = mean_na
    )

    FUN_mean_or_mode <- list(
        ".FUN_character" = get_mode,
        ".FUN_factor" = get_mode,
        ".FUN_logical" = get_mode,
        ".FUN_binary" = get_mode,
        ".FUN_numeric" = mean_na,
        ".FUN_integer" = mean_i,
        ".FUN_other" = mean_na
    )

    FUN_counterfactual <- stats::setNames(
        rep(list(identity), length(FUN_balanced)),
        names(FUN_balanced)
    )

    if (grid_type == "balanced") {
        FUN_list <- FUN_balanced
    } else if (grid_type == "mean_or_mode") {
        FUN_list <- FUN_mean_or_mode
    } else if (grid_type == "dataframe") {
        FUN_list <- FUN_mean_or_mode
    } else if (grid_type == "counterfactual") {
        FUN_list <- FUN_counterfactual
    }

    # FUN overrides defaults from grid_type but not manual specifications
    # order matters
    if (!is.null(FUN)) {
        FUN_list <- stats::setNames(rep(list(FUN), length(FUN_list)), names(FUN_list))
    }

    list2env(FUN_list, envir = environment())
    if (!is.null(FUN_character)) .FUN_character <- FUN_character
    if (!is.null(FUN_factor)) .FUN_factor <- FUN_factor
    if (!is.null(FUN_logical)) .FUN_logical <- FUN_logical
    if (!is.null(FUN_numeric)) .FUN_numeric <- FUN_numeric
    if (!is.null(FUN_integer)) .FUN_integer <- FUN_integer
    if (!is.null(FUN_binary)) .FUN_binary <- FUN_binary
    if (!is.null(FUN_other)) .FUN_other <- FUN_other

    values <- list()

    explicit_values <- list()
    explicit <- list(...)
    for (e in setdiff(names(explicit), "marginaleffects_internal")) {
        if (is.function(explicit[[e]])) {
            if (!e %in% colnames(newdata)) {
                warn_sprintf("The variable '%s' is not in the newdata.", e)
            }
            explicit_values[[e]] <- explicit[[e]](newdata[[e]])
        } else if (isTRUE(checkmate::check_atomic_vector(explicit[[e]]))) {
            explicit_values[[e]] <- unique(explicit[[e]])
        } else {
            stop_sprintf("The value for '%s' must be a vector or a function.", e)
        }
    }

    implicit_values <- list()
    implicit <- setdiff(colnames(newdata), names(explicit))

    # balanced grid should not balance on response otherwise we multiple the rows
    if (grid_type == "balanced") {
        implicit <- setdiff(implicit, hush(insight::find_response(model, flatten = TRUE)))
    }

    # some packages require the response variable to be included in the grid for predict()
    # implicit <- setdiff(implicit, hush(insight::find_response(model, flatten = TRUE)))
    for (i in implicit) {
        if (check_variable_class(variable_class, i, "binary")) {
            implicit_values[[i]] <- .FUN_binary(newdata[[i]])
        } else if (check_variable_class(variable_class, i, "character")) {
            implicit_values[[i]] <- .FUN_character(newdata[[i]])
        } else if (check_variable_class(variable_class, i, "logical")) {
            implicit_values[[i]] <- .FUN_logical(newdata[[i]])
        } else if (check_variable_class(variable_class, i, "categorical")) {
            implicit_values[[i]] <- .FUN_factor(newdata[[i]])
        } else if (check_variable_class(variable_class, i, "numeric")) {
            implicit_values[[i]] <- .FUN_numeric(newdata[[i]])
        } else if (check_variable_class(variable_class, i, "integer")) {
            implicit_values[[i]] <- .FUN_integer(newdata[[i]])
        } else {
            implicit_values[[i]] <- .FUN_other(newdata[[i]])
        }
    }

    # Process factors in explicit values
    for (e in names(explicit_values)) {
        modeldata_col <- NULL
        if (!is.null(modeldata) && e %in% colnames(modeldata)) {
            modeldata_col <- modeldata[[e]]
        }
        explicit_values[[e]] <- sanitize_datagrid_factor(
            explicit_values[[e]],
            newdata[[e]],
            variable_class,
            e,
            modeldata_col = modeldata_col
        )
    }

    out <- list(implicit = implicit_values, explicit = explicit_values)
    return(out)
}


sanitize_datagrid_factor <- function(values, newdata_col, variable_class, var_name, modeldata_col = NULL) {
    if (
        is.factor(newdata_col) ||
            is.factor(modeldata_col) ||
            isTRUE(check_variable_class(variable_class, var_name, "factor"))
    ) {
        level_source <- newdata_col
        if (!is.null(modeldata_col) && length(modeldata_col) > 0) {
            level_source <- modeldata_col
        }
        if (is.factor(level_source)) {
            levs <- levels(level_source)
        } else {
            levs <- as.character(sort(unique(level_source)))
        }
        values <- as.character(values)
        if (!all(values %in% c(levs, NA))) {
            msg <- sprintf(
                'The "%s" element corresponds to a factor variable. The values supplied must be one of the factor levels: %s.',
                var_name,
                toString(dQuote(levs, NULL))
            )
            stop(msg, call. = FALSE)
        }

        values <- factor(values, levels = levs)
    }
    return(values)
}

#' @rdname get_vcov
#' @export
get_vcov.orm <- function(model, vcov = NULL, ...) {
    if (!is.null(vcov) && !isTRUE(checkmate::check_flag(vcov))) {
        msg <- "The `vcov` argument is not supported for models of this class."
        stop_sprintf(msg)
    }
    vcov <- sanitize_vcov(model, vcov)
    out <- stats::vcov(model, intercepts = "all")
    return(out)
}


#' @rdname sanitize_model_specific
sanitize_model_specific.rms <- function(model, ...) {
    # Issue #1600
    md <- get_modeldata(model)
    ord <- sapply(md, function(x) is.factor(x) & inherits(x, "ordered"))
    if (any(ord)) {
        msg <- "Ordered factors sometimes cause issues with `rms` models. Please convert them to unordered factors and refit the model."
        warning(msg, call. = FALSE)
    }
    return(model)
}

#' @rdname sanitize_model_specific
#' @export
sanitize_model_specific.orm <- sanitize_model_specific.rms


#' @rdname sanitize_model_specific
#' @export
sanitize_model_specific.lrm <- sanitize_model_specific.rms


#' @rdname sanitize_model_specific
#' @export
sanitize_model_specific.ols <- sanitize_model_specific.rms


#' @rdname get_predict
#' @export
get_predict.rms <- function(
    model,
    newdata = insight::get_data(model),
    type = NULL,
    mfx = NULL,
    ...) {
    if (is.null(type)) {
        calling_function <- if (!is.null(mfx)) mfx@calling_function else "predictions"
        type <- sanitize_type(model, type, calling_function = calling_function)
    }
    if (inherits(newdata, "tbl_df")) {
        warning("Converting `newdata` from tibble to data.frame.", call. = FALSE)
        newdata <- as.data.frame(newdata)
    }

    # {rms} predict methods break on additional arguments
    out <- get_predict.default(model, newdata = newdata, type = type)
    out <- add_rowid(out, newdata)
    return(out)
}

#' @rdname get_predict
#' @export
get_predict.orm <- get_predict.rms


#' @rdname get_predict
#' @export
get_predict.lrm <- get_predict.rms


#' @rdname get_predict
#' @export
get_predict.ols <- get_predict.rms



### AUTODIFF: 
### DO NOT DO THIS BECAUSE get_model_matrix() doesn't work with these models

# #' @keywords internal
# #' @export
# get_autodiff_args.ols <- function(model, mfx) {
#     # no inheritance! Important to avoid breaking other models
#     if (!class(model)[1] == "ols") {
#         return(NULL)
#     }
#
#     if (!is.null(model$offset)) {
#         autodiff_warning("models with offsets")
#         return(NULL)
#     }
#
#     if (!is.null(model$penalty)) {
#         autodiff_warning("models with offsets")
#         return(NULL)
#     }
#
#     # Check type support
#     if (!mfx@type %in% c("lp")) {
#         autodiff_warning(sprintf("`type='%s'`", mfx@type))
#         return(NULL)
#     }
#
#     # If all checks pass, return supported arguments
#     out <- list(model_type = "linear")
#     return(out)
# }
#
#
# #' @keywords internal
# #' @export
# get_autodiff_args.lrm <- function(model, mfx) {
#     # no inheritance! Important to avoid breaking other models
#     if (!class(model)[1] == "lrm") {
#         return(NULL)
#     }
#
#     if (!is.null(model$offset)) {
#         autodiff_warning("models with offsets")
#         return(NULL)
#     }
#
#     if (!is.null(model$penalty)) {
#         autodiff_warning("models with offsets")
#         return(NULL)
#     }
#
#     # Check type support
#     if (!mfx@type %in% c("fitted", "lp")) {
#         autodiff_warning(sprintf("`type='%s'`", mfx@type))
#         return(NULL)
#     }
#
#     # If all checks pass, return supported arguments
#     mAD <- settings_get("mAD")
#     out <- list(
#         model_type = "glm",
#         family_type = mAD$glm$families$Family$BINOMIAL,
#         link_type = mAD$glm$families$Link$LOGIT
#     )
#     return(out)
# }

#' @include sanity_model.R
#' @rdname sanitize_model_specific
#' @export
sanitize_model_specific.brmsfit <- function(model, ...) {
    insight::check_if_installed("collapse", minimum_version = "1.9.0")
    # terms: brmsfit objects do not have terms immediately available
    te <- tryCatch(attr(stats::terms(stats::formula(model)$formula), "term.labels"), error = function(e) NULL)
    if (any(grepl("^factor\\(", te))) {
        stop("The `factor()` function cannot be used in the model formula of a `brmsfit` model. Please convert your variable to a factor before fitting the model, or use the `mo()` function to specify monotonic variables (see the `brms` vignette on monotonic variables).",
             call. = FALSE)
    }
    return(model)
}


#' @rdname get_coef
#' @export
get_coef.brmsfit <- function(model, ...) {
    out <- insight::get_parameters(model)
    out <- collapse::dapply(out, MARGIN = 2, FUN = collapse::fmedian)
    return(out)
}


#' @include get_predict.R
#' @rdname get_predict
#' @export
get_predict.brmsfit <- function(model,
                                newdata = insight::get_data(model),
                                type = "response",
                                ...) {

    checkmate::assert_choice(type, choices = c("response", "link", "prediction", "average"))

    if (type == "link") {
        insight::check_if_installed("rstantools")
        draws <- rstantools::posterior_linpred(
            model,
            newdata = newdata,
            ...)
    } else if (type == "response") {
        insight::check_if_installed("rstantools")
        draws <- rstantools::posterior_epred(
            model,
            newdata = newdata,
            ...)
    } else if (type == "prediction") {
        insight::check_if_installed("rstantools")
        draws <- rstantools::posterior_predict(
            model,
            newdata = newdata,
            ...)
    } else if (type == "average") {
        insight::check_if_installed("brms")
        draws <- brms::pp_average(
            model,
            newdata = newdata,
            summary = FALSE,
            ...)
    }
    
    if ("rowid_internal" %in% colnames(newdata)) {
        idx <- newdata[["rowid_internal"]]
    } else if ("rowid" %in% colnames(newdata)) {
        idx <- newdata[["rowid"]]
    } else {
        idx <- seq_len(nrow(newdata))
    }

    # resp_subset sometimes causes dimension mismatch 
    if (length(dim(draws)) == 2 && nrow(newdata) != ncol(draws)) {
        msg <- sprintf("Dimension mismatch: There are %s parameters in the posterior draws but %s observations in `newdata` (or the original dataset).",
                       ncol(draws), nrow(newdata))
        insight::format_error(msg)
    }


    # 1d outcome
    if (length(dim(draws)) == 2) {
        med <- collapse::dapply(draws, MARGIN = 2, FUN = collapse::fmedian)
        out <- data.frame(
            rowid = idx,
            group = "main_marginaleffect",
            estimate = med)

    # multi-dimensional outcome
    } else if (length(dim(draws)) == 3) {
        out <- apply(draws, c(2, 3), stats::median)
        levnames <- dimnames(draws)[[3]]
        if (is.null(levnames)) {
            colnames(out) <- seq_len(ncol(out))
        } else {
            colnames(out) <- levnames
        }
        out <- data.frame(
            rowid = rep(idx, times = ncol(out)),
            group = rep(colnames(out), each = nrow(out)),
            estimate = c(out))
        out$group <- group_to_factor(out$group, model)
    } else {
        stop("marginaleffects cannot extract posterior draws from this model. Please report this problem to the Bug tracker with a reporducible example: https://github.com/vincentarelbundock/marginaleffects/issues", call. = FALSE)
    }

    # group for multi-valued outcome
    if (length(dim(draws)) == 3) {
        draws <- lapply(1:dim(draws)[3], function(i) draws[, , i])
        draws <- do.call("cbind", draws)
    }
    attr(out, "posterior_draws") <- t(draws)

    return(out)
}


#' @include get_group_names.R
#' @rdname get_group_names
#' @export
get_group_names.brmsfit <- function(model, ...) {
    if (!is.null(model$family) && "cumulative" %in% model$family) {
        out <- unique(insight::get_response(model))
    } else {
        out <- "main_marginaleffect"
    }
    return(out)
}


#' @rdname get_vcov
#' @export
get_vcov.brmsfit <- function(model,
                             vcov = NULL,
                             ...) {
    if (!is.null(vcov) && !is.logical(vcov)) {
        insight::format_warning("The `vcov` argument is not supported for models of this class.")
    }
    return(NULL)
}

#' (EXPERIMENTAL) Bootstrap, Conformal, and Simulation-Based Inference
#'
#' @description
#' Warning: This function is experimental. It may be renamed, the user interface may change, or the functionality may migrate to arguments in other `marginaleffects` functions.
#'
#' Apply this function to a `marginaleffects` object to change the inferential method used to compute uncertainty estimates.
#'
#' @param x Object produced by one of the core `marginaleffects` functions, or a data frame suitable for the function supplied to the `estimator` argument.
#' @param method String
#' + "delta": delta method standard errors
#' + "boot" package
#' + "fwb": fractional weighted bootstrap
#' + "rsample" package
#' + "simulation" from a multivariate normal distribution (Krinsky & Robb, 1986)
#' + "conformal_split": prediction intervals using split conformal prediction (see Angelopoulos & Bates, 2022)
#' + "conformal_cv+": prediction intervals using cross-validation+ conformal prediction (see Barber et al., 2020)
#' + "conformal_full": prediction intervals using full conformal prediction (see Lei et al., 2018). **Warning**: This method is computationally expensive and typically much slower than split or CV+ methods.
#' + "conformal_quantile": prediction intervals using full conformal prediction (see Romano et al., 2020).
#' @param R Number of resamples, simulations, or cross-validation folds.
#' @param conf_type String: type of bootstrap interval to construct.
#' + `boot`: "perc", "norm", "basic", or "bca"
#' + `fwb`: "perc", "norm", "wald", "basic", "bc", or "bca"
#' + `rsample`: "perc" or "bca"
#' + `simulation`: "perc" or "wald"
#' @param data_train Data frame used to train/fit the model. If `NULL`, `marginaleffects` tries to extract the data from the original model object. Test data are inferred directly from the `newdata` supplied to the originating `marginaleffects` call (e.g., `predictions()`).
#' @param data_test Data frame make out of sample prediction. Only used for conformal inference. If `NULL`, the data frame supplied to `newdata` in the original `marginaleffects` call is used.
#' @param data_calib Data frame used for calibration in split conformal prediction.
#' @param conformal_score String. Warning: The `type` argument in `predictions()` must generate predictions which are on the same scale as the outcome variable. Typically, this means that `type` must be "response" or "probs".
#'   + "residual_abs" or "residual_sq" for regression tasks (numeric outcome)
#'   + "softmax" for classification tasks (when `predictions()` returns a `group` columns, such as multinomial or ordinal logit models.
#' @param estimator Function that accepts a data frame, fits a model, applies a `marginaleffects` function, and returns the object. Only supported with `method = "rsample"` or `method = "boot"`. When `method = "rsample"`, the output must include a "term" column. This is not always the case for `predictions()`, in which case users may have to create the column manually.
#' @param ...
#' + If `method = "boot"`, additional arguments are passed to `boot::boot()`.
#' + If `method = "fwb"`, additional arguments are passed to `fwb::fwb()`.
#' + If `method = "rsample"`, additional arguments are passed to `rsample::bootstraps()`, unless the user supplies a `group` argument, in which case all arguments are passed to `rsample::group_bootstraps()`.
#' + If `method = "conformal_full"`, additional arguments control the optimization process:
#'   - `var_multiplier`: multiplier for initial search bounds (default: 10)
#'   - `max_iter`: maximum iterations for root finding (default: 100)
#'   - `tolerance`: tolerance for root finding convergence (default: `.Machine$double.eps^0.25`)
#' + If `method = "conformal_quantile"`, additional arguments are passed to `quantregForest::quantregForest()` for fitting the quantile regression forest (e.g., `ntree`, `mtry`, `nodesize`, `nthreads`).
#' + Additional arguments are ignored for other conformal methods (`conformal_split`, `conformal_cv+`).
#' @details
#' When `method = "simulation"`, we conduct simulation-based inference following the method discussed in Krinsky & Robb (1986):
#' 1. Draw `R` sets of simulated coefficients from a multivariate normal distribution with mean equal to the original model's estimated coefficients and variance equal to the model's variance-covariance matrix (classical, "HC3", or other).
#' 2. Use the `R` sets of coefficients to compute `R` sets of estimands: predictions, comparisons, slopes, or hypotheses.
#' 3. Take quantiles of the resulting distribution of estimands to obtain a confidence interval (when `conf_type = "perc"`) and the standard deviation of simulated estimates to estimate the standard error (which is used for a Z-test and Wald confidence intervals when `conf_type = "wald"`).
#'
#' When `method = "fwb"`, drawn weights are supplied to the model fitting function's `weights` argument; if the model doesn't accept non-integer weights, this method should not be used. If weights were included in the original model fit, they are extracted by [weights()] and multiplied by the drawn weights. These weights are supplied to the `wts` argument of the estimation function (e.g., `comparisons()`).
#'
#' Warning: custom model classes are not supported by `inferences()` because they are not guaranteed to come with an appropriate `update()` method.
#'
#' @section References:
#'
#' Krinsky, I., and A. L. Robb. 1986. "On Approximating the Statistical Properties of Elasticities." Review of Economics and Statistics 68 (4): 715–9.
#'
#' King, Gary, Michael Tomz, and Jason Wittenberg. "Making the most of statistical analyses: Improving interpretation and presentation." American journal of political science (2000): 347-361
#'
#' Dowd, Bryan E., William H. Greene, and Edward C. Norton. "Computation of standard errors." Health services research 49.2 (2014): 731-750.
#'
#' Angelopoulos, Anastasios N., and Stephen Bates. 2022. "A Gentle Introduction to Conformal Prediction and Distribution-Free Uncertainty Quantification." arXiv. https://doi.org/10.48550/arXiv.2107.07511.
#'
#' Barber, Rina Foygel, Emmanuel J. Candes, Aaditya Ramdas, and Ryan J. Tibshirani. 2020. "Predictive Inference with the Jackknife+." arXiv. http://arxiv.org/abs/1905.02928.
#'
#' Lei, Jing, Max G'Sell, Alessandro Rinaldo, Ryan J. Tibshirani, and Larry Wasserman. 2018. "Distribution-Free Predictive Inference for Regression." Journal of the American Statistical Association 113 (523): 1094–1111.
#'
#' Romano, Yaniv, Evan Patterson, and Emmanuel Candes. 2020. "Conformalized quantile regression." Advances in neural information processing systems 32.
#'
#' @template parallel
#' @return
#' A `marginaleffects` object with simulation or bootstrap resamples and objects attached.
#' @examples
#' \dontrun{
#' library(magrittr)
#' set.seed(1024)
#' mod <- lm(Sepal.Length ~ Sepal.Width * Species, data = iris)
#'
#' # bootstrap
#' avg_predictions(mod, by = "Species") %>%
#'     inferences(method = "boot")
#'
#' avg_predictions(mod, by = "Species") %>%
#'     inferences(method = "rsample")
#'
#' # Fractional (bayesian) bootstrap
#' avg_slopes(mod, by = "Species") %>%
#'     inferences(method = "fwb") %>%
#'     get_draws("rvar") %>%
#'     data.frame()
#'
#' # Simulation-based inference
#' slopes(mod) %>%
#'     inferences(method = "simulation") %>%
#'     head()
#'
#' # Two-step estimation procedure: Propensity score + G-Computation
#' lalonde <- get_dataset("lalonde")
#' estimator <- function(data) {
#'     # Step 1: Estimate propensity scores
#'     fit1 <- glm(treat ~ age + educ + race, family = binomial, data = data)
#'     ps <- predict(fit1, type = "response")
#'     # Step 2: Fit weighted outcome model
#'     m <- lm(re78 ~ treat * (re75 + age + educ + race),
#'         data = data, weight = ps
#'     )
#'     # Step 3: Compute average treatment effect by G-computation
#'     avg_comparisons(m, variables = "treat", wts = ps, vcov = FALSE)
#' }
#' inferences(lalonde, method = "rsample", estimator = estimator)
#' }
#' @export
inferences <- function(
    x,
    method,
    R = 1000,
    conf_type = "perc",
    data_train = NULL,
    data_test = NULL,
    data_calib = NULL,
    conformal_score = "residual_abs",
    estimator = NULL,
    ...) {
    dots <- list(...)

    # keep these alternative arguments forever because they are in the book
    if ("conformal_calibration" %in% names(dots)) {
        data_calib = dots[["conformal_calibration"]]
    }
    if ("conformal_train" %in% names(dots)) {
        data_train = dots[["conformal_train"]]
    }
    if ("conformal_test" %in% names(dots)) {
        data_test = dots[["conformal_test"]]
    }

    mfx <- attr(x, "marginaleffects")

    # dummy mfx for `estimator` and no marginaleffects object
    if (!inherits(mfx, "marginaleffects_internal")) {
        mfx <- new_marginaleffects_internal(NULL, call("predictions"))
        if (isTRUE(checkmate::check_data_frame(x))) {
            mfx@modeldata <- x
            mfx@modeldata_available <- TRUE
        } else {
            mfx@modeldata_available <- FALSE
        }
    }

    # supported classes
    supported_classes <- c("predictions", "comparisons", "slopes", "hypotheses")
    if (is.null(estimator)) {
        has_supported_class <- any(vapply(
            supported_classes,
            function(cls) inherits(x, cls),
            logical(1)
        ))
        if (!has_supported_class) {
            msg <- "The `x` argument must be a `marginaleffects` object when supplying a function to the `estimator` argument."
            stop_sprintf(msg)
        }
    } else {
        x <- sanitize_estimator(x = x, estimator = estimator, method = method)
    }

    # inherit conf_level from the original object
    conf_level <- mfx@conf_level
    checkmate::assert_integerish(R, lower = 2)
    checkmate::assert_choice(
        method,
        choices = c(
            "delta",
            "boot",
            "fwb",
            "rsample",
            "simulation",
            "conformal_split",
            "conformal_cv+",
            "conformal_quantile",
            "conformal_full"
        )
    )

    if (is.null(data_train)) {
        if (!isTRUE(mfx@modeldata_available)) {
            checkmate::assert_data_frame(data_train, null.ok = FALSE)
        } else {
            data_train <- mfx@modeldata
        }
    }

    if (is.null(data_test)) {
        checkmate::assert_data_frame(mfx@newdata)
        data_test <- mfx@newdata
    }

    if (inherits(mfx@model, c("Learner", "model_fit", "workflow"))) {
        if (method == "simulation") {
            msg <- "Simulation-based inference is not supported for this class."
            stop_sprintf(msg)
        }
        checkmate::assert_data_frame(data_train, null.ok = FALSE)
    }

    # Issue #1501: `newdata` should only use the pre-evaluated `newdata` instead of bootstrapping datagrid()
    mfx <- attr(x, "marginaleffects")
    call_mfx <- mfx@call

    # Update call with pre-evaluated newdata if available
    if (!is.null(call_mfx)) {
        nd <- mfx@newdata
        # avoid error on two arguments matching `newdata`
        if (inherits(nd, "data.frame") && !"newdata" %in% ...names()) {
            call_mfx[["newdata"]] <- nd
        }

        # Update mfx object if available
        if (!is.null(mfx) && inherits(nd, "data.frame")) {
            mfx@newdata <- nd
            mfx@call <- call_mfx
            attr(x, "marginaleffects") <- mfx
        }
    }

    # default standard errors are Delta anyway
    if (method == "delta") {
        return(x)
    } else if (method == "boot") {
        out <- inferences_boot(x,
            R = R,
            conf_level = conf_level,
            conf_type = conf_type,
            estimator = estimator,
            data_train = data_train,
            mfx = mfx,
            ...)
    } else if (method == "fwb") {
        if (isTRUE("wts" %in% names(mfx@call)) && !isFALSE(mfx@call[["wts"]])) {
            stop_sprintf("The `fwb` method is not supported with the `wts` argument.")
        }
        out <- inferences_fwb(x,
            R = R,
            conf_level = conf_level,
            conf_type = conf_type,
            mfx = mfx,
            ...)
    } else if (method == "rsample") {
        out <- inferences_rsample(x,
            R = R,
            conf_level = conf_level,
            conf_type = conf_type,
            estimator = estimator,
            data_train = data_train,
            mfx = mfx,
            ...)
    } else if (method == "simulation") {
        out <- inferences_simulation(x,
            R = R,
            conf_level = conf_level,
            conf_type = conf_type,
            mfx = mfx,
            ...)
    } else if (isTRUE(grepl("conformal", method))) {
        sanity_inferences_conformal(
            mfx = mfx,
            score = conformal_score,
            method = method,
            data_calib = data_calib,
            R = R
        )

        conformal_fun <- switch(method,
            "conformal_split" = conformal_split,
            "conformal_cv+" = conformal_cv_plus,
            "conformal_quantile" = conformal_quantile,
            "conformal_full" = conformal_full,
            NULL
        )

        out <- conformal_fun(
            x,
            R = R,
            conf_level = conf_level,
            data_test = data_test,
            data_calib = data_calib,
            data_train = data_train,
            score = conformal_score,
            mfx = mfx,
            ...
        )
    }

    return(out)
}

#' @include set_coef.R
#' @rdname set_coef
#' @keywords internal
#' @export
set_coef.merMod <- function(model, coefs, ...) {
    # in 'merMod', predictions work the slot called "beta", which is unnamed
    # `fixef(model)` returns the same thing named
    beta <- methods::slot(model, "beta")
    beta[match(names(coefs), names(lme4::fixef(model)))] <- as.numeric(coefs)
    methods::slot(model, "beta") <- beta
    model
}


#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.merMod <- function(model, ...) {
    lme4::fixef(model)
}


#' @rdname get_predict
#' @export
get_predict.merMod <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    ...
) {
    get_predict.default(model, newdata = newdata, type = type, ...)
}


#' @rdname set_coef
#' @export
set_coef.lmerModLmerTest <- set_coef.merMod


#' @rdname get_coef
#' @export
get_coef.lmerModLmerTest <- get_coef.merMod


#' @rdname get_predict
#' @export
get_predict.lmerModLmerTest <- get_predict.merMod


#' @rdname set_coef
#' @export
set_coef.lmerMod <- set_coef.merMod


#' @rdname get_coef
#' @export
get_coef.lmerMod <- get_coef.merMod


#' @rdname get_predict
#' @export
get_predict.lmerMod <- get_predict.merMod


#' @rdname sanitize_model_specific
#' @export
sanitize_model_specific.merMod <- function(model, re.form, vcov = TRUE, ...) {
    if (!isFALSE(vcov) && (missing(re.form) || (!isTRUE(is.na(re.form))))) {
        msg <- "For this model type, `marginaleffects` only takes into account the uncertainty in fixed-effect parameters. This is often appropriate when `re.form=NA`, but may be surprising to users who set `re.form=NULL` (default) or to some other value. Call `options(marginaleffects_safe = FALSE)` to silence this warning."
        warn_sprintf(msg)
    }
    return(model)
}

#' find mode, preserve type, and pick an arbitrary value when multi-modal
#' https://stackoverflow.com/a/8189441/342331
#' @noRd
get_mode <- function(x) {
  ux <- unique(x)
  ux[which.max(tabulate(match(x, ux)))]
}

####################################################################
#  The functions below were copied from the `prediction` package. ##
#  Copyright: Thomas J. Leeper 2016-2018                          ##
#  MIT License                                                    ##
####################################################################

#' Compute the mode or mean of `x`
#' @param x extract the mean or the mode of vector or data.frame x depending on its type
#' @keywords internal
#' @noRd
#' @return numeric vector
get_mean_or_mode <- function(x) {
    UseMethod("get_mean_or_mode")
}

#' @export
get_mean_or_mode.default <- function(x) {
    mean(x)
}

#' @export
get_mean_or_mode.character <- function(x) {
    get_mode(x)
}

#' @export
get_mean_or_mode.factor <- function(x) {
    get_mode(x)
}

#' @export
get_mean_or_mode.logical <- function(x) {
    get_mode(x)
}

#' @export
get_mean_or_mode.data.frame <- function(x) {
    out <- list()
    for (n in names(x)) {
        # variables transformed to factor in formula are assigned a "factor"
        # TRUE attribute by insight::get_data
        if (isTRUE(attributes(x)$marginaleffects_variable_class[[n]] == "factor")) {
            out[[n]] <- get_mean_or_mode.factor(x[[n]])
        } else {
            out[[n]] <- get_mean_or_mode(x[[n]])
        }
    }
    return(out)
}

#' @rdname get_predict
#' @export
get_predict.mhurdle <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    ...) {
    out <- stats::predict(model, what = type, newdata = newdata)
    out <- data.table(estimate = out)
    out <- add_rowid(out, newdata)
    return(out)
}


#' @rdname get_vcov
#' @export
get_vcov.mhurdle <- function(model, vcov = NULL, ...) {
    if (!is.null(vcov) && !is.logical(vcov)) {
        stop_sprintf(
            "The `vcov` for this class of models must be TRUE or FALSE."
        )
    }
    vcov <- sanitize_vcov(model, vcov)
    out <- try(stats::vcov(model), silent = TRUE)
    if (inherits(out, "try-error")) {
        out <- tryCatch(model[["vcov"]], error = function(e) NULL)
    }
    return(out)
}

#' @include sanity_model.R
#' @rdname sanitize_model_specific
#' @keywords internal
sanitize_model_specific.plm <- function(model, ...) {
    if ("within" %in% model$args$model) {
        stop('The `plm::predict` function does not appear to support the `newdata` argument when `plm(model="within")`. Therefore, `marginaleffects` cannot support "within" models, even if it supports many other models produced by the `plm` package. You may want to try the `fixest` package instead.',
             call. = FALSE)
    }
    return(model)
}


#' @rdname sanitize_model_specific
sanitize_model_specific.plm <- function(model, ...) {
    insight::check_if_installed("insight", minimum_version = "0.17.1")
    return(model)
}

#' @keywords internal
#' @export
get_autodiff_args.ivreg <- function(model, mfx) {
    # no inheritance! Important to avoid breaking other models
    if (!class(model)[1] == "ivreg") {
        return(NULL)
    }

    if (!is.null(model$offset)) {
        autodiff_warning("models with offsets")
        return(NULL)
    }

    if (!is.null(mfx@wts)) {
        autodiff_warning("the `wts` argument")
        return(NULL)
    }

    # If all checks pass, return supported arguments
    out <- list(model_type = "linear")
    return(out)
}



#' @keywords internal
#' @export
get_autodiff_args.ivreg <- get_autodiff_args.lm

# Execute code silently
# Do not export to avoid conflict with modelsummary
hush <- function(code) {
    void <- utils::capture.output({
        out <- invisible(
            suppressMessages(
                suppressWarnings(
                    tryCatch(code, error = function(e) NULL))
            )
        )
    })
    return(out)
}

#' @importFrom generics tidy
#' @export
generics::tidy


#' @importFrom generics glance
#' @export
generics::glance


#' tidy helper
#'
#' @noRd
#' @export
tidy.comparisons <- function(x, ...) {
    insight::check_if_installed("tibble")
    out <- tibble::as_tibble(x)
    return(out)
}


#' tidy helper
#'
#' @noRd
#' @export
tidy.slopes <- tidy.comparisons


#' tidy helper
#'
#' @noRd
#' @export
tidy.predictions <- tidy.comparisons


#' tidy helper
#'
#' @noRd
#' @export
tidy.hypotheses <- tidy.comparisons


#' @noRd
#' @export
glance.slopes <- function(x, ...) {
    insight::check_if_installed("insight")
    insight::check_if_installed("modelsummary")
    mfx <- attr(x, "marginaleffects")
    model <- tryCatch(mfx@model, error = function(e) NULL)
    if (is.null(model) || isTRUE(checkmate::check_string(model))) {
        model <- tryCatch(mfx@call[["model"]], error = function(e) NULL)
    }
    gl <- suppressMessages(suppressWarnings(try(
        modelsummary::get_gof(model, ...),
        silent = TRUE
    )))
    if (inherits(gl, "data.frame")) {
        out <- data.frame(gl)
    } else {
        out <- NULL
    }
    vcov_type <- components(x, "vcov_type")
    if (is.null(out) && !is.null(vcov_type)) {
        out <- data.frame("vcov.type" = vcov_type)
    } else if (!is.null(out)) {
        out[["vcov.type"]] <- vcov_type
    }
    out <- tibble::as_tibble(out)
    return(out)
}


#' @noRd
#' @export
glance.predictions <- glance.slopes


#' @noRd
#' @export
glance.comparisons <- glance.slopes


#' @noRd
#' @export
glance.hypotheses <- glance.slopes

#' Print a marginaleffects object in knitr
#'
#' @keywords internal
#' @return A string with class `knit_asis` to be printed in Rmarkdown or Quarto documents.
#' @rawNamespace S3method(knitr::knit_print, marginaleffects)
knit_print.marginaleffects <- function(x, ...) {
    if (isTRUE(getOption("marginaleffects_print_style") == "tinytable")) {
        insight::check_if_installed("tinytable")
        x <- print(x, "tinytable")
        printfun <- utils::getFromNamespace("knit_print.tinytable", "tinytable")
        printfun(x)
    } else {
        print(x)
    }
}


#' Print `marginaleffects` objects
#'
#' @description
#' This function controls the text which is printed to the console when one of the core `marginalefffects` functions is called and the object is returned: `predictions()`, `comparisons()`, `slopes()`, `hypotheses()`, `avg_predictions()`, `avg_comparisons()`, `avg_slopes()`.
#'
#' All of those functions return standard data frames. Columns can be extracted by name, `predictions(model)$estimate`, and all the usual data manipulation functions work out-of-the-box:  `colnames()`, `head()`, `subset()`, `dplyr::filter()`, `dplyr::arrange()`, etc.
#'
#' Some of the data columns are not printed by default. You can disable pretty printing and print the full results as a standard data frame using the `style` argument or by applying `as.data.frame()` on the object. See examples below.
#'
#' @param x An object produced by one of the `marginaleffects` package functions.
#' @param style "summary", "data.frame", or "tinytable"
#' @param digits The number of digits to display.
#' @param p_eps p values smaller than this number are printed in "<0.001" style.
#' @param topn The number of rows to be printed from the beginning and end of tables with more than `nrows` rows.
#' @param nrows The number of rows which will be printed before truncation.
#' @param ncols The maximum number of column names to display at the bottom of the printed output.
#' @param type boolean: should the type be printed?
#' @param column_names boolean: should the column names be printed?
#' @param ... Other arguments are currently ignored.
#' @export
#' @examples
#' library(marginaleffects)
#' mod <- lm(mpg ~ hp + am + factor(gear), data = mtcars)
#' p <- predictions(mod, by = c("am", "gear"))
#' p
#'
#' subset(p, am == 1)
#'
#' print(p, style = "data.frame")
#'
#' data.frame(p)
#'
print.marginaleffects <- function(
    x,
    style = getOption("marginaleffects_print_style", default = "summary"),
    digits = getOption("marginaleffects_print_digits", default = 3),
    p_eps = getOption("marginaleffects_print_p_eps", default = 0.001),
    topn = getOption("marginaleffects_print_topn", default = 5),
    nrows = getOption("marginaleffects_print_nrows", default = 30),
    ncols = getOption("marginaleffects_print_ncols", default = 30),
    type = getOption("marginaleffects_print_type", default = TRUE),
    column_names = getOption(
        "marginaleffects_print_column_names",
        default = FALSE
    ),
    ...
) {
    checkmate::assert_number(digits)
    checkmate::assert_number(topn)
    checkmate::assert_number(nrows)
    checkmate::assert_choice(
        style,
        choices = c(
            "data.frame",
            "summary",
            "tinytable",
            "html",
            "latex",
            "markdown",
            "typst"
        )
    )

    mfx <- attr(x, "marginaleffects")

    if (isTRUE(style == "data.frame")) {
        print(as.data.frame(x))
        return(invisible(x))
    }

    print_columns_text <- print_type_text <- print_term_text <- print_contrast_text <- NULL

    out <- x

    nrows <- max(nrows, 2 * topn)

    if ("group" %in% colnames(out) && all(out$group == "main_marginaleffects")) {
        out$group <- NULL
    }

    # subset before rounding so that digits match top and bottom rows
    if (nrow(out) > nrows) {
        out <- rbind(utils::head(out, topn), utils::tail(out, topn))
        splitprint <- TRUE
    } else {
        splitprint <- FALSE
    }

    # round and replace NAs
    ps <- c("p.value", "p.value.nonsup", "p.value.noninf", "p.value.equiv", 
        "p.rope.conditional", "p.rope.unconditional")

    for (i in seq_along(out)) {
        if (colnames(out)[i] %in% ps) {
            out[[i]] <- format.pval(out[[i]], digits = digits, eps = p_eps)
        } else if (isTRUE("s.value" == colnames(out)[i])) {
            out[[i]] <- sprintf("%.1f", out[[i]])
        } else {
            out[[i]] <- format(out[[i]], digits = digits)
        }
    }

    conf_level <- if (is.null(mfx)) 0.95 else mfx@conf_level
    alpha <- 100 * (1 - conf_level)

    hypotheses_joint_label <- attr(x, "hypotheses_joint_label")
    if (is.null(hypotheses_joint_label)) {
        if (any(out[["df"]] < Inf)) {
            hypotheses_joint_label <- "t"
        } else {
            hypotheses_joint_label <- "z"
        }
    }

    # don't print df if they're all infinite
    if ("df" %in% colnames(out) && all(out$df == Inf, na.rm = TRUE)) {
        out[["df"]] <- NULL
    }

    # rename
    dict <- c(
        "term" = "Term",
        "group" = "Group",
        "contrast" = "Contrast",
        "hypothesis" = "Hypothesis",
        "value" = "Value",
        "by" = "By",
        "estimate" = "Estimate",
        "std.error" = "Std. Error",
        "statistic" = hypotheses_joint_label,
        "p.value" = sprintf("Pr(>|%s|)", hypotheses_joint_label),
        "s.value" = "S",
        "conf.low" = sprintf("%.1f %%", alpha / 2),
        "conf.high" = sprintf("%.1f %%", 100 - alpha / 2),
        "pred.low" = sprintf("Pred. %.1f %%", alpha / 2),
        "pred.high" = sprintf("Pred. %.1f %%", 100 - alpha / 2),
        "pred.set" = sprintf("Pred Set %.1f %%", 100 - alpha / 2),
        "rope" = "ROPE",
        "p.rope.conditional" = "Pr(ROPE|CI)",
        "p.rope.unconditional" = "Pr(ROPE)",
        "p.value.nonsup" = "p (NonSup)",
        "p.value.equiv" = "p (Equiv)",
        "df" = "Df",
        "df1" = "Df 1",
        "df2" = "Df 2",
        "rvar" = "rvar"
    )

    # explicitly given by user in `datagrid()` or `by` or `newdata`
    bycols <- "by"
    if (!is.null(mfx)) {
        bycols <- c(
            bycols,
            mfx@variable_names_by, 
            mfx@variable_names_by_hypothesis, 
            mfx@variable_names_datagrid)
    }

    # useless columns should not be printed
    useless <- c(
        # indices
        "rowid",
        "rowidcf",
        # user-supplied omissions
        getOption("marginaleffects_print_omit", default = NULL),
        # response variable
        if (!is.null(mfx)) mfx@variable_names_response else NULL
    )

    if ("term" %in% colnames(out) && length(unique(out$term)) == 1L) {
        print_term_text <- sprintf("Term: %s\n", out[["term"]][1L])
        useless <- c(useless, "term")
    }

    if ("contrast" %in% colnames(out) && length(unique(out$contrast)) == 1L) {
        print_contrast_text <- sprintf("Comparison: %s\n", out[["contrast"]][1L])
        useless <- c(useless, "contrast")
    }

    # Subset columns
    implicit <- if (!is.null(mfx) && is.data.frame(mfx@newdata)) attr(mfx@newdata, "implicit") else NULL
    idx <- c(
        bycols,
        names(dict),
        grep("^contrast_", colnames(x), value = TRUE)
    )
    start <- grep(
        "term|^contrast|group",
        c(names(dict), colnames(x)),
        value = TRUE
    )
    middle <- bycols
    end <- setdiff(intersect(names(dict), colnames(x)), c(start, middle))
    end <- c(end, implicit)
    idx <- c(start, middle, end)
    idx <- intersect(idx, colnames(out))
    idx <- setdiff(idx, useless)
    idx <- unique(idx)
    out <- data.table(out)[, ..idx, drop = FALSE]

    # rename columns
    old <- colnames(out)
    new <- gsub("^contrast_", "C: ", old)
    idx_match <- match(old, names(dict))
    new[!is.na(idx_match)] <- dict[idx_match[!is.na(idx_match)]]
    data.table::setnames(out, old = old, new = new)

    # Footnotes
    if (ncol(x) <= ncols && isTRUE(column_names)) {
        print_columns_text <- sprintf("Columns: %s\n", toString(colnames(x)))
    }

    if (isTRUE(type) && !is.null(mfx) && !is.null(mfx@type)) {
        print_type_text <- sprintf("Type: %s\n", mfx@type)
    }

    # avoid infinite recursion by stripping marginaleffect.summary class
    data.table::setDF(out)

    if (style %in% c("tinytable", "html", "latex", "typst", "markdown")) {
        insight::check_if_installed("tinytable")

        tab <- as.data.frame(out)

        if (isTRUE(splitprint)) {
            tab <- rbind(utils::head(tab, topn), utils::tail(tab, topn))
        }

        args <- list(x = tab)
        notes <- c(print_type_text, print_columns_text)
        if (!is.null(notes)) args$notes <- notes
        tab <- do.call(tinytable::tt, args)
        tab <- tinytable::format_tt(tab, escape = TRUE)

        if (isTRUE(splitprint)) {
            msg <- "%s rows omitted"
            msg <- sprintf(msg, nrow(x) - 2 * topn)
            msg <- stats::setNames(list(topn + 1), msg)
            tab <- tinytable::group_tt(tab, i = msg)
            tab <- tinytable::style_tt(tab, i = topn + 1, align = "c")
        }

        tab@output <- style
        if (style == "tinytable") {
            return(tab)
        }
        print(tab)
        return(invisible(tab))
    }

    # head
    cat("\n")
    print_head <- attr(x, "print_head")
    if (!is.null(print_head)) {
        cat(print_head, "\n")
    }

    # some commands do not generate average contrasts/mfx. E.g., `lnro` with `by`
    if (splitprint) {
        tmp <- utils::capture.output(print(out, row.names = FALSE))

        top <- paste(tmp[seq_len(topn + 1)], collapse = "\n")
        cat(top)

        msg <- "\n--- %s rows omitted. See ?print.marginaleffects ---\n"
        msg <- sprintf(msg, nrow(x) - 2 * topn)
        cat(msg)

        bottom <- paste(tmp[-seq_len(topn + 1)], collapse = "\n")
        cat(bottom)
    } else {
        print(out, row.names = FALSE)
    }
    cat("\n")

    cat(print_term_text)
    cat(print_type_text)
    cat(print_contrast_text)
    cat(print_columns_text)
    cat("\n")

    print_tail <- attr(x, "print_tail")
    if (!is.null(print_tail)) {
        cat(print_tail, "\n")
    }

    return(invisible(x))
}

#' @noRd
#' @export
print.hypotheses <- print.marginaleffects

#' @noRd
#' @export
print.predictions <- print.marginaleffects

#' @noRd
#' @export
print.comparisons <- print.marginaleffects

#' @noRd
#' @export
print.slopes <- print.marginaleffects

#' @noRd
#' @exportS3Method knitr::knit_print
knit_print.hypotheses <- knit_print.marginaleffects

#' @noRd
#' @exportS3Method knitr::knit_print
knit_print.predictions <- knit_print.marginaleffects

#' @noRd
#' @exportS3Method knitr::knit_print
knit_print.comparisons <- knit_print.marginaleffects

#' @noRd
#' @exportS3Method knitr::knit_print
knit_print.slopes <- knit_print.marginaleffects

get_by <- function(
    estimates,
    draws,
    newdata,
    by,
    byfun = NULL,
    verbose = TRUE,
    ...) {

    if (is.null(by) || isFALSE(by)) {
        out <- estimates
        attr(out, "posterior_draws") <- draws
        return(out)
    }

    missing <- setdiff(setdiff(colnames(by), "by"), colnames(estimates))
    if (length(missing) > 0) {
        idx <- intersect(c("rowid", "rowidcf", missing), colnames(newdata))
        estimates <- merge(estimates, newdata[, idx], sort = FALSE, all.x = TRUE)
    }

    if (isTRUE(by)) {
        regex <- "^term$|^group$|^contrast$|^contrast_"
        bycols <- grep(regex, colnames(estimates), value = TRUE)

    } else if (isTRUE(checkmate::check_character(by))) {
        bycols <- by

    } else if (isTRUE(checkmate::check_data_frame(by))) {
        idx <- setdiff(intersect(colnames(estimates), colnames(by)), "by")
        # harmonize column types
        for (v in colnames(by)) {
            if (isTRUE(is.character(estimates[[v]])) && isTRUE(is.numeric(by[[v]]))) {
                by[[v]] <- as.character(by[[v]])
            } else if (isTRUE(is.numeric(estimates[[v]])) && isTRUE(is.character(by[[v]]))) {
                by[[v]] <- as.numeric(by[[v]])
            }
        }
        estimates[by, by := by, on = idx]
        bycols <- "by"
    }

    if ("by" %in% colnames(estimates) && anyNA(estimates[["by"]])) {
        msg <- insight::format_message("The `by` data.frame does not cover all combinations of response levels and/or predictors. Some estimates will not be included in the aggregation.")
        if (isTRUE(verbose)) warning(msg, call. = FALSE)
        tmp <- !is.na(estimates[["by"]])
        draws <- draws[tmp, drop = FALSE] 
        estimates <- estimates[tmp, drop = FALSE] 
    }

    bycols <- intersect(unique(c("term", bycols)), colnames(estimates))

    # bayesian
    if (!is.null(draws)) {
        estimates <- average_draws(
            data = estimates,
            index = bycols,
            draws = draws,
            byfun = byfun)

    # frequentist
    } else {
        if (!is.null(byfun)) {
            estimates <- estimates[,
                .(estimate = byfun(estimate)),
                keyby = bycols]

        } else if ("marginaleffects_wts_internal" %in% colnames(newdata)) {
            estimates <- estimates[,
                .(estimate = stats::weighted.mean(
                    estimate,
                    marginaleffects_wts_internal,
                    na.rm = TRUE)),
                keyby = bycols]

        } else {
            estimates <- estimates[,
                .(estimate = mean(estimate, na.rm = TRUE)),
                keyby = bycols]
        }
    }

    return(estimates)
}

#' (Non-)Linear Tests for Null Hypotheses, Joint Hypotheses, Equivalence, Non Superiority, and Non Inferiority
#'
#' @description
#' Uncertainty estimates are calculated as first-order approximate standard errors for linear or non-linear functions of a vector of random variables with known or estimated covariance matrix. In that sense, [`hypotheses`] emulates the behavior of the excellent and well-established [car::deltaMethod] and [car::linearHypothesis] functions, but it supports more models; requires fewer dependencies; expands the range of tests to equivalence and superiority/inferiority; and offers convenience features like robust standard errors.
#'
#' To learn more, read the hypothesis tests vignette, visit the
#' package website:
#'
#' * <https://marginaleffects.com/chapters/hypothesis.html>
#' * <https://marginaleffects.com/>
#'
#' Warning #1: Tests are conducted directly on the scale defined by the `type` argument. For some models, it can make sense to conduct hypothesis or equivalence tests on the `"link"` scale instead of the `"response"` scale which is often the default.
#'
#' Warning #2: For hypothesis tests on objects produced by the `marginaleffects` package, it is safer to use the `hypothesis` argument of the original function.  Using `hypotheses()` may not work in certain environments, in lists, or when working programmatically with *apply style functions.
#'
#' Warning #3: The tests assume that the `hypothesis` expression is (approximately) normally distributed, which for non-linear functions of the parameters may not be realistic. More reliable confidence intervals can be obtained using the \code{inferences()} function with `method = "boot"`.
#'
#' @inheritParams comparisons
#' @param model Model object or object generated by the `comparisons()`, `slopes()`, or `predictions()` functions.
#' @param joint Joint test of statistical significance. The null hypothesis value can be set using the `hypothesis` argument.
#' - FALSE: Hypotheses are not tested jointly.
#' - TRUE: All parameters are tested jointly.
#' - String: A regular expression to match parameters to be tested jointly. `grep(joint, perl = TRUE)`
#' - Character vector of parameter names to be tested. Characters refer to the names of the vector returned by `marginaleffects::get_coef(object)`.
#' - Integer vector of indices. Which parameters positions to test jointly.
#' - Note: When using the `joint` argument, the `hypothesis` argument is limited to `NULL` or numeric values. Users can chain multiple `hypotheses()` for complex joint hypothesis tests.
#' @param conf_level NULL or numeric value between 0 and 1. Confidence level to use to build a confidence interval. When `NULL` and `model` was generated by `marginaleffects`, the confidence level is taken from the `conf_level` attribute of the model. Otherwise, the default value is 0.95.
#' @param joint_test A character string specifying the type of test, either "f" or "chisq". The null hypothesis is set by the `hypothesis` argument, with default null equal to 0 for all parameters.
#' @param df Degrees of freedom used to compute p values and confidence intervals.
#'   - A single numeric value between 1 and `Inf`, or a numeric vector with length equal to the number of rows in the output. When `df` is `Inf`, the normal distribution is used. When `df` is finite, the `t` distribution is used.
#'   - "residual": Calls [insight::get_df] to extract degrees of freedom from the model automatically.
#'   - When using `joint_test="f"`, the `df` argument should be a numeric vector of length 2.
#' @param multcomp Logical or string. If `TRUE` or string, apply multiple comparison adjustment to the p values and report family-wise confidence intervals. Valid strings: "holm", "hochberg", "hommel", "bonferroni", "BH", "BY", "fdr", "single-step", "Shaffer", "Westfall", "free". When `multcomp` is `TRUE`, the "holm" method is used.
#'
#' @section Joint hypothesis tests:
#' The test statistic for the joint Wald test is calculated as (R * theta_hat - r)' * inv(R * V_hat * R') * (R * theta_hat - r) / Q,
#' where theta_hat is the vector of estimated parameters, V_hat is the estimated covariance matrix, R is a Q x P matrix for testing Q hypotheses on P parameters,
#' r is a Q x 1 vector for the null hypothesis, and Q is the number of rows in R. If the test is a Chi-squared test, the test statistic is not normalized.
#'
#' The p-value is then calculated based on either the F-distribution (for F-test) or the Chi-squared distribution (for Chi-squared test).
#' For the F-test, the degrees of freedom are Q and (n - P), where n is the sample size and P is the number of parameters.
#' For the Chi-squared test, the degrees of freedom are Q.
#'
#' @template equivalence
#' @examples
#' mod <- lm(mpg ~ hp + wt + factor(cyl), data = mtcars)
#'
#' hypotheses(mod)
#'
#' # Test of equality between coefficients
#' hypotheses(mod, hypothesis = "hp = wt")
#'
#' # Non-linear function
#' hypotheses(mod, hypothesis = "exp(hp + wt) = 0.1")
#'
#' # Robust standard errors
#' hypotheses(mod, hypothesis = "hp = wt", vcov = "HC3")
#'
#' # b1, b2, ... shortcuts can be used to identify the position of the
#' # parameters of interest in the output of
#' hypotheses(mod, hypothesis = "b2 = b3")
#'
#' # wildcard
#' hypotheses(mod, hypothesis = "b* / b2 = 1")
#'
#' # term names with special characters have to be enclosed in backticks
#' hypotheses(mod, hypothesis = "`factor(cyl)6` = `factor(cyl)8`")
#'
#' mod2 <- lm(mpg ~ hp * drat, data = mtcars)
#' hypotheses(mod2, hypothesis = "`hp:drat` = drat")
#'
#' # predictions(), comparisons(), and slopes()
#' mod <- glm(am ~ hp + mpg, data = mtcars, family = binomial)
#' cmp <- comparisons(mod, newdata = "mean")
#' hypotheses(cmp, hypothesis = "b1 = b2")
#'
#' mfx <- slopes(mod, newdata = "mean")
#' hypotheses(cmp, hypothesis = "b2 = 0.2")
#'
#' pre <- predictions(mod, newdata = datagrid(hp = 110, mpg = c(30, 35)))
#' hypotheses(pre, hypothesis = "b1 = b2")
#'
#' # The `hypothesis` argument can be used to compute standard errors for fitted values
#' mod <- glm(am ~ hp + mpg, data = mtcars, family = binomial)
#'
#' f <- function(x) predict(x, type = "link", newdata = mtcars)
#' p <- hypotheses(mod, hypothesis = f)
#' head(p)
#'
#' f <- function(x) predict(x, type = "response", newdata = mtcars)
#' p <- hypotheses(mod, hypothesis = f)
#' head(p)
#'
#' # Complex aggregation
#' # Step 1: Collapse predicted probabilities by outcome level, for each individual
#' # Step 2: Take the mean of the collapsed probabilities by group and `cyl`
#' library(dplyr)
#' library(MASS)
#' library(dplyr)
#' library(magrittr)
#'
#' dat <- transform(mtcars, gear = factor(gear))
#' mod <- polr(gear ~ factor(cyl) + hp, dat)
#'
#' aggregation_fun <- function(x) {
#'     predictions(x, vcov = FALSE) %>%
#'         mutate(group = ifelse(group %in% c("3", "4"), "3 & 4", "5")) %>%
#'         summarize(estimate = sum(estimate), .by = c("rowid", "cyl", "group")) %>%
#'         summarize(estimate = mean(estimate), .by = c("cyl", "group")) %>%
#'         rename(term = cyl)
#' }
#'
#' hypotheses(mod, hypothesis = aggregation_fun)
#'
#' # Equivalence, non-inferiority, and non-superiority tests
#' mod <- lm(mpg ~ hp + factor(gear), data = mtcars)
#' p <- predictions(mod, newdata = "median")
#' hypotheses(p, equivalence = c(17, 18))
#'
#' mfx <- avg_slopes(mod, variables = "hp")
#' hypotheses(mfx, equivalence = c(-.1, .1))
#'
#' cmp <- avg_comparisons(mod, variables = "gear", hypothesis = ~pairwise)
#' hypotheses(cmp, equivalence = c(0, 10))
#'
#' # joint hypotheses: character vector
#' model <- lm(mpg ~ as.factor(cyl) * hp, data = mtcars)
#' hypotheses(model, joint = c("as.factor(cyl)6:hp", "as.factor(cyl)8:hp"))
#'
#' # joint hypotheses: regular expression
#' hypotheses(model, joint = "cyl")
#'
#' # joint hypotheses: integer indices
#' hypotheses(model, joint = 2:3)
#'
#' # joint hypotheses: different null hypotheses
#' hypotheses(model, joint = 2:3, hypothesis = 1)
#' hypotheses(model, joint = 2:3, hypothesis = 1:2)
#'
#' # joint hypotheses: marginaleffects object
#' cmp <- avg_comparisons(model)
#' hypotheses(cmp, joint = "cyl")
#'
#' # Multiple comparison adjustment
#' # p values and family-wise confidence intervals
#' cmp <- avg_comparisons(model)
#' hypotheses(cmp, multcomp = "hochberg")
#'
#' @export
hypotheses <- function(
    model = NULL,
    hypothesis = NULL,
    vcov = TRUE,
    conf_level = NULL,
    df = NULL,
    equivalence = NULL,
    joint = FALSE,
    joint_test = "f",
    multcomp = FALSE,
    numderiv = "fdforward",
    ...) {
    call <- construct_call(model, "hypotheses")

    # Early validation and setup
    if (is.null(model)) model <- ...get("model_perturbed")
    if (is.null(model)) stop_sprintf("`model` is missing.")
    sanity_multcomp(multcomp, hypothesis, joint)

    # Early returns for special cases - removed mice check, moved later

    if (!isFALSE(joint)) {
        return(joint_test(
            object = model,
            joint_index = joint,
            joint_test = joint_test,
            hypothesis = hypothesis,
            df = df,
            vcov = vcov
        ))
    }

    # Handle marginaleffects objects vs regular models
    internal_classes <- c("predictions", "comparisons", "slopes", "hypotheses")
    if (inherits(model, internal_classes)) {
        if (!isTRUE(checkmate::check_flag(vcov))) {
            stop_sprintf(
                "The `vcov` argument is not available when `model` is a `predictions`, `comparisons`, `slopes`, or `hypotheses` object. Please specify the type of standard errors in the initial `marginaleffects` call."
            )
        }
        mfx <- components(model, "all")
        if (!is.null(mfx) && !is.null(mfx@draws)) {
            stop_sprintf(
                "The `hypotheses()` function cannot be used to post-process `marginaleffects` objects that include draws from a bootstrap, simulation, or bayesian posterior distribution. Users should call the `get_draws()` function and process draws manually."
            )
        }
        if (is.null(conf_level)) {
            conf_level <- mfx@conf_level
        }
    }

    # init
    # in this function we need to sanitize_model() later
    if (inherits(model, "marginaleffects_internal")) {
        mfx <- model
    } else {
        # hypotheses() supports raw data frames
        if (!inherits(model, c(internal_classes, "data.frame"))) {
            model <- sanitize_model(model, call = call, wts = wts, vcov = vcov, by = by, ...)
        }
        mfx <- new_marginaleffects_internal(
            call = call,
            model = model
        )
    }

    # multiple imputation - moved here after mfx object creation
    if (inherits(model, c("mira", "amest"))) {
        return(process_imputation(mfx))
    }

    mfx@conf_level <- sanitize_conf_level(conf_level, ...)
    hypothesis_is_formula <- isTRUE(checkmate::check_formula(hypothesis))

    # Setup degrees of freedom and hypothesis
    df <- if (is.null(df)) Inf else df
    mfx <- add_degrees_of_freedom(mfx = mfx, df = df, vcov = vcov)
    mfx <- add_hypothesis(mfx, hypothesis)
    mfx <- add_numderiv(mfx, numderiv)

    vcov_false <- isFALSE(vcov)
    if (!isTRUE(checkmate::check_matrix(vcov))) {
        mfx@vcov_type <- get_vcov_label(vcov = vcov)
        vcov <- get_vcov(model = model, vcov = vcov)
        mfx@vcov_model <- vcov
    } else {
        mfx@vcov_type <- "matrix"
    }

    b <- get_hypotheses(
        model_perturbed = model,
        hypothesis = mfx@hypothesis,
        hypothesis_is_formula = hypothesis_is_formula,
        ...
    )

    # Store hypothesis group-by variables in S4 slot
    hypothesis_by_vars <- attr(b, "hypothesis_function_by")
    if (!is.null(hypothesis_by_vars)) {
        mfx@variable_names_by_hypothesis <- hypothesis_by_vars
    }

    # bayesian posterior
    mfx@draws <- attr(b, "posterior_draws")
    if (!is.null(mfx@draws)) {
        J <- NULL
        se <- rep(NA, length(b))

        # standard errors via delta method
    } else if (!vcov_false && isTRUE(checkmate::check_matrix(vcov))) {
        args <- list(
            mfx = mfx,
            model_perturbed = model,
            vcov = vcov,
            hypothesis = mfx@hypothesis,
            FUN = get_hypotheses,
            hypothesis_is_formula = hypothesis_is_formula
        )
        if (...length() > 0) {
            args <- utils::modifyList(args, list(...))
        }
        se <- do_call(get_se_delta, args)
        J <- attr(se, "jacobian")
        mfx@jacobian <- J
        attr(se, "jacobian") <- NULL
        mfx@draws <- NULL

        # no standard error
    } else {
        J <- mfx@draws <- NULL
        se <- rep(NA, length(b))
    }

    # Build output data frame
    hyplab <- attr(b, "label")
    if (is.null(hyplab) && !is.null(mfx@hypothesis)) {
        hyplab <- attr(mfx@hypothesis, "label")
    }

    if (!is.null(hyplab) && nrow(hyplab) == length(b)) {
        out <- cbind(hyplab, data.frame(estimate = b, std.error = se))
    } else {
        term_name <- if (!is.null(mfx@hypothesis)) "custom" else paste0("b", seq_along(b))
        out <- data.frame(term = term_name, estimate = b, std.error = se)
    }

    # Remove std.error column when not computing st.errors and in bootstrap
    if (vcov_false) {
        out$std.error <- NULL
    }

    out[["group"]] <- attr(b, "grouplab")

    out <- get_ci(out, mfx)

    if (!is.null(equivalence)) {
        out <- equivalence(
            out,
            df = mfx@df,
            equivalence = equivalence,
            draws = mfx@draws
        )
    }

    out <- sort_columns(out)

    data.table::setDF(out)
    class(out) <- c("hypotheses", class(out))

    # Add common attributes from mfx S4 slots
    out <- add_attributes(out, mfx)
    out <- prune_attributes(out)

    # must be after attributes for vcov
    out <- multcomp_test(
        out,
        multcomp = multcomp,
        conf_level = mfx@conf_level,
        df = mfx@df
    )

    # Issue #1102: hypotheses() should not be called twice on the same object
    attr(out, "hypotheses_call") <- TRUE

    return(out)
}

config_get <- function(x) {
  config <- config_read()
  return(config[[x]])
}


config_set <- function(x, value) {
  config <- config_read()
  config[[x]] <- value
  dn <- tools::R_user_dir(package = "marginaleffects", which = "config")
  fn <- file.path(dn, "config.rds")
  saveRDS(config, fn)
}


config_delete <- function() {
  dn <- tools::R_user_dir(package = "marginaleffects", which = "config")
  fn <- file.path(dn, "config.rds")
  if (file.exists(fn)) hush(unlink(fn))
  message("`marginaleffects` returned to default settings.", call. = FALSE)
}


config_read <- function() {
  dn <- tools::R_user_dir(package = "marginaleffects", which = "config")
  if (!dir.exists(dn)) dir.create(dn, recursive = TRUE)
  fn <- file.path(dn, "config.rds")
  if (!file.exists(fn)) {
    config <- list()
  } else {
    config <- readRDS(fn)
  }
  return(invisible(config))
}

# Similar to match.call() but works better in lapply, etc.
# May be slow due to multiple calls to eval().
construct_call <- function(model, calling_function, env = parent.frame(1L)) {
    # minimal call
    out <- list(name = calling_function, model = model)

    # known arguments
    arg_names_all <- names(formals(
        get(calling_function, pos = asNamespace("marginaleffects")),
        asNamespace(utils::packageName())
    ))
    arg_names <- setdiff(arg_names_all, c("model", "..."))
    arg <- lapply(arg_names, function(i) {
        eval(substitute(substitute(arg), list(arg = as.name(i))), envir = env)
    })
    out <- c(out, stats::setNames(arg, arg_names))

    # ellipsis
    if ("..." %in% arg_names_all) {
        out <- c(out, as.list(substitute(alist(...), env))[-1L])
    }

    call_out <- do.call("call", out, quote = TRUE)

    # append marginaleffects:: to ensure function comes from package
    if (!startsWith(calling_function, paste0(utils::packageName(), "::"))) {
        call_out[[1L]] <- str2lang(paste0(
            utils::packageName(),
            "::",
            calling_function
        ))
    }

    call_out
}

marginaleffects_settings <- new.env()

settings_cache <- function(setti) {
    out <- list()
    for (s in setti) {
        out[[s]] <- settings_get(s)
    }
    return(out)
}

settings_restore <- function(cache) {
    for (n in names(cache)) {
        settings_set(n, cache[[n]])
    }
}

settings_init <- function(settings = NULL) {
    settings_rm()

    default_settings <- list(
        marginaleffects_safefun_return1 = FALSE
    )

    checkmate::assert_list(settings, null.ok = TRUE, names = "unique")

    if (!is.null(settings)) {
        settings <- c(settings, default_settings)
    }

    for (i in seq_along(settings)) {
        settings_set(names(settings)[i], settings[[i]])
    }
}


# https://stackoverflow.com/questions/79786610/solved-r-package-loading-error-onattach-failed-in-attachnamespace
settings_read_persistent <- function() {
    dn <- tools::R_user_dir(package = "marginaleffects", which = "config")
    if (!dir.exists(dn)) dir.create(dn, recursive = TRUE)
    fn <- file.path(dn, "config.rds")
    if (!file.exists(fn)) {
        config <- list()
    } else {
        config <- tryCatch(readRDS(fn), error = function(err) {
          warning("marginaleffects: loading persistence failed, this can happen when ",
            sQuote(file.path(dn, "config.rds"), FALSE), " becomes corrupted.\n",
            "  We suggest you delete this file and/or restore it from another location if able.\n",
            "  The load failure: ", conditionMessage(err))
          list()
        })
    }
    return(invisible(config))
}


settings_get <- function(name) {
    # First check in-memory settings
    if (name %in% names(marginaleffects_settings)) {
        return(get(name, envir = marginaleffects_settings))
    }

    # Then check persistent storage
    persistent_config <- settings_read_persistent()
    if (name %in% names(persistent_config)) {
        return(persistent_config[[name]])
    }

    return(NULL)
}

settings_set <- function(name, value, persistent = FALSE) {
    if (persistent) {
        # Save to persistent storage
        config <- settings_read_persistent()
        config[[name]] <- value
        dn <- tools::R_user_dir(package = "marginaleffects", which = "config")
        fn <- file.path(dn, "config.rds")
        saveRDS(config, fn)
    } else {
        # Save to in-memory storage
        assign(name, value = value, envir = marginaleffects_settings)
    }
}

settings_rm <- function(name = NULL) {
    if (is.null(name)) {
        rm(list = names(marginaleffects_settings), envir = marginaleffects_settings)
    } else if (name %in% names(marginaleffects_settings)) {
        rm(list = name, envir = marginaleffects_settings)
    }
}

settings_delete <- function() {
    dn <- tools::R_user_dir(package = "marginaleffects", which = "config")
    fn <- file.path(dn, "config.rds")
    if (file.exists(fn)) hush(unlink(fn))
    message("`marginaleffects` returned to default settings.", call. = FALSE)
}

settings_equal <- function(name, comparison) {
    k <- settings_get(name)
    if (!is.null(k) && length(comparison) == 1 && k == comparison) {
        out <- TRUE
    } else if (!is.null(k) && length(comparison) > 1 && k %in% comparison) {
        out <- TRUE
    } else {
        out <- FALSE
    }
    return(out)
}

#' @rdname set_coef
#' @export
set_coef.flexsurvreg <- function(model, coefs, ...) {
    out <- model
    out$res[, 1] <- coefs
    out$coefficients <- coefs
    return(out)
}


#' @rdname get_predict
#' @export
get_predict.flexsurvreg <- function(model, newdata, type, ...) {
    preds <- stats::predict(
        object = model,
        newdata = newdata,
        type = type,
        ...
    )

    if (ncol(preds) == 1L) {
        if (names(preds) == ".pred") {
            gp <- unlist(lapply(preds$.pred, function(x) {
                x[, 1, drop = TRUE]
            }))
            val <- unlist(lapply(preds$.pred, function(x) {
                x[, 2, drop = TRUE]
            }))

            out <- data.table(
                group = as.vector(gp),
                estimate = as.vector(val)
            )
            out$group <- group_to_factor(out$group, model)
            return(out)
        }
        out <- data.table(estimate = as.vector(preds[, 1, drop = TRUE]))
        return(out)
    }

    out <- data.table(
        group = as.vector(preds[, 1, drop = TRUE]),
        estimate = as.vector(preds[, 2, drop = TRUE])
    )
    out <- add_rowid(out, newdata)
    out$group <- group_to_factor(out$group, model)
    out
}

#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.tobit <- set_coef.default
# Note: needed because otherwise set.coef.survreg() would be called,
# which requires different treatment (see insight:::get_parameters.tobit()
# vs insight:::get_parameters.survreg())

add_attributes <- function(out, mfx = NULL, ...) {
    # Always add all attributes from S4 slots
    if (!is.null(mfx)) attr(out, "marginaleffects") <- mfx

    dots <- list(...)
    for (n in names(dots)) {
        if (is.null(attr(out, n))) {
            attr(out, n) <- dots[[n]]
        }
    }

    return(out)
}


prune_attributes <- function(out) {
    if (isTRUE(getOption("marginaleffects_lean", default = FALSE))) {
        out <- prune(out, component = "all")
    } else {
        out <- prune(out, component = "modeldata")
    }
    return(out)
}

bootstrap_rsample <- function(model, INF_FUN, ...) {

    # attached by `inferences()`
    conf_type <- attr(model, "inferences_conf_type")
    checkmate::assert_choice(conf_type, choices = c("perc", "bca"))

    # attached by `inferences()`
    conf_type <- attr(model, "inferences_conf_type")
    checkmate::assert_choice(conf_type, choices = c("perc", "norm", "basic", "bca"))

    # bootstrap using the original data and call
    modcall <- insight::get_call(model)
    modeldata <- get_modeldata(model, additional_variables = FALSE)
    data.table::setDF(modeldata)

    # evaluate the {marginaleffects} call to get output without inferences()
    # use ... because arguments are not the same for different {marginaleffects} functions
    dots <- list(...)
    dots[["vcov"]] <- FALSE

    # avoid recursion
    attr(model, "inferences_method") <- NULL
    out <- do.call(INF_FUN, c(list(model), dots))

    # default confidence level may be implicit in original call, but we need numeric
    if (is.null(dots[["conf_level"]])) {
        conf_level <- 0.95
    } else {
        conf_level <- dots[["conf_level"]]
    }

    bootfun <- function(data, ...) {
        modcall[["data"]] <- data
        modboot <- eval(modcall)
        modboot <- eval(modboot)
        args <- c(list(modboot, modeldata = data$data), dots)
        out <- do.call(INF_FUN, args)
        out <- tidy(out)
        # `rsample` averages by `term` columns; we don't use it anyway and assume things line up
        out$term <- seq_len(nrow(out))
        return(out)
    }
    args <- attr(model, "inferences_dots")
    args[["data"]] <- modeldata
    args[["apparent"]] <- TRUE # require for "bca"
    splits <- do.call(rsample::bootstraps, args)
    splits$estimates <- lapply(splits$splits, bootfun)

    if (isTRUE(conf_type == "bca")) {
        ci <- rsample::int_bca(
            splits,
            statistics = estimates,
            .fn = bootfun,
            alpha = 1 - conf_level)
    } else {
        ci <- rsample::int_pctl(
            splits,
            statistics = estimates,
            alpha = 1 - conf_level)
    }

    out$conf.low <- ci$.lower
    out$conf.high <- ci$.upper

    attr(out, "inferences") <- splits
    draws <- lapply(splits$estimates, function(x) as.matrix(x[, "estimate", drop = FALSE]))
    draws[[length(draws)]] <- NULL # apparent=TRUE appended the original estimates to the end
    draws <- do.call("cbind", draws)
    colnames(draws) <- NULL
    attr(out, "posterior_draws") <- draws
    return(out)
}

add_hypothesis <- function(mfx, hypothesis) {
    checkmate::assert(
        checkmate::check_character(hypothesis, pattern = "=|<=|>="),
        checkmate::check_numeric(hypothesis),
        checkmate::check_formula(hypothesis),
        checkmate::check_matrix(hypothesis),
        checkmate::check_function(hypothesis),
        checkmate::check_null(hypothesis)
    )

    hypothesis_direction <- "="
    hnull <- 0

    # Issue #1453: ratio ~ should use hypothesis=1 as null
    if (isTRUE(checkmate::check_formula(hypothesis))) {
        if (isTRUE(identical("ratio", as.character(hypothesis)[2]))) {
            hnull <- 1
        }
    }

    if (isTRUE(checkmate::check_character(hypothesis))) {
        # Extract direction for each hypothesis string
        hypothesis_direction <- rep("=", length(hypothesis))
        hypothesis_direction[grepl("<=", hypothesis)] <- "<="
        hypothesis_direction[grepl(">=", hypothesis)] <- ">="

        # Collapse to single value if all are the same
        if (length(unique(hypothesis_direction)) == 1) {
            hypothesis_direction <- hypothesis_direction[1]
        }

        if (isTRUE(grepl("^=|^<=|^>=", hypothesis))) {
            hypothesis <- sub("^=|^<=|^>=", "", hypothesis)
            hypothesis <- trimws(hypothesis)
            hypothesis <- as.numeric(hypothesis)
        }
    }

    if (isTRUE(checkmate::check_character(hypothesis, pattern = "=|<=|>="))) {
        out <- paste(gsub("=|<=|>=", "-(", hypothesis), ")")
        attr(out, "label") <- hypothesis
        hypothesis <- out
    } else if (isTRUE(checkmate::check_matrix(hypothesis))) {
        attr(hypothesis, "label") <- colnames(hypothesis)
    } else if (isTRUE(checkmate::check_numeric(hypothesis, len = 1))) {
        hnull <- hypothesis
        hypothesis <- NULL
    }

    mfx@hypothesis <- hypothesis
    mfx@hypothesis_null <- hnull
    mfx@hypothesis_direction <- hypothesis_direction

    return(mfx)
}

sort_columns <- function(x, newdata = data.frame(), by = NULL) {
    if (!inherits(x, "data.table")) {
        data.table::setDT(x)
    }
    if (isTRUE(checkmate::check_character(by))) {
        bycols <- by
    } else if (isTRUE(checkmate::check_data_frame(by))) {
        bycols <- colnames(by)
    } else {
        bycols <- NULL
    }
    stubcols <- c(
        "rowid", "rowidcf", "term", "group", "hypothesis", "by",
        grep("^contrast", colnames(x), value = TRUE),
        bycols,
        "estimate", "std.error", "statistic", "p.value", "s.value", "conf.low", "conf.high",
        attr(newdata, "newdata_variables_datagrid"),
        "marginaleffects_wts",
        sort(grep("^predicted", colnames(newdata), value = TRUE)))
    cols <- intersect(stubcols, colnames(x))
    cols <- unique(c(cols, colnames(x)))
    x <- x[, ..cols]
    if ("group" %in% names(x) && all(x$group == "main_marginaleffect")) {
        x$group <- NULL
    }
    return(x)
}

#' @include sanity_model.R
#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.DirichletRegModel <- function(model, ...) {
    coefs <- model$coefficients
    return(coefs)
}


#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.DirichletRegModel <- function(model, coefs, ...) {
    model$coefficients <- coefs
    return(model)
}


#' @include get_predict.R
#' @rdname get_predict
#' @export
get_predict.DirichletRegModel <- function(model, newdata, ...) {
    pred <- stats::predict(model, newdata = newdata, ...)
    if (is.matrix(pred)) {
        if (is.null(colnames(pred))) {
            if ("varnames" %in% names(model)) {
                colnames(pred) <- model$varnames
            } else {
                colnames(pred) <- seq_len(ncol(pred))
            }
        }
        out <- data.frame(
            group = rep(colnames(pred), each = nrow(pred)),
            estimate = as.vector(pred)
        )
        out$group <- group_to_factor(out$group, model)
    } else if (is.atomic(pred)) {
        out <- data.frame(estimate = as.vector(pred))
    } else {
        stop("Unexpected prediction output type from DirichletReg::predict().")
    }
    out <- add_rowid(out, newdata)
    return(out)
}

#' @rdname sanitize_model_specific
#' @export
sanitize_model_specific.DirichletRegModel <- function(
    model,
    calling_function = "marginaleffects",
    ...
) {
    if (isTRUE(getOption("marginaleffects_safe", default = TRUE))) {
        warning(
            paste(
                "DirichletReg model support is experimental and numerical results",
                "have not been tested extensively. Contribute test cases at:",
                "https://github.com/vincentarelbundock/marginaleffects/issues/1636"
            ),
            call. = FALSE
        )
    }
    return(model)
}

#' Extract Posterior Draws or Bootstrap Resamples from `marginaleffects` Objects
#'
#' @param x An object produced by a `marginaleffects` package function, such as `predictions()`, `avg_slopes()`, `hypotheses()`, etc.
#' @param shape string indicating the shape of the output format:
#' * "long" (default): long format data frame
#' * "PxD" (fastest): Matrix with parameters as rows and draws as columns
#' * "DxP": Matrix with draws as rows and parameters as columns
#' * "rvar": Random variable datatype (see `posterior` package documentation).
#' @return A data.frame with `drawid` and `draw` columns.
#' @details
#' If DxP and PxD and the names returned by `coef(x)` are unique, `marginaleffects` sets parameter names to those names. Otherwise, it sets them to `b1`, `b2`, etc.
#' @export
get_draws <- function(x, shape = "long") {
    checkmate::assert_choice(shape, choices = c("long", "DxP", "PxD", "rvar"))

    # tidy.comparisons() sometimes already saves draws in a nice long format
    mfx <- attr(x, "marginaleffects")
    draws <- mfx@draws

    if (is.null(draws)) {
        warning(
            "This object does not include draws. The `posterior_draws` function only supports bayesian models produced by the `marginaleffects` or `predictions` functions of the `marginaleffects` package.",
            call. = FALSE
        )
        return(invisible(NULL))
    }

    if (nrow(draws) != nrow(x)) {
        stop(
            "The number of parameters in the object does not match the number of parameters for which posterior draws are available.",
            call. = FALSE
        )
    }

    if (shape %in% c("PxD", "DxP")) {
        parnames <- names(stats::coef(x))
        if (length(parnames) != nrow(draws) || anyDuplicated(parnames) > 0) {
            parnames <- paste0("b", seq_len(nrow(draws)))
        }
        row.names(draws) <- parnames
        colnames(draws) <- paste0("draw", seq_len(ncol(draws)))
    }

    if (shape == "PxD") {
        return(draws)
    }

    if (shape == "DxP") {
        return(t(draws))
    }

    if (shape == "rvar") {
        insight::check_if_installed("posterior")
        draws <- t(draws)
        if (mfx@draws_chains > 0) {
            x[["rvar"]] <- posterior::rvar(draws, nchains = mfx@draws_chains)
        } else {
            x[["rvar"]] <- posterior::rvar(draws)
        }
        return(x)
    }

    if (shape == "long") {
        draws <- data.table(draws)
        setnames(draws, as.character(seq_len(ncol(draws))))
        for (v in colnames(x)) {
            draws[[v]] <- x[[v]]
        }
        out <- melt(
            draws,
            id.vars = colnames(x),
            variable.name = "drawid",
            value.name = "draw"
        )
        cols <- unique(c("drawid", "draw", "rowid", colnames(out)))
        cols <- intersect(cols, colnames(out))
        setcolorder(out, cols)
        data.table::setDF(out)
        return(out)
    }
}


average_draws <- function(data, index, draws, byfun = NULL) {
    insight::check_if_installed("collapse", minimum_version = "2.0.18")

    w <- data[["marginaleffects_wts_internal"]]
    if (all(is.na(w))) {
        w <- NULL
    }

    if (is.null(index)) {
        index <- intersect(colnames(data), "type")
    }

    if (length(index) > 0) {
        if ("term" %in% colnames(data)) {
            Encoding(data[["term"]]) <- "UTF-8"
        }
        g <- collapse::GRP(data, by = index)

        if (is.null(byfun)) {
            draws <- collapse::fmean(
                draws,
                g = g,
                w = w,
                drop = FALSE
            )
        } else {
            draws <- collapse::BY(
                draws,
                g = g,
                FUN = byfun,
                drop = FALSE
            )
        }
        out <- data.table(
            g[["groups"]],
            average = collapse::dapply(draws, MARGIN = 1, FUN = collapse::fmedian)
        )
    } else {
        if (is.null(byfun)) {
            draws <- collapse::fmean(
                draws,
                w = w,
                drop = FALSE
            )
        } else {
            draws <- collapse::BY(
                draws,
                g = g,
                FUN = byfun,
                drop = FALSE
            )
        }
        out <- data.table(
            average = collapse::dapply(draws, MARGIN = 1, FUN = collapse::fmedian)
        )
    }

    setnames(out, old = "average", new = "estimate")
    attr(out, "posterior_draws") <- draws
    return(out)
}


#' alias to `get_draws()` keep forever for backward compatibility with JSS
#'
#' @inherit posterior_draws
#' @keywords internal
#' @export
posterior_draws <- get_draws

get_contrasts <- function(model,
                          newdata,
                          type,
                          variables,
                          original,
                          lo,
                          hi,
                          wts = FALSE,
                          by = NULL,
                          hypothesis = NULL,
                          cross = FALSE,
                          verbose = TRUE,
                          deltamethod = FALSE,
                          ...) {

    settings_init()

    # some predict() methods need data frames and will convert data.tables
    # internally, which can be very expensive if done many times. we do it once
    # here.
    data.table::setDF(lo)
    data.table::setDF(hi)
    data.table::setDF(original)

    # brms models need to be combined to use a single seed when sample_new_levels="gaussian"
    if (inherits(model, c("brmsfit", "bart"))) {

        if (!"rowid" %in% colnames(lo)) {
            lo$rowid <- hi$rowid <- seq_len(nrow(lo))
        }

        both <- rbindlist(list(lo, hi))

        pred_both <- myTryCatch(get_predict(
            model,
            type = type,
            newdata = both,
            ...))

        # informative error in case of allow.new.levels level breakage
        if (inherits(pred_both[["error"]], "simpleError")) {
            insight::format_error(pred_both[["error"]][["message"]])
        } else {
            pred_both <- pred_both[["value"]]
        }

        data.table::setDT(pred_both)
        pred_both[, "lo" := seq_len(.N) <= .N / 2, by = "group"]

        pred_lo <- pred_both[pred_both$lo, .(rowid, group, estimate), drop = FALSE]
        pred_hi <- pred_both[!pred_both$lo, .(rowid, group, estimate), drop = FALSE]
        data.table::setDF(pred_lo)
        data.table::setDF(pred_hi)

        draws <- attr(pred_both, "posterior_draws")
        draws_lo <- draws[pred_both$lo, , drop = FALSE]
        draws_hi <- draws[!pred_both$lo, , drop = FALSE]

        attr(pred_lo, "posterior_draws") <- draws_lo
        attr(pred_hi, "posterior_draws") <- draws_hi

    } else {
        pred_lo <- myTryCatch(get_predict(
            model,
            type = type,
            newdata = lo,
            ...))

        # tidymodels
        if (inherits(pred_lo$error, "rlang_error") &&
            isTRUE(grepl("the object should be", pred_lo$error$message))) {
                insight::format_error(pred_lo$error$message)
        } else {
            pred_lo <- pred_lo[["value"]]
        }

        pred_hi <- myTryCatch(get_predict(
            model,
            type = type,
            newdata = hi,
            ...))

        # otherwise we keep the full error object instead of extracting the value
        if (inherits(pred_hi$value, "data.frame")) {
            pred_hi <- pred_hi$value
        } else {
            pred_hi <- pred_hi$error
        }

    }

    # predict() takes up 2/3 of the wall time. This call is only useful when we
    # compute elasticities, or for the main estimate, not for standard errors,
    # so we probably save 1/3 of that 2/3.
    elasticities <- c(
        # "dydx", # useless and expensive
        "eyex",
        "eydx",
        "dyex",
        # "dydxavg", # useless and expensive
        "eyexavg",
        "eydxavg",
        "dyexavg")
    fun <- function(x) {
        out <- checkmate::check_choice(x$comparison, choices = elasticities)
        isTRUE(out)
    }
    tmp <- Filter(fun, variables)
    if (!isTRUE(deltamethod) || length(tmp) > 0) {
        pred_or <- myTryCatch(get_predict(
            model,
            type = type,
            newdata = original,
            ...))[["value"]]
    } else {
        pred_or <- NULL
    }

    # lots of indexing later requires a data.table
    data.table::setDT(original)

    if (!inherits(pred_hi, "data.frame") || 
        !inherits(pred_lo, "data.frame") || 
        !inherits(pred_or, c("data.frame", "NULL")) ||
        all(is.na(pred_lo$estimate))) {
        msg <- "Unable to compute predicted values with this model. This error can arise when `insight::get_data()` is unable to extract the dataset from the model object, or when the data frame was modified since fitting the model. You can try to supply a different dataset to the `newdata` argument."
        if (inherits(pred_hi, c("try-error", "error"))) {
            msg <-c(msg, "", "In addition, this error message was raised:", "", as.character(pred_hi)) 
        }
        msg <- c(msg, "", "Bug Tracker: https://github.com/vincentarelbundock/marginaleffects/issues")
        insight::format_error(msg)
    }

    # output data.frame
    out <- pred_lo
    data.table::setDT(out)

    # univariate outcome:
    # original is the "composite" data that we constructed by binding terms and
    # compute predictions. It includes a term column, which we need to
    # replicate for each group.
    out[, "marginaleffects_wts_internal" := NA_real_] # default (probably almost always overwritten)
    mult <- nrow(out) / nrow(original)
    regex <- "^term$|rowid_dedup|^group$|^contrast|^marginaleffects_wts_internal$"
    if (isTRUE(mult == 1)) {
        for (v in grep(regex, colnames(original), value = TRUE)) {
            out[, (v) := original[[v]]]
        }

    # group or multivariate outcomes
    } else if (isTRUE(mult > 1)) {
        for (v in grep(regex, colnames(original), value = TRUE)) {
            out[, (v) := rep(original[[v]], times = mult)]
        }

    # cross-contrasts or weird cases
    } else {
        out <- merge(out, newdata, by = "rowid", all.x = TRUE, sort = FALSE)
        if (isTRUE(nrow(out) == nrow(lo))) {
            tmp <- data.table(lo)[, .SD, .SDcols = patterns("^contrast|marginaleffects_wts_internal")]
            out <- cbind(out, tmp)
            idx <- c("rowid", grep("^contrast", colnames(out), value = TRUE), colnames(out))
            idx <- unique(idx)
            out <- out[, ..idx]
        }
    }

    if (!"term" %in% colnames(out)) {
        out[, "term" := "cross"]
    }


    # by
    if (isTRUE(checkmate::check_data_frame(by))) {
        bycols <- "by"
        data.table::setDT(by)
        tmp <- setdiff(intersect(colnames(out), colnames(by)), "by")

        if (length(tmp) == 0) {
            if (all(colnames(by) %in% c("by", colnames(newdata)))) {
                nd <- c("rowid", "rowid_dedup", setdiff(colnames(by), "by"))
                nd <- intersect(nd, colnames(newdata))
                nd <- newdata[, ..nd, drop = FALSE]
                bycol <- intersect(c("rowid", "rowid_dedup"), colnames(nd))
                out <- merge(out, nd, by = bycol, sort = FALSE)
                tmp <- setdiff(intersect(colnames(out), colnames(by)), "by")
            } else {
                insight::format_error("The column in `by` must be present in `newdata`.")
            }
        }

        # harmonize column types
        for (v in colnames(by)) {
            if (isTRUE(is.character(out[[v]])) && isTRUE(is.numeric(by[[v]]))) {
                by[[v]] <- as.character(by[[v]])
            } else if (isTRUE(is.numeric(out[[v]])) && isTRUE(is.character(by[[v]]))) {
                by[[v]] <- as.numeric(by[[v]])
            }
        }
        out[by, by := by, on = tmp]
        by <- "by"

    } else if (isTRUE(by)) {
        regex <- "^term$|^contrast_?|^group$"
        by <- grep(regex, colnames(out), value = TRUE)
        by <- unique(by)

    } else if (isTRUE(checkmate::check_character(by))) {
        regex <- "^term$|^contrast_?|^group$"
        by <- c(by, grep(regex, colnames(out), value = TRUE))
        by <- unique(by)
    }

    # comparison function could be different for different terms
    # sanitize_variables() ensures all functions are identical when there are cross
    fun_list <- sapply(names(variables), function(x) variables[[x]][["function"]])
    fun_list[["cross"]] <- fun_list[[1]]

    # elasticity requires the original (properly aligned) predictor values
    # this will discard factor variables which are duplicated, so in principle
    # it should be the "correct" size
    # also need `x` when `x` is in the signature of the `comparison` custom function

    FUN <- function(z) {
        (is.character(z$comparison) && z$comparison %in% elasticities) ||
        (is.function(z$comparison) && "x" %in% names(formals(z$comparison)))
    }
    elasticities <- Filter(FUN, variables)
    elasticities <- lapply(elasticities, function(x) x$name)

    if (length(elasticities) > 0) {
        # assigning a subset of "original" to "idx1" takes time and memory
        # better to do this here for most columns and add the "v" column only
        # in the loop
        if (!is.null(original)) {
            idx1 <- c("rowid", "rowidcf", "term", "group", grep("^contrast", colnames(original), value = TRUE))
            idx1 <- intersect(idx1, colnames(original))
            idx1 <- original[, ..idx1]
        }

        for (v in names(elasticities)) {
            idx2 <- unique(c("rowid", "term", "group", by, grep("^contrast", colnames(out), value = TRUE)))
            idx2 <- intersect(idx2, colnames(out))
            # discard other terms to get right length vector
            idx2 <- out[term == v, ..idx2]
            # original is NULL when cross=TRUE
            if (!is.null(original)) {
                # if not first iteration, need to remove previous "v" and "elast"
                if (v %in% colnames(idx1)) {
                    idx1[, (v) := NULL]
                }
                if ("elast" %in% colnames(idx1)) {
                    idx1[, elast := NULL]
                }
                idx1[, (v) := original[[v]]]
                setnames(idx1, old = v, new = "elast")
                on_cols <- intersect(colnames(idx1), colnames(idx2))
                idx2 <- unique(merge(idx2, idx1, by = on_cols, sort = FALSE)[, elast := elast])
            }
            elasticities[[v]] <- idx2$elast
        }
    }

    draws <- attr(pred_lo, "posterior_draws")

    # frequentist
    if (is.null(draws)) {
        draws_lo <- draws_hi <- draws_or <- NULL

    # bayes
    } else {
        draws_lo <- attr(pred_lo, "posterior_draws")
        draws_hi <- attr(pred_hi, "posterior_draws")
        draws_or <- attr(pred_or, "posterior_draws")
    }

    data.table::setDT(pred_hi)

    out[, predicted_lo := pred_lo[["estimate"]]]
    out[, predicted_hi := pred_hi[["estimate"]]]

    if (!is.null(pred_or)) {
        data.table::setDT(pred_or)
        out[, predicted := pred_or[["estimate"]]]
    } else {
        out[, predicted := NA_real_]
    }

    idx <- grep("^contrast|^group$|^term$|^type$|^comparison_idx$", colnames(out), value = TRUE)

    # when `by` is a character vector, we sometimes modify the comparison
    # function on the fly to use the `avg` version.  this is important and
    # convenient because some of the statistics are non-collapsible, so we can't
    # average them at the very end.  when `by` is a data frame, we do this only
    # at the very end.
    # TODO: What is the UI for this? Doesn't make sense to have different functions.
    if (isTRUE(checkmate::check_character(by))) {
        tmp <- intersect(colnames(newdata), c(by, colnames(out)))
        if (length(tmp) > 1) {
            tmp <- subset(newdata, select = tmp)
            out <- merge(out, tmp, all.x = TRUE, sort = FALSE)
            idx <- unique(c(idx, by))
        }
    }

    # we feed these columns to safefun(), even if they are useless for categoricals
    if (!"marginaleffects_wts_internal" %in% colnames(out))  out[, "marginaleffects_wts_internal" := NA]

    # safe version of comparison
    # unknown arguments
    # singleton vs vector
    # different terms use different functions
    safefun <- function(hi, lo, y, n, term, cross, wts, tmp_idx, newdata) {
        tn <- term[1]
        eps <- variables[[tn]]$eps
        # when cross=TRUE, sanitize_comparison enforces a single function
        if (isTRUE(cross)) {
            fun <- fun_list[[1]]
        } else {
            fun <- fun_list[[tn]]
        }
        args <- list(
            "hi" = hi,
            "lo" = lo,
            "y" = y,
            "eps" = eps,
            "w" = wts,
            "newdata" = newdata)

        # sometimes x is exactly the same length, but not always
        args[["x"]] <- elasticities[[tn]][tmp_idx]

        args <- args[names(args) %in% names(formals(fun))]
        con <- try(do.call("fun", args), silent = TRUE)

        if (!isTRUE(checkmate::check_numeric(con, len = n)) && !isTRUE(checkmate::check_numeric(con, len = 1))) {
            msg <- sprintf("The function supplied to the `comparison` argument must accept two numeric vectors of predicted probabilities of length %s, and return a single numeric value or a numeric vector of length %s, with no missing value.", n, n) #nolint
            insight::format_error(msg)
        }
        if (length(con) == 1) {
            con <- c(con, rep(NA_real_, length(hi) - 1))
            settings_set("marginaleffects_safefun_return1", TRUE)
        }
        return(con)
    }

    # need a temp index for group-by operations when elasticities is a vector of length equal to full rows of `out`
    tmp <- grep("^term$|^contrast|^group$", colnames(out), value = TRUE)
    if (length(tmp) > 0) {
        out[, tmp_idx := 1:.N, by = tmp]
    } else {
        out[, tmp_idx := 1:.N]
    }

    # bayesian
    if (!is.null(draws)) {
        # drop missing otherwise get_averages() fails when trying to take a
        # simple mean
        idx_na <- !is.na(out$predicted_lo)
        out <- stats::na.omit(out, cols = "predicted_lo")

        # TODO: performance is probably terrrrrible here, but splitting is
        # tricky because grouping rows are not always contiguous, and the order
        # of rows is **extremely** important because draws don't have the
        # indices that would allow us to align them back with `out`
        draws <- draws[idx_na, , drop = FALSE]

        if (isTRUE(checkmate::check_character(by, min.len = 1))) {
            by_idx <- out[, ..by]
            by_idx <- do.call(paste, c(by_idx, sep = "|"))
        } else {
            by_idx <- out$term
        }

        # loop over columns (draws) and term names because different terms could use different functions
        for (tn in unique(by_idx)) {
            for (i in seq_len(ncol(draws))) {
                idx <- by_idx == tn
                draws[idx, i] <- safefun(
                    hi = draws_hi[idx, i],
                    lo = draws_lo[idx, i],
                    y = draws_or[idx, i],
                    n = sum(idx),
                    term = out$term[idx],
                    cross = cross,
                    wts = out$marginaleffects_wts_internal[idx],
                    tmp_idx = out$tmp_idx[idx],
                    newdata = newdata)
            }
        }

        # function returns unique value
        idx <- !is.na(draws[, 1])
        draws <- draws[idx, , drop = FALSE]

        # if comparison returns a single value, then we padded with NA. That
        # also means we don't want `rowid` otherwise we will merge and have
        # useless duplicates.
        if (!all(idx)) {
            if (settings_equal("marginaleffects_safefun_return1", TRUE)) {
                out[, "rowid" := NULL]
            }
            out <- out[idx, , drop = FALSE]
        }

        FUN_CENTER <- getOption("marginaleffects_posterior_center", default = stats::median)
        out[, "estimate" := apply(draws, 1, FUN_CENTER)]

    # frequentist
    } else {
        out <- stats::na.omit(out, cols = "predicted_lo")
        # We want to write the "estimate" column in-place because it safer
        # than group-merge; there were several bugs related to this in the past.
        # safefun() returns 1 value and NAs when the function retunrs a
        # singleton.
        idx <- intersect(idx, colnames(out))
        out[, "estimate" := safefun(
            hi = predicted_hi,
            lo = predicted_lo,
            y = predicted,
            n = .N,
            term = term,
            cross = cross,
            wts = marginaleffects_wts_internal,
            tmp_idx = tmp_idx,
            newdata = newdata),
        keyby = idx]
        out[, tmp_idx := NULL]

        # if comparison returns a single value, then we padded with NA. That
        # also means we don't want `rowid` otherwise we will merge and have
        # useless duplicates.
        if (anyNA(out$estimate)) {
            if (settings_equal("marginaleffects_safefun_return1", TRUE)) {
                out[, "rowid" := NULL]
            }
        }
        out <- stats::na.omit(out, cols = "estimate")
    }

    # clean
    if ("rowid_dedup" %in% colnames(out)) {
        out[, "rowid_dedup" := NULL]
    }


    # averaging by groups
    # sometimes this work is already done
    # if `by` is a column name, then we have merged-in a data frame earlier
    auto_mean_fun_sub <- any(grepl("^mean\\(", unique(out$contrast)))
    if (nrow(out) > 1) {
        if (!auto_mean_fun_sub && !(is.null(by) || isFALSE(by)) && any(grepl("^contrast[_]?", colnames(out)))) {
            out <- get_by(
                out,
                draws = draws,
                newdata = newdata,
                by = by,
                verbose = verbose)
            draws <- attr(out, "posterior_draws")
        } else {
            bycols <- c(by, "group", "term", "^contrast[_]?")
            bycols <- paste(bycols, collapse = "|")
            bycols <- grep(bycols, colnames(out), value = TRUE)
        }
    }

    # issue #531: uncertainty estimates from get_predict() sometimes get retained, but they are not overwritten later by get_ci()
    # drop by reference for speed
    bad <- intersect(
        colnames(out),
        c("conf.low", "conf.high", "std.error", "statistic", "p.value"))
    if (length(bad) > 0) {
        out[, (bad) := NULL]
    }

    # before get_hypothesis
    attr(out, "posterior_draws") <- draws

    # hypothesis tests using the delta method
    out <- get_hypothesis(out, hypothesis, by = by, newdata = original, draws = draws)

    # reset settings
    settings_rm("marginaleffects_safefun_return1")

    # output
    attr(out, "original") <- original
    return(out)
}

#' @rdname set_coef
#' @export
set_coef.hetprob <- function(model, coefs, ...) {
    model[["estimate"]][names(coefs)] <- coefs
    model
}

#' @rdname set_coef
#' @export
set_coef.ivpml <- set_coef.hetprob

# This function is very strict.
sanity_dots <- function(model, calling_function = NULL, ...) {
    stop_deprecate <- c()
    if ("p_adjust" %in% ...names()) stop_deprecate("p_adjust", "inferences()")
    if ("transform_post" %in% ...names()) stop_deprecate("transform_post", "transform")
    if ("interaction" %in% ...names()) stop_deprecate("interaction", "cross")

    if (identical(calling_function, "slopes")) {
        if ("comparison" %in% ...names()) stop_deprecate("comparison")
        if ("transform" %in% ...names()) stop_deprecate("transform")
        if ("cross" %in% ...names()) stop_deprecate("cross")
        # rest will be handled by sanity_dots() in comparisons()
        return(invisible(NULL))
    }

    valid <- list()

    # mixed effects
    me <- c("include_random", "re.form", "allow.new.levels", "random.only")
    valid[["merMod"]] <- me
    valid[["lmerMod"]] <- me
    valid[["glmerMod"]] <- me
    valid[["lmerModLmerTest"]] <- me

    # bayes
    valid[["brmsfit"]] <- c(
        "draw_ids",
        "incl_autocor",
        "nlpar",
        "ndraws",
        "re_formula",
        "allow_new_levels",
        "sample_new_levels",
        "dpar",
        "resp"
    )
    valid[["brmsfit_multiple"]] <- valid[["brmsfit"]]

    # misc
    valid[["selection"]] <- c("part") # sampleSelection
    valid[["glmmTMB"]] <- c("re.form", "allow.new.levels", "zitype") # glmmTMB
    valid[["bam"]] <- c("exclude", "discrete") # mgcv
    valid[["gam"]] <- c("exclude", "discrete") # mgcv
    valid[["rlmerMod"]] <- c("re.form", "allow.new.levels")
    valid[["gamlss"]] <- c("what", "safe") # gamlss
    valid[["lme"]] <- c("level") # nlme::lme
    valid[["bife"]] <- c("alpha_new", "corrected") # nlme::lme
    valid[["process_error"]] <- c("times", "p", "start")
    valid[["flexsurvreg"]] <- c("times", "p", "start")

    # survival
    valid[["survreg"]] <- c("p")

    # WeightIt models
    valid[["ordinal_weightit"]] <- valid[["multinom_weightit"]] <- "values"

    white_list <- c(
        "conf.int",
        "modeldata",
        "internal_call",
        "df",
        "transform",
        "comparison",
        "side",
        "delta",
        "null",
        "equivalence",
        "draw",
        "flag", # internal dev
        "variables_grid", # backward compatibility in marginal_means()
        "at" # topmodels procast
    )

    model_class <- class(model)[1]

    good <- NULL
    if (model_class %in% names(valid)) {
        good <- valid[[model_class]]
    }

    backward_compatibility <- c("conf.level")
    good <- c(good, backward_compatibility)

    bad <- setdiff(...names(), c(good, white_list))
    if (length(bad) > 0) {
        if (model_class %in% names(valid)) {
            msg <- sprintf(
                "These arguments are not known to be supported for models of class `%s`: %s. These arguments are known to be valid: %s. All arguments are still passed to the model-specific prediction function, but users are encouraged to check if the argument is indeed supported by their modeling package. Please file a request on Github if you believe that an unknown argument should be added to the `marginaleffects` white list of known arguments, in order to avoid raising this warning: https://github.com/vincentarelbundock/marginaleffects/issues",
                model_class,
                toString(bad),
                toString(valid[[model_class]])
            )
        } else {
            msg <- sprintf(
                "These arguments are not known to be supported for models of class `%s`: %s. All arguments are still passed to the model-specific prediction function, but users are encouraged to check if the argument is indeed supported by their modeling package. Please file a request on Github if you believe that an unknown argument should be added to the `marginaleffects` white list of known arguments, in order to avoid raising this warning: https://github.com/vincentarelbundock/marginaleffects/issues",
                model_class,
                toString(bad)
            )
        }
        warning(msg, call. = FALSE)
    }
}

sanity_reserved <- function(mfx) {
    modeldata <- mfx@modeldata
    # reserved keywords
    # Issue #697: we used to allow "group", as long as it wasn't in
    # `variables`, but this created problems with automatic `by=TRUE`. Perhaps
    # I could loosen this, but there are many interactions, and the lazy way is
    # just to forbid entirely.
    # NOTE: if no modeldata is available, we use `newdata`, but that often has a
    # `rowid` column. This used to break the extensions.Rmd vignette.
    reserved <- c(
        "group",
        "term",
        "contrast",
        "estimate",
        "std.error",
        "statistic",
        "conf.low",
        "conf.high",
        "p.value",
        "p.value.nonsup",
        "p.value.noninf",
        "by"
    )
    bad <- unique(intersect(names(modeldata), reserved))
    if (length(bad) > 0) {
        msg <- c(
            "These variable names are forbidden to avoid conflicts with the outputs of `marginaleffects`:",
            toString(dQuote(bad, NULL)),
            "Please rename your variables before fitting the model."
        )
        stop_sprintf(msg)
    }
    return(invisible(NULL))
}

#' Define class unions
#' @keywords internal
#' @noRd
setClassUnion("matrixOrNULL", c("matrix", "NULL"))

#' @keywords internal
#' @noRd
setClassUnion("numericOrNULL", c("numeric", "NULL"))

#' @keywords internal
#' @noRd
setClassUnion("characterOrNULL", c("character", "NULL"))

#' @keywords internal
#' @noRd
setClassUnion("logicalOrNULL", c("logical", "NULL"))

#' @keywords internal
#' @noRd
setClassUnion("functionOrNULL", c("function", "NULL"))

#' Internal S4 class for marginaleffects
#'
#' This S4 class is used internally to hold common arguments passed between
#' functions to simplify the function signatures and reduce redundant argument passing.
#'
#' @slot by Aggregation/grouping specification
#' @slot byfun Function for aggregation when using by
#' @slot call The original function call
#' @slot calling_function The name of the calling function (comparisons, predictions, hypotheses)
#' @slot comparison Comparison function specification
#' @slot cross Boolean flag for cross-contrasts
#' @slot df The degrees of freedom
#' @slot eps Epsilon value for numerical derivatives
#' @slot jacobian The jacobian matrix or NULL
#' @slot model The fitted model object
#' @slot modeldata The model data frame
#' @slot newdata The new data frame for predictions
#' @slot type The sanitized type from sanitize_type()
#' @slot vcov_model The variance-covariance matrix
#' @slot wts The weights specification
#' @keywords internal
setClass(
    "marginaleffects_internal",
    slots = c(
        by = "ANY",
        byfun = "functionOrNULL",
        call = "ANY",
        call_model = "ANY",
        calling_function = "character",
        comparison = "ANY",
        conf_level = "numeric",
        cross = "logicalOrNULL",
        df = "ANY",
        draws = "matrixOrNULL",
        draws_chains = "numeric",
        eps = "numericOrNULL",
        hypothesis = "ANY",
        hypothesis_null = "ANY",
        hypothesis_direction = "ANY",
        inferences = "ANY",
        jacobian = "matrixOrNULL",
        model = "ANY",
        modeldata = "ANY", # TODO: lmerTest returns nfnGroupedData
        modeldata_available = "logical",
        newdata = "ANY", # Changed from "data.frame" to handle mira deferred processing
        numderiv = "list",
        type = "characterOrNULL",
        variables = "list",
        variable_class = "characterOrNULL",
        variable_names_datagrid = "characterOrNULL",
        variable_names_predictors = "characterOrNULL",
        variable_names_response = "characterOrNULL",
        variable_names_by = "characterOrNULL",
        variable_names_by_hypothesis = "characterOrNULL",
        variable_names_wts = "characterOrNULL",
        vcov_model = "ANY",
        vcov_type = "characterOrNULL",
        wts = "ANY"
    )
)

#' Constructor for marginaleffects_internal class
#'
#' @param model The fitted model object (required)
#' @param call The original function call (required)
#' @param vcov_model The variance-covariance matrix
#' @param df The degrees of freedom
#' @param wts The weights specification
#' @param type The sanitized type from sanitize_type()
#' @return An object of class marginaleffects_internal
#' @keywords internal
new_marginaleffects_internal <- function(
    model,
    call,
    by = FALSE,
    byfun = NULL,
    comparison = NULL,
    conf_level = 0.95,
    cross = FALSE,
    df = NULL,
    draws = NULL,
    draws_chains = 0,
    eps = NULL,
    hypothesis = NULL,
    hypothesis_null = NULL,
    hypothesis_direction = NULL,
    jacobian = NULL,
    modeldata = NULL,
    numderiv = list("fdforward"),
    type = NULL,
    variables = list(),
    variable_names_by = character(),
    variable_names_by_hypothesis = character(),
    vcov_model = NULL,
    vcov_type = NULL,
    wts = NULL) {
    # For mice objects, modeldata handling is deferred to process_imputation()
    if (is.null(modeldata)) {
        if (inherits(model, c("mira", "amest"))) {
            modeldata <- data.frame() # placeholder - actual processing happens in process_imputation()
        } else {
            modeldata <- get_modeldata(model, additional_variables = TRUE)
        }
    }

    variable_class <- detect_variable_class(modeldata, model = model)

    variable_names_response <- hush(insight::find_response(model,
        combine = TRUE, component = "all", flatten = TRUE))

    if (is.null(variable_names_response)) {
        variable_names_response <- character(0)
    }

    variable_names_wts <- hush(insight::find_weights(model))

    variable_names_predictors <- hush(insight::find_predictors(model, flatten = TRUE, verbose = FALSE))

    # Extract calling function from call
    calling_function <- extract_calling_function(call)

    call_model <- tryCatch(insight::get_call(model), error = function(e) NULL)

    methods::new(
        "marginaleffects_internal",
        by = by,
        byfun = byfun,
        call = call,
        call_model = call_model,
        calling_function = calling_function,
        comparison = comparison,
        conf_level = conf_level,
        cross = cross,
        df = df,
        draws = draws,
        draws_chains = draws_chains,
        eps = eps,
        hypothesis = hypothesis,
        hypothesis_null = hypothesis_null,
        hypothesis_direction = hypothesis_direction,
        jacobian = jacobian,
        model = model,
        modeldata = modeldata,
        modeldata_available = TRUE,
        newdata = data.frame(),
        numderiv = numderiv,
        type = type,
        variables = variables,
        variable_class = variable_class,
        variable_names_by = variable_names_by,
        variable_names_by_hypothesis = variable_names_by_hypothesis,
        variable_names_datagrid = character(),
        variable_names_predictors = variable_names_predictors,
        variable_names_response = variable_names_response,
        variable_names_wts = variable_names_wts,
        vcov_model = vcov_model,
        vcov_type = vcov_type,
        wts = wts
    )
}

#' Extract calling function name from call object
#'
#' @param call The call object from which to extract the function name
#' @return Character string of the calling function ("comparisons", "predictions", or "hypotheses")
#' @keywords internal
extract_calling_function <- function(call) {
    if (is.call(call)) {
        func_name <- as.character(call[[1]])
        # Handle namespaced calls like marginaleffects::predictions
        if (length(func_name) > 1) {
            func_name <- func_name[length(func_name)]
        }
        # Map avg_ functions to their base function
        if (startsWith(func_name, "avg_")) {
            func_name <- sub("^avg_", "", func_name)
        }
        # Return the appropriate function name if it's one of the main functions
        if (func_name %in% c("comparisons", "predictions", "hypotheses")) {
            return(func_name)
        }
    }
    # Default fallback
    return("unknown")
}

#' Get a named model matrix
#'
#' @inheritParams slopes
#' @rdname get_model_matrix
#' @keywords internal
#' @export
get_model_matrix <- function(model, newdata, mfx = NULL) {
    UseMethod("get_model_matrix", model)
}


#' @rdname get_model_matrix
#' @keywords internal
#' @export
get_model_matrix.default <- function(model, newdata, mfx = NULL) {
    # some models require the response variable first value only allows us to
    # handle `newdata="balanced"` and friends. This is a hack, but it probably
    # doesn't matter. There can be many response variables, so we loop.
    if (!is.null(mfx)) {
        dv <- mfx@variable_names_response
        for (d in dv) {
            if (!d %in% colnames(newdata)) {
                if (nrow(newdata) == nrow(mfx@modeldata)) {
                    newdata[[d]] <- mfx@modeldata[[d]]
                } else {
                    # For different sized newdata, use first value as placeholder
                    newdata[[d]] <- mfx@modeldata[[d]][1]
                }
            }
        }
    }

    # faster
    if (class(model)[1] %in% c("lm", "glm")) {
        out <- stats::model.matrix(model, data = newdata)
        # more general
    } else {
        out <- hush(insight::get_modelmatrix(model, data = newdata))
    }

    beta <- get_coef(model)
    if (!isTRUE(nrow(out) == nrow(newdata)) || !isTRUE(ncol(out) == length(beta))) {
        return(NULL)
    } else {
        return(out)
    }
}


#' Add model matrix attribute to newdata
#' @param mfx marginaleffects object
#' @param newdata data frame to add attributes to
#' @keywords internal
#' @noRd
add_model_matrix_attribute <- function(mfx, newdata = NULL) {
    model <- mfx@model

    # predictions() only passes mfx; comparisons() passes mfx and hi/lo
    if (is.null(newdata)) {
        newdata <- mfx@newdata
    }

    if (nrow(newdata) == 0) {
        return(newdata)
    }

    # supported models (no inheritance)
    supported <- c("lm", "glm", "rq", "ols", "lrm", "ivreg")
    if (!isTRUE(class(model)[1] %in% supported)) {
        return(newdata)
    }

    # stats::model.matrix creates all-0 columns with splines::bs() and other functions
    funs <- grep("factor\\(|\\(Intercept|bs\\(", colnames(newdata), value = TRUE)
    if (length(funs) > 0) {
        return(newdata)
    }

    # we don't support offsets, so revert to stats::predict()
    if (!is.null(model[["offset"]])) {
        return(newdata)
    }

    # subset variables for listwise deletion
    vars <- unlist(mfx@variable_names_predictors, use.names = FALSE)
    vars <- c(vars, unlist(mfx@variable_names_response, use.names = FALSE))
    vars <- intersect(vars, colnames(newdata))

    nd <- as.data.frame(newdata)[, vars, drop = FALSE]

    MM <- hush(get_model_matrix(model, newdata = nd, mfx = mfx))

    attr(newdata, "marginaleffects_model_matrix") <- MM
    return(newdata)
}

#' @rdname get_predict
#' @export
get_predict.mlogit <- function(model, newdata, ...) {
    mat <- stats::predict(model, newdata = as.data.frame(newdata))
    if (isTRUE(checkmate::check_atomic_vector(mat))) {
        out <- data.table(group = names(mat), estimate = mat)
    } else {
        out <- data.table(
            # only for sorting; delete after
            marginaleffects_internal_rowid = rep(seq_len(nrow(mat)), rep = ncol(mat)),
            group = rep(colnames(mat), each = nrow(mat)),
            estimate = as.vector(mat)
        )
    }
    setkey(out, marginaleffects_internal_rowid, group)
    # only for sorting; delete after
    out[, marginaleffects_internal_rowid := NULL]
    if ("term" %in% colnames(newdata)) {
        out[, "term" := newdata[["term"]]]
    }
    # do not convert to factor because DV is often "yes" or "no" while the "group" is outcome levels.
    # out$group <- group_to_factor(out$group, model)
    out <- add_rowid(out, newdata)
    return(out)
}


#' @include sanity_model.R
#' @rdname sanitize_model_specific
#' @keywords internal
sanitize_model_specific.mlogit <- function(model, calling_function = NULL, ...) {
    if (!is.null(calling_function) && !identical(calling_function, "predictions")) {
        msg <- "`predictions()` and `avg_predictions()` are supported for `mlogit` models, but the other `marginaleffects` functions are not. The reason is that the data structure for these models is one observation-choice per row, which complicates internal handling by `marginaleffects`."
        stop_sprintf(msg)
    }
    return(model)
}

get_unique_index <- function(x, term_only = FALSE) {
    idx <- c("term", "contrast", grep("^contrast_", colnames(x), value = TRUE))
    idx <- intersect(unique(idx), colnames(x))
    if (length(idx) == 0) {
        return(NULL)
    }
    if (length(idx) == 1) {
        return(x[[idx]])
    }
    out <- x[, idx, drop = FALSE]
    for (i in ncol(out):2) {
        if (length(unique(out[[i]])) == 1) {
            out[[i]] <- NULL
        }
    }
    out <- apply(out, 1, toString)
    return(out)
}


warn_once <- function(msg, id) {
    if (!isTRUE(getOption(id, default = TRUE)) || !isTRUE(getOption("marginaleffects_safe", default = TRUE))) {
        return(invisible())
    }
    msg <- paste(msg, "This warning appears once per session.")
    # warn_sprintf(msg, call. = FALSE)
    warning(msg, call. = FALSE)
    opts <- list(FALSE)
    names(opts) <- id
    options(opts)
}


# Cross join a list of data tables
# Source: https://github.com/Rdatatable/data.table/issues/1717#issuecomment-545758165
cjdt <- function(dtlist) {
    Reduce(
        function(DT1, DT2) cbind(DT1, DT2[rep(seq_len(.N), each = nrow(DT1))]),
        dtlist
    )
}


# recurse up. mostly useful for `tinytest`
# this is dumb, but it's late and i don't feel like thinking about this
evalup <- function(xcall) {
    out <- myTryCatch(eval(xcall))
    if (inherits(out$error, "simpleError")) {
        msg <- out$error$message
        out <- NULL
    } else {
        msg <- NULL
        out <- out$value
    }
    for (i in 1:10) {
        if (is.null(out)) {
            out <- hush(eval(xcall, parent.frame(i)))
        }
    }
    if (is.null(out) && !is.null(msg)) {
        stop(msg)
    }
    return(out)
}


merge_by_rowid <- function(x, y) {
    # return data
    # very import to avoid sorting, otherwise bayesian draws won't fit predictions
    # merge only with rowid; not available for hypothesis
    mergein <- setdiff(colnames(y), colnames(x))
    if ("rowid" %in% colnames(x) && "rowid" %in% colnames(y) && length(mergein) > 0) {
        idx <- c("rowid", mergein)
        if (!data.table::is.data.table(y)) {
            data.table::setDT(y)
            tmp <- y[, ..idx]
        } else {
            tmp <- y[, ..idx]
        }
        # TODO: this breaks in mclogit. maybe there's a more robust merge
        # solution for weird grouped data. But it seems fine because
        # `predictions()` output does include the original predictors.
        out <- tryCatch(
            merge(x, tmp, by = "rowid", sort = FALSE),
            error = function(e) x
        )
    } else {
        out <- x
    }
    return(out)
}

# faster than all(x %in% 0:1)
is_binary <- function(x) {
    isTRUE(checkmate::check_integerish(
        x,
        null.ok = TRUE,
        upper = 1,
        lower = 0,
        any.missing = TRUE,
        all.missing = FALSE
    ))
}


sub_named_vector <- function(x, y) {
    # issue 1005
    xlab <- gsub("^`|`$", "", names(x))
    ylab <- gsub("^`|`$", "", names(y))

    idx <- match(ylab, xlab)
    if (length(stats::na.omit(idx)) > 0) {
        x[stats::na.omit(idx)] <- y[!is.na(idx)]
    } else if (length(y) == length(x)) {
        return(y)
    } else {
        stop(
            "set_coef() substitution error. Please report on Github with a reproducible example: https://github.com/vincentarelbundock/marginaleffects/issues",
            call. = FALSE
        )
    }

    return(x)
}


group_to_factor <- function(group, model) {
    dv <- try(insight::get_response(model), silent = TRUE)
    if (inherits(dv, "factor")) {
        if (length(unique(group)) == nlevels(dv)) {
            group <- factor(group, levels(dv))
        }
    }
    return(group)
}


...get <- function(x, ifnotfound = NULL) {
    eval(
        quote(
            if (!anyNA(.m1 <- match(.x, ...names())) && !is.null(.m2 <- ...elt(.m1))) {
                .m2
            } else {
                .ifnotfound
            }),
        pairlist(.x = x[1L], .ifnotfound = ifnotfound),
        parent.frame(1L)
    )
}


...mget <- function(x) {
    found <- match(x, eval(quote(...names()), parent.frame(1L)))
    not_found <- is.na(found)
    if (all(not_found)) {
        return(list())
    }
    stats::setNames(
        lapply(found[!not_found], function(z) {
            eval(
                quote(...elt(.z)),
                pairlist(.z = z),
                parent.frame(3L)
            )
        }),
        x[!not_found]
    )
}

stop_deprecate <- function(old, new = NULL) {
    if (is.null(new)) {
        msg <- sprintf("The `%s` argument is not supported.", old)
    } else {
        msg <- sprintf("The `%s` argument is not supported. Please use `%s` instead.", old, new)
    }
    stop(msg, call. = FALSE)
}


stop_sprintf <- function(msg, ...) {
    dots <- list(...)
    if (length(dots) > 0) {
        msg <- sprintf(msg, ...)
    }
    stop(msg, call. = FALSE)
}


warn_sprintf <- function(msg, ...) {
    if (!isTRUE(getOption("marginaleffects_safe", default = TRUE))) {
        return(invisible())
    }
    dots <- list(...)
    if (length(dots) > 0) {
        msg <- sprintf(msg, ...)
    }
    warning(msg, call. = FALSE)
}


# Like do.call() but avoids inlining argument values into the call expression.
# When do.call(fn, list(huge_model, huge_data)) errors, R stores the entire
# inlined data in sys.calls(), and IDEs (RStudio/Positron) try to deparse it,
# causing multi-minute hangs. This builds a call with symbol references instead.
# See: https://github.com/vincentarelbundock/marginaleffects/issues/1663
do_call <- function(what, args) {
    call_env <- new.env(parent = parent.frame())
    call_env[[".what"]] <- what
    arg_names <- names(args) %||% rep("", length(args))
    call_list <- vector("list", length(args) + 1L)
    call_list[[1L]] <- as.symbol(".what")
    for (i in seq_along(args)) {
        sym_name <- sprintf(".arg%d", i)
        call_env[[sym_name]] <- args[[i]]
        call_list[[i + 1L]] <- as.symbol(sym_name)
    }
    names(call_list) <- c("", arg_names)
    eval(as.call(call_list), call_env)
}

equivalence <- function(x, equivalence = NULL, df = Inf, draws = NULL, ...) {
    if (is.null(equivalence)) {
        return(x)
    }

    if (!is.null(draws)) {
        # `draws` is an Parameters x Draws matrix.

        # Pr(Equivalence)
        idx_equiv <- draws > equivalence[1] & draws < equivalence[2]
        x$p.rope.unconditional <- rowMeans(idx_equiv)

        # Pr(Equivalence|CI)
        rope <- draws
        rope[rope < x$conf.low | rope > x$conf.high] <- NA
        rope <- rope > equivalence[1] & rope < equivalence[2]
        x$p.rope.conditional <- rowMeans(rope, na.rm = TRUE)

        return(x)
    }

    if (!all(c("estimate", "std.error") %in% colnames(x))) {
        msg <- "The `equivalence` argument is not supported with models for which `marginaleffects` does not estimate a standard error."
        stop_sprintf(msg)
    }

    checkmate::assert_numeric(equivalence, min.len = 1, max.len = 2)
    if (length(equivalence) == 1) {
        equivalence <- c(equivalence, equivalence)
    }

    delta <- abs(diff(equivalence)) / 2

    # definitions from `emmeans`, with a different user interface based on symmetric "equivalence"
    x$statistic.noninf <- (x$estimate - equivalence[1]) / x$std.error
    x$statistic.nonsup <- (x$estimate - equivalence[2]) / x$std.error

    ## keep this in case we return to the emmeans-style user interface
    # x$statistic.inf <- (x$estimate - null + delta) / x$std.error
    # x$statistic.sup <- (x$estimate - null - delta) / x$std.error

    if (is.infinite(df)) {
        x$p.value.noninf <- stats::pnorm(x$statistic.noninf, lower.tail = FALSE)
        x$p.value.nonsup <- stats::pnorm(x$statistic.nonsup, lower.tail = TRUE)
    } else {
        x$p.value.noninf <- stats::pt(
            x$statistic.noninf,
            lower.tail = FALSE,
            df = x[["df"]]
        )
        x$p.value.nonsup <- stats::pt(
            x$statistic.nonsup,
            lower.tail = TRUE,
            df = x[["df"]]
        )
    }
    x$p.value.equiv <- pmax(x$p.value.nonsup, x$p.value.noninf)

    return(x)
}

#' @rdname get_coef
#' @export
get_coef.brmultinom <- function(model, ...) {
    out <- insight::get_parameters(model)
    out <- stats::setNames(out$Estimate,
                           sprintf("%s:%s", out$Response, out$Parameter))
    return(out)
}


#' @include methods_nnet.R
#' @rdname get_predict
#' @export
get_predict.brmultinom <- get_predict.multinom


#' @include get_group_names.R
#' @include methods_nnet.R
#' @rdname get_group_names
#' @export
get_group_names.bracl <- get_group_names.multinom


#' @rdname get_coef
#' @export
get_coef.bracl <- function(model, ...) {
    stats::coef(model)
}

process_imputation <- function(mfx) {
    insight::check_if_installed("mice")

    model <- mfx@model
    mfxcall <- mfx@call

    micedata <- tryCatch(get_modeldata_mids(model), error = function(e) NULL)

    # issue #1269: transforms must be applied after pooling
    if ("transform" %in% names(mfxcall)) {
        tr <- eval.parent(mfxcall[["transform"]])
        mfxcall[["transform"]] <- NULL
    } else {
        tr <- NULL
    }

    if (inherits(model, "mira")) {
        modellist <- model$analyses
    } else if (inherits(model, "amest")) {
        # amest already in the right format
        modellist <- model
    } else {
        stop_sprintf("MI class not implemented yet.")
    }

    mfxlist <- vector("list", length(modellist))
    for (i in seq_along(modellist)) {
        calltmp <- mfxcall
        calltmp[["model"]] <- modellist[[i]]

        # Handle deferred newdata processing for calls like subset(treat == 1)
        if ("newdata" %in% names(calltmp) && rlang::is_call(calltmp[["newdata"]])) {
            newdata_call <- calltmp[["newdata"]]
            if (rlang::call_name(newdata_call) == "subset") {
                # Add the model data as the 'x' argument for subset
                if (!"x" %in% rlang::call_args_names(newdata_call)) {
                    newdata_call <- rlang::call_modify(newdata_call, x = micedata[[i]])
                }
            } else if (rlang::call_name(newdata_call) == "filter") {
                # Add the model data as the '.data' argument for filter
                if (!".data" %in% rlang::call_args_names(newdata_call)) {
                    newdata_call <- rlang::call_modify(newdata_call, .data = micedata[[i]])
                }
            }
            # Evaluate the newdata call with the individual model's data
            calltmp[["newdata"]] <- eval(newdata_call)
        }

        mfxlist[[i]] <- eval.parent(calltmp)

        # save the S4 structure of a single model to host the eventual output
        if (i == 1) {
            out <- mfxlist[[i]]
        }

        # needed for mice::pool(); after saving `out`
        mfxlist[[i]]$term <- seq_len(nrow(mfxlist[[i]]))

        # Needed for mice::pool() to get dfcom, even when lean = TRUE
        attr(mfxlist[[i]], "marginaleffects")@model <- modellist[[i]]
        class(mfxlist[[i]]) <- c("marginaleffects_mids", class(mfxlist[[i]]))
    }

    mipool <- mice::pool(mfxlist)

    ti <- mice::tidy(mipool)
    for (col in c("estimate", "std.error", "statistic", "p.value", "conf.low", "conf.high", "df")) {
        if (col %in% colnames(ti)) {
            out[[col]] <- ti[[col]]
        } else {
            out[[col]] <- NULL
        }
    }

    # Extract confidence level with fallback
    conf_level <- mfxcall[["conf_level"]]
    if (is.null(conf_level)) {
        conf_level <- 0.95 # Default fallback
    }

    out$df <- mipool$pooled$df

    # TODO: change this once we store the metadata in its final place
    out <- get_ci(out, attr(mfxlist[[1]], "marginaleffects"))

    out <- backtransform(out, sanitize_transform(tr))

    return(out)
}

#' tidy helper
#'
#' @noRd
#' @export
tidy.marginaleffects_mids <- function(x, ...) {
    if (!"std.error" %in% colnames(x)) {
        stop_sprintf(
            'The output does not include a `std.error` column. Some models do not generate standard errors when estimates are backtransformed (e.g., GLM models). One solution is to use `type="response"` for those models.'
        )
    }
    out <- as.data.frame(x[, c("estimate", "std.error")])
    out$term <- seq_len(nrow(out))
    return(out)
}


#' glance helper
#'
#' @noRd
#' @export
glance.marginaleffects_mids <- function(x, ...) {
    data.frame()
}

#' @rdname set_coef
#' @export
set_coef.rlmerMod <- function(model, coefs, ...) {
    model@beta <- coefs
    model
}


#' @rdname get_predict
#' @export
get_predict.rlmerMod <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    mfx = NULL,
    ...) {
    args <- list(...)
    # some predict methods raise warnings on unused arguments
    unused <- c(
        "type",
        "normalize_dydx",
        "step_size",
        "numDeriv_method",
        "conf_level",
        "conf.level",
        "internal_call",
        "return_format",
        "vcov",
        "eps",
        "modeldata",
        "marginaleffects",
        "mfx"
    )
    args <- args[setdiff(names(args), unused)]
    args[["object"]] <- model
    args[["newdata"]] <- newdata
    out <- data.table(estimate = do_call(stats::predict, args))
    out <- add_rowid(out, newdata)
    return(out)
}

#' Slopes (aka Partial derivatives, Marginal Effects, or Trends)
#'
#' @description
#' Partial derivative of the regression equation with respect to a regressor of interest.
#'
#' * `slopes()`: unit-level (conditional) estimates.
#' * `avg_slopes()`: average (marginal) estimates.
#'
#' The `newdata` argument and the `datagrid()` function can be used to control where statistics are evaluated in the predictor space: "at observed values", "at the mean", "at representative values", etc.
#'
#' See the slopes vignette and package website for worked examples and case studies:
#'
#' * [https://marginaleffects.com/chapters/slopes.html](https://marginaleffects.com/chapters/slopes.html)
#' * [https://marginaleffects.com/](https://marginaleffects.com/)
#'
#' @details
#' A "slope" or "marginal effect" is the partial derivative of the regression equation
#' with respect to a variable in the model. This function uses automatic
#' differentiation to compute slopes for a vast array of models,
#' including non-linear models with transformations (e.g., polynomials).
#' Uncertainty estimates are computed using the delta method.
#'
#' @param model Model object
#' @param variables Focal variables
#' * `NULL`: compute slopes or comparisons for all the variables in the model object (can be slow).
#' * Character vector: subset of variables (usually faster).
#' @param newdata Grid of predictor values at which we evaluate the slopes.
#' + Warning: Please avoid modifying your dataset between fitting the model and calling a `marginaleffects` function. This can sometimes lead to unexpected results.
#' + `NULL` (default): Unit-level slopes for each observed value in the dataset (empirical distribution). The dataset is retrieved using [insight::get_data()], which tries to extract data from the environment. This may produce unexpected results if the original data frame has been altered since fitting the model.
#' + [datagrid()] call to specify a custom grid of regressors. For example:
#'   - `newdata = datagrid(cyl = c(4, 6))`: `cyl` variable equal to 4 and 6 and other regressors fixed at their means or modes.
#'   - See the Examples section and the [datagrid()] documentation.
#' + [subset()] call with a single argument to select a subset of the dataset used to fit the model, ex: `newdata = subset(treatment == 1)`
#' + [dplyr::filter()] call with a single argument to select a subset of the dataset used to fit the model, ex: `newdata = filter(treatment == 1)`
#' + string:
#'   - "mean": Slopes evaluated when each predictor is held at its mean or mode.
#'   - "median": Slopes evaluated when each predictor is held at its median or mode.
#'   - "balanced": Slopes evaluated on a balanced grid with every combination of categories and numeric variables held at their means.
#'   - "tukey": Slopes evaluated at Tukey's 5 numbers.
#'   - "grid": Slopes evaluated on a grid of representative numbers (Tukey's 5 numbers and unique values of categorical predictors).
#' @param vcov Type of uncertainty estimates to report (e.g., for robust standard errors). Acceptable values:
#'  * FALSE: Do not compute standard errors. This can speed up computation considerably.
#'  * TRUE: Unit-level standard errors using the default `vcov(model)` variance-covariance matrix.
#'  * String which indicates the kind of uncertainty estimates to return.
#'    - Heteroskedasticity-consistent: `"HC"`, `"HC0"`, `"HC1"`, `"HC2"`, `"HC3"`, `"HC4"`, `"HC4m"`, `"HC5"`. See `?sandwich::vcovHC`
#'    - Heteroskedasticity and autocorrelation consistent: `"HAC"`
#'    - Mixed-Models degrees of freedom: "satterthwaite", "kenward-roger"
#'    - Other: `"NeweyWest"`, `"KernHAC"`, `"OPG"`. See the `sandwich` package documentation.
#'  * One-sided formula which indicates the name of cluster variables (e.g., `~unit_id`). This formula is passed to the `cluster` argument of the `sandwich::vcovCL` function.
#'  * Square covariance matrix
#'  * Function which returns a covariance matrix (e.g., `stats::vcov(model)`)
#' @param conf_level numeric value between 0 and 1. Confidence level to use to build a confidence interval.
#' @param type string indicates the type (scale) of the predictions used to
#' compute contrasts or slopes. This can differ based on the model
#' type, but will typically be a string such as: "response", "link", "probs",
#' or "zero". When an unsupported string is entered, the model-specific list of
#' acceptable values is returned in an error message. When `type` is `NULL`, the
#' first entry in the error message is used by default.
#' @param slope string indicates the type of slope or (semi-)elasticity to compute:
#' - "dydx": dY/dX
#' - "eyex": dY/dX * Y / X
#' - "eydx": dY/dX * Y
#' - "dyex": dY/dX / X
#' - Y is the predicted value of the outcome; X is the observed value of the predictor.
#' @param wts logical, string or numeric: weights to use when computing average predictions, contrasts or slopes. These weights only affect the averaging in `avg_*()` or with the `by` argument, and not unit-level estimates. See `?weighted.mean`
#' + string: column name of the weights variable in `newdata`. When supplying a column name to `wts`, it is recommended to supply the original data (including the weights variable) explicitly to `newdata`.
#' + numeric: vector of length equal to the number of rows in the original data or in `newdata` (if supplied).
#' + FALSE: Equal weights.
#' + TRUE: Extract weights from the fitted object with `insight::find_weights()` and use them when taking weighted averages of estimates. Warning: `newdata=datagrid()` returns a single average weight, which is equivalent to using `wts=FALSE`
#' @param hypothesis specify a hypothesis test or custom contrast using a number , formula, string equation, vector, matrix, or function.
#' + Number: The null hypothesis used in the computation of Z and p (before applying `transform`).
#' + String: Equation to specify linear or non-linear hypothesis tests. If the terms in `coef(object)` uniquely identify estimates, they can be used in the formula. Otherwise, use `b1`, `b2`, etc. to identify the position of each parameter. The `b*` wildcard can be used to test hypotheses on all estimates. If a named vector is used, the names are used as labels in the output. Examples:
#'   - `hp = drat`
#'   - `hp + drat = 12`
#'   - `b1 + b2 + b3 = 0`
#'   - `b* / b1 = 1`
#' + Formula: `lhs ~ rhs | group`
#'   + `lhs`
#'     - `ratio`
#'     - `difference`
#'     - Leave empty for default value
#'   + `rhs`
#'     - `pairwise` and `revpairwise`: pairwise differences between estimates in each row.
#'     - `reference`: differences between the estimates in each row and the estimate in the first row.
#'     - `sequential`: difference between an estimate and the estimate in the next row.
#'     - `meandev`: difference between an estimate and the mean of all estimates.
#'     - `meanotherdev: difference between an estimate and the mean of all other estimates, excluding the current one.
#'     - `poly`: polynomial contrasts, as computed by the `stats::contr.poly()` function.
#'     - `helmert`: Helmert contrasts, as computed by the `stats::contr.helmert()` function. Contrast 2nd level to the first, 3rd to the average of the first two, and so on.
#'     - `trt_vs_ctrl`: difference between the mean of estimates (except the first) and the first estimate.
#'     - `I(fun(x))`: custom function to manipulate the vector of estimates `x`. The function `fun()` can return multiple (potentially named) estimates.
#'   + `group` (optional)
#'     - Column name of `newdata`. Conduct hypothesis tests withing subsets of the data.
#'   + Examples:
#'      + `~ poly`
#'      + `~ sequential | groupid`
#'      + `~ reference`
#'      + `ratio ~ pairwise`
#'      + `difference ~ pairwise | groupid`
#'      + `~ I(x - mean(x)) | groupid`
#'      + `~ I(\(x) c(a = x[1], b = mean(x[2:3]))) | groupid`
#' + Matrix or Vector: Each column is a vector of weights. The the output is the dot product between these vectors of weights and the vector of estimates. The matrix can have column names to label the estimates.
#' + Function:
#'   - Accepts an argument `x`: object produced by a `marginaleffects` function or a data frame with column `rowid` and `estimate`
#'   - Returns a data frame with columns `term` and `estimate` (mandatory) and `rowid` (optional).
#'   - The function can also accept optional input arguments: `newdata`, `by`, `draws`.
#'   - This function approach will not work for Bayesian models or with bootstrapping. In those cases, it is easy to use `get_draws()` to extract and manipulate the draws directly.
#' + See the Examples section below and the vignette: [https://marginaleffects.com/chapters/hypothesis.html](https://marginaleffects.com/chapters/hypothesis.html)
#' @param df Degrees of freedom used to compute p values and confidence intervals. A single numeric value between 1 and `Inf`. When `df` is `Inf`, the normal distribution is used. When `df` is finite, the `t` distribution is used. See [insight::get_df] for a convenient function to extract degrees of freedom. Ex: `slopes(model, df = insight::get_df(model))`
#' @param eps NULL or numeric value which determines the step size to use when
#' calculating numerical derivatives: (f(x+eps)-f(x))/eps. When `eps` is
#' `NULL`, the step size is 0.0001 multiplied by the difference between
#' the maximum and minimum values of the variable with respect to which we
#' are taking the derivative. Changing `eps` may be necessary to avoid
#' numerical problems in certain models.
#' @param numderiv string or list of strings indicating the method to use to for the numeric differentiation used in to compute delta method standard errors.
#' + "fdforward": finite difference method with forward differences
#' + "fdcenter": finite difference method with central differences (default)
#' + "richardson": Richardson extrapolation method
#' + Extra arguments can be specified by passing a list to the `numDeriv` argument, with the name of the method first and named arguments following, ex: `numderiv=list("fdcenter", eps = 1e-5)`. When an unknown argument is used, `marginaleffects` prints the list of valid arguments for each method.
#' @param ... Additional arguments are passed to the `predict()` method
#' supplied by the modeling package.These arguments are particularly useful
#' for mixed-effects or bayesian models (see the online vignettes on the
#' `marginaleffects` website). Available arguments can vary from model to
#' model, depending on the range of supported arguments by each modeling
#' package. See the "Model-Specific Arguments" section of the
#' `?slopes` documentation for a non-exhaustive list of available
#' arguments.
#' @inheritParams comparisons
#'
#' @details
#' Numerical derivatives for the `slopes` function are calculated
#' using a simple epsilon difference approach: \eqn{\partial Y / \partial X = (f(X + \varepsilon/2) - f(X-\varepsilon/2)) / \varepsilon}{dY/dX = (f(X + e/2) - f(X-e/2)) / e},
#' where f is the `predict()` method associated with the model class, and
#' \eqn{\varepsilon}{e} is determined by the `eps` argument.
#' @template references
#' @template deltamethod
#' @template model_specific_arguments
#' @template bayesian
#' @template equivalence
#' @template type
#' @template parallel
#' @template order_of_operations
#' @template options
#' @template return
#' @examplesIf interactive() || isTRUE(Sys.getenv("R_DOC_BUILD") == "true")
#' # Unit-level (conditional) Marginal Effects
#' mod <- glm(am ~ hp * wt, data = mtcars, family = binomial)
#' mfx <- slopes(mod)
#' head(mfx)
#'
#' # Average Marginal Effect (AME)
#' avg_slopes(mod, by = TRUE)
#'
#'
#' # Marginal Effect at the Mean (MEM)
#' slopes(mod, newdata = datagrid())
#'
#' # Marginal Effect at User-Specified Values
#' # Variables not explicitly included in `datagrid()` are held at their means
#' slopes(mod, newdata = datagrid(hp = c(100, 110)))
#'
#' # Group-Average Marginal Effects (G-AME)
#' # Calculate marginal effects for each observation, and then take the average
#' # marginal effect within each subset of observations with different observed
#' # values for the `cyl` variable:
#' mod2 <- lm(mpg ~ hp * cyl, data = mtcars)
#' avg_slopes(mod2, variables = "hp", by = "cyl")
#'
#' # Marginal Effects at User-Specified Values (counterfactual)
#' # Variables not explicitly included in `datagrid()` are held at their
#' # original values, and the whole dataset is duplicated once for each
#' # combination of the values in `datagrid()`
#' mfx <- slopes(mod,
#'   newdata = datagrid(
#'     hp = c(100, 110),
#'     grid_type = "counterfactual"))
#' head(mfx)
#'
#' # Heteroskedasticity robust standard errors
#' mfx <- slopes(mod, vcov = sandwich::vcovHC(mod))
#' head(mfx)
#'
#' # hypothesis test: is the `hp` marginal effect at the mean equal to the `drat` marginal effect
#' mod <- lm(mpg ~ wt + drat, data = mtcars)
#'
#' slopes(
#'   mod,
#'   newdata = "mean",
#'   hypothesis = "wt = drat")
#'
#' # same hypothesis test using row indices
#' slopes(
#'   mod,
#'   newdata = "mean",
#'   hypothesis = "b1 - b2 = 0")
#'
#' # same hypothesis test using numeric vector of weights
#' slopes(
#'   mod,
#'   newdata = "mean",
#'   hypothesis = c(1, -1))
#'
#' # two custom contrasts using a matrix of weights
#' lc <- matrix(
#'   c(
#'     1, -1,
#'     2, 3),
#'   ncol = 2)
#' colnames(lc) <- c("Contrast A", "Contrast B")
#' slopes(
#'   mod,
#'   newdata = "mean",
#'   hypothesis = lc)
#'
#' @export
slopes <- function(model,
                   newdata = NULL,
                   variables = NULL,
                   type = NULL,
                   by = FALSE,
                   vcov = TRUE,
                   conf_level = 0.95,
                   slope = "dydx",
                   wts = FALSE,
                   hypothesis = NULL,
                   equivalence = NULL,
                   df = Inf,
                   eps = NULL,
                   numderiv = "fdforward",
                   ...) {
  dots <- list(...)

  # very early, before any use of newdata
  # if `newdata` is a call to `typical` or `counterfactual`, insert `model`
  scall <- rlang::enquo(newdata)
  newdata <- sanitize_newdata_call(scall, newdata, model, by = by)

  # build call: match.call() doesn't work well in *apply()
  call_attr <- c(
    list(
      name = "slopes",
      model = model,
      newdata = newdata,
      variables = variables,
      type = type,
      vcov = vcov,
      by = by,
      conf_level = conf_level,
      slope = slope,
      wts = wts,
      hypothesis = hypothesis,
      df = df,
      eps = eps),
    list(...))
  call_attr <- do.call("call", call_attr)

  # slopes() does not support a named list of variables like comparisons()
  checkmate::assert_character(variables, null.ok = TRUE)

  # slope
  valid <- c("dydx", "eyex", "eydx", "dyex", "dydxavg", "eyexavg", "eydxavg", "dyexavg")
  checkmate::assert_choice(slope, choices = valid)

  # sanity checks and pre-processing
  model <- sanitize_model(model = model, newdata = newdata, wts = wts, vcov = vcov, by = by, calling_function = "marginaleffects", ...)
  sanity_dots(model = model, calling_function = "marginaleffects", ...)
  type <- sanitize_type(model = model, type = type, calling_function = "slopes")

  ############### sanity checks are over

  # Bootstrap
  out <- inferences_dispatch(
    INF_FUN = slopes,
    model = model, newdata = newdata, vcov = vcov, variables = variables, type = type,
    conf_level = conf_level,
    by = by,
    wts = wts, slope = slope, hypothesis = hypothesis, ...)
  if (!is.null(out)) {
    return(out)
  }

  out <- comparisons(
    model,
    newdata = newdata,
    variables = variables,
    vcov = vcov,
    conf_level = conf_level,
    type = type,
    wts = wts,
    hypothesis = hypothesis,
    equivalence = equivalence,
    df = df,
    by = by,
    eps = eps,
    numderiv = numderiv,
    comparison = slope,
    cross = FALSE,
    # secret arguments
    internal_call = TRUE,
    ...)

  # class
  data.table::setDF(out)
  class(out) <- setdiff(class(out), "comparisons")
  class(out) <- c("slopes", "marginaleffects", class(out))
  return(out)
}




#' Average slopes (aka Average partial derivatives, marginal effects, or trends)
#' @describeIn slopes Average slopes
#' @export
#'
avg_slopes <- function(model,
                       newdata = NULL,
                       variables = NULL,
                       type = NULL,
                       by = TRUE,
                       vcov = TRUE,
                       conf_level = 0.95,
                       slope = "dydx",
                       wts = FALSE,
                       hypothesis = NULL,
                       equivalence = NULL,
                       df = Inf,
                       eps = NULL,
                       numderiv = "fdforward",
                       ...) {
  # order of the first few paragraphs is important
  # if `newdata` is a call to `typical` or `counterfactual`, insert `model`
  # should probably not be nested too deeply in the call stack since we eval.parent() (not sure about this)
  scall <- rlang::enquo(newdata)
  newdata <- sanitize_newdata_call(scall, newdata, model, by = by)


  # Bootstrap
  out <- inferences_dispatch(
    INF_FUN = avg_slopes,
    model = model, newdata = newdata, vcov = vcov, variables = variables, type = type,
    conf_level = conf_level, by = by,
    wts = wts, slope = slope, hypothesis = hypothesis, ...)
  if (!is.null(out)) {
    return(out)
  }


  out <- slopes(
    model = model,
    newdata = newdata,
    variables = variables,
    type = type,
    vcov = vcov,
    conf_level = conf_level,
    by = by,
    slope = slope,
    wts = wts,
    hypothesis = hypothesis,
    equivalence = equivalence,
    df = df,
    eps = eps,
    numderiv = numderiv,
    ...)

  return(out)
}

#' Get a named vector of coefficients from a model object
#'
#' Mostly for internal use, but can be useful because the output is consistent across model classes.
#' @inheritParams slopes
#' @return A named vector of coefficients. The names must match those of the variance matrix.
#' @rdname get_coef
#' @export
get_coef <- function(model, ...) {
  UseMethod("get_coef", model)
}

#' @rdname get_coef
#' @export
get_coef.default <- function(model, ...) {
  ## faster
  # out <-  stats::coef(model)

  # more general
  out <- insight::get_parameters(model, component = "all")

  out <- stats::setNames(out$Estimate, out$Parameter)
  return(out)
}


#' @export
get_coef.predictions <- function(model, ...) {
  stats::coef(model)
}

#' @export
get_coef.comparisons <- get_coef.predictions

#' @export
get_coef.slopes <- get_coef.predictions

sanitize_conf_level <- function(conf_level, ...) {
    # periods in arg name are bad style because of s3, but we want to accept both because `broom` uses `conf_level`
    if ("conf.level" %in% ...names()) {
        conf_level <- ...elt(match("conf.level", ...names())[1L])
    }

    if (is.null(conf_level)) conf_level <- 0.95

    checkmate::assert(
        checkmate::check_numeric(conf_level, len = 1),
        checkmate::check_true(conf_level > 0),
        checkmate::check_true(conf_level < 1),
        combine = "and"
    )

    return(conf_level)
}

#' @noRd

#' @noRd
#' @export
vcov.comparisons <- function(object, ...) {
    # align J and V: This might be a problematic hack, but I have not found examples yet.
    V <- components(object, "vcov_model")
    J <- components(object, "jacobian")
    aligned <- align_jacobian_vcov(J, V, object, ...)
    aligned$J %*% aligned$V %*% t(aligned$J)
}


#' @noRd
#' @export
vcov.predictions <- vcov.comparisons


#' @noRd
#' @export
vcov.hypotheses <- vcov.comparisons


#' @noRd
#' @export
vcov.slopes <- vcov.comparisons


#' @export
#' @noRd
coef.comparisons <- function(object, ...) {
    if (!is.null(object$estimate)) {
        out <- object$estimate
        if (is.null(names(out))) {
            lab <- tryCatch(get_labels(object), error = function(e) NULL)
            if (length(lab) == length(out)) {
                out <- stats::setNames(out, lab)
            }
        }
        return(out)
    } else {
        stop("The input object does not contain an 'estimate' element.")
    }
}


#' @export
#' @noRd
coef.slopes <- coef.comparisons


#' @export
#' @noRd
coef.predictions <- coef.comparisons


#' @export
#' @noRd
coef.hypotheses <- coef.comparisons


#' @export
#' @noRd
df.residual.comparisons <- function(object, ...) {
    out <- tryCatch(
        stats::df.residual(components(object, "model")),
        error = function(e) NULL
    )
    if (is.null(out)) out <- Inf
    return(out)
}


#' @export
#' @noRd
df.residual.predictions <- df.residual.comparisons


#' @export
#' @noRd
df.residual.slopes <- df.residual.comparisons

matrix_apply_column <- function(x, FUN, by = NULL) {
    if (inherits(by, "data.frame") && nrow(by) == nrow(x)) {
        x <- cbind(data.table::data.table(by), data.table::data.table(x))
        data.table::setDT(x)
        bycols <- colnames(by)
        x <- x[, lapply(.SD, FUN), keyby = bycols]
        cols <- setdiff(colnames(x), colnames(by))
        x <- x[, ..cols]
        x <- as.matrix(x)
        dimnames(x) <- NULL
    } else {
        x <- apply(x, MARGIN = 2, FUN = FUN, simplify = TRUE)
        if (is.null(dim(x))) {
            x <- matrix(x, nrow = 1)
        }
    }
    return(x)
}

# fancy way to catch the call so that get_averages(slopes()) does not evaluate twice
# and is fast
recall <- function(x, ...) {


    funs <- c("comparisons", "slopes", "predictions", "hypotheses", "avg_predictions", "avg_comparisons", "avg_slopes")

    # 2-step estimation with already evaluated & assigned call
    if (!is.call(x)) {
        # unsupported evaluated object: return `NULL`
        if (!inherits(x, funs)) {
            return(NULL)
        }

        # retrieve call
        mc <- attr(x, "call")
        if (!is.call(mc)) {
            msg <- sprintf("Call could not be retrieved from object of class %s.", class(x)[1])
            insight::format_error(msg)
        }

    # unsupported call: return `NULL`
    } else {
        if (!as.character(x[1]) %in% funs) {
            return(NULL)
        }
        mc <- x
    }

    dots <- list(...)

    # don't overwrite certain arguments
    if ("hypothesis" %in% names(mc) && "hypothesis" %in% names(dots)) {
        if (is.null(dots[["hypothesis"]])) {
            dots[["hypothesis"]] <- NULL
        }
    }

    # safe to work with original objects when available
    objs <- c("newdata", "model")
    for (obj in objs) {
        if (!is.null(attr(x, "call")[[obj]])) {
            dots[[obj]] <- attr(x, "call")[[obj]]
        }
    }

    # overwrite previous arguments
    for (n in names(dots)) {
        # named NULL should not remove the corresponding argument from the call
        if (is.null(dots[[n]])) {
            mc[n] <- list(NULL)
        } else {
            mc[[n]] <- dots[[n]]
        }
    }

    out <- eval(mc)

    return(out)
}

sanity_predict_vector <- function(pred, model, newdata, type) {
    if (
        !isTRUE(checkmate::check_atomic_vector(pred)) &&
            !isTRUE(checkmate::check_array(pred, d = 1))
    ) {
        msg <- sprintf(
            '`predict(model, type = "%s")` was called on a model of class `%s`, but this command did not produce the expected outcome: A numeric vector of length %s. This can sometimes happen when users try compute a marginal effect for some models with grouped or multivariate outcome which are not supported yet by `marginaleffects` package. Please consult your modeling package documentation to learn what alternative `type` arguments are accepted by the `predict` method, or file a feature request on Github:  https://github.com/vincentarelbundock/marginaleffects/issues',
            type,
            class(model)[1],
            nrow(newdata)
        )
        stop(msg, call. = FALSE)
    }
}


sanity_predict_numeric <- function(pred, model, newdata, type) {
    if (!isTRUE(checkmate::check_numeric(pred))) {
        msg <- sprintf(
            '`predict(model, type = "%s")` was called on a model of class `%s`, but this command did not produce the expected outcome: A numeric vector of length %s. This can sometimes happen when users try compute a marginal effect for an outcome type which is unsupported, or which cannot be differentiated. Please consult your modeling package documentation to learn what alternative `type` arguments are accepted by the `predict` method.',
            type,
            class(model)[1],
            nrow(newdata)
        )
        stop(msg, call. = FALSE)
    }
}

wmean <- function(x, w) {
    stats::weighted.mean(x, w)
}

comparison_function_dict <- list(
    # default = difference between predictions
    "difference" = function(hi, lo) hi - lo,
    "differenceavg" = function(hi, lo) mean(hi - lo),
    "differenceavgwts" = function(hi, lo, w) wmean(hi - lo, w),

    # slopes and elasticities
    "dydx" = function(hi, lo, eps) (hi - lo) / eps,
    "eyex" = function(hi, lo, eps, y, x) (hi - lo) / eps * (x / y),
    "eydx" = function(hi, lo, eps, y, x) ((hi - lo) / eps) / y,
    "dyex" = function(hi, lo, eps, x) ((hi - lo) / eps) * x,

    # average slopes and elasticities
    "dydxavg" = function(hi, lo, eps) mean((hi - lo) / eps),
    "eyexavg" = function(hi, lo, eps, y, x) mean((hi - lo) / eps * (x / y)),
    "eydxavg" = function(hi, lo, eps, y, x) mean(((hi - lo) / eps) / y),
    "dyexavg" = function(hi, lo, eps, x) mean(((hi - lo) / eps) * x),
    "dydxavgwts" = function(hi, lo, eps, w) wmean((hi - lo) / eps, w),
    "eyexavgwts" = function(hi, lo, eps, y, x, w) wmean((hi - lo) / eps * (x / y), w),
    "eydxavgwts" = function(hi, lo, eps, y, x, w) wmean(((hi - lo) / eps) / y, w),
    "dyexavgwts" = function(hi, lo, eps, x, w) wmean(((hi - lo) / eps) * x, w),

    # ratios
    "ratio" = function(hi, lo) hi / lo,
    "ratioavg" = function(hi, lo) mean(hi) / mean(lo),
    "ratioavgwts" = function(hi, lo, w) wmean(hi, w) / wmean(lo, w),

    "lnratio" = function(hi, lo) log(hi / lo),
    "lnratioavg" = function(hi, lo) log(mean(hi) / mean(lo)),
    "lnratioavgwts" = function(hi, lo, w) log(wmean(hi, w) / wmean(lo, w)),

    "lnor" = function(hi, lo) log((hi / (1 - hi)) / (lo / (1 - lo))),
    "lnoravg" = function(hi, lo) log((mean(hi) / (1 - mean(hi))) / (mean(lo) / (1 - mean(lo)))),
    "lnoravgwts" = function(hi, lo, w) log((wmean(hi, w) / (1 - wmean(hi, w))) / (wmean(lo, w) / (1 - wmean(lo, w)))),

    # others
    "lift" = function(hi, lo) (hi - lo) / lo,
    "liftavg" = function(hi, lo) (mean(hi - lo)) / mean(lo),
    "liftavgwts" = function(hi, lo, w) (wmean(hi - lo, w)) / wmean(lo, w),

    "expdydx" = function(hi, lo, eps) ((exp(hi) - exp(lo)) / exp(eps)) / eps,
    "expdydxavg" = function(hi, lo, eps) mean(((exp(hi) - exp(lo)) / exp(eps)) / eps),
    "expdydxavgwts" = function(hi, lo, eps, w) wmean(((exp(hi) - exp(lo)) / exp(eps)) / eps, w)
)

comparison_label_dict <- list(
    "difference" = "%s - %s",
    "differenceavg" = "%s - %s",
    "differenceavgwts" = "%s - %s",

    "dydx" = "dY/dX",
    "eyex" = "eY/eX",
    "eydx" = "eY/dX",
    "dyex" = "dY/eX",

    "dydxavg" = "dY/dX",
    "eyexavg" = "eY/eX",
    "eydxavg" = "eY/dX",
    "dyexavg" = "dY/eX",

    "dydxavg" = "dY/dX",
    "eyexavg" = "eY/eX",
    "eydxavg" = "eY/dX",
    "dyexavg" = "dY/eX",
    "dydxavgwts" = "dY/dX",
    "eyexavgwts" = "eY/eX",
    "eydxavgwts" = "eY/dX",
    "dyexavgwts" = "dY/eX",

    "ratio" = "%s / %s",
    "ratioavg" = "mean(%s) / mean(%s)",
    "ratioavgwts" = "mean(%s) / mean(%s)",

    "lnratio" = "ln(%s / %s)",
    "lnratioavg" = "ln(mean(%s) / mean(%s))",
    "lnratioavgwts" = "ln(mean(%s) / mean(%s))",

    "lnor" = "ln(odds(%s) / odds(%s))",
    "lnoravg" = "ln(odds(%s) / odds(%s))",
    "lnoravgwts" = "ln(odds(%s) / odds(%s))",

    "lift" = "lift",
    "liftavg" = "liftavg",
    "liftavgwts" = "liftavgwts",

    "expdydx" = "exp(dY/dX)"
)

sanity_comparison <- function(comparison) {
    # wts versions are used internally but not available directly to users
    valid <- names(comparison_function_dict)
    valid <- valid[!grepl("wts$", valid)]
    checkmate::assert(
        checkmate::check_choice(comparison, choices = valid),
        checkmate::check_function(comparison))
}


sanitize_transform <- function(x) {
    good <- c("exp", "ln")
    # issue #1115: sanitize_transform() wraps `transform` into a named list, so the assertion may fail when using `inferences()`
    if (isTRUE(checkmate::check_list(x, names = "named"))) {
        checkmate::assert(
            checkmate::check_choice(x[[1]], choices = good, null.ok = TRUE),
            checkmate::check_function(x[[1]]))
        x <- x[[1]]
    } else {
        checkmate::assert(
            checkmate::check_choice(x, choices = good, null.ok = TRUE),
            checkmate::check_function(x))
    }

    if (is.null(x)) {
        return(x)
    } else if (is.function(x)) {
        out <- list(x)
        names(out) <- deparse(substitute(x))
    } else if (x == "exp") {
        out <- list("exp" = exp)
    } else if (x == "ln") {
        out <- list("ln" = log)
    }

    return(out)
}

get_modeldata_mids <- function(model, envir = parent.frame()) {
    # If the user passed a mids directly
    if (inherits(model, "mids")) {
        imp <- model
    } else {
        # Try to recover the mids object from the model's call
        cl <- tryCatch(
            {
                if (!is.null(model$call)) model$call else stats::getCall(model)
            },
            error = function(e) NULL)

        if (is.null(cl) || is.null(cl$data)) {
            stop("Cannot find a 'data' argument in the model call. Pass a 'mids' or 'mira' object.")
        }

        data_expr <- cl$data

        # Environments to try, in order
        envs <- list(
            # user-provided search env
            envir,
            # environment attached to the call (if any)
            attr(cl, ".Environment"),
            # formula/terms env (for lm/glm fits)
            tryCatch(environment(stats::formula(model)), error = function(e) NULL),
            tryCatch(environment(stats::terms(model)), error = function(e) NULL),
            # current and global as fallbacks
            environment(),
            .GlobalEnv
        )

        imp <- NULL
        for (ev in envs) {
            if (is.null(ev)) next
            obj <- tryCatch(eval(data_expr, envir = ev), error = function(e) NULL)
            if (inherits(obj, "mids")) {
                imp <- obj
                break
            }
        }

        if (is.null(imp)) {
            stop("Could not locate the original 'mids' object used to fit the model.")
        }
    }

    # Return list of completed datasets
    lapply(seq_len(imp$m), function(i) mice::complete(imp, action = i))
}

hypothesis_function <- function(x, newdata, hypothesis, by) {
    draws <- attr(x, "posterior_draws")
    if (!is.null(draws)) {
        msg <- "The `hypothesis` argument does not support function for models with draws. You can use `get_draws()` to extract draws and manipulate them directly instead."
        insight::format_error(msg)
    }

    if ("rowid" %in% colnames(x) && "rowid" %in% colnames(newdata)) {
        x <- merge(x, newdata, all.x = TRUE, by = intersect(colnames(x), colnames(newdata)))
    } else if (isTRUE(nrow(x) == nrow(newdata))) {
        x <- cbind(x, newdata)
    }

    attr(x, "variables_datagrid") <- attr(newdata, "variables_datagrid")
    attr(x, "by") <- if (is.character(by)) by else names(by)

    argnames <- names(formals(hypothesis))
    if (!"x" %in% argnames) insight::format_error("The `hypothesis` function must accept an `x` argument.")
    if (!all(argnames %in% c("x", "draws"))) {
        msg <- "The allowable arguments for the `hypothesis` function are: `x` and `draws`"
        insight::format_error(msg)
    }
    args <- list(x = x, newdata = newdata, by = by, draws = draws)
    args <- args[names(args) %in% argnames]
    out <- do.call(hypothesis, args)
    at <- attr(out, "hypothesis_function_by")

    # sanity
    msg <- "The `hypothesis` argument function must return a data frame with `term` (or `hypothesis`) and `estimate` columns."
    if (inherits(out, "data.frame")) {
        if (!all(c("term", "estimate") %in% colnames(out)) && !all(c("hypothesis", "estimate") %in% colnames(out))) {
            insight::format_error(msg)
        }
    } else if (isTRUE(checkmate::check_numeric(out))) {
        if (isTRUE(checkmate::check_data_frame(x, nrows = length(out))) && "term" %in% colnames(out)) {
            out <- data.frame(term = out$term, estimate = out)
        } else {
            out <- data.frame(term = seq_along(out), estimate = out)
        }
    } else {
        insight::format_error(msg)
    }

    data.table::setDT(out)
    attr(out, "hypothesis_function_by") <- at
    return(out)
}

get_modeldata <- function(model, ...) {
    modeldata <- insight::get_data(model, verbose = FALSE, additional_variables = TRUE)
    modeldata <- unpack_matrix_1col(modeldata)
    if (is.null(modeldata)) {
        return(NULL)
    }
    # Avoid mutating a data.table returned by reference (would affect the model's
    # original data.table object).
    modeldata <- data.table::copy(modeldata)
    data.table::setDF(modeldata)
    return(modeldata)
}

#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.betareg <- function(model, coefs, ...) {
    # coefs are split between mean coefs (which can be length 0) and precision coefs
    # (which must be length > 0 and always start with "(phi)_" due to get_coef.betareg(),
    # to match with get_varcov(., component = "all") output). In betareg object, these
    # are stored as two elements in a list, with precision coefs lacking the "(phi)_"
    # prefix, so we remove it.
    mu <- !grepl("\\(phi\\)|^Log\\(nu\\)", names(coefs))
    if (length(mu) > 0) {
        if ("mean" %in% names(model$coefficients)) {
            model[["coefficients"]][["mean"]] <- coefs[mu]
        } else if ("mu" %in% names(model$coefficients)) {
            model[["coefficients"]][["mu"]] <- coefs[mu]
        }
    }

    phi <- startsWith(names(coefs), "(phi)")
    if (length(phi) > 0) {
        if ("precision" %in% names(model$coefficients)) {
            model[["coefficients"]][["precision"]] <- coefs[phi]
        } else if ("phi" %in% names(model$coefficients)) {
            model[["coefficients"]][["phi"]] <- coefs[phi]
        }
    }

    nu <- startsWith(names(coefs), "Log(nu)")
    if (length(nu) > 0 && "nu" %in% names(model[["coefficients"]])) {
        model[["coefficients"]][["nu"]] <- coefs[nu]
    }

    return(model)
}

#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.betareg <- function(model, ...) {
    out <- model$coefficients
    for (n in names(out)) {
        if (n %in% c("phi", "precision")) {
            names(out[[n]]) <- sprintf("(phi)_%s", names(out[[n]]))
        }
    }
    out <- stats::setNames(unlist(out), unlist(lapply(out, names)))
    return(out)
}


#' @include get_predict.R
#' @rdname get_predict
#' @export
get_predict.betareg <- function(model, newdata, type = "response", ...) {
    dots <- list(...)
    args <- list(
        model,
        newdata = newdata,
        type = type
    )
    args[["at"]] <- dots[["at"]]
    out <- do_call(stats::predict, args)
    out <- data.table(estimate = out)
    out <- add_rowid(out, newdata)
    return(out)
}


#' @rdname sanitize_model_specific
sanitize_model_specific.betareg <- function(model, ...) {
    insight::check_if_installed("insight", minimum_version = "0.17.1")
    return(model)
}

#' @rdname get_predict
#' @export
get_predict.MCMCglmm <- function(
    model,
    newdata,
    type = "response",
    mfx = NULL,
    newparams = NULL,
    ndraws = 1000,
    se.fit = NULL,
    ...) {
    ndraws_mod <- nrow(model$VCV)
    if (ndraws < ndraws_mod) {
        idx <- sample.int(ndraws_mod, ndraws)
    } else {
        idx <- seq_len(ndraws_mod)
    }
    # hack: predict() appears to require the response in `newdata`, but the value does not appear to maek a difference
    nd <- newdata
    dvname <- insight::find_response(model)
    nd[, dvname] <- 1000
    draws <- lapply(
        idx,
        function(i) stats::predict(model, newdata = nd, it = i, ...)
    )
    draws <- do.call("cbind", draws)
    out <- data.table(estimate = apply(draws, MARGIN = 1, FUN = stats::median))
    out <- add_rowid(out, newdata)
    attr(out, "posterior_draws") <- draws
    return(out)
}


#' @rdname get_vcov
#' @export
get_vcov.MCMCglmm <- function(model, vcov = NULL, ...) {
    if (!is.null(vcov) && !is.logical(vcov)) {
        warn_sprintf(
            "The `vcov` argument is not supported for models of this class."
        )
    }
    vcov <- sanitize_vcov(model, vcov)
    return(NULL)
}

get_model_matrix_attribute <- function(model, newdata = NULL) {
    # supported models (no inheritance)
    if (!isTRUE(class(model)[1] %in% c("lm", "glm", "rq"))) {
        return(newdata)
    }
    
    # stats::model.matrix creates all-0 columns with splines::bs() and other functions
    # this may be too aggressive, but it avoids all functions
    flag <- any(grepl("\\(", setdiff(names(get_coef(model)), "(Intercept)")))
    if (isTRUE(flag)) {
        return(newdata)
    }

    # we don't support offsets, so revert to stats::predict()
    if (!is.null(model[["offset"]])) {
        return(newdata)
    }

    # subset variables for listwise deletion
    vars <- unlist(insight::find_predictors(model, verbose = FALSE), use.names = FALSE)
    vars <- c(vars, unlist(insight::find_response(model), use.names = FALSE))
    vars <- intersect(vars, colnames(newdata))

    MM <- hush(get_model_matrix(model, newdata = data.frame(newdata)[, vars]))

    attr(newdata, "marginaleffects_model_matrix") <- MM
    return(newdata)
}

#' Method to raise model-specific warnings and errors
#'
#' @inheritParams slopes
#' @return A warning, an error, or nothing
#' @rdname sanitize_model_specific
#' @keywords internal
sanitize_model_specific <- function(model, ...) {
    UseMethod("sanitize_model_specific", model)
}


#' @rdname sanitize_model_specific
sanitize_model_specific.default <- function(
  model,
  vcov = NULL,
  calling_function = "marginaleffects",
  ...
) {
    return(model)
}


sanity_model_supported_class <- function(model, custom = TRUE) {
    checkmate::assert_character(
        getOption("marginaleffects_model_classes", default = NULL),
        null.ok = TRUE
    )

    if (isTRUE(custom)) {
        custom_classes <- getOption("marginaleffects_model_classes", default = NULL)
        custom_classes <- as.list(custom_classes)
    } else {
        custom_classes <- NULL
    }

    supported <- append(
        custom_classes,
        list(
            "marginaleffects_internal",
            "afex_aov",
            "amest", # package: Amelia
            "bart", # package: dbarts
            "betareg",
            "bglmerMod",
            "blmerMod",
            # "bife",
            "biglm",
            c("bigglm", "biglm"),
            "brglmFit",
            "brmsfit",
            c("bracl", "brmultinom", "brglmFit"),
            c("brnb", "negbin", "glm"),
            c("clogit", "coxph"),
            "clm",
            c("clmm2", "clm2"),
            "coxph",
            "coxph_weightit",
            "crch",
            "DirichletRegModel",
            "flexsurvreg", # package: flexsurv
            "fixest",
            "flic",
            "flac",
            c("Gam", "glm", "lm"), # package: gam
            c("gam", "glm", "lm"), # package: mgcv
            c("gamlss", "gam", "glm", "lm"), # package: gamlss
            c("geeglm", "gee", "glm"),
            c("Gls", "rms", "gls"),
            "glm",
            "gls",
            "glmerMod",
            "glmrob",
            "glmmTMB",
            "glmgee",
            c("glmmPQL", "lme"),
            "glimML",
            "glmx",
            "glm_weightit",
            "hetprob",
            "hurdle",
            "hxlr",
            "ivreg",
            "iv_robust",
            "ivpml",
            "lda",
            "Learner",
            "lm",
            "lme",
            "lmerMod",
            "lmerModLmerTest",
            "lmrob",
            "lmRob",
            "lm_robust",
            "lmtree", # partykit
            "glmtree", # partykit
            # "logitr",
            "loess",
            "logistf",
            c("lrm", "lm"),
            c("lrm", "rms", "glm"),
            c("mblogit", "mclogit"),
            c("mclogit", "lm"),
            "MCMCglmm",
            "mhurdle",
            "mira",
            "mlogit",
            "model_fit",
            c("multinom", "nnet"),
            "multinom_weightit",
            "mvgam",
            c("negbin", "glm", "lm"),
            "nls",
            c("ols", "rms", "lm"),
            "ordinal_weightit",
            c("orm", "rms"),
            c("oohbchoice", "dbchoice"),
            "phylolm",
            "phyloglm",
            c("plm", "panelmodel"),
            "polr",
            "stpm2",
            "gsm",
            "pstpm2",
            "aft",
            "Rchoice",
            "rendo.base",
            "rlmerMod",
            "rq",
            c("scam", "glm", "lm"),
            c("selection", "selection", "list"),
            "speedglm",
            "speedlm",
            "stanreg",
            "survreg",
            "svyolr",
            "systemfit",
            c("tobit", "survreg"),
            "tobit1",
            "truncreg",
            "workflow",
            "zeroinfl"
        )
    )
    flag <- FALSE
    for (sup in supported) {
        if (!is.null(sup) && isTRUE(all(sup %in% class(model)))) {
            flag <- TRUE
        }
    }
    if (isFALSE(flag)) {
        support <- toString(sort(unique(sapply(supported, function(x) x[1]))))
        msg <- sprintf('Models of class "%s" are not supported. Supported model classes include:\n\n%s.\n\nRequest support for new models on the issue tracker: https://github.com/vincentarelbundock/marginaleffects', class(model)[1], paste(support, collapse = ", "))

        # default length is 1000. We only want to override that if the user has not explicitly set a different value
        op <- getOption("warning.length")
        if (is.numeric(op) && op == 1000) {
            on.exit(options(warning.length = op))
            options(warning.length = 2000)
        }

        stop(msg, call. = FALSE)
    }
}


# Function that operates on mfx objects
sanitize_model <- function(model, call, newdata = NULL, vcov = NULL, by = FALSE, ...) {
    # Extract calling_function from mfx@call
    calling_function <- extract_calling_function(call)

    # Sanitize the model
    model <- sanitize_model_specific(
        model,
        vcov = vcov,
        newdata = newdata,
        by = by,
        calling_function = calling_function,
        ...
    )
    sanity_model_supported_class(model)

    # Assign the sanitized model to the slot
    return(model)
}

sanitize_newdata_call <- function(scall, newdata = NULL, mfx = NULL, by = NULL) {
    # For mice objects, defer newdata processing to process_imputation()
    if (!is.null(mfx) && inherits(mfx@model, c("mira", "amest"))) {
        # Return the call as-is, will be processed later with actual model data
        return(scall)
    }

    if (rlang::quo_is_call(scall)) {
        df <- FALSE
        if (grepl("^datagrid", rlang::call_name(scall))) {
            if (!"model" %in% rlang::call_args_names(scall)) {
                scall <- rlang::call_modify(scall, model = mfx@model)
            }
        } else if (isTRUE(rlang::call_name(scall) == "data.frame")) {
            df <- TRUE
        } else if (isTRUE(rlang::call_name(scall) == "subset")) {
            argnames <- rlang::call_args_names(scall)
            if (!"x" %in% argnames && length(argnames) == 1) {
                scall <- rlang::call_modify(scall, x = mfx@modeldata)
            }
        } else if (isTRUE(rlang::call_name(scall) == "filter")) {
            argnames <- rlang::call_args_names(scall)
            if (!".data" %in% argnames && length(argnames) == 1) {
                scall <- rlang::call_modify(scall, .data = mfx@modeldata)
            }
        } else if (rlang::call_name(scall) %in% "visualisation_matrix") {
            if (!"x" %in% rlang::call_args_names(scall)) {
                scall <- rlang::call_modify(scall, x = mfx@modeldata)
            }
        }
        out <- rlang::eval_tidy(scall)
        # newdata=data.frame() all columns must be printed as explicit in print.R
        if (isTRUE(df)) {
            attr(out, "implicit") <- unique(c(attr(out, "implicit"), colnames(out)))
        }
    } else {
        out <- newdata
    }
    return(out)
}


sanitize_newdata <- function(mfx, newdata, by, wts) {
    model <- mfx@model
    modeldata <- mfx@modeldata

    # For mice objects, defer processing - newdata might be a call
    if (inherits(model, c("mira", "amest"))) {
        # Return placeholder that will be handled in process_imputation()
        return(list(
            "newdata" = newdata, # could be a call like subset(treat == 1)
            "modeldata" = modeldata
        ))
    }

    # Input validation (skip for mice objects)
    checkmate::assert(
        checkmate::check_data_frame(newdata, null.ok = TRUE),
        checkmate::check_choice(
            newdata,
            choices = c("mean", "median", "tukey", "grid", "balanced")
        ),
        combine = "or"
    )

    # Process 'by' argument for datagrid calls
    if (isTRUE(checkmate::check_data_frame(by))) {
        by_vars <- setdiff(colnames(by), "by")
    } else if (isTRUE(checkmate::check_flag(by))) {
        by_vars <- NULL
    } else {
        by_vars <- by
    }

    # Build datagrid arguments
    args <- list(model = model)
    for (b in by_vars) {
        args[[b]] <- unique
    }

    # Process newdata input
    if (is.null(newdata)) {
        # NULL -> modeldata
        newdata <- modeldata
    } else if (identical(newdata, "mean")) {
        # string -> datagrid()
        newdata <- do.call("datagrid", args)
    } else if (identical(newdata, "median")) {
        args[["FUN_numeric"]] <- args[["FUN_integer"]] <- args[["FUN_logical"]] <-
            function(x) stats::median(x, na.rm = TRUE)
        newdata <- do.call("datagrid", args)
    } else if (identical(newdata, "tukey")) {
        args[["FUN_numeric"]] <- function(x) stats::fivenum(x, na.rm = TRUE)
        newdata <- do.call("datagrid", args)
    } else if (identical(newdata, "grid")) {
        args[["FUN_numeric"]] <- function(x) stats::fivenum(x, na.rm = TRUE)
        args[["FUN_factor"]] <- args[["FUN_character"]] <- args[["FUN_logical"]] <- unique
        newdata <- do.call("datagrid", args)
    } else if (identical(newdata, "balanced")) {
        # grid with all unique values of categorical variables, and numerics at their means
        args[["grid_type"]] <- "balanced"
        newdata <- do.call("datagrid", args)
        # Issue #580: outcome should not duplicate grid rows
        # there can be multiple response variables
        dv <- mfx@variable_names_response
        for (d in dv) {
            if (isTRUE(d %in% colnames(newdata))) {
                newdata[[d]] <- get_mean_or_mode(newdata[[d]])
                newdata <- unique(newdata)
            }
        }
    }

    # Validate that we have a data.frame
    if (!inherits(newdata, "data.frame")) {
        msg <- "Unable to extract the data from model of class `%s`. This can happen in a variety of cases, such as when a `marginaleffects` package function is called from inside a user-defined function, or using an `*apply()`-style operation on a list. Please supply a data frame explicitly via the `newdata` argument."
        msg <- sprintf(msg, class(model)[1])
        stop_sprintf(msg)
    }

    # Process matrix columns
    # Issue #1327: matrix columns with single column breaks rbindlist(). See `scale()`
    newdata <- unpack_matrix_1col(newdata)

    # Issue #363: unpacking matrix columns works with {mgcv} but breaks {mclogit}
    if (inherits(model, "gam")) {
        newdata <- unpack_matrix_cols(newdata)
    }

    # Add rowid column for tracking
    if (!"rowid" %in% colnames(newdata)) {
        newdata$rowid <- seq_len(nrow(newdata))
    }

    # Add weights column if needed
    newdata <- add_wts_column(wts, newdata, model)

    # Convert to data.table
    # setDT() raises warnings and breaks mergin with mlogit index columns
    # but as.data.table() breaks a bunch of other models, for reasons I can't
    # quite figure out
    if (inherits(model, "mlogit")) {
        out <- as.data.table(newdata)
    } else {
        data.table::setDT(newdata)
    }

    safe <- isTRUE(getOption("marginaleffects_safe", default = TRUE))
    nc <- ncol(newdata)
    if (safe && isTRUE(nc > 100)) {
        warn_sprintf("The `newdata` data frame has %s columns. This can slow down computation and increase memory costs. Please supply a data frame with fewer columns to the `newdata` argument, or set `options('marginaleffects_safe'=FALSE)` to silence this warning.", nc)
    }

    return(newdata)
}


add_wts_column <- function(wts, newdata, model) {
    # weights must be available in the `comparisons()` function, NOT in
    # `tidy()`, because comparisons will often duplicate newdata for
    # multivariate outcomes and the like. We need to track which row matches
    # which.
    if (isFALSE(wts)) {
        return(newdata)
    }

    if (isTRUE(wts)) {
        wtsname <- insight::find_weights(model)
        if (
            !is.character(wtsname) ||
                length(wtsname) != 1 ||
                !wtsname %in% colnames(newdata)
        ) {
            msg <- "Unable to retrieve weights automatically from the model. Please specify `wts` argument explicitly."
            stop_sprintf(msg)
        } else {
            newdata[["marginaleffects_wts_internal"]] <- newdata[[wtsname]]
            return(newdata)
        }
    }

    flag1 <- isTRUE(checkmate::check_string(wts)) &&
        isTRUE(wts %in% colnames(newdata))
    flag2 <- isTRUE(checkmate::check_numeric(wts, len = nrow(newdata)))
    if (!flag1 && !flag2) {
        msg <- sprintf(
            "The `wts` argument must be a numeric vector of length %s, or a string which matches one of the `colnames()` in the data frame that you supplied to the `newdata`, or in the `marginaleffects` objects.",
            nrow(newdata)
        )
        stop(msg, call. = FALSE)
    }

    # weights: before sanitize_variables
    if (isTRUE(checkmate::check_string(wts))) {
        newdata[["marginaleffects_wts_internal"]] <- newdata[[wts]]
    } else {
        newdata[["marginaleffects_wts_internal"]] <- wts
    }

    return(newdata)
}



add_newdata <- function(
    mfx,
    scall,
    newdata = NULL,
    by = FALSE,
    wts = FALSE,
    cross = NULL,
    comparison = NULL,
    byfun = NULL) {
    # For mice objects, defer all processing to process_imputation()
    if (inherits(mfx@model, c("mira", "amest"))) {
        # Store the raw newdata (could be a call) for later processing
        mfx@newdata <- scall
        return(mfx)
    }

    # Step 1: Handle quoted calls to datagrid, subset, etc.
    newdata <- sanitize_newdata_call(scall, newdata, mfx = mfx, by = by)

    # Step 2: Streamlined newdata processing (combines sanitize_newdata + build_newdata)
    newdata <- sanitize_newdata(
        mfx = mfx,
        newdata = newdata,
        by = by,
        wts = wts
    )

    # Step 3: Extract numeric weights from newdata and store in @wts slot
    if ("marginaleffects_wts_internal" %in% colnames(newdata)) {
        mfx@wts <- newdata[["marginaleffects_wts_internal"]]
    } else {
        mfx@wts <- NULL
    }

    # Store processed newdata in the mfx object
    mfx@newdata <- newdata

    # Merge variable class info from newdata to handle workflows that drop raw predictors
    newdata_vclass <- detect_variable_class(newdata, model = mfx@model)
    if (length(newdata_vclass) > 0) {
        if (length(mfx@variable_class) == 0) {
            mfx@variable_class <- newdata_vclass
        } else {
            missing <- setdiff(names(newdata_vclass), names(mfx@variable_class))
            if (length(missing) > 0) {
                mfx@variable_class[missing] <- newdata_vclass[missing]
            }
        }
    }

    # Extract and store variable_names_datagrid attribute from newdata
    datagrid_vars <- attr(newdata, "variable_names_datagrid")
    if (!is.null(datagrid_vars)) {
        mfx@variable_names_datagrid <- datagrid_vars
    }

    # if `modeldata` is unavailable, we default to `newdata`
    flag1 <- isTRUE(checkmate::check_data_frame(mfx@modeldata, min.rows = 1))
    flag2 <- isTRUE(checkmate::check_data_frame(newdata, min.rows = 1))
    if (!flag1 && flag2) {
        mfx@modeldata <- newdata
        mfx@modeldata_available <- FALSE
        mfx@variable_class <- detect_variable_class(newdata, model = mfx@model)
    }

    # Return the updated mfx object
    return(mfx)
}


dedup_newdata <- function(
    mfx,
    newdata,
    by,
    wts,
    comparison = "difference",
    cross = FALSE,
    byfun = NULL) {
    # init
    model <- mfx@model

    # issue #1113: elasticities or custom functions should skip dedup because it is difficult to align x and y
    elasticities <- c("eyexavg", "eydxavg", "dyexavg")
    if (
        isTRUE(checkmate::check_choice(comparison, elasticities)) ||
            isTRUE(checkmate::check_function(comparison))
    ) {
        return(newdata)
    }

    elasticities <- c("eyex", "eydx", "dyex")
    if (!isFALSE(by) && isTRUE(checkmate::check_choice(comparison, elasticities))) {
        return(data.table(newdata))
    }

    flag <- isTRUE(checkmate::check_string(comparison, pattern = "avg"))
    if (
        !flag &&
            (isFALSE(by) || # weights only make sense when we are marginalizing
                !isFALSE(wts) ||
                !is.null(byfun) ||
                !isFALSE(cross) ||
                isFALSE(getOption("marginaleffects_dedup", default = TRUE)))
    ) {
        return(newdata)
    }

    vclass <- attr(newdata, "marginaleffects_variable_class")

    # copy to allow mod by reference later without overwriting newdata
    out <- data.table(newdata)

    dv <- mfx@variable_names_response
    if (isTRUE(checkmate::check_string(dv)) && dv %in% colnames(out)) {
        out[, (dv) := NULL]
        vclass <- vclass[names(vclass) != dv]
    }

    categ <- c("factor", "character", "logical", "strata", "cluster", "binary")
    if (!all(vclass %in% categ)) {
        return(newdata)
    }

    cols <- colnames(out)
    out <- out[, .("marginaleffects_wts_internal" = .N), by = cols]

    out[["rowid_dedup"]] <- seq_len(nrow(out))
    attr(out, "marginaleffects_variable_class") <- vclass

    return(out)
}

#' Plot Conditional or Marginal Comparisons
#'
#' @description
#' Plot comparisons on the y-axis against values of one or more predictors (x-axis, colors/shapes, and facets).
#'
#' The `by` argument is used to plot marginal comparisons, that is, comparisons made on the original data, but averaged by subgroups. This is analogous to using the `by` argument in the `comparisons()` function.
#'
#' The `condition` argument is used to plot conditional comparisons, that is, comparisons made on a user-specified grid. This is analogous to using the `newdata` argument and `datagrid()` function in a `comparisons()` call. All variables whose values are not specified explicitly are treated as usual by `datagrid()`, that is, they are held at their mean or mode (or rounded mean for integers). This includes grouping variables in mixed-effects models, so analysts who fit such models may want to specify the groups of interest using the `condition` argument, or supply model-specific arguments to compute population-level estimates. See details below.
#'
#' See the "Plots" vignette and website for tutorials and information on how to customize plots:
#'
#' * https://marginaleffects.com/bonus/plot.html
#' * https://marginaleffects.com
#'
#' @param variables Name of the variable whose contrast we want to plot on the y-axis.
#' @param draw `TRUE` returns a `ggplot2` plot. `FALSE` returns a `data.frame` of the underlying data.
#' @param newdata When `newdata` is `NULL`, the grid is determined by the `condition` argument. When `newdata` is not `NULL`, the argument behaves in the same way as in the `predictions()` function. Note that the `condition` argument builds its own grid, so the `newdata` argument is ignored if the `condition` argument is supplied.
#' @inheritParams comparisons
#' @inheritParams plot_slopes
#' @inheritParams slopes
#' @template model_specific_arguments
#' @return A `ggplot2` object
#' @export
#' @examplesIf interactive() || isTRUE(Sys.getenv("R_DOC_BUILD") == "true")
#' mod <- lm(mpg ~ hp * drat * factor(am), data = mtcars)
#'
#' plot_comparisons(mod, variables = "hp", condition = "drat")
#'
#' plot_comparisons(mod, variables = "hp", condition = c("drat", "am"))
#'
#' plot_comparisons(mod, variables = "hp", condition = list("am", "drat" = 3:5))
#'
#' plot_comparisons(mod, variables = "am", condition = list("hp", "drat" = range))
#'
#' plot_comparisons(mod, variables = "am", condition = list("hp", "drat" = "threenum"))
#'
#' # marginal comparisons
#' plot_comparisons(mod, variables = "hp", by = "am")
#'
#' # marginal comparisons on a counterfactual grid
#' plot_comparisons(mod,
#'     variables = "hp",
#'     by = "am",
#'     newdata = datagrid(am = 0:1, grid_type = "counterfactual")
#' )
#'
plot_comparisons <- function(
    model,
    variables = NULL,
    condition = NULL,
    by = NULL,
    newdata = NULL,
    type = NULL,
    vcov = NULL,
    conf_level = 0.95,
    wts = FALSE,
    comparison = "difference",
    transform = NULL,
    rug = FALSE,
    gray = getOption("marginaleffects_plot_gray", default = FALSE),
    draw = TRUE,
    ...) {
    if (inherits(model, c("mira", "amest"))) {
        msg <- "This function does not support multiple imputation. Call `comparisons()` or `avg_comparisons()` instead. These functions return easy to plot data frames."
        stop_sprintf(msg)
    }

    # init
    call <- construct_call(model, "comparisons")
    model <- sanitize_model(model, call = call, newdata = newdata, wts = wts, vcov = vcov, by = by, ...)
    mfx <- new_marginaleffects_internal(
        call = call,
        model = model
    )

    if (inherits(mfx@model, "mira") && is.null(newdata)) {
        msg <- "Please supply a data frame to the `newdata` argument explicitly."
        stop_sprintf(msg)
    }

    # order of the first few paragraphs is important
    scall <- rlang::enquo(newdata)
    newdata <- sanitize_newdata_call(scall, newdata, mfx = mfx, by = by)
    if (!isFALSE(wts) && is.null(by)) {
        stop_sprintf("The `wts` argument requires a `by` argument.")
    }

    checkmate::assert_character(
        by,
        null.ok = TRUE,
        max.len = 3,
        min.len = 1,
        names = "unnamed"
    )
    if ((!is.null(condition) && !is.null(by)) || (is.null(condition) && is.null(by))) {
        msg <- "One of the `condition` and `by` arguments must be supplied, but not both."
        stop_sprintf(msg)
    }

    # sanity check
    checkmate::assert(
        checkmate::check_character(variables, names = "unnamed"),
        checkmate::check_list(variables, names = "unique"),
        .var.name = "variables"
    )

    # mlr3 and tidymodels
    if (is.null(mfx@modeldata) || nrow(mfx@modeldata) == 0) {
        mfx@modeldata <- newdata
    }

    # conditional
    if (!is.null(condition)) {
        condition <- sanitize_condition(
            mfx = mfx,
            condition = condition,
            variables = variables
        )
        v_x <- condition$condition1
        v_color <- condition$condition2
        v_facet_1 <- condition$condition3
        v_facet_2 <- condition$condition4
        datplot <- comparisons(
            mfx@model,
            newdata = condition$newdata,
            type = type,
            vcov = vcov,
            conf_level = conf_level,
            by = FALSE,
            wts = wts,
            variables = variables,
            comparison = comparison,
            transform = transform,
            cross = FALSE,
            modeldata = mfx@modeldata,
            ...
        )
    }

    # marginal
    if (!is.null(by)) {
        newdata <- sanitize_newdata(
            mfx = mfx,
            newdata = newdata,
            by = by,
            wts = wts
        )
        datplot <- comparisons(
            mfx@model,
            by = by,
            newdata = newdata,
            type = type,
            vcov = vcov,
            conf_level = conf_level,
            variables = variables,
            wts = wts,
            comparison = comparison,
            transform = transform,
            cross = FALSE,
            modeldata = mfx@modeldata,
            ...
        )
        v_x <- by[[1]]
        v_color <- hush(by[[2]])
        v_facet_1 <- hush(by[[3]])
        v_facet_2 <- hush(by[[4]])
    }

    datplot <- plot_preprocess(
        datplot,
        v_x = v_x,
        v_color = v_color,
        v_facet_1 = v_facet_1,
        v_facet_2 = v_facet_2,
        condition = condition,
        mfx = mfx
    )
    attr(datplot$estimate, "label") <- "Comparison"

    # return immediately if the user doesn't want a plot
    if (isFALSE(draw)) {
        out <- as.data.frame(datplot)
        return(out)
    }

    # ggplot2
    insight::check_if_installed("ggplot2", minimum_version = "4.0.0")
    p <- plot_build(
        datplot,
        v_x = v_x,
        v_color = v_color,
        v_facet_1 = v_facet_1,
        v_facet_2 = v_facet_2,
        gray = gray,
        rug = rug,
        mfx = mfx
    )

    return(p)
}

#' `tinytest` helper
#' 
#' @export
#' @keywords internal
expect_slopes <- function(
    object,
    n_unique = NULL,
    pct_na = 5,
    se = TRUE,
    ...) {

    insight::check_if_installed("tinytest")

    object <- hush(slopes(object, ...))

    diff <- ""

    # class
    fail_class <- !isTRUE(checkmate::check_class(object, "slopes"))
    if (fail_class) {
        msg <- sprintf("Wrong class: `%s`.", class(object)[1])
        diff <- c(diff, msg)
    }

    # tidy()
    tid <- hush(tidy(object))
    fail_tidy <- !isTRUE(checkmate::check_data_frame(tid))
    if (fail_tidy) {
        msg <- "tidy() failed to return a data frame."
        diff <- c(diff, msg)
    }

    # na
    fail_na <- isTRUE(hush(mean(is.na(object$estimate)) * 100 > pct_na))
    if (fail_na) {
        msg <- sprintf("More than %s of missing values.", pct_na)
        diff <- c(diff, msg)
    }

    # unique
    fail_unique <- isTRUE(hush(length(unique(object$estimate)) < n_unique - 1))
    if (fail_unique) {
        msg <- sprintf("Fewer than %s unique values.", n_unique)
        diff <- c(diff, msg)
    }

    # unique
    if (isTRUE(se) && !"std.error" %in% colnames(object)) {
        msg <- sprintf("Fewer than %s unique values.", n_unique)
        diff <- c(diff, msg)
        fail_se <- TRUE
    } else {
        fail_se <- FALSE
    }

    # diff message
    diff <- paste(diff, collapse = "\n")

    # pass/fail
    fail <- fail_class || fail_tidy || fail_na || fail_unique || fail_se

    # tinytest object
    out <- tinytest::tinytest(
        result = !fail,
        call = sys.call(sys.parent(1)),
        diff = diff)

    return(out)
}





#' `tinytest` helper
#' 
#' @export
#' @keywords internal
expect_predictions <- function(object,
                               se = TRUE,
                               n_row = NULL,
                               n_col = NULL) {

    insight::check_if_installed("tinytest")

    diff <- ""

    # class
    fail_class <- !isTRUE(checkmate::check_class(object, "predictions"))
    if (fail_class) {
        msg <- sprintf("Wrong class: `%s`.", class(object)[1])
        diff <- c(diff, msg)
    }

    # std.error
    if (isTRUE(se) && !"std.error" %in% colnames(object)) {
        msg <- "No standard error."
        diff <- c(diff, msg)
        fail_se <- TRUE
    } else {
        fail_se <- FALSE
    }

    # rows and cols
    if (isTRUE(n_row > nrow(object))) {
        msg <- sprintf("Number of rows: %s", nrow(object))
        diff <- c(diff, msg)
        fail_row <- TRUE
    } else {
        fail_row <- FALSE
    }

    if (isTRUE(n_col > ncol(object))) {
        msg <- sprintf("Number of columns: %s", ncol(object))
        diff <- c(diff, msg)
        fail_col <- TRUE
    } else {
        fail_col <- FALSE
    }

    # diff message
    diff <- paste(diff, collapse = "\n")

    # pass/fail
    fail <- fail_class || fail_se || fail_row || fail_col

    # tinytest object
    out <- tinytest::tinytest(
        result = !fail,
        call = sys.call(sys.parent(1)),
        diff = diff)

    return(out)
}


#' `tinytest` helper
#' 
#' @export
#' @keywords internal
expect_margins <- function(results,
                           margins_object,
                           se = TRUE,
                           tolerance = 1e-5,
                           verbose = FALSE) {

    insight::check_if_installed("tinytest")

    is_equal <- function(x, y) {
        all(abs((x - y) / x) < tolerance)
    }

    results$type <- NULL

    margins_object <- data.frame(margins_object)
    term_names <- unique(results$term)

    flag <- TRUE

    # dydx
    for (tn in term_names) {
        unknown <- results[results$term == tn, "estimate"]
        lab <- paste0("dydx_", tn)
        if (lab %in% colnames(margins_object)) {
            known <- as.numeric(margins_object[, lab])
            tmp <- is_equal(known, unknown)
            if (isFALSE(tmp)) {
                flag <- FALSE
                if (isTRUE(verbose)) print(sprintf("dydx: %s", tn))
            }
        }
    }

    # std.error
    if (isTRUE(se) && "std.error" %in% colnames(results)) {
        for (tn in term_names) {
            lab_se <- paste0("SE_dydx_", tn)
            lab_var <- paste0("Var_dydx_", tn)
            if (lab_se %in% colnames(margins_object)) {
                unknown <- results[results$term == tn, "std.error"]
                known <- as.numeric(margins_object[, lab_se])
                tmp <- is_equal(known, unknown)
                if (isFALSE(tmp)) {
                    flag <- FALSE
                    if (isTRUE(verbose)) print(sprintf("se: %s", tn))
                }
            } else if (lab_var %in% colnames(margins_object)) {
                unknown <- results[results$term == tn, "std.error"]
                known <- sqrt(as.numeric(margins_object[, lab_var]))
                tmp <- is_equal(known, unknown)
                if (isFALSE(tmp)) {
                    flag <- FALSE
                    if (isTRUE(verbose)) print(sprintf("Var: %s", tn))
                }
            } else {
                flag <- FALSE
                if (isTRUE(verbose)) print(sprintf("missing column: %s", lab))
            }
        }
    }

    # tinytest object
    out <- tinytest::tinytest(
        result = flag,
        call = sys.call(sys.parent(1)),
        diff = diff)

    return(out)
}

#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.svyolr <- function(model, ...) {
    out <- insight::get_parameters(model)
    out <- stats::setNames(out$Estimate, out$Parameter)

    return(out)
}

#' @include set_coef.R
#' @rdname set_coef
#' @export

set_coef.svyolr <- function(model, coefs, ...) {
    # in basic model classes coefficients are named vector
    idx <- match(names(model$coefficients), names(coefs))
    model[["coefficients"]] <- coefs[idx]

    # thresholds can be named either "1|2" or "Intercept: 1|2"
    zeta_names <- names(model$zeta)
    idx <- match(zeta_names, names(coefs))
    if (anyNA(idx)) {
        zeta_alt <- paste("Intercept:", zeta_names)
        idx_alt <- match(zeta_alt, names(coefs))
        idx <- ifelse(is.na(idx), idx_alt, idx)
    }
    model[["zeta"]] <- coefs[idx]
    model
}

#' @include get_group_names.R
#' @rdname get_group_names
#' @export
get_group_names.svyolr <- function(model, ...) {
    resp <- insight::get_response(model)
    if (is.factor(resp)) {
        out <- levels(resp)
    } else {
        out <- unique(resp)
    }
    return(out)
}

#' @include get_predict.R
#' @rdname get_predict
#' @export
get_predict.svyolr <- function(
    model,
    newdata = insight::get_data(model),
    type = "probs",
    mfx = NULL,
    ...) {
    calling_function <- if (!is.null(mfx)) mfx@calling_function else "predictions"
    type <- sanitize_type(model, type, calling_function = calling_function)

    # hack: 1-row newdata returns a vector, so get_predict.default
    # does not learn about groups
    hack <- FALSE
    if (nrow(newdata) == 1) {
        hack <- TRUE
        newdata <- newdata[c(1, 1), , drop = FALSE]
    }

    out <- get_predict.default(model, newdata = newdata, type = type, ...)

    if (hack) {
        out <- out[1, , drop = FALSE]
    }

    out <- add_rowid(out, newdata)

    return(out)
}


#' @include get_predict.R
#' @rdname get_predict
#' @export
get_predict.svyglm <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    mfx = NULL,
    newparams = NULL,
    ndraws = NULL,
    se.fit = FALSE,
    ...) {
    estimate <- stats::predict(
        model,
        newdata = newdata,
        type = type,
        se.fit = se.fit
    )
    out <- data.frame(estimate = as.numeric(estimate))
    row.names(out) <- NULL
    return(out)
}


#' @include sanity_model.R
#' @rdname sanitize_model_specific
#' @export
sanitize_model_specific.svyolr <- function(
    model,
    wts = FALSE,
    by = FALSE,
    ...) {
    if (isFALSE(wts) && !isFALSE(by)) {
        warning(
            "With models of this class, it is normally good practice to specify weights using the `wts` argument. Otherwise, weights will be ignored in the computation of quantities of interest.",
            call. = FALSE
        )
    }
    return(model)
}


#' @include sanity_model.R
#' @rdname sanitize_model_specific
#' @export
sanitize_model_specific.svyglm <- sanitize_model_specific.svyolr

get_contrast_data_factor <- function(model,
                                     newdata,
                                     variable,
                                     cross,
                                     first_cross,
                                     modeldata,
                                     ...) {

    if (is.factor(newdata[[variable$name]])) {
        levs <- levels(newdata[[variable$name]])
        convert_to_factor <- TRUE

    } else if (get_variable_class(newdata, variable$name, "binary")) {
        levs <- variable$value
        convert_to_factor <- FALSE

    } else {
        msg <- "The `%s` variable is treated as a categorical (factor) variable, but the original data is of class %s. It is safer and faster to convert such variables to factor before fitting the model and calling a `marginaleffects` function." 
        msg <- sprintf(msg, variable$name, class(newdata[[variable$name]])[1])
        warn_once(msg, "marginaleffects_warning_factor_on_the_fly_conversion")
        if (is.factor(modeldata[[variable$name]])) {
            levs <- levels(modeldata[[variable$name]])
            convert_to_factor <- TRUE
        } else {
            levs <- sort(unique(modeldata[[variable$name]]))
            convert_to_factor <- FALSE
        }
    }

    # string shortcuts
    flag <- checkmate::check_choice(variable$value, c("reference", "revreference", "pairwise", "revpairwise", "sequential", "revsequential", "all", "minmax"))
    if (isTRUE(flag)) {
        levs_idx <- contrast_categories_shortcuts(levs, variable, interaction)

    # custom data frame or function
    } else if (isTRUE(checkmate::check_function(variable$value)) || isTRUE(checkmate::check_data_frame(variable$value))) {
        out <- contrast_categories_custom(variable, newdata)
        return(out)

    # vector of two values
    } else if (isTRUE(checkmate::check_atomic_vector(variable$value, len = 2))) {
        if (is.character(variable$value)) {
            tmp <- modeldata[[variable$name]]
            if (!all(variable$value %in% as.character(tmp))) {
                msg <- "Some of the values supplied to the `variables` argument were not found in the dataset."
                insight::format_error(msg)
            }
            idx <- match(variable$value, as.character(tmp))
            levs_idx <- data.table::data.table(lo = tmp[idx[1]], hi = tmp[idx[[2]]])
        } else if (is.numeric(variable$value)) {
            tmp <- newdata[[variable$name]]
            if (convert_to_factor) {
                levs_idx <- data.table::data.table(
                    lo = factor(as.character(variable$value[1]), levels = levels(tmp)),
                    hi = factor(as.character(variable$value[2]), levels = levels(tmp)))
            } else {
                levs_idx <- data.table::data.table(lo = variable$value[1], hi = variable$value[2])
            }
        } else {
            levs_idx <- data.table::data.table(lo = variable$value[1], hi = variable$value[2])
        }
    }


    tmp <- contrast_categories_processing(first_cross, levs_idx, levs, variable, newdata)
    lo <- tmp[[1]]
    hi <- tmp[[2]]
    original <- tmp[[3]]

    if (is.factor(newdata[[variable$name]]) || isTRUE(convert_to_factor)) {
        lo[[variable$name]] <- factor(lo[["marginaleffects_contrast_lo"]], levels = levs)
        hi[[variable$name]] <- factor(hi[["marginaleffects_contrast_hi"]], levels = levs)
    } else {
        lo[[variable$name]] <- lo[["marginaleffects_contrast_lo"]]
        hi[[variable$name]] <- hi[["marginaleffects_contrast_hi"]]
    }

    contrast_label <- hi$marginaleffects_contrast_label
    contrast_null <- hi$marginaleffects_contrast_hi == hi$marginaleffects_contrast_lo

    tmp <- !grepl("^marginaleffects_contrast", colnames(lo))
    lo <- lo[, tmp, with = FALSE]
    hi <- hi[, tmp, with = FALSE]

    out <- list(rowid = original$rowid,
                lo = lo,
                hi = hi,
                original = original,
                ter = rep(variable$name, nrow(lo)), # lo can be different dimension than newdata
                lab = contrast_label,
                contrast_null = contrast_null)
    return(out)
}


contrast_categories_shortcuts <- function(levs, variable, interaction) {
    # index contrast orders based on variable$value
    if (isTRUE(variable$value %in% c("reference", "revreference"))) {
        # null contrasts are interesting with interactions
        if (!isTRUE(interaction)) {
            levs_idx <- data.table::data.table(lo = levs[1], hi = levs[2:length(levs)])
        } else {
            levs_idx <- data.table::data.table(lo = levs[1], hi = levs)
        }

    } else if (isTRUE(variable$value %in% c("pairwise", "revpairwise"))) {
        levs_idx <- CJ(lo = levs, hi = levs, sorted = FALSE)
        # null contrasts are interesting with interactions
        if (!isTRUE(interaction)) {
            levs_idx <- levs_idx[levs_idx$hi != levs_idx$lo, ]
            levs_idx <- levs_idx[match(levs_idx$lo, levs) < match(levs_idx$hi, levs), ]
        }

    } else if (isTRUE(variable$value %in% c("sequential", "revsequential"))) {
        levs_idx <- data.table::data.table(lo = levs[1:(length(levs) - 1)],
                                           hi = levs[2:length(levs)])

    } else if (isTRUE(variable$value == "all")) {
        levs_idx <- CJ(lo = levs, hi = levs, sorted = FALSE)

    } else if (isTRUE(variable$value == "minmax")) {
        levs_idx <- data.table::data.table(lo = levs[1], hi = levs[length(levs)])
    }

    if (isTRUE(variable$value %in% c("revreference", "revpairwise", "revsequential"))) {
        levs_idx <- levs_idx[, .(lo = hi, hi = lo)]
    }

    return(levs_idx)
}



contrast_categories_df <- function(variable) {
    # manual data frame
    if (all(c("low", "high") %in% colnames(variable$value))) {
        low <- variable$value$low
        high <- variable$value$high
    } else if (all(c("lo", "hi") %in% colnames(variable$value))) {
        low <- variable$value$low
        high <- variable$value$high
    } else {
        low <- variable$value[[1]]
        high <- variable$value[[2]]
    }
    levs_idx <- data.table::data.table(
        lo = low,
        hi = high
    )
    return(levs_idx)
}


contrast_categories_processing <- function(first_cross, levs_idx, levs, variable, newdata) {
    # internal option applied to the first of several contrasts when
    # interaction=TRUE to avoid duplication. when only the first contrast
    # flips, we get a negative sign, but if first increases and second
    # decreases, we get different total effects.
    if (isTRUE(first_cross)) {
        idx <- match(levs_idx$hi, levs) >= match(levs_idx$lo, levs)
        if (sum(idx) > 0) {
            levs_idx <- levs_idx[idx, , drop = FALSE]
        }
    }

    levs_idx$isNULL <- levs_idx$hi == levs_idx$lo
    levs_idx$label <- suppressWarnings(tryCatch(
        sprintf(variable$label, levs_idx$hi, levs_idx$lo),
        error = function(e) variable$label))
    levs_idx <- stats::setNames(levs_idx, paste0("marginaleffects_contrast_", colnames(levs_idx)))
    if (!"marginaleffects_contrast_label" %in% colnames(levs_idx) || all(levs_idx$marginaleffects_contrast_label == "custom")) {
        levs_idx[, "marginaleffects_contrast_label" := paste0(marginaleffects_contrast_hi, ", ", marginaleffects_contrast_lo)]
    }

    lo <- hi <- cjdt(list(newdata, levs_idx))
    original <- data.table::rbindlist(rep(list(newdata), nrow(levs_idx)))
    return(list(lo, hi, original))
}





contrast_categories_custom <- function(variable, newdata) {
    original <- newdata
    if (!"rowid" %in% colnames(original)) {
        original$rowid <- seq_len(nrow(original))
    }
    hi <- lo <- original
    if (isTRUE(checkmate::check_function(variable$value))) {
        variables_df <- variable$value(newdata[[variable$name]])
    } else if (isTRUE(checkmate::check_data_frame(variable$value))) {
        variables_df <- variable$value
    }
    checkmate::assert_data_frame(variables_df, nrows = nrow(original))
    if (all(c("low", "high") %in% colnames(variables_df))) {
        lo[[variable$name]] <- variables_df[["low"]]
        hi[[variable$name]] <- variables_df[["high"]]
    } else if (all(c("lo", "hi") %in% colnames(variables_df))) {
        lo[[variable$name]] <- variables_df[["lo"]]
        hi[[variable$name]] <- variables_df[["hi"]]
    } else {
        lo[[variable$name]] <- variables_df[[1]]
        hi[[variable$name]] <- variables_df[[2]]
    }
    out <- list(
        rowid = original$rowid,
        lo = lo,
        hi = hi,
        original = original,
        ter = rep(variable$name, nrow(lo)), # lo can be different dimension than newdata
        lab = "custom",
        contrast_null = rep(FALSE, nrow(lo))
    )
    return(out)
}

#' Comparisons Between Predictions Made With Different Regressor Values
#'
#' @description
#' Predict the outcome variable at different regressor values (e.g., college
#' graduates vs. others), and compare those predictions by computing a difference,
#' ratio, or some other function. `comparisons()` can return many quantities of
#' interest, such as contrasts, differences, risk ratios, changes in log odds, lift,
#' slopes, elasticities, etc.
#'
#' * `comparisons()`: unit-level (conditional) estimates.
#' * `avg_comparisons()`: average (marginal) estimates.
#'
#' `variables` identifies the focal regressors whose "effect" we are interested in. `comparison` determines how predictions with different regressor values are compared (difference, ratio, odds, etc.). The `newdata` argument and the `datagrid()` function control where statistics are evaluated in the predictor space: "at observed values", "at the mean", "at representative values", etc.
#'
#' See the comparisons chapter on the package website for worked examples and case studies:
#'
#' * [https://marginaleffects.com/chapters/comparisons.html](https://marginaleffects.com/chapters/comparisons.html)
#' * [https://marginaleffects.com/](https://marginaleffects.com/)
#'
#' @inheritParams slopes
#' @inheritParams predictions
#' @param variables Focal variables
#' * `NULL`: compute comparisons for all the variables in the model object (can be slow).
#' * Character vector: subset of variables (usually faster).
#' * Named list: names identify the subset of variables of interest, and values define the type of contrast to compute. Acceptable values depend on the variable type:
#'   - Factor or character variables:
#'     * "reference": Each factor level is compared to the factor reference (base) level
#'     * "all": All combinations of observed levels
#'     * "sequential": Each factor level is compared to the previous factor level
#'     * "pairwise": Each factor level is compared to all other levels
#'     * "minmax": The highest and lowest levels of a factor.
#'     * "revpairwise", "revreference", "revsequential": inverse of the corresponding hypotheses.
#'     * Vector of length 2 with the two values to compare.
#'     * Data frame with the same number of rows as `newdata`, with two columns of "lo" and "hi" values to compare.
#'     * Function that accepts a vector and returns a data frame with two columns of "lo" and "hi" values to compare. See examples below.
#'   - Logical variables:
#'     * NULL: contrast between TRUE and FALSE
#'     * Data frame with the same number of rows as `newdata`, with two columns of "lo" and "hi" values to compare.
#'     * Function that accepts a vector and returns a data frame with two columns of "lo" and "hi" values to compare. See examples below.
#'   - Numeric variables:
#'     * Numeric of length 1: Forward contrast for a gap of `x`, computed between the observed value and the observed value plus `x`. Users can set a global option to get a "center" or "backward" contrast instead: `options(marginaleffects_contrast_direction="center")`
#'     * Numeric vector of length 2: Contrast between the largest and the smallest elements of the `x` vector.
#'     * Data frame with the same number of rows as `newdata`, with two columns of "lo" and "hi" values to compare.
#'     * Function that accepts a vector and returns a data frame with two columns of "lo" and "hi" values to compare. See examples below.
#'     * "iqr": Contrast across the interquartile range of the regressor.
#'     * "sd": Contrast across one standard deviation around the regressor mean.
#'     * "2sd": Contrast across two standard deviations around the regressor mean.
#'     * "minmax": Contrast between the maximum and the minimum values of the regressor.
#'   - Examples:
#'     + `variables = list(gear = "pairwise", hp = 10)`
#'     + `variables = list(gear = "sequential", hp = c(100, 120))`
#'     + `variables = list(hp = function(x) data.frame(low = x - 5, high = x + 10))`
#'     + See the Examples section below for more.
#' @param newdata Grid of predictor values at which we evaluate the comparisons.
#' + Warning: Avoid modifying your dataset between fitting the model and calling a `marginaleffects` function. This can sometimes lead to unexpected results.
#' + `NULL` (default): Unit-level contrasts for each observed value in the dataset (empirical distribution). The dataset is retrieved using [insight::get_data()], which tries to extract data from the environment. This may produce unexpected results if the original data frame has been altered since fitting the model.
#' + data frame: Unit-level contrasts for each row of the `newdata` data frame.
#' + string:
#'   - "mean": Contrasts at the Mean. Contrasts when each predictor is held at its mean or mode.
#'   - "median": Contrasts at the Median. Contrasts when each predictor is held at its median or mode.
#'   - "balanced": Contrasts evaluated on a balanced grid with every combination of categories and numeric variables held at their means.
#'   - "tukey": Contrasts at Tukey's 5 numbers.
#'   - "grid": Contrasts on a grid of representative numbers (Tukey's 5 numbers and unique values of categorical predictors).
#' + [datagrid()] call to specify a custom grid of regressors. For example:
#'   - `newdata = datagrid(cyl = c(4, 6))`: `cyl` variable equal to 4 and 6 and other regressors fixed at their means or modes.
#'   - `newdata = datagrid(mpg = fivenum)`: `mpg` variable held at Tukey's five numbers (using the `fivenum` function), and other regressors fixed at their means or modes.
#'   - See the Examples section and the [datagrid] documentation.
#' + [subset()] call with a single argument to select a subset of the dataset used to fit the model, ex: `newdata = subset(treatment == 1)`
#' + [dplyr::filter()] call with a single argument to select a subset of the dataset used to fit the model, ex: `newdata = filter(treatment == 1)`
#' @param comparison How should pairs of predictions be compared? Difference, ratio, odds ratio, or user-defined functions.
#' * string: shortcuts to common contrast functions. Supported shortcuts strings: `r toString(names(marginaleffects:::comparison_function_dict))`
#' * function: accept two equal-length numeric vectors of adjusted predictions (`hi` and `lo`) and returns a vector of contrasts of the same length, or a unique numeric value.
#' * See the "Comparison functions" section below for a list of common transformations and the definitions of available shortcuts.
#' @param transform string or function. Transformation applied to unit-level estimates and confidence intervals just before the function returns results. Functions must accept a vector and return a vector of the same length. Support string shortcuts: "exp", "ln"
#' @param equivalence Numeric vector of length 2: bounds used for the two-one-sided test (TOST) of equivalence, and for the non-inferiority and non-superiority tests. For bayesian models, this report the proportion of posterior draws in the interval and the ROPE. See Details section below.
#' @param by Aggregate unit-level estimates (aka, marginalize, average over). Valid inputs:
#'   - `FALSE`: return the original unit-level estimates.
#'   - `TRUE`: aggregate estimates for each term.
#'   - Character vector of column names in `newdata` or in the data frame produced by calling the function without the `by` argument.
#'   - Data frame with a `by` column of group labels, and merging columns shared by `newdata` or the data frame produced by calling the same function without the `by` argument.
#'   - See examples below.
#'   - For more complex aggregations, you can use the `FUN` argument of the `hypotheses()` function. See that function's documentation and the Hypothesis Test vignettes on the `marginaleffects` website.
#' @param cross
#' * `FALSE`: Contrasts represent the change in adjusted predictions when one predictor changes and all other variables are held constant.
#' * `TRUE`: Contrasts represent the changes in adjusted predictions when all the predictors specified in the `variables` argument are manipulated simultaneously (a "cross-contrast").
#' @template references
#' @template deltamethod
#' @template model_specific_arguments
#' @template comparison_functions
#' @template bayesian
#' @template equivalence
#' @template order_of_operations
#' @template parallel
#' @template options
#' @template return
#' @template type
#'
#' @examplesIf interactive() || isTRUE(Sys.getenv("R_DOC_BUILD") == "true")
#' # Linear model
#' tmp <- mtcars
#' tmp$am <- as.logical(tmp$am)
#' mod <- lm(mpg ~ am + factor(cyl), tmp)
#' avg_comparisons(mod, variables = list(cyl = "reference"))
#' avg_comparisons(mod, variables = list(cyl = "sequential"))
#' avg_comparisons(mod, variables = list(cyl = "pairwise"))
#'
#' # GLM with different scale types
#' mod <- glm(am ~ factor(gear), data = mtcars)
#' avg_comparisons(mod, type = "response")
#' avg_comparisons(mod, type = "link")
#'
#' # Contrasts at the mean
#' comparisons(mod, newdata = "mean")
#'
#' # Contrasts between marginal means
#' comparisons(mod, newdata = "balanced")
#'
#' # Contrasts at user-specified values
#' comparisons(mod, newdata = datagrid(am = 0, gear = tmp$gear))
#' comparisons(mod, newdata = datagrid(am = unique, gear = max))
#'
#' m <- lm(mpg ~ hp + drat + factor(cyl) + factor(am), data = mtcars)
#' comparisons(m, variables = "hp", newdata = datagrid(FUN_factor = unique, FUN_numeric = median))
#'
#' # Numeric contrasts
#' mod <- lm(mpg ~ hp, data = mtcars)
#' avg_comparisons(mod, variables = list(hp = 1))
#' avg_comparisons(mod, variables = list(hp = 5))
#' avg_comparisons(mod, variables = list(hp = c(90, 100)))
#' avg_comparisons(mod, variables = list(hp = "iqr"))
#' avg_comparisons(mod, variables = list(hp = "sd"))
#' avg_comparisons(mod, variables = list(hp = "minmax"))
#'
#' # using a function to specify a custom difference in one regressor
#' dat <- mtcars
#' dat$new_hp <- 49 * (dat$hp - min(dat$hp)) / (max(dat$hp) - min(dat$hp)) + 1
#' modlog <- lm(mpg ~ log(new_hp) + factor(cyl), data = dat)
#' fdiff <- function(x) data.frame(x, x + 10)
#' avg_comparisons(modlog, variables = list(new_hp = fdiff))
#'
#' # Adjusted Risk Ratio
#' mod <- glm(vs ~ mpg, data = mtcars, family = binomial)
#' avg_comparisons(mod, comparison = "lnratioavg", transform = exp)
#'
#' # Adjusted Risk Ratio: Manual specification of the `comparison`
#' avg_comparisons(
#'   mod,
#'   comparison = function(hi, lo) log(mean(hi) / mean(lo)),
#'   transform = exp)
#' # cross contrasts
#' mod <- lm(mpg ~ factor(cyl) * factor(gear) + hp, data = mtcars)
#' avg_comparisons(mod, variables = c("cyl", "gear"), cross = TRUE)
#'
#' # variable-specific contrasts
#' avg_comparisons(mod, variables = list(gear = "sequential", hp = 10))
#'
#' # hypothesis test: is the `hp` marginal effect at the mean equal to the `drat` marginal effect
#' mod <- lm(mpg ~ wt + drat, data = mtcars)
#'
#' comparisons(
#'   mod,
#'   newdata = "mean",
#'   hypothesis = "wt = drat")
#'
#' # same hypothesis test using row indices
#' comparisons(
#'   mod,
#'   newdata = "mean",
#'   hypothesis = "b1 - b2 = 0")
#'
#' # same hypothesis test using numeric vector of weights
#' comparisons(
#'   mod,
#'   newdata = "mean",
#'   hypothesis = c(1, -1))
#'
#' # two custom contrasts using a matrix of weights
#' lc <- matrix(
#'   c(
#'     1, -1,
#'     2, 3),
#'   ncol = 2)
#' comparisons(
#'   mod,
#'   newdata = "mean",
#'   hypothesis = lc)
#'
#' # Effect of a 1 group-wise standard deviation change
#' # First we calculate the SD in each group of `cyl`
#' # Second, we use that SD as the treatment size in the `variables` argument
#' library(dplyr)
#' mod <- lm(mpg ~ hp + factor(cyl), mtcars)
#' tmp <- mtcars %>%
#'   group_by(cyl) %>%
#'   mutate(hp_sd = sd(hp))
#' avg_comparisons(mod,
#'   variables = list(hp = function(x) data.frame(x, x + tmp$hp_sd)),
#'   by = "cyl")
#'
#' # `by` argument
#' mod <- lm(mpg ~ hp * am * vs, data = mtcars)
#' comparisons(mod, by = TRUE)
#'
#' mod <- lm(mpg ~ hp * am * vs, data = mtcars)
#' avg_comparisons(mod, variables = "hp", by = c("vs", "am"))
#'
#' library(nnet)
#' mod <- multinom(factor(gear) ~ mpg + am * vs, data = mtcars, trace = FALSE)
#' by <- data.frame(
#'   group = c("3", "4", "5"),
#'   by = c("3,4", "3,4", "5"))
#' comparisons(mod, type = "probs", by = by)
#'
#' @export
comparisons <- function(
    model,
    newdata = NULL,
    variables = NULL,
    comparison = "difference",
    type = NULL,
    vcov = TRUE,
    by = FALSE,
    conf_level = 0.95,
    transform = NULL,
    cross = FALSE,
    wts = FALSE,
    hypothesis = NULL,
    equivalence = NULL,
    df = Inf,
    eps = NULL,
    numderiv = "fdforward",
    ...
) {
    # init
    if (inherits(model, "marginaleffects_internal")) {
        mfx <- model
    } else {
        # pass through ... to avoid calling `get_modeldata()` in inferences()
        if ("modeldata" %in% ...names()) {
            modeldata <- ...get("modeldata")
        } else {
            modeldata <- NULL
        }
        call <- construct_call(model, "comparisons")
        model <- sanitize_model(model, 
            call = call, 
            newdata = newdata, 
            wts = wts, 
            vcov = vcov, 
            by = by, ...)
        mfx <- new_marginaleffects_internal(
            call = call,
            model = model,
            modeldata = modeldata,
            by = by,
            comparison = comparison,
            cross = cross,
            eps = eps
        )
    }

    scall <- rlang::enquo(newdata)
    mfx <- add_newdata(mfx, 
        scall, 
        newdata = newdata, 
        by = by, 
        wts = wts, 
        cross = cross, 
        comparison = comparison)

    # sanity checks
    dots <- list(...)
    sanity_dots(mfx@model, ...)

    # multiple imputation
    if (inherits(mfx@model, c("mira", "amest"))) {
        out <- process_imputation(mfx)
        return(out)
    }

    # inferences() dispatch
    methods <- c("rsample", "boot", "fwb", "simulation")
    if (isTRUE(checkmate::check_choice(vcov, methods))) {
        inferences_method <- vcov
        vcov <- FALSE
    } else {
        inferences_method <- NULL
    }

    # very early, before any use of newdata
    newdata <- mfx@newdata

    # misc
    mfx@conf_level <- sanitize_conf_level(conf_level, ...)
    mfx <- add_numderiv(mfx, numderiv)
    checkmate::assert_number(eps, lower = 1e-10, null.ok = TRUE)

    # misc sanitation
    mfx <- add_by(mfx, by)
    sanity_reserved(mfx)

    cross <- sanitize_cross(cross, variables, mfx@model)
    mfx@type <- sanitize_type(
        model = mfx@model,
        type = type,
        calling_function = mfx@calling_function
    )
    sanity_comparison(comparison)

    # transforms
    transform <- sanitize_transform(transform)

    # after sanitize_newdata
    mfx <- add_variables(
        variables = variables,
        mfx = mfx
    )

    # get dof before transforming the vcov arg
    # add_degrees_of_freedom() produces a weird warning on non lmerMod. We can skip them
    # because get_vcov() will produce an informative error later.
    mfx <- add_degrees_of_freedom(mfx = mfx, df = df, by = by, hypothesis = hypothesis, vcov = vcov)

    mfx@vcov_type <- get_vcov_label(vcov)
    mfx@vcov_model <- get_vcov(mfx@model, vcov = vcov, type = mfx@type, ...)

    predictors <- mfx@variables

    mfx <- add_hypothesis(mfx, hypothesis)

    ############### sanity checks are over

    # compute contrasts and standard errors
    args <- list(
        mfx = mfx,
        variables = predictors,
        cross = cross
    )
    dots[["modeldata"]] <- NULL # dont' pass twice
    args <- utils::modifyList(args, dots)
    contrast_data <- do_call(get_comparisons_data, args)

    # Add model matrices to hi/lo data
    contrast_data$hi <- add_model_matrix_attribute_data(mfx, contrast_data$hi)
    contrast_data$lo <- add_model_matrix_attribute_data(mfx, contrast_data$lo)

    # Try autodiff early exit
    autodiff_result <- NULL
    if (isTRUE(settings_get("autodiff")) &&
        !isFALSE(vcov) &&
        isTRUE(checkmate::check_matrix(mfx@vcov_model))) {

        # Add internal group ID for robust matching with Python results
        # This ensures we can merge results back correctly regardless of sort order
        if (is.character(by)) {
            bycols <- unique(c(by, "term", "contrast"))
            bycols <- intersect(bycols, colnames(contrast_data$original))
            if (length(bycols) > 0) {
                # Create group IDs based on unique combinations, sorted to match finite diff output order
                # Finite diff sorts by (term, user-specified variables, then contrast)
                # So reorder bycols to put term first, then user variables, then contrast
                bycols_sorted <- c("term", setdiff(bycols, c("term", "contrast")), "contrast")
                bycols_sorted <- intersect(bycols_sorted, bycols)  # Keep only columns that exist
                group_lookup <- unique(contrast_data$original[, bycols, with = FALSE])
                # Sort by the reordered columns to match finite diff order
                data.table::setorderv(group_lookup, bycols_sorted)
                group_lookup[, marginaleffects_group_id := seq_len(.N)]
                # Add group ID to original data
                contrast_data$original <- merge(
                    contrast_data$original,
                    group_lookup,
                    by = bycols,
                    sort = FALSE
                )
            }
        }

        autodiff_result <- jax_comparisons(
            mfx = mfx,
            vcov_matrix = mfx@vcov_model,
            hi = contrast_data$hi,
            lo = contrast_data$lo,
            original = contrast_data$original,
            ...
        )
    }

    if (!is.null(autodiff_result)) {
        # SUCCESS: Use autodiff results and skip computation pipeline

        # Build result data.frame
        if (isFALSE(mfx@by)) {
            # Row-level results - merge with original data
            cmp <- data.table::copy(contrast_data$original)
            cmp[, estimate := autodiff_result$estimate]
            cmp[, std.error := autodiff_result$std.error]
        } else {
            # Aggregated results
            cmp <- data.table::data.table(
                estimate = autodiff_result$estimate,
                std.error = autodiff_result$std.error
            )

            # Add grouping columns
            if (isTRUE(mfx@by)) {
                # by=TRUE: aggregate by term/contrast
                bycols <- intersect(c("term", "contrast"), colnames(contrast_data$original))
                if (length(bycols) > 0) {
                    group_cols <- unique(contrast_data$original[, bycols, with = FALSE])
                    cmp <- cbind(group_cols, cmp)
                }
            } else if (is.character(mfx@by)) {
                # by=character: aggregate by user-specified + term/contrast
                bycols <- unique(c(mfx@by, "term", "contrast"))
                bycols <- intersect(bycols, colnames(contrast_data$original))
                if (length(bycols) > 0) {
                    # Create lookup table with group IDs, sorted to match finite diff output order
                    # Finite diff sorts by (term, user-specified variables, then contrast)
                    bycols_sorted <- c("term", setdiff(bycols, c("term", "contrast")), "contrast")
                    bycols_sorted <- intersect(bycols_sorted, bycols)  # Keep only columns that exist
                    group_lookup <- unique(contrast_data$original[, bycols, with = FALSE])
                    data.table::setorderv(group_lookup, bycols_sorted)
                    group_lookup[, marginaleffects_group_id := seq_len(.N)]

                    # Python returns results in sorted order (by group_id which we passed)
                    # So results are already in the correct order!
                    cmp <- data.table::data.table(
                        estimate = autodiff_result$estimate,
                        std.error = autodiff_result$std.error
                    )
                    cmp[, marginaleffects_group_id := seq_len(.N)]

                    # Merge to add back the grouping columns
                    cmp <- merge(group_lookup, cmp, by = "marginaleffects_group_id", sort = FALSE)
                    cmp[, marginaleffects_group_id := NULL]  # Remove internal ID
                }
            }
        }

        # Store jacobian
        mfx@jacobian <- autodiff_result$jacobian

    } else {
        # FALLBACK: Use standard computation pipeline

        args <- list(
            mfx = mfx,
            variables = predictors,
            type = mfx@type,
            original = contrast_data[["original"]],
            hi = contrast_data[["hi"]],
            lo = contrast_data[["lo"]],
            by = by,
            cross = cross,
            hypothesis = mfx@hypothesis
        )
        args <- utils::modifyList(args, dots)
        cmp <- do_call(get_comparisons, args)

        # hypothesis formula names are attached in by() in get_predictions()
        mfx@variable_names_by <- unique(c(
            mfx@variable_names_by,
            attr(cmp, "hypothesis_function_by")))

        # bayesian posterior
        mfx@draws <- attr(cmp, "posterior_draws")

        # standard errors via delta method
        if (is.null(mfx@draws) &&
            !isFALSE(vcov) &&
            isTRUE(checkmate::check_matrix(mfx@vcov_model))) {

            idx <- intersect(colnames(cmp), c("group", "term", "contrast"))
            idx <- cmp[, (idx), drop = FALSE]
            fun <- function(...) {
                get_comparisons(..., verbose = FALSE)$estimate
            }
            args <- list(
                mfx = mfx,
                model_perturbed = mfx@model,
                vcov = mfx@vcov_model,
                type = mfx@type,
                FUN = fun,
                index = idx,
                variables = predictors,
                hypothesis = mfx@hypothesis,
                hi = contrast_data$hi,
                lo = contrast_data$lo,
                original = contrast_data$original,
                estimates = cmp,
                numderiv = numderiv
            )
            args <- utils::modifyList(args, dots)
            se <- do_call(get_se_delta, args)
            mfx@jacobian <- attr(se, "jacobian")
            cmp$std.error <- as.vector(as.numeric(se)) # drop attributes
            mfx@draws <- NULL
        }
    }

    # Common path for both autodiff and fallback

    # merge original data back in
    if (isFALSE(by) && "rowid" %in% colnames(cmp)) {
        if ("rowid" %in% colnames(mfx@newdata)) {
            idx <- c(
                "rowid",
                "rowidcf",
                "term",
                "contrast",
                "by",
                setdiff(colnames(contrast_data$original), colnames(cmp))
            )
            idx <- intersect(idx, colnames(contrast_data$original))
            tmp <- contrast_data$original[, ..idx, drop = FALSE]
            # contrast_data is duplicated to compute contrasts for different terms or pairs
            bycols <- intersect(colnames(tmp), colnames(cmp))
            idx <- duplicated(tmp, by = bycols)
            tmp <- tmp[!idx]
            cmp <- merge(cmp, tmp, all.x = TRUE, by = bycols, sort = FALSE)
            # HACK: relies on NO sorting at ANY point
        } else {
            idx <- setdiff(colnames(contrast_data$original), colnames(cmp))
            cmp <- data.table(cmp, contrast_data$original[, ..idx])
        }
    }

    # meta info
    cmp <- get_ci(cmp, mfx)

    # clean rows and columns
    # WARNING: we cannot sort rows at the end because `get_hypothesis()` is
    # applied in the middle, and it must already be sorted in the final order,
    # otherwise, users cannot know for sure what is going to be the first and
    # second rows, etc.
    cmp <- sort_columns(cmp, mfx@newdata, by)

    # equivalence tests
    cmp <- equivalence(cmp, equivalence = equivalence, df = mfx@df, draws = mfx@draws, ...)

    # after draws attribute
    cmp <- backtransform(cmp, transform, draws = mfx@draws)
    new_draws <- attr(cmp, "posterior_draws") # important!
    if (!is.null(new_draws)) mfx@draws <- new_draws

    # remove weights column (now handled by add_attributes)
    cmp[["marginaleffects_wts_internal"]] <- NULL

    out <- cmp

    data.table::setDF(out)

    class(out) <- c("comparisons", class(out))

    # before add_attributes()
    if (inherits(mfx@model, "brmsfit")) {
        insight::check_if_installed("brms")
        mfx@draws_chains <- brms::nchains(mfx@model) 
    }

    # class before prune
    out <- add_attributes(out, mfx)
    out <- prune_attributes(out)


    if (!is.null(inferences_method)) {
        out <- inferences(out, method = inferences_method)
    }

    return(out)
}


#' Average comparisons
#' @describeIn comparisons Average comparisons
#' @export
#'
avg_comparisons <- function(
    model,
    newdata = NULL,
    variables = NULL,
    type = NULL,
    vcov = TRUE,
    by = TRUE,
    conf_level = 0.95,
    comparison = "difference",
    transform = NULL,
    cross = FALSE,
    wts = FALSE,
    hypothesis = NULL,
    equivalence = NULL,
    df = Inf,
    eps = NULL,
    numderiv = "fdforward",
    ...
) {
    call_attr <- construct_call(model, "comparisons")

    out <- eval.parent(call_attr)

    return(out)
}

get_ci <- function(
    x,
    conf_level,
    df = NULL,
    draws = NULL,
    vcov = TRUE,
    null_hypothesis = 0,
    model = NULL,
    ...) {

    checkmate::assert_number(null_hypothesis)

    if (!is.null(draws)) {
        out <- get_ci_draws(
            x,
            conf_level = conf_level,
            draws = draws,
            model = model)
        return(out)
    }

    required <- c("estimate", "std.error")
    if (!inherits(x, "data.frame") || !all(required %in% colnames(x))) {
        return(x)
    }

    normal <- FALSE
    if (!"df" %in% colnames(x)) {
        if (identical(df, Inf)) {
            normal <- TRUE

        # 1 or matching length
        } else if (length(df) %in% c(1, nrow(x))) {
            x[["df"]] <- df
            normal <- FALSE

        # multiple, such as rbind() contrast terms
        } else if (length(df) < nrow(x) && "rowid" %in% colnames(x)) {
            rowids <- unique(x$rowid)
            if (length(rowids) == length(df)) {
                rowids <- data.table(rowid = rowids, df = df)
                x <- merge(x, rowids, all.x = TRUE, by = "rowid", sort = FALSE)
            } else {
                insight::format_error("The degrees of freedom argument was ignored.")
            }

        # mismatch
        } else {
            insight::format_error("Satterthwaite and Kenward-Roger corrections are not supported in this command.")
        }
    }

    p_overwrite <-  !"p.value" %in% colnames(x) ||
                    null_hypothesis != 0 ||
                    identical(vcov, "satterthwaite") ||
                    identical(vcov, "kenward-roger")

    z_overwrite <- !"statistic" %in% colnames(x) ||
                   null_hypothesis != 0 ||
                   p_overwrite

    ci_overwrite <- !"conf.low" %in% colnames(x) &&
                    "std.error" %in% colnames(x)

    if (z_overwrite) {
        x[["statistic"]] <- (x[["estimate"]] - null_hypothesis) / x[["std.error"]]
        if (normal) {
            x[["p.value"]] <- 2 * stats::pnorm(-abs(x$statistic))
        } else {
            x[["p.value"]] <- 2 * stats::pt(-abs(x$statistic), df = x[["df"]])
        }
    }

    if (ci_overwrite) {
        alpha <- 1 - conf_level
        if (normal) {
            critical <- abs(stats::qnorm(alpha / 2))
        } else {
            critical <- abs(stats::qt(alpha / 2, df = x[["df"]]))
        }
        x[["conf.low"]] <- x[["estimate"]] - critical * x[["std.error"]]
        x[["conf.high"]] <- x[["estimate"]] + critical * x[["std.error"]]
    }

    # s-value
    if ("p.value" %in% colnames(x)) {
        x$s.value <- -log2(x$p.value)
    }

    return(x)
}


get_ci_draws <- function(x, conf_level, draws, model = NULL) {
    

    checkmate::check_number(conf_level, lower = 1e-10, upper = 1 - 1e-10)
    critical <- (1 - conf_level) / 2

    # faster known case
    if (inherits(model, "inferences_simulation")) {
        insight::check_if_installed("collapse", minimum_version = "1.9.0")
        CIs <- collapse::dapply(draws, MARGIN = 1, FUN = collapse::fquantile, probs = c(critical, 1 - critical))
        x$std.error <- collapse::dapply(draws, MARGIN = 1, FUN = collapse::fsd)
        x$conf.low <- CIs[, 1]
        x$conf.high <- CIs[, 2]
        return(x)

    } else if (identical("eti", getOption("marginaleffects_posterior_interval", default = "eti")) &&
        identical("median", getOption("marginaleffects_posterior_center", default = "median"))) {
        insight::check_if_installed("collapse", minimum_version = "1.9.0")
        # Issue #1017
        if (nrow(draws) > 0) {
            CIs <- collapse::dapply(draws, MARGIN = 1, FUN = collapse::fquantile, probs = c(critical, 0.5, 1 - critical))
            x$estimate <- CIs[, 2]
            x$conf.low <- CIs[, 1]
            x$conf.high <- CIs[, 3]
        }
        return(x)
    }

    # faster known case
    if (identical("eti", getOption("marginaleffects_posterior_interval", default = "eti")) &&
        identical("mean", getOption("marginaleffects_posterior_center", default = "median"))) {
        insight::check_if_installed("collapse", minimum_version = "1.9.0")
        Bs <- collapse::dapply(draws, MARGIN = 1, FUN = collapse::fmean)
        CIs <- collapse::dapply(draws, MARGIN = 1, FUN = collapse::fquantile, probs = c(critical, 1 - critical))
        x$estimate <- Bs
        x$conf.low <- CIs[, 1]
        x$conf.high <- CIs[, 2]
        return(x)
    }

    # option name change
    FUN_INTERVAL <- getOption("marginaleffects_posterior_interval")
    if (is.null(FUN_INTERVAL)) {
        FUN_INTERVAL <- getOption("marginaleffects_credible_interval", default = "eti")
    }
    checkmate::assert_choice(FUN_INTERVAL, choices = c("eti", "hdi"))
    if (FUN_INTERVAL == "hdi") {
        FUN_INTERVAL <- get_hdi
    } else {
        FUN_INTERVAL <- get_eti
    }

    FUN_CENTER <- getOption("marginaleffects_posterior_center", default = stats::median)

    checkmate::assert(
        checkmate::check_choice(FUN_CENTER, choices = c("mean", "median")),
        checkmate::check_function(FUN_CENTER)
    )

    if (identical(FUN_CENTER, "mean")) {
        FUN_CENTER <- mean
    } else if (identical(FUN_CENTER, "median")) {
        FUN_CENTER <- stats::median
    }

    # necessary for `hypothesis_apply()` when attributes are added by collapse::BY or collapse::dapply
    colnames(draws) <- row.names(draws) <- NULL

    CIs <- t(apply(draws, 1, FUN_INTERVAL, credMass = conf_level))
    Bs <- apply(draws, 1, FUN_CENTER)
    # comparison returns a single value
    if (nrow(x) < nrow(CIs)) {
        CIs <- unique(CIs)
        Bs <- unique(Bs)
    }
    x[["estimate"]] <- Bs
    x[["conf.low"]] <- CIs[, "lower"]
    x[["conf.high"]] <- CIs[, "upper"]

    return(x)
}


get_eti <- function(object, credMass = 0.95, ...) {
  checkmate::assert_numeric(object)
  checkmate::assert_number(credMass)
  checkmate::assert_true(credMass > 0)
  checkmate::assert_true(credMass < 1)
  critical <- (1 - credMass) / 2
  out <- stats::quantile(object, probs = c(critical, 1 - critical))
  out <- stats::setNames(out, c("lower", "upper"))
  return(out)
}


# this is only used for tests to match emmeans. we use ETI as default for bayesian models.
get_hdi <- function(object, credMass = 0.95, ...) {
    result <- c(NA_real_, NA_real_)
    if (is.numeric(object)) {
        attributes(object) <- NULL
        x <- sort.int(object, method = "quick") # removes NA/NaN, but not Inf
        n <- length(x)
        if (n > 0) {
            # exclude <- ceiling(n * (1 - credMass)) # Not always the same as...
            exclude <- n - floor(n * credMass) # Number of values to exclude
            low.poss <- x[1:exclude] # Possible lower limits...
            upp.poss <- x[(n - exclude + 1):n] # ... and corresponding upper limits
            best <- which.min(upp.poss - low.poss) # Combination giving the narrowest interval
            if (length(best)) {
                result <- c(low.poss[best], upp.poss[best])
            } else {
                tmp <- range(x)
                if (length(tmp) == 2) {
                    result <- tmp
                }
            }
        }
    }
    names(result) <- c("lower", "upper")
    return(result)
}

get_contrast_data <- function(model,
                              newdata,
                              variables,
                              cross,
                              modeldata = NULL,
                              ...) {

    lo <- hi <- ter <- lab <- original <- rowid <- list()

    # after variable class assignment
    if (is.null(modeldata)) {
        modeldata <- attr(newdata, "newdata_modeldata")
    }
    # sometimes needed for extensions when get_data doesn't work
    if (is.null(modeldata) || nrow(modeldata) == 0) {
        modeldata <- newdata
    }
    
    # safety need for extensions not supported by `insight`
    variable_classes <- attr(newdata, "newdata_variable_class")
    if (length(variable_classes) == 0) {
        newdata <- set_variable_class(newdata, model)
        variable_classes <- attr(newdata, "marginaleffects_variable_class")
    }
    if (length(attr(modeldata, "marginaleffects_variable_class")) == 0) {
        modeldata <- set_variable_class(modeldata, model)
    }

    if (any(c("factor", "character") %in% variable_classes)) {
        first_cross <- names(variable_classes[variable_classes %in% c("factor", "character")])[1]
    } else {
        first_cross <- NULL
    }

    # must use `as.data.table()` because `setDT()` does not handle columns with
    # more dimensions (e.g., "idx" in {mlogit})
    newdata <- as.data.table(newdata)

    for (v in variables) {
        args <- list(
            model = model,
            newdata = newdata,
            variable = v,
            cross = cross,
            first_cross = identical(v$name, first_cross),
            modeldata = modeldata)
        args <- append(args, list(...))

        # logical and character before factor used to be important; but I don't think so anymore
        if (get_variable_class(modeldata, v$name, "logical")) {
            fun <- get_contrast_data_logical
        } else if (get_variable_class(modeldata, v$name, "character")) {
            fun <- get_contrast_data_character
        } else if (get_variable_class(modeldata, v$name, "categorical")) {
            fun <- get_contrast_data_factor
        } else if (get_variable_class(modeldata, v$name, "numeric")) {
            fun <- get_contrast_data_numeric
        } else {
            msg <- sprintf("Class of the `%s` variable is class is not supported.", v$name)
            stop(msg, call. = FALSE)
        }

        tmp <- do.call("fun", args)

        lo[[v$name]] <- tmp$lo
        if (isTRUE(cross)) {
            lo[[v$name]][[paste0("null_contrast_", v$name)]] <- tmp$contrast_null
        }
        hi[[v$name]] <- tmp$hi
        ter[[v$name]] <- tmp$ter
        lab[[v$name]] <- tmp$lab
        original[[v$name]] <- tmp$original
        rowid[[v$name]] <- tmp$rowid
    }

    clean <- function(x) {
        for (col in colnames(x)) {
            # tobit1 introduces AsIs columns
            if (inherits(x[[col]], "AsIs")) {
                x[[col]] <- as.numeric(x[[col]])
            }

            # plm creates c("pseries", "numeric"), but when get_contrast_data
            # assigns +1 numeric, we lose the inheritance
            if (inherits(x[[col]], "pseries")) {
                x[[col]] <- as.numeric(x[[col]])
            }

            # strip labelled data which break rbindlist()
            cl <- class(x[[col]])
            if (length(cl) == 2 && cl[1] == "labelled") {
                class(x[[col]]) <- class(x[[col]])[2]
            }

            # # mlogit uses a `newdata` with one row per unit-choice and returns
            # # an `idx` column with the choice label in second position
            # if (inherits(model, "mlogit") && inherits(x[[col]], "idx")) {
            #     x[[col]] <- NULL
            # }
        }
        return(x)
    }


    lo <- lapply(lo, clean)
    hi <- lapply(hi, clean)
    original <- lapply(original, clean)


    # single contrast
    if (!isTRUE(cross)) {
        lo <- rbindlist(lo, fill = TRUE, ignore.attr = TRUE)
        hi <- rbindlist(hi, fill = TRUE, ignore.attr = TRUE)
        original <- rbindlist(original, fill = TRUE, ignore.attr = TRUE)
        # long names to avoid user-supplied colname conflict
        marginaleffects_ter <- unlist(ter, use.names = FALSE)
        marginaleffects_lab <- unlist(lab, use.names = FALSE)
        lo[, "term" := marginaleffects_ter]
        hi[, "term" := marginaleffects_ter]
        original[, "term" := marginaleffects_ter]
        lo[, "contrast" := marginaleffects_lab]
        hi[, "contrast" := marginaleffects_lab]
        original[, "contrast" := marginaleffects_lab]

    # cross contrast
    } else {
        # drop variables for which we have contrasts
        for (i in seq_along(lo)) {
            if (i == 1) {
                # keep rowid and original data only in one of the datasets
                idx_lo <- setdiff(names(variables), names(lo)[i])
                idx_hi <- setdiff(names(variables), names(hi)[i])
                idx_or <- setdiff(names(variables), names(hi)[i])
            } else {
                # exclude rowid and variables excluded from `variables`, for
                # which we do not compute cross-contrasts
                contrast_null <- grep("rowid|^null_contrast_", colnames(lo[[i]]), value = TRUE)
                idx_lo <- c(setdiff(names(lo[[i]]), c(contrast_null, names(variables))),
                    setdiff(names(variables), names(lo)[[i]]))
                idx_hi <- c(setdiff(names(hi[[i]]), c(contrast_null, names(variables))),
                    setdiff(names(variables), names(hi)[[i]]))

                idx_or <- c(setdiff(names(original[[i]]), c(contrast_null, names(variables))),
                    setdiff(names(variables), names(original)[[i]]))
            }
            lo[[i]] <- data.table(lo[[i]])[, !..idx_lo]
            hi[[i]] <- data.table(hi[[i]])[, !..idx_hi]
            original[[i]] <- data.table(original[[i]])[, !..idx_or]
            lo[[i]][[paste0("contrast_", names(lo)[i])]] <- lab[[i]]
            hi[[i]][[paste0("contrast_", names(hi)[i])]] <- lab[[i]]
            original[[i]][[paste0("contrast_", names(original)[i])]] <- lab[[i]]
        }

        fun <- function(x, y) merge(x, y, all = TRUE, allow.cartesian = TRUE, sort = FALSE)
        lo <- Reduce("fun", lo)
        hi <- Reduce("fun", hi)
        original <- Reduce("fun", original)

        # faster to rbind, but creates massive datasets. need cartesian join on rowid
        # lo <- cjdt(lo)
        # hi <- cjdt(hi)

        # if there are fewer null_contrast_* columns, then there is at least
        # one always non-null variable type, so we keep everything
        idx <- grepl("^null_contrast_", colnames(lo))
        idx_df <- lo[, ..idx]
        lo <- lo[, !..idx]
        if (sum(idx) == length(variables)) {
            idx <- rowSums(idx_df) < ncol(idx_df)
            lo <- lo[idx]
            hi <- hi[idx]
            original <- original[idx]
        }
    }
    
    # get_predict() is much faster if we only build the model matrix once
    lo <- get_model_matrix_attribute(model, lo)
    hi <- get_model_matrix_attribute(model, hi)
    original <- get_model_matrix_attribute(model, original)

    out <- list(lo = lo, hi = hi, original = original)

    return(out)
}

condition_shortcuts <- function(x, tr, shortcuts) {
    if (identical(tr, "threenum")) {
        m <- mean(x, na.rm = TRUE)
        s <- stats::sd(x, na.rm = TRUE)
        out <- c(m - s, m, m + s)
    } else if (identical(tr, "fivenum")) {
        out <- stats::fivenum(x, na.rm = TRUE)
    } else if (identical(tr, "minmax")) {
        out <- c(
            min(x, na.rm = TRUE),
            max(x, na.rm = TRUE))
    } else if (identical(tr, "quartile")) {
        out <- stats::quantile(x, probs = c(0.25, 0.5, 0.75), na.rm = TRUE)
    }
    return(out)
}


sanitize_condition <- function(model, condition, variables = NULL, modeldata = NULL) {
    # allow multiple conditions and/or effects
    checkmate::assert(
        checkmate::check_character(condition, min.len = 1, max.len = 4),
        checkmate::check_list(condition, min.len = 1, max.len = 4))

    # c("a", "b") or list("a", "b") -> named list of NULLs
    flag1 <- isTRUE(checkmate::check_character(condition))
    flag2 <- isTRUE(checkmate::check_list(condition, names = "unnamed")) &&
             all(sapply(condition, function(x) isTRUE(checkmate::check_string(x))))
    if (flag1 || flag2) {
        condition <- stats::setNames(rep(list(NULL), length(condition)), unlist(condition))
    }

    # validity of the list
    for (i in seq_along(condition)) {
        if (identical(names(condition)[i], "")) {
            if (!isTRUE(checkmate::check_character(condition[[i]], len = 1))) {
                msg <- "The `condition` argument must be a character vector or a named list."
                insight::format_error(msg)
            } else {
                names(condition)[i] <- condition[[i]]
                tmp <- stats::setNames(list(NULL), names(condition)[i])
                condition <- utils::modifyList(condition, tmp, keep.null = TRUE)
            }
        }
    }

    # get data to know over what range of values we should plot
    if (is.null(modeldata) && isTRUE(checkmate::check_character(condition))) {
        dat <- get_modeldata(model, additional_variables = condition)
    } else if (is.null(modeldata) && isTRUE(checkmate::check_list(condition))) {
        dat <- get_modeldata(model, additional_variables = names(condition))
    } else {
        dat <- modeldata
    }
    resp <- insight::get_response(model)
    respname <- insight::find_response(model)
    
    flag <-  checkmate::check_true(all(names(condition) %in% c(colnames(dat), "group")))
    if (!isTRUE(flag)) {
        msg <- sprintf("Entries in the `condition` argument must be element of: %s",
                       paste(colnames(dat), collapse = ", "))
        insight::format_error(msg)
    }


    # condition names
    condition1 <- names(condition)[[1]]
    condition2 <- hush(names(condition)[[2]])
    condition3 <- hush(names(condition)[[3]])
    condition4 <- hush(names(condition)[[4]])

    # build typical dataset with a sequence of values over "condition" range
    at_list <- list()

    shortcuts <- c("threenum", "fivenum", "minmax", "quartile")

    # condition 1: x-axis
    if (is.null(condition[[1]])) {
        if (get_variable_class(dat, condition1, "binary")) {
            at_list[[condition1]] <- 0:1
        } else if (is.numeric(dat[[condition1]]) && !get_variable_class(dat, condition1, "categorical")) {
            at_list[[condition1]] <- seq(
                min(dat[[condition1]], na.rm = TRUE),
                max(dat[[condition1]], na.rm = TRUE),
                length.out = 50)
        } else {
            at_list[[condition1]] <- factor(unique(dat[[condition1]]))
        }
    } else {
        if (isTRUE(checkmate::check_choice(condition[[1]], shortcuts))) {
            at_list[[condition1]] <- condition_shortcuts(dat[[condition1]], condition[[1]], shortcuts)
        } else {
            at_list[[condition1]] <- condition[[1]]
        }
    }

    # condition 2: color
    if (length(condition) > 1) {
        # defaults
        if (is.null(condition[[2]])) {
            #binary
            if (get_variable_class(dat, condition2, "binary")) {
              at_list[[condition2]] <- condition[[2]] <- 0:1
            # numeric default = Tukey's 5 numbers
            } else if (is.numeric(dat[[condition2]])) {
              condition[[2]] <- "fivenum"
            # other default = unique values
            } else if (condition2 %in% colnames(dat)) {
              condition[[2]] <- unique(dat[[condition2]])
            }
        }
        # known string shortcuts
        if (isTRUE(checkmate::check_choice(condition[[2]], shortcuts))) {
            at_list[[condition2]] <- condition_shortcuts(
                dat[[condition2]], condition[[2]], shortcuts
            )
        # user-supplied
        } else {
            at_list[[condition2]] <- condition[[2]]
        }
    }

    # condition 3: facet_1
    if (length(condition) > 2) {
        if (is.null(condition[[3]])) {
            if (is.numeric(dat[[condition3]])) {
                at_list[[condition3]] <- stats::fivenum(dat[[condition3]])
            } else {
                at_list[[condition3]] <- unique(dat[[condition3]])
            }
        } else {
            if (isTRUE(checkmate::check_choice(condition[[3]], shortcuts))) {
                at_list[[condition3]] <- condition_shortcuts(dat[[condition3]], condition[[3]], shortcuts)
            } else {
                at_list[[condition3]] <- condition[[3]]
            }
        }
    }
    
    # condition 4: facet_2
    if (length(condition) > 3) {
      if (is.null(condition[[4]])) {
        if (is.numeric(dat[[condition4]])) {
          at_list[[condition4]] <- stats::fivenum(dat[[condition4]])
        } else {
          at_list[[condition4]] <- unique(dat[[condition4]])
        }
      } else {
        if (isTRUE(checkmate::check_choice(condition[[4]], shortcuts))) {
          at_list[[condition4]] <- condition_shortcuts(dat[[condition4]], condition[[4]], shortcuts)
        } else {
          at_list[[condition4]] <- condition[[4]]
        }
      }
    }

    at_list[["model"]] <- model
    at_list[["newdata"]] <- dat
    
    if (isTRUE(checkmate::check_list(variables))) {
        flag <- all(names(variables) %in% names(condition))
    } else {
        flag <- all(variables %in% names(condition))
    }
    if (!flag) {
        # sometimes we use the same condition as effect (e.g., GAM vignette),
        # but otherwise we don't want it at all
        if (isTRUE(checkmate::check_character(variables))) {
            dups <- setdiff(variables, names(condition))
            for (d in dups) {
                at_list[[d]] <- NULL
            }
        } else {
            at_list[[names(variables)]] <- NULL
        }
    }

    # mlr3 and tidymodels are not supported by `insight::find_variables()`, so we need to create a grid based on all the variables supplied in `newdata`
    if (inherits(at_list$model, "Learner") || 
        inherits(at_list$model, "model_fit") ||
        inherits(at_list$model, "workflow") ) {
        at_list$model <- NULL
    }

    # create data
    nd <- do.call("datagrid", at_list)

    out <- list(
        "modeldata" = dat,
        "newdata" = nd, 
        "resp" = resp,
        "respname" = respname,
        "condition" = condition, 
        "condition1" = condition1, 
        "condition2" = condition2, 
        "condition3" = condition3,
        "condition4" = condition4) 
    return(out)
}

#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.hurdle <- function(model, coefs, ...) {
    # in pscl::hurdle, coefficients are held in a named list:
    # model$coefficients. Each element of the list is a vector which
    # corresponds to one equation (e.g., "zero" or "response"). When calling
    # "coef(model)", the equation label is prefixed to the term name with an
    # underscore.
    out <- model
    for (lab in names(out$coefficients)) {
        idx <- paste0(lab, "_", names(out$coefficients[[lab]]))
        idx <- match(idx, names(coefs))
        # probably too conservative
        if (anyNA(idx)) {
            stop("Mismatched coefficients names. Please check the `marginaleffects::`set_coef.hurdle` or `set_coef.zeroinfl` function.",
                 call. = FALSE)
        }
        out$coefficients[[lab]] <- stats::setNames(coefs[idx], names(out$coefficients[[lab]]))
    }
    return(out)
}


#' @rdname set_coef
#' @export
set_coef.zeroinfl <- set_coef.hurdle


#' @rdname get_group_names
#' @export
get_group_names.hurdle <- function(model, type = "count", ...) {
    if (type == "prob") {
        out <- colnames(stats::predict(model, type = "prob"))
    } else {
        out <- "main_marginaleffect"
    } 
    return(out)
}

sanity_inferences_conformal <- function(mfx, score, method, data_calib, R) {
    checkmate::assert_choice(
        score,
        choices = c("residual_abs", "residual_sq", "softmax")
    )

    if (is.null(mfx) || !inherits(mfx, "marginaleffects_internal")) {
        stop_sprintf("Conformal prediction requires an object produced by `predictions()`.")
    }

    if (!identical(mfx@calling_function, "predictions")) {
        stop_sprintf("Conformal prediction currently supports outputs from `predictions()` only.")
    }

    if (!isFALSE(mfx@by)) {
        stop_sprintf("Conformal prediction requires `by = FALSE` in the original `predictions()` call.")
    }

    call <- mfx@call
    if (is.null(call)) {
        stop_sprintf("Unable to inspect the original `predictions()` call for conformal inference.")
    }

    call_args <- as.list(call)[-1]
    call_names <- names(call_args)

    # Detect ellipsis/unnamed arguments
    if (any(is.na(call_names) | call_names == "")) {
        stop_sprintf("Conformal prediction does not support additional unnamed arguments in `predictions()`.")
    }

    formals_pred <- as.list(formals(marginaleffects::predictions))
    formal_names <- names(formals_pred)

    # Reject arguments passed through ... (not formal arguments)
    extra_args <- setdiff(call_names, formal_names)
    if (length(extra_args) > 0) {
        stop_sprintf(
            "Conformal prediction only supports `predictions()` with `newdata`. Remove the following arguments: %s",
            paste(extra_args, collapse = ", ")
        )
    }

    # Only select arguments may differ from defaults
    allowed_args <- c("model", "newdata", "conf_level", "vcov", "...")
    restricted <- setdiff(formal_names, allowed_args)

    for (arg in restricted) {
        current <- if (arg %in% call_names) call_args[[arg]] else formals_pred[[arg]]
        default <- formals_pred[[arg]]
        if (!identical(current, default)) {
            stop_sprintf(
                "Conformal prediction only supports raw `predictions()` calls; please remove the `%s` argument.",
                arg
            )
        }
    }

    if (method %in% c("conformal_split", "conformal_quantile")) {
        checkmate::assert_data_frame(data_calib, null.ok = FALSE)
    }

    if (method == "conformal_cv+") {
        checkmate::assert_integerish(R, upper = 25)
    }

    if (method == "conformal_full" && score == "softmax") {
        msg <- "The 'softmax' score is not supported for full conformal prediction. Use 'residual_abs' or 'residual_sq' for regression tasks."
        stop_sprintf(msg)
    }

    return(invisible(NULL))
}


sanitize_estimator <- function(x, estimator, method) {
    checkmate::assert_function(estimator)
    checkmate::assert_data_frame(x)
    cl <- c("data.frame", "data.table", "tbl_df", "tbl")
    if (!any(cl %in% class(x)[1])) {
        msg <- "The `x` argument must be a raw data frame when using the `estimator` argument."
        stop_sprintf(msg)
    }
    if (!isTRUE(checkmate::check_choice(method, c("rsample", "boot")))) {
        stop_sprintf("The `estimator` argument is only supported when `method` is \"rsample\" or \"boot\".")
    }
    x <- estimator(x)
    cl <- c("predictions", "comparisons", "slopes", "hypotheses")
    if (!any(cl %in% class(x))) {
        msg <- sprintf("The `estimator` function must return a `marginaleffects` object.")
        stop_sprintf(msg)
    }
    return(x)
}

get_conformal_score <- function(x, score) {
    response_name <- insight::find_response(attr(x, "model"))
    response <- x[[response_name]]
    if (!is.numeric(response) && score != "softmax") {
      insight::format_error('The response must be numeric. Did you want to use `conformal_score="softmax"`?')
    }
    if (score == "residual_abs") {
      out <- abs(response - x$estimate)
    } else if (score == "residual_sq") {
      out <- (response - x$estimate)^2
    } else if(score == "softmax") {
      model <- attr(x, "model")
      response <- x[[insight::find_response(model)]]
      if (is.numeric(response) && is_binary(response)) {
        # See p.4 of Angelopoulos, Anastasios N., and Stephen Bates. 2022. “A
        # Gentle Introduction to Conformal Prediction and Distribution-Free
        # Uncertainty Quantification.” arXiv.
        # https://doi.org/10.48550/arXiv.2107.07511.
        # 1 minus the softmax output of the true class
        out <- ifelse(response == 1, 1 - x$estimate, x$estimate)
      } else if ("group" %in% colnames(x)) {
        # HACK: is this fragile? I think `group` should always be character.
        idx <- as.character(response) == as.character(x$group)
        out <- 1 - x$estimate[idx]
      } else {
        insight::format_error("Failed to compute the conformity score.")
      }
    }
    return(out)
}


get_conformal_bounds <- function(x, score, conf_level) {
    model <- attr(x, "model")
    response_name <- insight::find_response(model)
    response <- x[[response_name]]
    d <- min(score[score > stats::quantile(score, probs = conf_level)])
    if ("group" %in% colnames(x)) {
      q <- stats::quantile(score, probs = (length(score) + 1) * conf_level / length(score))
      out <- x[x$estimate > (1 - q),]
      data.table::setDT(out)
      out <- out[, .(pred.set = list(unique(group))), by = c("rowid", response_name)]
      setorder(out, rowid)
      data.table::setDF(out)
      class(out) <- c("predictions", class(out))
      attr(out, "variables_datagrid") <- response_name
      return(out)
    } else {
      # continuous outcome: conformity half-width
      x$pred.low <- x$estimate - d
      x$pred.high <- x$estimate + d
    }
    return(x)
}


conformal_split <- function(x, test, calibration, score, conf_level, ...) {
    # calibration
    # use original model---fitted on the training set---to make predictions in the calibration set
    # p_calib is the `predictions()` call, which we re-evaluate on newdata=calibration
    p_calib <- attr(x, "call")
    p_calib[["newdata"]] <- calibration
    p_calib[["vcov"]] <- FALSE # faster
    p_calib <- eval(p_calib)
    score <- get_conformal_score(p_calib, score = score)

    # test
    # use original model to make predictions in the test set
    p_test <- attr(x, "call")
    p_test[["newdata"]] <- test
    p_test <- eval(p_test)

    # bounds
    out <- get_conformal_bounds(p_test, score = score, conf_level = conf_level)

    return(out)
}


conformal_cv_plus <- function(x, test, R, score, conf_level, ...) {
    # cross-validation
    train <- get_modeldata(attr(x, "model"))
    idx <- sample.int(nrow(train), nrow(train))
    idx <- split(idx, ceiling(seq_along(idx) / (length(idx) / R)))
    scores <- NULL
    for (i in idx) {
        data_cv <- train[-i,]
        # re-fit the original model on training sets withholding the CV fold
        model_cv <- stats::update(attr(x, "model"), data = data_cv)
        # use the updated model to make out-of-fold predictions
        # call_cv is the `predictions()` call, which we re-evaluate in-fold: newdata=train[i,]
        call_cv <- attr(x, "call")
        call_cv[["model"]] <- model_cv
        call_cv[["newdata"]] <- train[i,]
        call_cv[["vcov"]] <- FALSE # faster
        pred_cv <- eval(call_cv)
        # save the scores form each fold
        scores <- c(scores, get_conformal_score(pred_cv, score = score))
    }

    # test
    out <- attr(x, "call")
    out[["newdata"]] <- test
    out <- eval(out)

    # bounds
    out <- get_conformal_bounds(out, score = scores, conf_level = conf_level)

    return(out)
}

#' @include get_predict.R
#' @rdname get_predict
#' @export
get_predict.bart <- function(model, newdata = NULL, ...) {
    args <- c(
        list(
            object = model,
            newdata = newdata
        ),
        list(...)
    )
    p <- do_call(stats::predict, args)
    p_med <- collapse::fmedian(p)
    out <- data.table(
        group = "main_marginaleffect",
        estimate = p_med)
    out <- add_rowid(out, newdata)
    attr(out, "posterior_draws") <- t(p)
    return(out)
}


#' @include sanity_model.R
#' @rdname sanitize_model_specific
#' @export
sanitize_model_specific.bart <- function(model, calling_function, ...) {
    insight::check_if_installed("collapse", minimum_version = "1.9.0")
    if (
        !isTRUE(
            as.character(insight::get_call(model))[1] %in% c("bart2", "dbarts::bart2")
        )
    ) {
        msg <- "`marginaleffects` only supports models estimated using the formula interface in `bart2()` function, not the matrix input in `bart()`."
        stop_sprintf(msg)
    }
    if (calling_function == "hypotheses") {
        msg <- "`marginaleffects` does not support hypothesis tests for models of class `bart`."
        stop_sprintf(msg)
    }
    return(model)
}


#' @rdname get_vcov
#' @export
get_vcov.bart <- function(model, vcov = NULL, ...) {
    if (!is.null(vcov) && !is.logical(vcov)) {
        warn_sprintf(
            "The `vcov` argument is not supported for models of this class."
        )
    }
    vcov <- sanitize_vcov(model, vcov)
    return(NULL)
}

#' Internal function to set coefficients
#'
#' Set the coefficients in a model to different values and return the modified object (internal function)
#'
#' @rdname set_coef
#' @param model object to modify
#' @param coefs vector of coefficients to insert in the model object
#' @export
#' @keywords internal
#' @return Model object of the same class as the `model` argument, but with
#'   different stored coefficients.
#' @details To compute the variance of marginal effects we need to take the
#' Jacobian with
# respect to the model coefficients. These functions manipulate model objects
# to change the coefficients stored internally, which changes the output of the
# `predict()` function.
set_coef <- function(model, coefs, ...) {
    UseMethod("set_coef")
}

#' @rdname set_coef
#' @export
set_coef.default <- function(model, coefs, ...) {
    # in basic model classes coefficients are named vector
    # in ordinal::clm models, there are sometimes duplicates, so name matching doesn't work
    # sometimes names are backticked, others not Issue #1005
    a <- gsub("`", "", names(coefs))
    b <- gsub("`", "", names(model$coefficients))
    flag <- length(model[["coefficients"]]) == length(coefs) && all(a == b)
    if (flag) {
        model[["coefficients"]] <- coefs
    } else {
        model[["coefficients"]][names(coefs)] <- coefs
    }
    model
}

#' Refit a marginaleffects object with new data
#'
#' @param object A marginaleffects object (predictions, comparisons, or slopes)
#' @param ... Additional arguments passed to methods
#' @return A marginaleffects object
#' @export
refit <- function(object, ...) {
    UseMethod("refit")
}

#' @rdname refit
#' @param data Optional data frame to refit the underlying model
#' @param newdata Optional data frame to re-evaluate the marginaleffects call
#' @param vcov Optional logical or variance-covariance matrix specification to pass to the marginaleffects call
#' @details
#' If `data` is supplied, the underlying model is refitted using that data.
#' If `newdata` is supplied, the marginaleffects call is re-evaluated with the new data.
#' Both can be supplied together to refit the model and make predictions on new data.
#' @export
refit.marginaleffects <- function(object, data = NULL, newdata = NULL, vcov = NULL, ...) {
    # If both are NULL, return unchanged
    if (is.null(data) && is.null(newdata)) {
        return(object)
    }

    mfx <- attr(object, "marginaleffects")
    if (is.null(mfx)) {
        stop("Object does not have marginaleffects attributes")
    }

    model <- mfx@model

    fit_again <- function(model, data) {
        # Try stats::update first
        model <- tryCatch(
            stats::update(model, data = data),
            error = function(e) NULL
        )
        # Fallback: modify call and re-evaluate
        if (is.null(model)) {
            if (is.call(mfx@call_model) && "data" %in% names(mfx@call_model)) {
                call_new <- mfx@call_model
                call_new$data <- data
                model <- try(eval(call_new), silent = TRUE)
                if (inherits(model, "try-error")) {
                    stop("Failed to refit the model.", call. = FALSE)
                }
            } else {
                stop("Failed to refit model: no update method available", call. = FALSE)
            }
        }
        return(model)
    }

    # Step 1: Refit model if data is supplied
    if (!is.null(data)) {
        # For workflows, tidymodels provides its own fit.workflow method
        if (inherits(model, "workflow")) {
            model <- generics::fit(model, data = data)
        } else if (inherits(model, "model_fit")) {
            model <- fit_again(model[["fit"]], data = data)
        } else {
            model <- fit_again(model, data = data)
        }
    }

    # Step 2: Re-evaluate marginaleffects call
    call_new <- mfx@call
    call_new[["model"]] <- model

    if (!is.null(newdata)) {
        call_new[["newdata"]] <- newdata
    }

    if (!is.null(vcov)) {
        call_new[["vcov"]] <- vcov
    }

    result <- eval(call_new)
    return(result)
}

#' @rdname refit
#' @export
refit.predictions <- refit.marginaleffects

#' @rdname refit
#' @export
refit.comparisons <- refit.marginaleffects

#' @rdname refit
#' @export
refit.slopes <- refit.marginaleffects

#' @rdname refit
#' @export
refit.hypotheses <- refit.marginaleffects

get_contrast_data_character <- function(model,
                                        newdata,
                                        variable,
                                        cross,
                                        first_cross,
                                        modeldata,
                                        ...) {

    # factors store all levels, but characters do not, so we need to extract the
    # original data from the model.
    tmp <- modeldata

    # unsupported by insight (e.g., numpyro)
    if (is.null(tmp)) {
        tmp <- newdata
    }

    levs <- sort(unique(tmp[[variable$name]]))

    # string shortcuts
    flag <- checkmate::check_choice(variable$value, c("reference", "revreference", "pairwise", "revpairwise", "sequential", "revsequential", "all", "minmax"))
    if (isTRUE(flag)) {
        levs_idx <- contrast_categories_shortcuts(levs, variable, interaction)

    # custom data frame or function
    } else if (isTRUE(checkmate::check_function(variable$value)) || isTRUE(checkmate::check_data_frame(variable$value))) {
        out <- contrast_categories_custom(variable, newdata)
        return(out)

    # vector of two values
    } else if (isTRUE(checkmate::check_atomic_vector(variable$value, len = 2))) {
        if (is.character(variable$value)) {
            tmp <- modeldata[[variable$name]]
            if (!all(variable$value %in% as.character(tmp))) {
                msg <- "Some of the values supplied to the `variables` argument were not found in the dataset."
                insight::format_error(msg)
            }
            idx <- match(variable$value, as.character(tmp))
            levs_idx <- data.table::data.table(lo = tmp[idx[1]], hi = tmp[idx[[2]]])
        } else if (is.numeric(variable$value)) {
            tmp <- newdata[[variable$name]]
            levs_idx <- data.table::data.table(
                lo = as.character(variable$value[1]),
                hi = as.character(variable$value[2]))
        } else {
            levs_idx <- data.table::data.table(lo = variable$value[1], hi = variable$value[2])
        }
    }

    tmp <- contrast_categories_processing(first_cross, levs_idx, levs, variable, newdata)
    lo <- tmp[[1]]
    hi <- tmp[[2]]
    original <- tmp[[3]]

    lo[[variable$name]] <- lo[["marginaleffects_contrast_lo"]]
    hi[[variable$name]] <- hi[["marginaleffects_contrast_hi"]]
    contrast_label <- hi$marginaleffects_contrast_label
    contrast_null <- hi$marginaleffects_contrast_hi == hi$marginaleffects_contrast_lo

    tmp <- !grepl("^marginaleffects_contrast", colnames(lo))
    lo <- lo[, tmp, with = FALSE]
    hi <- hi[, tmp, with = FALSE]


    out <- list(rowid = original$rowid,
                lo = lo,
                hi = hi,
                original = original,
                ter = rep(variable$name, nrow(lo)), # lo can be different dimension than newdata
                lab = contrast_label,
                contrast_null = contrast_null)
    return(out)
}

# https://stackoverflow.com/questions/9439256/how-can-i-handle-r-cmd-check-no-visible-binding-for-global-variable-notes-when#comment20826625_12429344
# 2012 hadley says "globalVariables is a hideous hack and I will never use it"
# 2014 hadley updates his own answer with globalVariables as one of "two solutions"
#' @import data.table
utils::globalVariables(c(
    ".",
    "..by",
    "..bycols",
    "..cl",
    "..col_newdata",
    "..col_x",
    "..cols",
    "..hypothesis_by",
    "..idx",
    "..idx1",
    "..idx2",
    "..idx_by",
    "..idx_hi",
    "..idx_lo",
    "..idx_or",
    "..keepers",
    "..lab",
    "..lab_cols",
    "..label_columns",
    "..nd",
    "..nonfocal",
    "..tmp",
    "..to_keep",
    "..variables_grid",
    "..w",
    ".data",
    "N",
    "aa",
    "chunk",
    "comparison",
    "comparison_idx",
    "condition_variable",
    "conf.high",
    "conf.low",
    "conf_int",
    "conf_level",
    "contrast",
    "draw",
    "dydx",
    "elast",
    "eps",
    "estimate",
    "estimates",
    "formatBicLL",
    "gn",
    "group",
    "hi",
    "lo",
    "logLik",
    "m",
    "mAD",
    "marginaleffects_contrast_hi",
    "marginaleffects_contrast_lo",
    "marginaleffects_function",
    "marginaleffects_group_id",
    "marginaleffects_internal_label",
    "marginaleffects_internal_rowid",
    "marginaleffects_lab",
    "marginaleffects_ter",
    "marginaleffects_unique_labels",
    "marginaleffects_wts_internal",
    "marginalmean",
    "model",
    "p.value",
    "part",
    "patterns",
    "position",
    "pred_hi",
    "pred_lo",
    "predicted",
    "predicted_hi",
    "predicted_lo",
    "predicted_or",
    "results",
    "rowid",
    "rowidcf",
    "rowidunique",
    "section",
    "stars_note",
    "statistic",
    "statistic1",
    "std.error",
    "term",
    "ticks",
    "tmp_grp",
    "tmp_idx",
    "value",
    "vcov",
    "where",
    "wts"
))

#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.polr <- function(model, ...) {
    out <- insight::get_parameters(model)
    out <- stats::setNames(out$Estimate, out$Parameter)
    names(out) <- gsub("Intercept: ", "", names(out))
    return(out)
}


#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.polr <- function(model, coefs, ...) {
    # in basic model classes coefficients are named vector
    idx <- match(names(model$coefficients), names(coefs))
    model[["coefficients"]] <- coefs[idx]
    idx <- match(names(model$zeta), names(coefs))
    model[["zeta"]] <- coefs[idx]
    model
}


#' @include get_group_names.R
#' @rdname get_group_names
#' @export
get_group_names.polr <- function(model, ...) {
    resp <- insight::get_response(model)
    if (is.factor(resp)) {
        out <- levels(resp)
    } else {
        out <- unique(resp)
    }
    return(out)
}


#' @include get_predict.R
#' @rdname get_predict
#' @export
get_predict.polr <- function(
    model,
    newdata = insight::get_data(model),
    type = "probs",
    mfx = NULL,
    ...) {
    calling_function <- if (!is.null(mfx)) mfx@calling_function else "predictions"
    type <- sanitize_type(model, type, calling_function = calling_function)

    # hack: 1-row newdata returns a vector, so get_predict.default does
    # not learn about groups
    hack <- FALSE
    if (nrow(newdata) == 1) {
        hack <- TRUE
        newdata <- newdata[c(1, 1), , drop = FALSE]
    }

    out <- get_predict.default(model, newdata = newdata, type = type, ...)

    if (hack) {
        out <- out[seq_len(nrow(out)) %% 2 == 1, , drop = FALSE]
        newdata <- newdata[seq_len(nrow(newdata)) %% 2 == 1, , drop = FALSE]
    }

    out <- add_rowid(out, newdata)

    return(out)
}


#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.glmmPQL <- function(model, coefs, ...) {
    model[["coefficients"]][["fixed"]][names(coefs)] <- coefs
    model
}


#' @rdname get_predict
#' @export
get_predict.glmmPQL <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    mfx = NULL,
    ...) {
    out <- stats::predict(model, newdata = newdata, type = type, ...)
    out <- data.table(estimate = out)
    out <- add_rowid(out, newdata)
    return(out)
}


#' @rdname get_vcov
#' @export
get_vcov.lda <- function(model, ...) {
    return(NULL)
}



#' @rdname get_predict
#' @export
get_predict.lda <- function(
    model,
    newdata = insight::get_data(model),
    type = "class",
    ...) {
    out <- stats::predict(model, newdata = newdata)
    if (type == "class") {
        out <- data.table(estimate = out$class)
    } else if (type == "posterior") {
        out <- data.table::melt(data.table::data.table(out$posterior),
            variable.name = "group", value.name = "estimate")
    }
    out <- add_rowid(out, newdata)
    return(out)
}

get_hypothesis <- function(
    x,
    hypothesis,
    by = NULL,
    newdata = NULL,
    draws = NULL) {
    deprecated <- c("pairwise", "revpairwise", "sequential", "revsequential", "reference", "meandev", "meanotherdev", "revreference")
    if (isTRUE(checkmate::check_choice(hypothesis, deprecated))) {
        msg <- "This string is deprecated for use in the `hypothesis` argument. Use the formula interface instead. Ex: `hypothesis=~reference`"
        stop(msg, call. = FALSE)
    }

    # otherwise setDT strips out the marginaleffects-specific classes
    x <- data.table::copy(x)

    vec <- isTRUE(checkmate::check_atomic_vector(hypothesis)) &&
        isTRUE(checkmate::check_numeric(hypothesis))
    mat <- isTRUE(checkmate::check_matrix(hypothesis))

    if (is.null(hypothesis)) {
        data.table::setDT(x)
        return(x)
    } else if (isTRUE(checkmate::check_formula(hypothesis))) {
        out <- hypothesis_formula(x, newdata = newdata, hypothesis = hypothesis, by = by)
        return(out)
    } else if (isTRUE(checkmate::check_function(hypothesis))) {
        out <- hypothesis_function(x, newdata = newdata, hypothesis = hypothesis, by = by)
        return(out)
    } else if (vec || mat) {
        out <- hypothesis_matrix(x, hypothesis = hypothesis)
        return(out)
    } else if (is.character(hypothesis)) {
        out <- hypothesis_string(x, hypothesis = hypothesis)
        return(out)
    }

    stop("`hypothesis` is broken. Please report this bug with a reproducible example: https://github.com/vincentarelbundock/marginaleffects/issues.", call. = FALSE)
}

get_contrast_data_logical <- function(model,
                                      newdata,
                                      variable,
                                      ...) {

    # custom data frame or function
    if (isTRUE(checkmate::check_function(variable$value)) || isTRUE(checkmate::check_data_frame(variable$value))) {
        out <- contrast_categories_custom(variable, newdata)
        return(out)
    }

    lo <- hi <- newdata
    lo[[variable$name]] <- FALSE
    hi[[variable$name]] <- TRUE
    lab <- suppressWarnings(tryCatch(
        sprintf(variable$label, TRUE, FALSE),
        error = function(e) variable$label))
    out <- list(rowid = seq_len(nrow(newdata)),
        lo = lo,
        hi = hi,
        original = newdata,
        ter = rep(variable$name, nrow(newdata)),
        lab = rep(lab, nrow(newdata)))
    return(out)
}

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

eigenMatMult <- function(A, B) {
    .Call('_marginaleffects_eigenMatMult', PACKAGE = 'marginaleffects', A, B)
}

sanity_multcomp <- function(multcomp, hypothesis, joint) {
    # I don't know how to adjust p values for a different null in `multcomp::glht()`
    if (!isFALSE(multcomp) && isTRUE(checkmate::check_number(hypothesis))) {
        msg <- "The `multcomp` argument is not available when `hypothesis` is a number."
        stop(msg, call. = FALSE)
    }
    if (!isFALSE(multcomp) && !isFALSE(joint)) {
        msg <- "The `multcomp` argument cannot be used with the `joint` argument."
        stop_sprintf(msg)
    }
    return(invisible(NULL))
}

#' @rdname get_predict
#' @export
get_predict.biglm <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    mfx = NULL,
    ...) {
    calling_function <- if (!is.null(mfx)) mfx@calling_function else "predictions"
    type <- sanitize_type(model, type, calling_function = calling_function)
    out <- stats::predict(
        model,
        newdata = newdata,
        type = type
    )
    out <- as.vector(out)
    out <- data.frame(
        rowid = seq_along(out),
        estimate = out
    )
    return(out)
}


#' @rdname get_vcov
#' @export
get_vcov.biglm <- function(model, vcov = NULL, ...) {
    return(FALSE)
}


#' @rdname sanitize_model_specific
sanitize_model_specific.biglm <- function(model, vcov = NULL, ...) {
    if (!isFALSE(vcov)) {
        warn_sprintf(c(
            "The `vcov` argument is not supported for this model type. Set `vcov=FALSE` to silence this warning, and visit this link to learn why standard errors for this model are not yet supported and how you can help:",
            "https://github.com/vincentarelbundock/marginaleffects/issues/387"
        ))
    }
    return(model)
}

#' Get levels of the outcome variable in grouped or multivariate models
#' 
#' @inheritParams slopes
#' @return A character vector
#' @rdname get_group_names
#' @keywords internal
#' @export
get_group_names <- function (model, ...) {
    UseMethod("get_group_names", model)
}


#' @rdname get_group_names
#' @export
get_group_names.default <- function(model, ...) {
    return("main_marginaleffect")
}

sanitize_hypothesis_formula <- function(hypothesis) {
    hypothesis <- Formula::Formula(hypothesis)

    if (!length(hypothesis)[1] %in% 0:1) {
        stop(
            "The left-hand side of the `hypothesis` formula must be of length 0 or 1.",
            call. = FALSE
        )
    }

    if (!length(hypothesis)[2] %in% 1:2) {
        stop(
            "The right-hand side of the `hypothesis` formula must have 1 or two parts, separated by |",
            call. = FALSE
        )
    }

    if (length(hypothesis)[2] == 1) {
        by <- NULL
    } else {
        by <- all.vars(stats::formula(hypothesis, lhs = 0, rhs = 2))
    }
    lhs <- all.vars(stats::formula(hypothesis, rhs = 0))
    rhs <- all.vars(stats::formula(hypothesis, lhs = 0, rhs = 1))

    # Custom functions
    if (identical(rhs, "x")) {
        asis <- attr(stats::terms(hypothesis), "term.labels")
        asis <- grep("^I\\(", asis, value = TRUE)
        if (length(asis) == 1) {
            if (length(lhs) != 0) {
                stop(
                    "The left-hand side of `hypothesis` must be empty when using custom functions.",
                    call. = FALSE
                )
            }
            asis <- sub("^I\\((.*)\\)$", "\\1", asis)
            out <- list(
                lhs = "arbitrary_function",
                rhs = asis,
                group = by
            )
            return(out)
        } else if (length(asis) > 1) {
            stop(
                "Only one custom function is supported at a time on the right-hand side of the `hypothesis` formula.",
                call. = FALSE
            )
        }
    }

    valid <- c(
        "reference",
        "revreference",
        "sequential",
        "pairwise",
        "revpairwise",
        "meandev",
        "meanotherdev",
        "poly",
        "helmert",
        "trt_vs_ctrl"
    )
    checkmate::assert_choice(
        rhs,
        valid,
        .var.name = "Right-hand side of `hypothesis` formula"
    )

    # dot product weights
    if (rhs %in% c("poly", "helmert")) {
        if (length(lhs) == 1 && lhs != "dotproduct") {
            stop(
                "The left-hand size of the `hypothesis` formula must be empty or `dotproduct`.",
                call. = FALSE
            )
        } else {
            lhs <- "dotproduct"
        }

        # differences and ratios
    } else {
        if (length(lhs) == 1 && !lhs %in% c("difference", "ratio")) {
            stop(
                "The left-hand size of the `hypothesis` formula must be empty, `difference`, or `ratio`.",
                call. = FALSE
            )
        } else if (length(lhs) == 0) {
            lhs <- "difference"
        }
    }

    out <- list(
        rhs = rhs,
        lhs = lhs,
        group = by
    )

    return(out)
}

get_labels <- function(x, idx = NULL, by = NULL, wrap_parens = FALSE, hypothesis_by = NULL) {
    if (!is.data.frame(x) && !is.vector(x)) {
        return(NULL)
    }

    if (is.data.frame(x)) {
        x <- data.table::data.table(x)

        # Identify relevant columns
        lab_cols <- grep("^term$|^group$|^contrast$|^contrast_|^value$|^by$", names(x), value = TRUE)
        if (isTRUE(checkmate::check_character(by))) {
            lab_cols <- unique(c(lab_cols, by))
        }

        lab_cols <- setdiff(lab_cols, hypothesis_by)

        # Filter out columns with more than one unique value, within `hypothesis_by`
        lab_cols <- Filter(function(col) length(unique(x[[col]])) > 1, lab_cols)

        if (length(lab_cols) == 0) {
            # Default labels if no meaningful columns found
            labels <- paste0("b", seq_len(nrow(x)))
        } else {
            # Create labels by pasting unique combinations of selected columns
            lab_df <- x[, ..lab_cols]
            labels <- apply(lab_df, 1, paste, collapse = " ")

            # duplicated labels (within groups) revert to b1, b2, ...
            uniq <- TRUE
            if (isTRUE(checkmate::check_character(hypothesis_by, min.len = 1))) {
                uniq <- x[, ..hypothesis_by]
                uniq[, marginaleffects_unique_labels := labels]
                uniq <- uniq[,
                    .(marginaleffects_uniq_labels = anyDuplicated(marginaleffects_unique_labels) > 0),
                    by = hypothesis_by]
                uniq <- !any(uniq$V1)
            } else if (anyDuplicated(labels) > 0) {
                uniq <- FALSE
            }

            if (!isTRUE(uniq)) {
                labels <- paste0("b", seq_len(nrow(x)))
            }
        }
    } else if (is.vector(x)) {
        # For vectors
        if (!is.null(names(x))) {
            labels <- names(x)
        } else {
            labels <- paste0("b", seq_along(x))
        }
    } else {
        return(NULL)
    }

    # Wrap labels in parentheses if requested
    if (wrap_parens && any(grepl("-", labels))) {
        labels <- sprintf("(%s)", labels)
    }

    # Subset labels if idx is provided
    if (!is.null(idx)) {
        labels <- labels[idx]
    }

    return(labels)
}

inferences_fwb <- function(x, R = 1000, conf_level = 0.95, conf_type = "perc", mfx = NULL, ...) {
    insight::check_if_installed("fwb", minimum_version = "0.5.0")

    out <- x
    call_mfx <- mfx@call
    call_mfx[["vcov"]] <- FALSE

    # Get modeldata from mfx object
    modeldata <- mfx@modeldata

    # Ensure parameters are embedded in the call, not just references
    if (!is.null(mfx@newdata)) {
        call_mfx[["newdata"]] <- mfx@newdata
    }
    if (!is.null(mfx@comparison)) {
        call_mfx[["comparison"]] <- mfx@comparison
    }

    bootfun <- function(data, w, ...) {
        # If model has weights, multiply them by random weights
        if (!is.null(w0 <- stats::weights(mfx@model))) {
            w <- w * w0
        }

        # Update the model's call and evaluate
        call_mod <- insight::get_call(mfx@model)
        call_mod[["weights"]] <- w
        boot_mod <- eval.parent(call_mod)

        # Update marginaleffects call
        call_mfx[["model"]] <- boot_mod
        call_mfx[["modeldata"]] <- data
        call_mfx[["wts"]] <- w
        boot_mfx <- eval.parent(call_mfx)
        return(boot_mfx$estimate)
    }

    args <- list("data" = modeldata, "statistic" = bootfun, R = R)
    args <- c(args, list(...))

    # fwb default verbose is TRUE
    if (!"verbose" %in% names(args)) {
        args[["verbose"]] <- FALSE
    }

    if (
        isTRUE(getOption("marginaleffects_parallel_inferences", default = FALSE)) &&
            !"cl" %in% names(args)
    ) {
        args[["cl"]] <- "future"
    }

    B <- do.call(fwb::fwb, args)

    # Extract SEs and CIs
    fwb_summary <- tidy(summary(B, conf = conf_level, ci.type = conf_type, p.value = TRUE))

    out$std.error <- fwb_summary$std.error
    out$conf.low <- fwb_summary$conf.low
    out$conf.high <- fwb_summary$conf.high

    cols <- setdiff(names(out), "df")

    if ("p.value" %in% names(fwb_summary)) {
        out$p.value <- fwb_summary$p.value
        out$s.value <- -log2(out$p.value)
    } else {
        cols <- setdiff(cols, c("s.value", "p.value"))
    }

    if ("statistic" %in% names(fwb_summary)) {
        out$statistic <- fwb_summary$statistic
    } else {
        cols <- setdiff(cols, "statistic")
    }

    out <- out[, cols, drop = FALSE]

    mfx@draws <- t(B$t)
    mfx@inferences <- B
    attr(out, "marginaleffects") <- mfx

    return(out)
}

# Conformalized Quantile Regression (CQR)
#
# This implementation is based on the split conformal inference method described by
# Romano et al. (2019): "Conformalized quantile regression." Advances in neural
# information processing systems 32.
#
# Reference implementation from the `probably` package (MIT License, 2025-11-10):
# Copyright holders: Max Kuhn [aut, cre], Davis Vaughan [aut], Edgar Ruiz [aut],
# Posit Software, PBC [cph, fnd] (ROR ID: https://ror.org/00cvxb145)
# URL: https://github.com/tidymodels/probably
#
# The implementation here is adapted to fit the marginaleffects package API and
# uses quantreg::rq instead of quantregForest for quantile estimation.

conformal_quantile <- function(x, data_test, data_train, data_calib, conf_level = 0.95,
                               mfx = NULL, ...) {
    if (is.null(mfx)) {
        mfx <- attr(x, "marginaleffects")
    }

    # Get model information
    model <- mfx@model
    response_name <- insight::find_response(model)

    # Check response is numeric
    if (!is.numeric(data_train[[response_name]])) {
        stop_sprintf("Quantile conformal inference requires a numeric response variable.")
    }

    # Check calibration has response
    if (!response_name %in% colnames(data_calib)) {
        stop_sprintf(
            "Calibration data must include the response variable '%s'.",
            response_name
        )
    }

    # Step 1: Fit quantile regression forest on training data
    insight::check_if_installed("quantregForest", reason = "to fit quantile regression forest")

    # Compute quantile levels for CQR
    # For coverage (1-alpha), use quantiles at alpha/2 and 1-alpha/2
    # e.g., for 90% coverage (conf_level=0.9): alpha=0.1, quantiles at 0.05 and 0.95
    alpha <- 1 - conf_level

    # Determine predictor columns (exclude response)
    # Use the order from training data consistently across all datasets
    predictor_cols <- setdiff(colnames(data_train), response_name)

    # Prepare predictor matrix (exclude response)
    # Convert to data.frame to avoid data.table subsetting issues
    x_train <- as.data.frame(data_train)[, predictor_cols, drop = FALSE]
    y_train <- data_train[[response_name]]

    # Filter out arguments not relevant to quantregForest::quantregForest
    dots <- list(...)
    # Remove arguments from inferences() that aren't for quantregForest()
    dots <- dots[!names(dots) %in% c("R", "conf_type", "estimator", "score")]

    # Fit quantile regression forest
    qrf <- tryCatch(
        do.call(
            quantregForest::quantregForest,
            c(list(x = x_train, y = y_train), dots)
        ),
        error = function(e) {
            stop_sprintf("Failed to fit quantile regression forest: %s", conditionMessage(e))
        })

    # Step 2: Get quantile predictions on calibration set
    # Use same predictor columns in same order as training
    x_calib <- as.data.frame(data_calib)[, predictor_cols, drop = FALSE]
    q_calib <- stats::predict(qrf, newdata = x_calib, what = c(alpha / 2, 1 - alpha / 2))
    q_calib_low <- q_calib[, 1]
    q_calib_high <- q_calib[, 2]

    # Step 3: Compute conformity scores on calibration set
    # E_i = max(q_low(X_i) - Y_i, Y_i - q_high(X_i))
    y_calib <- data_calib[[response_name]]
    conformity_scores <- pmax(
        q_calib_low - y_calib,
        y_calib - q_calib_high
    )

    # Step 4: Compute inflated empirical quantile
    # Following Romano et al. (2019) and reference implementations:
    # Quantiles at alpha/2 and 1-alpha/2, conformity quantile at (1-alpha)
    # Reference: conformal_quantile_reference.R line 142
    m <- length(conformity_scores)
    k <- ceiling((1 - alpha) * (m + 1))
    k <- min(max(k, 1), m) # clamp to [1, m]

    # Get the k-th smallest score
    Q <- sort(conformity_scores, partial = k)[k]

    # Step 5: Get quantile predictions on test set
    # Use same predictor columns in same order as training
    x_test <- as.data.frame(data_test)[, predictor_cols, drop = FALSE]
    q_test <- stats::predict(qrf, newdata = x_test, what = c(alpha / 2, 1 - alpha / 2))
    q_test_low <- q_test[, 1]
    q_test_high <- q_test[, 2]

    # Step 6: Inflate the quantile predictions
    # C(x*) = [q_low(x*) - Q, q_high(x*) + Q]
    out <- refit(x, newdata = data_test, vcov = FALSE)

    out$pred.low <- q_test_low - Q
    out$pred.high <- q_test_high + Q

    return(out)
}

sanitize_vcov <- function(model, vcov) {
    # TRUE generates a warning in `insight::get_varcov` for some models
    if (isTRUE(checkmate::check_flag(vcov))) {
        return(NULL)
    }

    # no vcov matrix for bayesian models
    if (inherits(model, c("brmsfit", "stanreg", "bart"))) {
        return(NULL)
    }

    # strings should be case-insensitive
    vcov_strings <- c(
        "stata",
        "robust",
        "HC",
        "HC0",
        "HC1",
        "HC2",
        "HC3",
        "HC4",
        "HC4m",
        "HC5",
        "HAC",
        "NeweyWest",
        "kernHAC",
        "OPG",
        "satterthwaite",
        "kenward-roger"
    )
    if (
        isTRUE(checkmate::check_choice(
            hush(tolower(vcov)),
            choices = tolower(vcov_strings)
        ))
    ) {
        idx <- match(tolower(vcov), tolower(vcov_strings))
        return(vcov_strings[idx])
    }

    checkmate::assert(
        checkmate::check_null(vcov),
        checkmate::check_function(vcov),
        checkmate::check_matrix(vcov),
        checkmate::check_formula(vcov),
        checkmate::check_choice(vcov, choices = vcov_strings)
    )

    out <- vcov

    if (isTRUE(checkmate::check_function(out))) {
        out <- hush(out(model))
    }

    if (isTRUE(checkmate::check_matrix(out))) {
        if (ncol(out) != nrow(out)) {
            stop("The `vcov` matrix must be square.", call. = FALSE)
        }
    }

    return(out)
}

#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.glmgee <- function(model, ...) {
    b <- model$coefficients
    b <- stats::setNames(as.vector(b), row.names(b))
    return(b)
}

#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.glmgee <- function(model, coefs, ...) {
    out <- model
    idx <- match(row.names(out$coefficients), names(coefs))
    out$coefficients[, 1] <- coefs[idx]
    return(out)
}

#' @include get_predict.R
#' @rdname get_predict
#' @export
get_predict.glmgee <- function(model, newdata, ...) {
    Yhat <- stats::predict(model, newdata = newdata, type = "response")
    out <- data.frame(
        rowid = seq_len(nrow(Yhat)),
        estimate = as.vector(Yhat))
    return(out)
}

# The reason why I don't supply these functions is that there is too much ambiguity about which kind of plot people are expecting. The `plot_predictions()` function has a bunch of extra arguments, and makes assumptions based on the order of the user-supplied `condition` values, for example. Having another `plot.predictions()` method in addition would require me to add all those arguments, or to make opinionated decisions that would require a ton of `if/else` based on the original `predictions()` call's `by`, `newdata` grid, and others. This is hard, and almost certainly unsatisfactory. Also, it duplicates the plotting user-interface. It feels better to steer all users to a single point of entry for plots.

#' @export
plot.predictions <- function(x, ...) {
    insight::format_error("Please use the `plot_predictions()` function.")
}

#' @export
plot.comparisons <- function(x, ...) {
    insight::format_error("Please use the `plot_comparisons()` function.")
}

#' @export
plot.slopes <- function(x, ...) {
    insight::format_error("Please use the `plot_slopes()` function.")
}

#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.gamlss <- function(model, ...) {
    dots <- list(...)

    if (is.null(dots$what)) stop("Argument `what` indicating the parameter of interest is missing.")

    out <- stats::coef(model, what = dots$what)

    return(out)
}


#' @include get_predict.R
#' @rdname get_predict
#' @export
get_predict.gamlss <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    ...
) {
    # Get predictions

    dots <- list(...)

    if (is.null(dots$what)) {
        msg <- sprintf(
            "Please specifiy a `what` argument with one of these values: %s",
            toString(model$parameter)
        )
        stop(msg, call. = FALSE)
    }

    # if (!isTRUE(checkmate::check_flag(vcov, null.ok = TRUE))) {
    #   msg <- "The `vcov` argument is not supported for models of this class."
    #   stop(msg, call. = FALSE)
    # }

    # predict.gamlss() breaks when `newdata` includes unknown variables
    origindata <- insight::get_data(model)
    originvars <- colnames(origindata)
    newdata <- as.data.frame(newdata)
    index <- which(colnames(newdata) %in% originvars)
    tmp <- newdata[, index]
    hush(
        out <- predict_gamlss(
            model,
            newdata = tmp,
            type = type,
            data = origindata,
            ...
        )
    )

    out <- data.table(estimate = out)
    out <- add_rowid(out, newdata)
    return(out)
}

#' @include get_vcov.R
#' @rdname get_vcov
#' @export
get_vcov.gamlss <- function(model, ...) {
    dots <- list(...)

    if (is.null(dots$what)) {
        msg <- sprintf(
            "Please specifiy a `what` argument with one of these values: %s",
            toString(model$parameter)
        )
        stop(msg, call. = FALSE)
    }

    p <- match(dots$what, model$parameters)

    vc <- stats::vcov(model, what = dots$what)
    index <- which(cumsum(rownames(vc) == "(Intercept)") == p)
    out <- vc[index, index, drop = FALSE]

    return(out)
}

#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.gamlss <- function(model, coefs, ...) {
    dots <- list(...)

    if (is.null(dots$what)) stop("Argument `what` indicating the parameter of interest is missing.")

    p <- paste0(dots$what, ".coefficients")
    model[[p]][names(coefs)] <- coefs
    out <- model

    return(out)
}


# Modified predict method from the R-package gamlss
# Renamed with underscore to avoid conflict with exported method from the package.
predict_gamlss <- function(
    object,
    what = c("mu", "sigma", "nu", "tau"),
    parameter = NULL,
    newdata = NULL,
    type = c("link", "response", "terms"),
    safe = TRUE,
    terms = NULL,
    se.fit = FALSE,
    data = NULL,
    ...
) {
    concat <- function(..., names = NULL) {
        tmp <- list(...)
        if (is.null(names)) {
            names <- names(tmp)
        }
        if (is.null(names)) {
            names <- sapply(as.list(match.call()), deparse)[-1]
        }
        if (any(sapply(tmp, is.matrix) | sapply(tmp, is.data.frame))) {
            len <- sapply(tmp, function(x) c(dim(x), 1)[1])
            len[is.null(len)] <- 1
            data <- rbind(...)
        } else {
            len <- lengths(tmp)
            data <- unlist(tmp)
        }
        namelist <- factor(rep(names, len), levels = names)
        return(data.table(data, source = namelist))
    }
    if (is.null(newdata)) {
        predictor <- gamlss::lpred(
            object,
            what = what,
            type = type,
            terms = terms,
            se.fit = se.fit,
            ...
        )
        return(predictor)
    }
    if (se.fit) {
        warning(
            " se.fit = TRUE is not supported for new data values at the moment \n"
        )
    }
    if (!(inherits(newdata, "data.frame"))) {
        stop("newdata must be a data frame ")
    }
    what <- if (!is.null(parameter)) {
        match.arg(
            parameter,
            choices = c(
                "mu",
                "sigma",
                "nu",
                "tau"
            )
        )
    } else {
        match.arg(what)
    }
    type <- match.arg(type)
    Call <- object$call
    data <- data1 <- if (is.null(data)) {
        if (!is.null(Call$data)) {
            eval(Call$data)
        } else {
            stop("define the original data using the option data")
        }
    } else {
        data
    }

    data <- data[match(names(newdata), names(data))]
    data <- concat(data, newdata)
    parform <- stats::formula(object, what)
    if (length(parform) == 3) {
        parform[2] <- NULL
    }
    Terms <- stats::terms(parform)
    offsetVar <- if (!is.null(off.num <- attr(Terms, "offset"))) {
        eval(attr(Terms, "variables")[[off.num + 1]], data)
    }
    m <- stats::model.frame(
        Terms,
        data,
        xlev = object[[paste(what, "xlevels", sep = ".")]]
    )
    X <- stats::model.matrix(Terms, data, contrasts = object$contrasts)
    y <- object[[paste(what, "lp", sep = ".")]]
    w <- object[[paste(what, "wt", sep = ".")]]
    onlydata <- data$source == "data"
    smo.mat <- object[[paste(what, "s", sep = ".")]]
    if (!is.null(off.num)) {
        y <- (y - offsetVar[onlydata])
    }
    if (!is.null(smo.mat)) {
        n.smooths <- dim(smo.mat)[2]
        y <- (y - smo.mat %*% rep(1, n.smooths))
    }

    # Modified from the original prediction function.
    if (safe) {
        refit <- stats::lm.wfit(X[onlydata, , drop = FALSE], y, w)
        if (
            abs(sum(stats::resid(refit))) > 0.1 ||
                abs(sum(
                    stats::coef(object, what = what) - stats::coef(refit),
                    na.rm = TRUE
                )) >
                    1e-05
        ) {
            warning(paste(
                "There is a discrepancy  between the original and the re-fit",
                " \n used to achieve 'safe' predictions \n ",
                sep = ""
            ))
        }
        coef <- refit$coef
    } else {
        coef <- stats::coef(object, what = what)
    }

    nX <- dimnames(X)
    rownames <- nX[[1]][!onlydata]
    nrows <- sum(!onlydata)
    nac <- is.na(coef)
    assign.coef <- attr(X, "assign")
    collapse <- type != "terms"
    Xpred <- X[!onlydata, ]
    Xpred <- matrix(Xpred, nrow = nrows)
    if (!collapse) {
        aa <- attr(X, "assign")
        ll <- attr(Terms, "term.labels")
        if (attr(Terms, "intercept") > 0) {
            ll <- c("(Intercept)", ll)
        }
        aaa <- factor(aa, labels = ll)
        asgn <- split(order(aa), aaa)
        hasintercept <- attr(Terms, "intercept") > 0
        p <- refit$qr$rank
        p1 <- seq(len = p)
        piv <- refit$qr$pivot[p1]
        if (hasintercept) {
            asgn$"(Intercept)" <- NULL
            avx <- colMeans(X[onlydata, ])
            termsconst <- sum(avx[piv] * coef[piv])
        }
        nterms <- length(asgn)
        pred <- matrix(ncol = nterms, nrow = nrows)
        dimnames(pred) <- list(rownames(newdata), names(asgn))
        if (hasintercept) {
            Xpred <- sweep(Xpred, 2, avx)
        }
        unpiv <- rep.int(0, NCOL(Xpred))
        unpiv[piv] <- p1
        for (i in seq(1, nterms, length = nterms)) {
            iipiv <- asgn[[i]]
            ii <- unpiv[iipiv]
            iipiv[ii == 0] <- 0
            pred[, i] <- if (any(iipiv > 0)) {
                Xpred[, iipiv, drop = FALSE] %*% coef[iipiv]
            } else {
                0
            }
        }
        attr(pred, "constant") <- if (hasintercept) {
            termsconst
        } else {
            0
        }
        if (!is.null(terms)) {
            pred <- pred[, terms, drop = FALSE]
        }
    } else {
        pred <- drop(Xpred[, !nac, drop = FALSE] %*% coef[!nac])
        if (!is.null(off.num) && collapse) {
            pred <- pred + offsetVar[!onlydata]
        }
    }
    if (!is.null(smo.mat)) {
        cat("new prediction", "\n")
        smooth.labels <- dimnames(smo.mat)[[2]]
        pred.s <- array(
            0,
            c(nrows, n.smooths),
            list(
                names(pred),
                dimnames(smo.mat)[[2]]
            )
        )
        smooth.calls <- lapply(m[smooth.labels], attr, "call")
        data <- subset(m, onlydata, drop = FALSE)
        attr(data, "class") <- NULL
        new.m <- subset(m, !onlydata, drop = FALSE)
        attr(new.m, "class") <- NULL
        residuals <- if (!is.null(off.num)) {
            object[[paste(what, "wv", sep = ".")]] -
                object[[paste(what, "lp", sep = ".")]] +
                offsetVar[onlydata]
        } else {
            object[[paste(what, "wv", sep = ".")]] -
                object[[paste(what, "lp", sep = ".")]]
        }
        for (TT in smooth.labels) {
            if (is.matrix(m[[TT]])) {
                nm <- names(attributes(m[[TT]]))
                attributes(data[[TT]]) <- attributes(m[[TT]])[nm[
                    -c(
                        1,
                        2
                    )
                ]]
            } else {
                attributes(data[[TT]]) <- attributes(m[[TT]])
            }
            Call <- smooth.calls[[TT]]
            Call$xeval <- substitute(new.m[[TT]], list(TT = TT))
            z <- residuals + smo.mat[, TT]
            pred.s[, TT] <- eval(Call)
        }
        if (type == "terms") {
            pred[, smooth.labels] <- pred[, smooth.labels] +
                pred.s[, smooth.labels]
        } else {
            pred <- drop(pred + pred.s %*% rep(1, n.smooths))
        }
    }
    if (type == "response") {
        if (methods::is(eval(parse(text = object$family[[1]])), "gamlss.family")) {
            pred <- eval(parse(text = object$family[[1]]))[[paste(
                what,
                "linkinv",
                sep = "."
            )]](pred)
        } else {
            pred <- gamlss.dist::gamlss.family(eval(parse(
                text = paste(
                    stats::family(object)[1],
                    "(",
                    what,
                    ".link=",
                    eval(parse(text = (paste("object$", what, ".link", sep = "")))),
                    ")",
                    sep = ""
                )
            )))[[paste(what, "linkinv", sep = ".")]](pred)
        }
    }
    pred
}


#' @include sanity_model.R
#' @rdname sanitize_model_specific
#' @export
sanitize_model_specific.gamlss <- function(model, calling_function, ...) {
    if (calling_function == "hypotheses") {
        msg <- "`marginaleffects` does not support hypothesis tests for models of this class."
        stop_sprintf(msg)
    }
    return(model)
}

sanitize_hypothesis <- function(hypothesis, ...) {
    checkmate::assert(
        checkmate::check_character(hypothesis, pattern = "="),
        checkmate::check_numeric(hypothesis),
        checkmate::check_formula(hypothesis),
        checkmate::check_matrix(hypothesis),
        checkmate::check_function(hypothesis),
        checkmate::check_null(hypothesis))

    hnull <- 0

    if (isTRUE(checkmate::check_character(hypothesis, pattern = "="))) {
        out <- paste(gsub("=", "-(", hypothesis), ")")
        attr(out, "label") <- hypothesis
        hypothesis <- out
    } else if (isTRUE(checkmate::check_matrix(hypothesis))) {
        attr(hypothesis, "label") <- colnames(hypothesis)
    } else if (isTRUE(checkmate::check_numeric(hypothesis, len = 1))) {
        hnull <- hypothesis
        hypothesis <- NULL
    }

    out <- list(
        "hypothesis" = hypothesis,
        "hypothesis_null" = hnull
    )

    return(out)
}

#' @rdname get_predict
#' @export
get_predict.rq <- function(
    model,
    newdata = insight::get_data(model),
    type = NULL,
    ...) {
    # type argument of the method is used to specify confidence interval type
    insight::check_if_installed("quantreg")

    MM <- attr(newdata, "marginaleffects_model_matrix")
    if (isTRUE(checkmate::check_matrix(MM))) {
        beta <- get_coef(model)
        out <- drop(MM %*% beta)
        if (isTRUE(checkmate::check_numeric(out, len = nrow(newdata)))) {
            out <- data.table(estimate = out)
        } else {
            out <- data.table(estimate = out)
        }
    } else {
        out <- quantreg::predict.rq(model, newdata = newdata, ...)
        out <- data.table(estimate = out)
    }

    out <- add_rowid(out, newdata)
    return(out)
}


#' @include sanity_model.R
#' @rdname sanitize_model_specific
#' @keywords internal
sanitize_model_specific.rqs <- function(model, ...) {
    stop(
        "`marginaleffects` only supports `quantreg::rq` models with a single `tau` value.",
        call. = FALSE
    )
}

# #' @rdname get_model_matrix
# #' @keywords internal
# #' @export
# get_model_matrix.rq <- function(object, newdata) {
#         tt <- stats::terms(object)
#         Terms <- delete.response(tt)
#         m <- model.frame(Terms, newdata, na.action = na.pass, xlev = object$xlevels)
#         if (!is.null(cl <- attr(Terms, "dataClasses")))
#             stats::.checkMFClasses(cl, m)
#         X <- model.matrix(Terms, m, contrasts.arg = object$contrasts)
#         if (!isTRUE(nrow(X) == nrow(newdata))) {
#             return(NULL)
#         } else {
#             return(X)
#         }
# }

#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.data.frame <- function(model, ...) {
    checkmate::assert_data_frame(model)
    if (!"estimate" %in% colnames(model)) {
        insight::format_error("The model object is a data.frame but doesn't contain the column 'estimate'. Make sure these columns are present")
    }

    out <- model$estimate
    if ("term" %in% colnames(model)) {
        names(out) <- model$term
    } else {
        names(out) <- seq_along(out)
    }

    return(out)
}


#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.data.frame <- function(model, coefs, ...) {
  model$estimate = coefs
  return(model)
}

get_predictions <- function(
    mfx,
    type,
    model_perturbed = NULL, # important for perturbed model
    by = NULL,
    byfun = byfun,
    hypothesis = NULL,
    verbose = TRUE,
    hi = NULL, # sink hole for shared comparisons/predictions call
    lo = NULL, # sink hole avoids pushing these variables through ... in get_predict()
    original = NULL, # sink hole
    ...) {
    newdata <- mfx@newdata

    # sometimes we want the perturbed coefficients model supplied by get_se_delta().
    model <- if (is.null(model_perturbed)) mfx@model else model_perturbed

    out <- get_predict_error(
        model,
        newdata = newdata,
        type = type,
        mfx = mfx,
        ...
    )

    # this needs to be a data.table
    data.table::setDT(out)

    if (
        !"rowid" %in% colnames(out) &&
            "rowid" %in% colnames(mfx@newdata) &&
            nrow(out) == nrow(mfx@newdata)
    ) {
        out$rowid <- mfx@newdata$rowid
    }

    # extract attributes before setDT
    draws <- attr(out, "posterior_draws")

    # TODO: find a cheaper way to do this, but it's tricky
    # variables can come from:
    # - by: characters, data.frame, TRUE, groups
    # - wts
    # - hypothesis multi-part formulae
    cols <- setdiff(colnames(newdata), colnames(out))
    if (inherits(mfx@model, "mlogit")) {
        out <- merge_by_rowid(out, newdata)
    } else {
        nd <- subset(newdata, select = cols)
        out <- cbind(out, nd)
    }

    # unpad factors before `by` and `hypothesis`
    tmp <- unpad(out, draws)
    list2env(tmp, envir = environment())


    # by: auto group
    if (isTRUE(checkmate::check_character(by))) {
        by <- intersect(c("group", by), colnames(out))
    }

    # averaging by groups
    out <- get_by(
        out,
        draws = draws,
        newdata = newdata,
        by = by,
        byfun = byfun,
        verbose = verbose,
        ...
    )

    draws <- attr(out, "posterior_draws")

    # hypothesis tests using the delta method
    out <- get_hypothesis(
        out,
        hypothesis = hypothesis,
        by = by,
        newdata = newdata,
        draws = draws
    )

    return(out)
}

#' @rdname get_group_names
#' @export
get_group_names.mblogit <- function(model, type, ...) {
    out <- get_predict(model, type = type)
    if ("group" %in% colnames(out)) {
        out <- unique(out$group)
    } else {
        out <- "main_marginaleffects"
    }
    return(out)
}

#' @include sanity_model.R
#' @rdname sanitize_model_specific
#' @export
sanitize_model_specific.mblogit <- function(model, calling_function = "marginaleffects", ...) {
    if (calling_function == "marginaleffects") {
        variables <- insight::find_variables(model, flatten = TRUE, verbose = FALSE)
        dat <- insight::get_data(model)
        dat <- dat[, intersect(variables, colnames(dat))]
        flag <- any(sapply(dat, is.character))
        if (isTRUE(flag)) {
            stop("Cannot compute marginal effects for models of class `mblogit` when the data includes character variables. Please convert character variables to factors in the dataset before fitting the model, and call `marginaleffects` again.", call. = FALSE)
        }
    }
    return(model)
}


#' @rdname get_predict
#' @export
get_predict.mblogit <- function(model,
                                newdata = insight::get_data(model),
                                type = "response",
                                ...) {
    out <- suppressMessages(
        get_predict.multinom(
            model = model,
            newdata = newdata,
            type = type,
            ...))
    return(out)
}


#' @rdname get_coef
#' @export
get_coef.mblogit <- function(model, ...) {
    # get_coef.default uses `insight::get_parameters`, but does not combine
    # Response and Parameter columns. This also matches naming scheme from
    # `vcov`
    stats::coef(model)
}

get_comparisons <- function(
    mfx,
    type,
    variables,
    original,
    lo,
    hi,
    model_perturbed = NULL,
    by = NULL,
    byfun = NULL,
    hypothesis = NULL,
    cross = FALSE,
    verbose = TRUE,
    ...) {
    newdata <- mfx@newdata

    # get_se_delta() needs perturbed coefficients model
    model <- if (is.null(model_perturbed)) mfx@model else model_perturbed

    predictions <- predictions_hi_lo(model, lo, hi, type, ...)
    list2env(predictions, environment())

    # needs to be a data.table
    out <- data.table(pred_lo)

    # predict() takes up 2/3 of the wall time. This call is only useful when we
    # compute elasticities, or for the main estimate, not for standard errors,
    # so we probably save 1/3 of that 2/3.
    elasticities <- c(
        # "dydx", # useless and expensive
        "eyex",
        "eydx",
        "dyex",
        # "dydxavg", # useless and expensive
        "eyexavg",
        "eydxavg",
        "dyexavg"
    )
    fun <- function(x) {
        out <- checkmate::check_choice(x$comparison, choices = elasticities)
        isTRUE(out)
    }
    n_elasticities <- length(Filter(fun, variables))

    # need y-hat for elasticities
    # need to use the perturbed model, otherwise SEs are far from Stata
    custom_fun <- any(sapply(variables, function(x) identical(x$label, "custom")))
    if (n_elasticities > 0 || custom_fun) {
        pred_or <- get_predict_error(
            model,
            type = type,
            newdata = original,
            ...)
        out[, predicted := pred_or[["estimate"]]]
    } else {
        pred_or <- NULL
    }

    # TODO: find a cheaper way to do this, but it's tricky
    # variables can come from:
    # - by: characters, data.frame, TRUE, groups
    # - wts
    # - hypothesis multi-part formulae
    cols <- setdiff(colnames(original), colnames(out))
    out <- cbind(out, original[, ..cols])

    if (isTRUE(cross)) {
        out <- merge(out, newdata, by = "rowid", all.x = TRUE, sort = FALSE)
        if (isTRUE(nrow(out) == nrow(lo))) {
            tmp <- lo[,
                .SD,
                .SDcols = patterns("^contrast|marginaleffects_wts_internal")
            ]
            out <- cbind(out, tmp)
            idx <- c(
                "rowid",
                grep("^contrast", colnames(out), value = TRUE),
                colnames(out)
            )
            idx <- unique(idx)
            out <- out[, ..idx]
        }
    }

    if (!"term" %in% colnames(out)) {
        out[, "term" := "cross"]
    }

    # by
    if (isTRUE(checkmate::check_data_frame(by))) {
        tmp <- setdiff(intersect(colnames(out), colnames(by)), "by")
        # harmonize column types
        for (v in colnames(by)) {
            if (isTRUE(is.character(out[[v]])) && isTRUE(is.numeric(by[[v]]))) {
                by[[v]] <- as.character(by[[v]])
            } else if (isTRUE(is.numeric(out[[v]])) && isTRUE(is.character(by[[v]]))) {
                by[[v]] <- as.numeric(by[[v]])
            }
        }
        out[by, by := by, on = tmp]
        # only `by` because we give complete flexibility for user to aggregate across terms, groups, contrasts, etc.
        # that requires more work when building the `by` data frame, but it's more flexible
        by <- "by"
    } else if (isTRUE(by)) {
        regex <- "^term$|^contrast_?|^group$"
        by <- unique(grep(regex, colnames(out), value = TRUE))
    } else if (isTRUE(checkmate::check_character(by))) {
        regex <- "^term$|^contrast_?|^group$"
        by <- unique(c(by, grep(regex, colnames(out), value = TRUE)))
    }

    # comparison function could be different for different terms
    # sanitize_variables() ensures all functions are identical when there are cross
    fun_list <- sapply(names(variables), function(x) variables[[x]][["function"]])
    fun_list[["cross"]] <- fun_list[[1]]

    # elasticity requires the original (properly aligned) predictor values
    # this will discard factor variables which are duplicated, so in principle
    # it should be the "correct" size
    # also need `x` when `x` is in the signature of the `comparison` custom function

    elasticities <- prepare_elasticities(variables, original, out, by, elasticities)

    draws <- attr(pred_lo, "posterior_draws")
    draws_lo <- attr(pred_lo, "posterior_draws")
    draws_hi <- attr(pred_hi, "posterior_draws")
    draws_or <- attr(pred_or, "posterior_draws")

    out[, predicted_lo := pred_lo[["estimate"]]]
    out[, predicted_hi := pred_hi[["estimate"]]]


    if (!is.null(pred_or)) {
        out[, predicted := pred_or[["estimate"]]]
    } else {
        out[, predicted := NA_real_]
    }

    idx <- grep(
        "^contrast|^group$|^term$|^type$|^comparison_idx$",
        colnames(out),
        value = TRUE
    )


    # when `by` is a character vector, we sometimes modify the comparison
    # function on the fly to use the `avg` version.  this is important and
    # convenient because some of the statistics are non-collapsible, so we can't
    # average them at the very end.  when `by` is a data frame, we do this only
    # at the very end.
    # TODO: What is the UI for this? Doesn't make sense to have different functions.
    if (isTRUE(checkmate::check_character(by))) {
        tmp <- intersect(colnames(newdata), c(by, colnames(out)))
        if (length(tmp) > 1) {
            tmp <- subset(newdata, select = tmp)
            # Issue #1638: unable to merge back some list columns
            out <- tryCatch(merge(out, tmp, all.x = TRUE, sort = FALSE), error = function(e) {warning(e);out})
            idx <- unique(c(idx, by))
        }
    }

    # we feed these columns to safefun(), even if they are useless for categoricals
    if (!"marginaleffects_wts_internal" %in% colnames(out)) {
        out[, "marginaleffects_wts_internal" := NA]
    }


    # need a temp index for group-by operations when elasticities is a vector of length equal to full rows of `out`
    tmp <- grep("^term$|^contrast|^group$", colnames(out), value = TRUE)
    if (length(tmp) > 0) {
        out[, tmp_idx := seq_len(.N), by = tmp]
    } else {
        out[, tmp_idx := seq_len(.N)]
    }

    if (!is.null(draws)) {
        result <- compare_hi_lo_bayesian(
            out = out,
            draws = draws,
            draws_hi = draws_hi,
            draws_lo = draws_lo,
            draws_or = draws_or,
            by = by,
            cross = cross,
            variables = variables,
            fun_list = fun_list,
            elasticities = elasticities,
            newdata = newdata
        )
        out <- result$out
        draws <- result$draws
    } else {
        out <- compare_hi_lo_frequentist(
            out = out,
            idx = idx,
            cross = cross,
            variables = variables,
            fun_list = fun_list,
            elasticities = elasticities,
            newdata = newdata
        )
    }

    # clean
    if ("rowid_dedup" %in% colnames(out)) {
        out[, "rowid_dedup" := NULL]
    }

    # averaging by groups
    # sometimes this work is already done
    # if `by` is a column name, then we have merged-in a data frame earlier
    auto_mean_fun_sub <- any(grepl("^mean\\(", unique(out$contrast)))
    if (!auto_mean_fun_sub && any(grepl("^contrast[_]?", colnames(out)))) {
        out <- get_by(
            out,
            draws = draws,
            newdata = newdata,
            by = by,
            verbose = verbose
        )
        draws <- attr(out, "posterior_draws")
    }

    # before get_hypothesis
    attr(out, "posterior_draws") <- draws

    # hypothesis tests using the delta method
    out <- get_hypothesis(
        out,
        hypothesis,
        by = by,
        newdata = original,
        draws = draws
    )

    # reset settings
    settings_rm("marginaleffects_safefun_return1")

    return(out)
}


predictions_hi_lo <- function(model, lo, hi, type, ...) {
    # brms models need to be combined to use a single seed when sample_new_levels="gaussian"
    if (inherits(model, c("brmsfit", "bart"))) {
        if (!"rowid" %in% colnames(lo)) {
            lo$rowid <- hi$rowid <- seq_len(nrow(lo))
        }

        both <- rbindlist(list(lo, hi))

        pred_both <- get_predict_error(
            model,
            type = type,
            newdata = both,
            ...
        )

        pred_both[, "lo" := seq_len(.N) <= .N / 2, by = "group"]

        pred_lo <- pred_both[pred_both$lo, .(rowid, group, estimate), drop = FALSE]
        pred_hi <- pred_both[!pred_both$lo, .(rowid, group, estimate), drop = FALSE]

        draws <- attr(pred_both, "posterior_draws")
        draws_lo <- draws[pred_both$lo, , drop = FALSE]
        draws_hi <- draws[!pred_both$lo, , drop = FALSE]

        attr(pred_lo, "posterior_draws") <- draws_lo
        attr(pred_hi, "posterior_draws") <- draws_hi
    } else {
        pred_lo <- get_predict_error(
            model,
            type = type,
            newdata = lo,
            ...
        )

        pred_hi_result <- myTryCatch(get_predict(
            model,
            type = type,
            newdata = hi,
            ...
        ))

        # otherwise we keep the full error object instead of extracting the value
        if (inherits(pred_hi_result$value, "data.frame")) {
            pred_hi <- pred_hi_result$value
        } else {
            pred_hi <- pred_hi_result$error
        }
    }

    return(list(pred_lo = pred_lo, pred_hi = pred_hi))
}


compare_hi_lo_bayesian <- function(out, draws, draws_hi, draws_lo, draws_or, by, cross, variables, fun_list, elasticities, newdata) {
    # drop missing otherwise get_averages() fails when trying to take a
    # simple mean
    idx_na <- !is.na(out$predicted_lo)
    # Build a single index and reuse it so `out` and the posterior draws remain aligned
    # while removing rows with missing `predicted_lo`.
    out <- out[idx_na]

    # TODO: performance is probably terrrrrible here, but splitting is
    # tricky because grouping rows are not always contiguous, and the order
    # of rows is **extremely** important because draws don't have the
    # indices that would allow us to align them back with `out`
    draws <- draws[idx_na, , drop = FALSE]

    if (isTRUE(checkmate::check_character(by, min.len = 1))) {
        by_idx <- out[, ..by]
        by_idx <- do.call(paste, c(by_idx, sep = "|"))
    } else {
        by_idx <- out$term
    }

    # loop over columns (draws) and term names because different terms could use different functions
    for (tn in unique(by_idx)) {
        for (i in seq_len(ncol(draws))) {
            idx <- by_idx == tn
            draws[idx, i] <- compare_hi_lo(
                hi = draws_hi[idx, i],
                lo = draws_lo[idx, i],
                y = draws_or[idx, i],
                n = sum(idx),
                term = out$term[idx],
                cross = cross,
                wts = out$marginaleffects_wts_internal[idx],
                tmp_idx = out$tmp_idx[idx],
                newdata = newdata,
                variables = variables,
                fun_list = fun_list,
                elasticities = elasticities
            )
        }
    }

    # function returns unique value
    idx <- !is.na(draws[, 1])
    draws <- draws[idx, , drop = FALSE]

    # if comparison returns a single value, it means we are using a special shortcut comparison function.
    # to do this, we padded with NA. That means we don't want `rowid` or covariates otherwise they will be misleading
    # since misaligned. But we do need the marginaleffects internal columns and by
    if (!all(idx)) {
        if (settings_equal("marginaleffects_safefun_return1", TRUE)) {
            cols <- grep("^estimate$|^group$|^term$|^contrast_?|^marginaleffects_wts_internal$|^by$",
                colnames(out),
                value = TRUE)
            if (isTRUE(checkmate::check_character(by, min.len = 1))) {
                cols <- unique(c(cols, by))
            }
            out <- subset(out, select = cols)
        }
        out <- out[idx, , drop = FALSE]
    }

    FUN_CENTER <- getOption(
        "marginaleffects_posterior_center",
        default = stats::median
    )

    out[, "estimate" := apply(draws, 1, FUN_CENTER)]

    return(list(out = out, draws = draws))
}


compare_hi_lo_frequentist <- function(out, idx, cross, variables, fun_list, elasticities, newdata) {
    # When survey weights add all-NA aux columns, stats::na.omit() (which ignores
    # `cols`) was zeroing out `out`. Build one index and reuse it so downstream
    # operations rely on the same filtered rows.
    idx_pred <- !is.na(out$predicted_lo)
    out <- out[idx_pred]
    # We want to write the "estimate" column in-place because it safer
    # than group-merge; there were several bugs related to this in the past.
    # safefun() returns 1 value and NAs when the function retunrs a
    # singleton.
    idx <- intersect(idx, colnames(out))
    out[,
        "estimate" := compare_hi_lo(
            hi = predicted_hi,
            lo = predicted_lo,
            y = predicted,
            n = .N,
            term = term,
            cross = cross,
            wts = marginaleffects_wts_internal,
            tmp_idx = tmp_idx,
            newdata = newdata,
            variables = variables,
            fun_list = fun_list,
            elasticities = elasticities
        ),
        keyby = idx
    ]
    out[, tmp_idx := NULL]

    # if comparison returns a single value, it means we are using a special shortcut comparison function.
    # to do this, we padded with NA. That means we don't want `rowid` or covariates otherwise they will be misleading
    # since misaligned. But we do need the marginaleffects internal columns and by
    if (anyNA(out$estimate)) {
        if (settings_equal("marginaleffects_safefun_return1", TRUE)) {
            # important to avoid merging
            idx <- c(idx, grep("^estimate$|^contrast|^group$|^term$|^marginaleffects_wts_internal$", colnames(out), value = TRUE))
            idx <- unique(intersect(idx, colnames(out)))
            out <- subset(out, select = idx)
        }
    }
    out <- out[!is.na(estimate)]

    return(out)
}


compare_hi_lo <- function(hi, lo, y, n, term, cross, wts, tmp_idx, newdata, variables, fun_list, elasticities) {
    if (n == 0 || length(hi) == 0) {
        return(numeric(0))
    }
    tn <- term[1]
    eps <- variables[[tn]]$eps
    # when cross=TRUE, sanitize_comparison enforces a single function
    if (isTRUE(cross)) {
        fun <- fun_list[[1]]
    } else {
        fun <- fun_list[[tn]]
    }
    args <- list(
        "hi" = hi,
        "lo" = lo,
        "y" = y,
        "eps" = eps,
        "w" = wts,
        "newdata" = newdata
    )

    # sometimes x is exactly the same length, but not always
    args[["x"]] <- elasticities[[tn]][tmp_idx]

    args <- args[names(args) %in% names(formals(fun))]
    con <- try(do_call(fun, args), silent = TRUE)

    if (
        !isTRUE(checkmate::check_numeric(con, len = n)) &&
            !isTRUE(checkmate::check_numeric(con, len = 1))
    ) {
        msg <- sprintf(
            "The function supplied to the `comparison` argument must accept two numeric vectors of predicted probabilities of length %s, and return a single numeric value or a numeric vector of length %s, with no missing value.",
            n,
            n
        ) # nolint
        stop_sprintf(msg)
    }
    if (length(con) == 1) {
        con <- c(con, rep(NA_real_, length(hi) - 1))
        settings_set("marginaleffects_safefun_return1", TRUE)
    }
    return(con)
}



prepare_elasticities <- function(variables, original, out, by, elasticities) {
    FUN <- function(z) {
        (is.character(z$comparison) && z$comparison %in% elasticities) ||
            (is.function(z$comparison) && "x" %in% names(formals(z$comparison)))
    }
    elasticities <- Filter(FUN, variables)

    if (length(elasticities) > 0) {
        # assigning a subset of "original" to "idx1" takes time and memory
        # better to do this here for most columns and add the "v" column only
        # in the loop
        if (!is.null(original)) {
            idx1 <- c(
                "rowid",
                "rowidcf",
                "term",
                "group",
                grep("^contrast", colnames(original), value = TRUE)
            )
            idx1 <- intersect(idx1, colnames(original))
            idx1 <- original[, ..idx1]
        }

        for (v in names(elasticities)) {
            idx2 <- unique(c(
                "rowid",
                "term",
                "group",
                by,
                grep("^contrast", colnames(out), value = TRUE)
            ))
            idx2 <- intersect(idx2, colnames(out))
            # discard other terms to get right length vector
            idx2 <- out[term == v, ..idx2]
            # original is NULL when cross=TRUE
            if (!is.null(original)) {
                # if not first iteration, need to remove previous "v" and "elast"
                if (v %in% colnames(idx1)) {
                    idx1[, (v) := NULL]
                }
                if ("elast" %in% colnames(idx1)) {
                    idx1[, elast := NULL]
                }
                idx1[, (v) := original[[v]]]
                setnames(idx1, old = v, new = "elast")
                on_cols <- intersect(colnames(idx1), colnames(idx2))
                idx2 <- unique(merge(idx2, idx1, by = on_cols, sort = FALSE)[
                    ,
                    elast := elast
                ])
            }
            elasticities[[v]] <- idx2$elast
        }
    }
    return(elasticities)
}

#' Plot Conditional or Marginal Slopes
#'
#' @description
#' Plot slopes on the y-axis against values of one or more predictors (x-axis, colors/shapes, and facets).
#'
#' The `by` argument is used to plot marginal slopes, that is, slopes made on the original data, but averaged by subgroups. This is analogous to using the `by` argument in the `slopes()` function.
#'
#' The `condition` argument is used to plot conditional slopes, that is, slopes computed on a user-specified grid. This is analogous to using the `newdata` argument and `datagrid()` function in a `slopes()` call. All variables whose values are not specified explicitly are treated as usual by `datagrid()`, that is, they are held at their mean or mode (or rounded mean for integers). This includes grouping variables in mixed-effects models, so analysts who fit such models may want to specify the groups of interest using the `condition` argument, or supply model-specific arguments to compute population-level estimates. See details below.

#' See the "Plots" vignette and website for tutorials and information on how to customize plots:
#'
#' * https://marginaleffects.com/bonus/plot.html
#' * https://marginaleffects.com
#'
#' @param variables Name of the variable whose marginal effect (slope) we want to plot on the y-axis.
#' @param condition Conditional slopes
#' + Character vector (max length 4): Names of the predictors to display.
#' + Named list (max length 4): List names correspond to predictors. List elements can be:
#'   - Numeric vector
#'   - Function which returns a numeric vector or a set of unique categorical values
#'   - Shortcut strings for common reference values: "minmax", "quartile", "threenum"
#' + 1: x-axis. 2: color/shape. 3: facet (wrap if no fourth variable, otherwise cols of grid). 4: facet (rows of grid).
#' + Numeric variables in positions 2 and 3 are summarized by Tukey's five numbers `?stats::fivenum`.
#' @param rug TRUE displays tick marks on the axes to mark the distribution of raw data.
#' @param gray FALSE grayscale or color plot
#' @param draw `TRUE` returns a `ggplot2` plot. `FALSE` returns a `data.frame` of the underlying data.
#' @param newdata When `newdata` is `NULL`, the grid is determined by the `condition` argument. When `newdata` is not `NULL`, the argument behaves in the same way as in the `predictions()` function. Note that the `condition` argument builds its own grid, so the `newdata` argument is ignored if the `condition` argument is supplied.
#' @inheritParams slopes
#' @template model_specific_arguments
#' @return A `ggplot2` object
#' @export
#' @examplesIf interactive() || isTRUE(Sys.getenv("R_DOC_BUILD") == "true")
#' mod <- lm(mpg ~ hp * drat * factor(am), data = mtcars)
#'
#' plot_slopes(mod, variables = "hp", condition = "drat")
#'
#' plot_slopes(mod, variables = "hp", condition = c("drat", "am"))
#'
#' plot_slopes(mod, variables = "hp", condition = list("am", "drat" = 3:5))
#'
#' plot_slopes(mod, variables = "am", condition = list("hp", "drat" = range))
#'
#' plot_slopes(mod, variables = "am", condition = list("hp", "drat" = "threenum"))
#'
#' # marginal slopes
#' plot_slopes(mod, variables = "hp", by = "am")
#'
#' # marginal slopes on a counterfactual grid
#' plot_slopes(mod,
#'     variables = "hp",
#'     by = "am",
#'     newdata = datagrid(am = 0:1, grid_type = "counterfactual")
#' )
#'
plot_slopes <- function(
    model,
    variables = NULL,
    condition = NULL,
    by = NULL,
    newdata = NULL,
    type = NULL,
    vcov = NULL,
    conf_level = 0.95,
    wts = FALSE,
    slope = "dydx",
    rug = FALSE,
    gray = getOption("marginaleffects_plot_gray", default = FALSE),
    draw = TRUE,
    ...) {
    if ("effect" %in% ...names()) {
        if (is.null(variables)) {
            variables <- ...elt(match("effect", ...names())[1L])
        } else {
            stop_sprintf(
                "The `effect` argument has been renamed to `variables`."
            )
        }
    }

    if (inherits(model, c("mira", "amest"))) {
        msg <- "This function does not support multiple imputation. Call `slopes()` or `avg_slopes()` instead. These functions return easy to plot data frames."
        stop_sprintf(msg)
    }

    # init
    call <- construct_call(model, "comparisons")
    model <- sanitize_model(model, call = call, newdata = newdata, wts = wts, vcov = vcov, by = by, ...)
    mfx <- new_marginaleffects_internal(
        call = call,
        model = model
    )

    if (inherits(mfx@model, "mira") && is.null(newdata)) {
        msg <- "Please supply a data frame to the `newdata` argument explicitly."
        stop_sprintf(msg)
    }

    # order of the first few paragraphs is important
    # if `newdata` is a call to `typical` or `counterfactual`, insert `model`
    # should probably not be nested too deeply in the call stack since we eval.parent() (not sure about this)
    scall <- rlang::enquo(newdata)
    newdata <- sanitize_newdata_call(scall, newdata, mfx = mfx)

    valid <- c("dydx", "eyex", "eydx", "dyex")
    checkmate::assert_choice(slope, choices = valid)

    out <- plot_comparisons(
        mfx@model,
        variables = variables,
        condition = condition,
        by = by,
        newdata = newdata,
        type = type,
        vcov = vcov,
        conf_level = conf_level,
        wts = wts,
        draw = draw,
        rug = rug,
        gray = gray,
        comparison = slope,
        ...
    )

    if (inherits(out, "ggplot")) {
        attr(out$data$estimate, "label") <- "Slope"
    }

    return(out)
}

bootstrap_boot <- function(model, INF_FUN, ...) {
    # attached by `inferences()`
    conf_type <- attr(model, "inferences_conf_type")
    checkmate::assert_choice(conf_type, choices = c("perc", "norm", "basic", "bca"))

    # bootstrap using the original data and call
    modcall <- insight::get_call(model)
    modeldata <- get_modeldata(model, additional_variables = FALSE)

    # evaluate the {marginaleffects} call to get output without inferences()
    # use ... because arguments are not the same for different {marginaleffects} functions
    dots <- list(...)
    dots[["vcov"]] <- FALSE

    # avoid recursion
    attr(model, "inferences_method") <- NULL
    out <- do.call(INF_FUN, c(list(model), dots))

    # default confidence level may be implicit in original call, but we need numeric
    if (is.null(dots[["conf_level"]])) {
        conf_level <- 0.95
    } else {
        conf_level <- dots[["conf_level"]]
    }

    bootfun <- function(data, indices) {
        d <- data[indices, , drop = FALSE]
        modcall[["data"]] <- d
        modboot <- eval(modcall)
        modboot <- eval(modboot)
        args <- c(list(modboot, modeldata = d), dots)
        out <- do.call(INF_FUN, args)$estimate
        return(out)
    }
    args <- list("data" = modeldata, "statistic" = bootfun)
    args <- c(args, attr(model, "inferences_dots"))
    args <- args[unique(names(args))]
    B <- do.call(boot::boot, args)

    # print.boot prints an ugly nested call
    t <- matrix(B$t, nrow = nrow(B$t))
    op <- cbind(
        apply(t, 2L, mean, na.rm = TRUE),
        sqrt(apply(t, 2L, function(t.st) stats::var(t.st[!is.na(t.st)]))))
    out$std.error <- op[, 2]

    # extract from weird boot.ci() list (inspired from `broom::tidy.broom` under MIT)
    ci_list <- lapply(seq_along(B$t0),
        boot::boot.ci,
        boot.out = B,
        conf = conf_level,
        type = conf_type)
    pos <- pmatch(conf_type, names(ci_list[[1]]))
    if (conf_type == "norm") {
        cols <- 2:3
    } else {
        cols <- 4:5
    }
    ci <- lapply(ci_list, function(x) x[[pos]])
    ci <- do.call("rbind", ci)[, cols]

    # add CI to original {marginaleffects} call
    if (is.matrix(ci)) {
        out$conf.low <- ci[, 1]
        out$conf.high <- ci[, 2]
    } else {
        out$conf.low <- ci[1]
        out$conf.high <- ci[2]
    }

    attr(out, "inferences") <- B
    attr(out, "posterior_draws") <- t(B$t)
    return(out)
}

#' Create a data.frame with all factor or character levels
#'
#' `model.matrix` breaks when `newdata` includes a factor
#' variable, but not all levels are present in the data. This is bad for us
#' because we often want to get predictions with one (or few) rows, where some
#' factor levels are inevitably missing.
#' @keywords internal
complete_levels <- function(x, character_levels = NULL) {

    checkmate::assert_data_frame(x)

    # fixest returned an empty list()
    if (is.null(character_levels) || length(character_levels) == 0) {
        return(data.frame())
    }

    # store variables with missing factors or characters
    vault <- list()
    for (v in colnames(x)) {
        if (is.factor(x[[v]])) {
            if (!all(levels(x[[v]]) %in% x[[v]])) {
                vault[[v]] <- factor(levels(x[[v]]), levels = levels(x[[v]]))
            }
        } else if (is.character(x[[v]])) {
            if (v %in% names(character_levels)) {
                vault[[v]] <- character_levels[[v]]
            }
        }
    }

    # create padding
    if (length(vault) > 0) {
        # HACK: Some models use a character variable with many levels (e.g.,
        # mixed-effects groups). This creates a massive padding dataset, and making
        # predictions can become very expensive.
        if (isTRUE(sum(lengths(vault)) > 100)) return(data.frame())

        padding <- utils::head(x, 1)
        data.table::setDT(padding)
        for (v in names(vault)) {
            padding[[v]] <- NULL
        }
        fun <- data.table::CJ
        gr <- do.call("fun", vault)
        padding <- cjdt(list(padding, gr))
        to_keep <- colnames(x)
        padding[, ..to_keep]
        setcolorder(padding, to_keep)
        data.table::setDF(padding)
    } else {
        padding <- data.frame()
    }

    padding$rowid <- -1 * padding$rowid

    return(padding)
}

get_hypotheses <- function(model_perturbed, hypothesis, hypothesis_is_formula, newparams = NULL, ...) {
    if (isTRUE(checkmate::check_numeric(model_perturbed))) {
        out <- data.frame(term = seq_along(model_perturbed), estimate = model_perturbed)
    } else if (inherits(model_perturbed, "data.frame")) {
        out <- model_perturbed
        if (!"estimate" %in% colnames(out)) {
            msg <- "`hypothesis` function must return a data.frame with a column named `estimate`."
            stop_sprintf(msg)
        }
        if (!"term" %in% colnames(out)) {
            n <- tryCatch(names(stats::coef(model_perturbed)), error = function(e) NULL)
            if (is.null(n)) {
                n <- paste0("b", seq_len(nrow(out)))
            }
            out$term <- n
        }
        if (!all(c("term", "estimate") %in% colnames(out))) {
            msg <- "`hypothesis` function must return a data.frame with two columns named `term` and `estimate`."
            stop_sprintf(msg)
        }

        # unknown model
    } else if (!is.function(hypothesis)) {
        out <- insight::get_parameters(model_perturbed, ...)
        if ("Component" %in% colnames(out) && !anyNA(out$Component)) {
            out$Parameter <- sprintf("%s_%s", out$Component, out$Parameter)
        } else if ("Response" %in% colnames(out) && !anyNA(out$Response)) {
            out$Parameter <- sprintf("%s_%s", out$Response, out$Parameter)
        }
        colnames(out)[1:2] <- c("term", "estimate")

        # glmmTMB
        if (!is.null(newparams)) {
            out$estimate <- newparams
        }
    } else if (hypothesis_is_formula) {
        beta <- get_coef(model_perturbed)
        out <- data.table::data.table(estimate = beta, term = names(beta))

        # unknown model but user-supplied hypothesis function
    } else {
        out <- model_perturbed
    }


    tmp <- get_hypothesis(out, hypothesis = hypothesis)

    # must be a data.table
    data.table::setDT(tmp)

    out <- tmp$estimate

    hypothesis_function_by <- attr(tmp, "hypothesis_function_by")

    # labels
    lab <- c("hypothesis", "term", hypothesis_function_by)
    lab <- intersect(lab, colnames(tmp))
    if (length(lab) > 0) {
        lab <- tmp[, ..lab]
        attr(out, "label") <- lab
    }

    if ("group" %in% colnames(tmp)) {
        attr(out, "grouplab") <- tmp[["group"]]
    }

    attr(out, "hypothesis_function_by") <- hypothesis_function_by
    return(out)
}

#' @rdname set_coef
#' @export
set_coef.stpm2 <- function(model, coefs, ...) {
    insight::check_if_installed("rstpm2")
    model@fullcoef <- coefs
    model
}

#' @rdname set_coef
#' @export
set_coef.pstpm2 <- set_coef.stpm2

#' @rdname set_coef
#' @export
set_coef.gsm <- set_coef.stpm2


#' @rdname set_coef
#' @export
set_coef.aft <- set_coef.stpm2


#' @rdname get_vcov
#' @export
get_vcov.stpm2 <- function(model, ...) {
    insight::check_if_installed("rstpm2")
    rstpm2::vcov(model)
}

#' @rdname get_vcov
#' @export
get_vcov.pstpm2 <- get_vcov.stpm2

#' @rdname get_vcov
#' @export
get_vcov.gsm <- get_vcov.stpm2

#' @rdname get_vcov
#' @export
get_vcov.aft <- get_vcov.stpm2


#' @rdname get_predict
#' @export
get_predict.stpm2 <- function(model, newdata = NULL, ...) {
    insight::check_if_installed("rstpm2")
    pred <- rstpm2::predict(model, newdata = newdata, ...)
    sanity_predict_vector(pred = pred, model = model, newdata = newdata)
    sanity_predict_numeric(pred = pred, model = model, newdata = newdata)
    out <- data.frame(
        rowid = seq_len(nrow(newdata)),
        estimate = pred)
    return(out)
}

#' @rdname get_predict
#' @export
get_predict.pstpm2 <- get_predict.stpm2

#' @rdname get_predict
#' @export
get_predict.gsm <- get_predict.stpm2

#' @rdname get_predict
#' @export
get_predict.aft <- get_predict.stpm2

multcomp_test <- function(
    object,
    multcomp = FALSE,
    conf_level = 0.95,
    df = Inf
) {
    valid <- c(
        "holm",
        "hochberg",
        "hommel",
        "bonferroni",
        "BH",
        "BY",
        "fdr",
        "single-step",
        "Shaffer",
        "Westfall",
        "free"
    )
    checkmate::assert(
        checkmate::check_choice(multcomp, choices = valid),
        checkmate::check_flag(multcomp)
    )

    if (isFALSE(multcomp)) {
        return(object)
    }

    if (isTRUE(multcomp)) multcomp <- "holm"

    insight::check_if_installed("multcomp")

    k <- multcomp::glht(object, df = df)
    k <- summary(k, test = multcomp::adjusted(type = multcomp))
    k <- stats::confint(k, level = conf_level)
    object$p.value <- k$test$pvalues
    object$conf.low <- k$confint[, 2, drop = TRUE]
    object$conf.high <- k$confint[, 3, drop = TRUE]
    object$s.value <- -log2(object$p.value)

    return(object)
}

inferences_rsample <- function(x, R = 1000, conf_level = 0.95, conf_type = "perc", estimator = NULL, data_train = NULL, mfx = NULL, ...) {
    insight::check_if_installed("rsample")

    out <- x

    if (!is.null(estimator)) {
        bootfun <- function(split, ...) {
            d <- rsample::analysis(split)
            result <- estimator(d)
            return(result)
        }
    } else {
        bootfun <- function(split, ...) {
            d <- rsample::analysis(split)
            boot_mfx <- refit(x, data = d, vcov = FALSE)
            out <- data.frame(boot_mfx)
            return(out)
        }
    }

    # rsample::int_bca/pctl collapse estimates when term is duplicated because of term/contrast/by unique
    bootfun_term <- function(split, ...) {
        out <- bootfun(split, ...)
        # This is a hack to avoid the issue of rsample::int_bca/pctl collapsing estimates when term is duplicated because of term/contrast/by unique
        # Warning: assumes that we always return estimates in the same order as the original {marginaleffects} call.
        # data.frame() to remove super heavy attributes (model, data, etc.)
        # as.character() because `rsample` assumes character `term`. Reported here:
        # https://github.com/tidymodels/rsample/issues/574
        out <- data.frame(
            term = as.character(seq_len(nrow(out))),
            estimate = out$estimate
        )
        return(out)
    }

    args <- list("apparent" = TRUE)
    args[["times"]] <- R
    args[["data"]] <- data_train

    args <- c(args, list(...))
    if ("group" %in% ...names()) {
        splits <- do.call(rsample::group_bootstraps, args)
    } else {
        splits <- do.call(rsample::bootstraps, args)
    }

    if (isTRUE(getOption("marginaleffects_parallel_inferences", default = FALSE))) {
        insight::check_if_installed("future.apply")
        pkg <- getOption("marginaleffects_parallel_packages", default = NULL)
        pkg <- unique(c("marginaleffects", pkg))
        splits$results <- future.apply::future_lapply(
            splits$splits,
            bootfun_term,
            future.seed = TRUE,
            future.packages = pkg
        )
    } else {
        splits$results <- lapply(splits$splits, bootfun_term)
    }

    if (isTRUE(conf_type == "bca")) {
        ci <- rsample::int_bca(
            splits,
            statistics = results,
            .fn = bootfun_term,
            alpha = 1 - conf_level
        )
    } else {
        ci <- rsample::int_pctl(
            splits,
            statistics = results,
            alpha = 1 - conf_level
        )
    }

    # hack: rsample only supports character `term`
    # https://github.com/tidymodels/rsample/issues/574
    ci$term <- as.numeric(ci$term)
    ci <- ci[order(ci$term), ]

    out$conf.low <- ci$.lower
    out$conf.high <- ci$.upper

    cols <- setdiff(names(out), c("p.value", "std.error", "statistic", "s.value", "df"))
    out <- out[, cols, drop = FALSE]

    if (all(out$term == seq_len(nrow(out)))) {
        out$term <- NULL
    }

    draws <- lapply(
        splits$results,
        function(x) as.matrix(x[, "estimate", drop = FALSE])
    )
    draws[[length(draws)]] <- NULL # apparent=TRUE appended the original estimates to the end
    draws <- do.call("cbind", draws)
    colnames(draws) <- NULL

    mfx <- attr(x, "marginaleffects")
    mfx@draws <- draws
    mfx@inferences <- splits
    attr(out, "marginaleffects") <- mfx

    return(out)
}

#' Get predicted values from a model object (internal function)
#'
#' @return A data.frame of predicted values with a number of rows equal to the
#' number of rows in `newdata` and columns "rowid" and "estimate". A "group"
#' column is added for multivariate models or models with categorical outcomes.
#' @rdname get_predict
#' @inheritParams slopes
#' @keywords internal
#' @export
get_predict <- function(
  model,
  newdata,
  type,
  mfx = NULL,
  newparams = NULL,
  ndraws = NULL,
  se.fit = NULL,
  ...
) {
    UseMethod("get_predict", model)
}


#' @rdname get_predict
#' @export
get_predict.default <- function(
  model,
  newdata = insight::get_data(model),
  type = "response",
  mfx = NULL,
  newparams = NULL,
  ndraws = NULL,
  se.fit = NULL,
  ...
) {
    dots <- list(...)

    if (is.null(type)) {
        type <- sanitize_type(model = model, type = type)
    }

    # some predict methods raise warnings on unused arguments
    unused <- c(
        "normalize_dydx",
        "eps",
        "numDeriv_method",
        "internal_call",
        "draw",
        "modeldata",
        "flag",
        "marginaleffects",
        "mfx"
    )
    dots <- dots[setdiff(names(dots), unused)]

    # first argument in the predict methods is not always named "x" or "model"
    dots[["newdata"]] <- newdata
    dots[["type"]] <- type
    if (!is.null(newparams)) {
        dots[["newparams"]] <- newparams
    }
    args <- c(list(model), dots)

    # `pred` is a secret argument called by `predict.lm` to turn a numeric vector into a data frame with correct `rowid`
    if ("pred" %in% ...names()) {
        pred <- dots[["pred"]]
    } else {
        fun <- stats::predict
        pred <- suppressWarnings(do_call(fun, args))
    }

    # 1-d array to vector (e.g., {mgcv})
    if (is.array(pred) && length(dim(pred)) == 1) {
        pred <- as.vector(pred)
    }

    # 1-d array to vector (e.g., Gam from {gam})
    if (
        is.array(pred) &&
            length(dim(pred)) == 3 &&
            dim(pred)[1] == 1 &&
            dim(pred)[2] == 1 &&
            dim(pred)[3] > 1
    ) {
        pred <- as.vector(pred)
    }

    # phylolm
    if (isTRUE(checkmate::check_matrix(pred, ncols = 1))) {
        pred <- drop(pred)
    }

    # atomic vector
    if (isTRUE(checkmate::check_atomic_vector(pred))) {
        # strip weird attributes added by some methods (e.g., predict.svyglm)
        # as.numeric is slow with large objects and we can't use is.numeric
        # to run it conditionally because objects of class "svystat" are
        # already numeric
        class(pred) <- "numeric"
        out <- data.frame(estimate = pred)

        # matrix with outcome levels as columns
    } else if (is.matrix(pred)) {
        if (is.null(colnames(pred))) {
            colnames(pred) <- seq_len(ncol(pred))
        }
        # internal calls always includes "rowid" as a column in `newdata`
        out <- data.frame(
            group = rep(colnames(pred), each = nrow(pred)),
            estimate = c(pred)
        )
        out$group <- group_to_factor(out$group, model)
    } else {
        stop(
            sprintf(
                "Unable to extract predictions of type %s from a model of class %s. Please report this problem, along with reproducible code and data on Github: https://github.com/vincentarelbundock/marginaleffects/issues",
                type,
                class(model)[1]
            ),
            call. = FALSE
        )
    }

    out <- add_rowid(out, newdata)

    return(out)
}


get_predict_error <- function(model, newdata = NULL, ...) {
    pred_result <- myTryCatch(get_predict(model, newdata = newdata, ...))

    if (inherits(pred_result$value, "data.frame")) {
        out <- pred_result$value
        return(out)
    } else {
        if (
            inherits(pred_result$error, "rlang_error") &&
                isTRUE(grepl("the object should be", pred_result$error$message))
        ) {
            stop_sprintf(pred_result$error$message)
        }

        msg <- "Unable to compute predicted values with this model. This error can arise when `insight::get_data()` is unable to extract the dataset from the model object, or when the original data frame (usually in the global environment) was modified since fitting the model. Make sure the original data object is unchanged and/or try to supply a different dataset to the `newdata` argument."
        if (!is.null(pred_result$error)) {
            msg <- c(
                msg,
                "\n\n",
                "In addition, this error message was raised:",
                "\n",
                pred_result$error$message
            )
        }
        msg <- c(
            msg,
            "\n\n",
            "Bug Tracker: https://github.com/vincentarelbundock/marginaleffects/issues"
        )
        stop_sprintf(msg)
    }
}

# Full conformal prediction
#
# WARNING: This method is computationally expensive as it refits the model
# for each combination of test observation and trial value. It is typically
# several orders of magnitude slower than split or CV+ conformal methods.
# Use sparingly on small test sets or consider conformal_split/conformal_cv+ instead.
#
# Reference implementation from the `probably` package (MIT License, 2025-11-10):
# Copyright holders: Max Kuhn [aut, cre], Davis Vaughan [aut], Edgar Ruiz [aut],
# Posit Software, PBC [cph, fnd] (ROR ID: https://ror.org/00cvxb145)
# URL: https://github.com/tidymodels/probably
#
# The implementation here is adapted to fit the marginaleffects package API.

# Conformity test for a single trial value - returns difference for uniroot
conformity_test_full <- function(trial_y,
                                 x_test,
                                 model,
                                 data_train,
                                 response_name,
                                 conf_level,
                                 mfx) {
    # Augment training data with test point and trial outcome
    x_test[[response_name]] <- trial_y

    # Ensure x_test has same columns as data_train in same order
    # Reorder columns to match data_train
    x_test <- x_test[, colnames(data_train), drop = FALSE]

    # Ensure all column types in x_test match data_train before rbind
    # This prevents type coercion issues with recipes/workflows
    for (col in colnames(data_train)) {
        target_class <- class(data_train[[col]])[1]
        current_class <- class(x_test[[col]])[1]
        if (target_class != current_class) {
            x_test[[col]] <- methods::as(x_test[[col]], target_class)
        }
    }

    augmented_data <- rbind(data_train, x_test)
    n <- nrow(data_train)

    # Refit model and get predictions using refit.predictions()
    # This handles workflows, stats::update, and call re-evaluation
    pred_aug <- tryCatch(
        {
            # Create a minimal predictions object to use refit() on
            # We need the mfx attribute for refit() to work
            dummy_pred <- data.frame(estimate = 1)
            class(dummy_pred) <- c("predictions", "data.frame")
            attr(dummy_pred, "marginaleffects") <- mfx

            # Use refit() to refit model and make predictions
            refit.predictions(dummy_pred, data = augmented_data, newdata = augmented_data)
        },
        error = function(e) {
            if (getOption("marginaleffects_conformal_debug", FALSE)) {
                warning("Refit/prediction failed: ", conditionMessage(e))
            }
            NULL
        })

    if (is.null(pred_aug) || !inherits(pred_aug, "predictions")) {
        return(NA_real_)
    }

    # Compute absolute residuals for all n+1 points
    residuals <- abs(augmented_data[[response_name]] - pred_aug$estimate)

    # Get residual for new point (last observation)
    resid_new <- residuals[n + 1]

    # Compute quantile of training residuals (excluding new point)
    quantile_val <- stats::quantile(residuals[-length(residuals)], probs = conf_level)

    # Return difference for root finding
    # When this crosses zero, we've found the boundary
    return(resid_new - unname(quantile_val))
}


conformal_full <- function(x,
                           data_test = NULL,
                           data_train = NULL,
                           conf_level = 0.95,
                           var_multiplier = 10,
                           max_iter = 100,
                           tolerance = .Machine$double.eps^0.25,
                           mfx = NULL,
                           ...) {
    dots <- list(...)
    if (is.null(data_test) && "test" %in% names(dots)) {
        data_test <- dots$test
    }

    if (is.null(mfx)) {
        mfx <- attr(x, "marginaleffects")
    }

    checkmate::assert_data_frame(data_test, null.ok = FALSE)
    checkmate::assert_data_frame(data_train, null.ok = FALSE)

    # Get model
    model <- mfx@model
    response_name <- insight::find_response(model)

    # Check that training data has predictors (more than just response)
    if (ncol(data_train) <= 1) {
        msg <- "Full conformal inference requires the original training data with predictors. For tidymodels workflows, pass the training data explicitly: predictions(mod, newdata = data_test) |> inferences(method = 'conformal_full', data_train = data_train)"
        stop_sprintf(msg)
    }

    # Check response is numeric
    if (!is.numeric(data_train[[response_name]])) {
        stop_sprintf("Full conformal inference requires a numeric response variable.")
    }

    # Get initial predictions and training residuals (for bounds estimation)
    pred_test <- predictions(model, newdata = data_test, vcov = FALSE)
    pred_train <- predictions(model, newdata = data_train, vcov = FALSE)
    train_residuals <- data_train[[response_name]] - pred_train$estimate
    trial_half_width <- var_multiplier * max(abs(train_residuals))

    # Function to compute bounds for a single test observation
    compute_bounds_single <- function(i) {
        x_test <- data_test[i, , drop = FALSE]
        data.table::setDF(x_test) # Convert to data.frame for easier handling
        # Remove marginaleffects-added columns like rowid
        keep_cols <- !colnames(x_test) %in% c("rowid", "rowidcf")
        if (any(keep_cols)) {
            x_test <- x_test[, keep_cols, drop = FALSE]
        }
        pred_value <- pred_test$estimate[i]

        # Get trial bounds
        bounds <- c(pred_value - trial_half_width, pred_value + trial_half_width)

        # Find upper bound using uniroot
        upper <- tryCatch(
            stats::uniroot(
                conformity_test_full,
                c(pred_value, bounds[2]),
                maxiter = max_iter,
                tol = tolerance,
                extendInt = "upX",
                x_test = x_test,
                model = model,
                data_train = data_train,
                response_name = response_name,
                conf_level = conf_level,
                mfx = mfx
            ),
            error = function(e) {
                insight::format_warning(sprintf(
                    "Failed to find upper bound for observation %d: %s", i, conditionMessage(e)
                ))
                NULL
            })

        # Find lower bound using uniroot
        lower <- tryCatch(
            stats::uniroot(
                conformity_test_full,
                c(bounds[1], pred_value),
                maxiter = max_iter,
                tol = tolerance,
                extendInt = "downX",
                x_test = x_test,
                model = model,
                data_train = data_train,
                response_name = response_name,
                conf_level = conf_level,
                mfx = mfx
            ),
            error = function(e) {
                insight::format_warning(sprintf(
                    "Failed to find lower bound for observation %d: %s", i, conditionMessage(e)
                ))
                NULL
            })

        pred_low <- if (!is.null(lower)) lower$root else NA_real_
        pred_high <- if (!is.null(upper)) upper$root else NA_real_

        return(c(pred.low = pred_low, pred.high = pred_high))
    }

    # Compute bounds for all test observations
    if (isTRUE(getOption("marginaleffects_parallel_inferences", default = FALSE))) {
        # Use future for parallelization if available
        insight::check_if_installed("future.apply")
        pkg <- getOption("marginaleffects_parallel_packages", default = NULL)
        pkg <- unique(c("marginaleffects", pkg))
        results <- future.apply::future_lapply(
            seq_len(nrow(data_test)),
            compute_bounds_single,
            future.seed = TRUE,
            future.packages = pkg)
    } else {
        results <- lapply(seq_len(nrow(data_test)), compute_bounds_single)
    }

    # Combine results
    bounds_df <- do.call(rbind, lapply(results, function(x) {
        data.frame(pred.low = x["pred.low"], pred.high = x["pred.high"])
    }))

    # Add bounds to output
    pred_test$pred.low <- bounds_df$pred.low
    pred_test$pred.high <- bounds_df$pred.high

    return(pred_test)
}

sanitize_cross <- function(cross, variables, model) {
    # cross: flip NULL to TRUE if there are interactions in the formula
    # and FALSE otherwise

    checkmate::assert_flag(cross, null.ok = FALSE)

    if (isTRUE(cross) && is.null(variables)) {
        msg <- "When `cross = TRUE`, you must use the `variables` argument to specify which variables should be interacted."
        insight::format_error(msg)
    }

    if (isTRUE(checkmate::check_flag(cross))) {
        return(cross)
    }

    inter <- try(insight::find_interactions(model, flatten = TRUE), silent = TRUE)

    # variables is length 1 means we don't want an interaction
    if (length(variables) > 1 && isTRUE(length(inter) > 0)) {
        return(TRUE)
    } else {
        return(FALSE)
    }

}

hypothesis_matrix <- function(x, hypothesis) {
    if (isTRUE(checkmate::check_atomic_vector(hypothesis))) {
        hypothesis <- matrix(hypothesis, ncol = 1)
    }
    checkmate::assert_matrix(hypothesis, nrows = nrow(x))

    if (is.null(colnames(hypothesis))) {
        colnames(hypothesis) <- rep("custom", ncol(hypothesis))
    }

    # bayesian
    draws <- attr(x, "posterior_draws")
    if (!is.null(draws)) {
        draws <- t(as.matrix(hypothesis)) %*% draws
        out <- data.table(
            term = colnames(hypothesis),
            tmp = apply(draws, 1, stats::median))
        setnames(out, old = "tmp", new = "estimate")
        attr(out, "posterior_draws") <- draws

        # frequentist
    } else {
        out <- data.table(
            term = colnames(hypothesis),
            tmp = as.vector(x[["estimate"]] %*% hypothesis))
        setnames(out, old = "tmp", new = "estimate")
    }

    out <- out[out$term != "1 - 1", , drop = FALSE]
    return(out)
}

#' @rdname get_predict
#' @export
get_predict.glimML <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    ...) {
    insight::check_if_installed("aod")

    out <- aod::predict(model, newdata = newdata, type = type, ...)
    out <- data.table(estimate = out)
    out <- add_rowid(out, newdata)

    return(out)
}


#' @rdname set_coef
#' @export
set_coef.glimML <- function(model, coefs, ...) {
    # in basic model classes coefficients are named vector
    model@fixed.param[names(coefs)] <- coefs
    model
}


#' @rdname get_vcov
#' @export
get_vcov.glimML <- function(model, vcov = NULL, ...) {
    insight::check_if_installed("aod")
    if (!is.null(vcov) && !is.logical(vcov)) {
        stop("The `vcov` argument is not supported for this kind of model.")
    }
    vcov <- sanitize_vcov(model, vcov)
    aod::vcov(model)
}


#' @rdname sanitize_model_specific
sanitize_model_specific.glimML <- function(model, ...) {
    mdat <- get_modeldata(model)
    cl <- detect_variable_class(mdat, model = model)
    if (isTRUE("character" %in% cl)) {
        stop_sprintf(
            "This function does not support character predictors. Please convert them to factors before fitting the model."
        )
    }
    return(model)
}

inferences_simulation <- function(x, R = 1000, conf_level = 0.95, conf_type = "perc", mfx = NULL, ...) {
    insight::check_if_installed("mvtnorm")

    checkmate::assert_choice(
        conf_type,
        choices = c(
            "perc",
            "wald"
        )
    )

    out <- x
    model <- mfx@model
    call_mfx <- mfx@call
    call_mfx[["vcov"]] <- FALSE
    # avoid calling get_modeldata() repeatedly via ...
    call_mfx[["modeldata"]] <- mfx@modeldata

    B <- get_coef(model)

    # respect robust vcov from the first call
    V <- mfx@vcov_model
    if (!isTRUE(checkmate::check_matrix(V))) {
        V <- get_vcov(model)
    }

    # Draw R sets of coefficients from multivariate normal
    coefmat <- mvtnorm::rmvnorm(R, B, V)

    # avoid recursion
    args <- list(...)
    args[["vcov"]] <- FALSE

    # Function to compute estimates with simulated coefficients
    inner_fun <- function(i = NULL) {
        mod_tmp <- set_coef(model, coefmat[i, ])
        call_mfx[["model"]] <- mod_tmp
        boot_mfx <- eval.parent(call_mfx)
        return(boot_mfx$estimate)
    }

    # Compute estimates for each set of coefficients
    # do not use simulation mean as point estimate
    # https://doi.org/10.1017/psrm.2023.8
    draws <- lapply(seq_len(nrow(coefmat)), inner_fun)
    draws <- do.call("cbind", draws)

    # Compute confidence intervals
    out$std.error <- apply(draws, 1, stats::sd)
    alpha <- 1 - conf_level

    if (conf_type == "perc") {
        out$conf.low <- apply(draws, 1, stats::quantile, probs = alpha / 2, names = FALSE)
        out$conf.high <- apply(draws, 1, stats::quantile, probs = 1 - alpha / 2, names = FALSE)

        cols <- setdiff(names(out), c("p.value", "std.error", "statistic", "s.value", "df"))
    } else if (conf_type == "wald") {
        out$statistic <- out$estimate / out$std.error

        critical <- abs(stats::qnorm(alpha / 2))

        out$conf.low <- out$estimate - critical * out$std.error
        out$conf.high <- out$estimate + critical * out$std.error

        out$p.value <- 2 * stats::pnorm(-abs(out$statistic))
        out$s.value <- -log2(out$p.value)

        cols <- setdiff(names(out), "df")
    }

    # Drop unnecessary columns
    out <- out[, cols, drop = FALSE]

    mfx@draws <- draws
    attr(out, "marginaleffects") <- mfx

    return(out)
}

#' @importFrom generics components
#' @export
generics::components


#' Extract components from marginaleffects objects
#'
#' @param object A marginaleffects object (predictions, comparisons, slopes, or hypotheses)
#' @param component Character string specifying which component to extract. Must be a valid
#'   slot name from the internal S4 object. If `NULL` (the default), `components()` prints a message with
#'   all available component names. Common components include: "model", "newdata",
#'   "modeldata", "call", "jacobian", "vcov_model", "type", "by", "comparison", "variables", etc.
#' @param ... Ignored.
#' @return The requested component from the mfx object
#' @details This function provides access to the internal components stored in the `mfx`
#'   attribute of marginaleffects objects. The `mfx` attribute contains an S4 object of
#'   class "marginaleffects_internal" with various slots containing model information,
#'   data, and computational details used by the marginaleffects functions.
#'
#' Warning: the internal slot names are not considered part of the public API and may change
#' without warning in future versions of the marginaleffects package.
#' @export
components.marginaleffects <- function(object, component = NULL, ...) {
    mfx <- attr(object, "marginaleffects", exact = TRUE)

    if (is.null(mfx)) {
        stop("No mfx attribute found in object")
    }

    if (identical(component, "all")) {
        return(mfx)
    }

    # Get all slot names from the S4 object
    valid_components <- sort(methods::slotNames(mfx))

    if (is.null(component)) {
        msg <- checkmate::check_choice(component, choices = valid_components)
        msg <- sub(".*\\{", "The `component` argument must be one of {'all',", msg)
        message(msg)
        return(invisible(NULL))
    }

    checkmate::assert_choice(component, choices = valid_components)

    return(methods::slot(mfx, component))
}

#' @export
components.predictions <- components.marginaleffects

#' @export
components.hypotheses <- components.marginaleffects

#' @export
components.slopes <- components.marginaleffects

#' @export
components.comparisons <- components.marginaleffects

#' @rdname get_coef
#' @export
get_coef.systemfit <- function(model, ...) {
    out <- stats::coef(model)
    return(out)
}

#' @rdname get_vcov
#' @export
get_vcov.systemfit <- function(model, ...) {
    vcov <- sanitize_vcov(model, vcov)
    out <- stats::vcov(model)
    return(out)
}

#' @rdname get_vcov
#' @export
get_predict.systemfit <- function(model, newdata = NULL, type = NULL, ...) {
    out <- stats::predict(model, newdata = newdata, ...)
    colnames(out) <- sub("\\.pred$", "", colnames(out))
    out <- melt(data.table(out), variable.name = "group", value.name = "estimate")
    data.table::setDF(out)
    out <- add_rowid(out, newdata)
    return(out)
}


#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.systemfit <- function(model, coefs, ...) {
    out <- model
    out$coefficients <- coefs
    for (i in seq_along(model$eq)) {
        eq <- model$eq[[i]]
        b <- coefs[grepl(paste0("^eq", i, "_"), names(coefs))]
        out$eq[[i]]$coefficients <- b
    }
    return(out)
}

#' @include get_predict.R
#' @rdname get_predict
#' @keywords internal
#' @export
get_predict.glmmTMB <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    mfx = NULL,
    newparams = NULL,
    ...) {
    # hack to avoid re-optimization
    # see https://github.com/vincentarelbundock/marginaleffects/issues/1064
    b_vec <- model$obj$env$parList()$b
    if (length(b_vec) > 0) {
        model$modelInfo$map$b <- factor(rep(NA, length(b_vec)))
    }

    np <- model$fit$par
    if (!is.null(newparams)) {
        np[seq_along(newparams)] <- newparams
    }

    out <- get_predict.default(
        model = model,
        newdata = newdata,
        type = type,
        allow.new.levels = TRUE, # otherwise we get errors in marginal_means()
        newparams = np,
        ...
    )
    return(out)
}


#' @include get_vcov.R
#' @rdname get_vcov
#' @export
get_vcov.glmmTMB <- function(model, vcov, ...) {
    vcov <- sanitize_vcov(model, vcov)

    # Extract the full covariance matrix
    out <- insight::get_varcov(model, vcov = vcov, component = "full")

    # Extract the fixed-effect coefficient names from get_coef
    coef_names <- names(get_coef.glmmTMB(model))

    # Handle dispersion and conditional terms
    cleaned_coef_names <- gsub("^cond~", "", coef_names) # Remove cond~ for conditional terms
    cleaned_coef_names <- gsub("^disp~", "d~", cleaned_coef_names) # Map disp~ to d~ for dispersion terms

    # The 'upper cutoff' and 'lower cutoff' will remain in both, so no removal

    # Get the current row and column names from the covariance matrix
    current_names <- rownames(out)

    # Match cleaned coef_names with current names in the covariance matrix
    matched_indices <- match(current_names, cleaned_coef_names)

    # Replace row/column names only where there is a valid match
    valid_indices <- which(!is.na(matched_indices))

    if (length(valid_indices) > 0) {
        # Apply the correct names from coef_names to matched rows/columns in the covariance matrix
        rownames(out)[valid_indices] <- coef_names[matched_indices[valid_indices]]
        colnames(out)[valid_indices] <- coef_names[matched_indices[valid_indices]]
    } else {
        warning(
            "No matching terms found between the covariance matrix and fixed-effect coefficients."
        )
    }

    return(out)
}


#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.glmmTMB <- function(model, ...) {
    # Extract the fixed-effect coefficients
    out <- unlist(glmmTMB::fixef(model))

    # Apply the gsub logic to rename terms (cond~, disp~, etc.)
    names(out) <- gsub("^(cond|zi|disp)\\.", "\\1~", names(out))

    # No removal of "lower cutoff" and "upper cutoff" - they remain in place
    return(out)
}


#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.glmmTMB <- function(model, coefs, ...) {
    # use predict(`newparams`) for this kind of model
    return(model)
}

#' @rdname sanitize_model_specific
sanitize_model_specific.glmmTMB <- function(model, vcov = TRUE, re.form, ...) {
    if (identical(vcov, "HC0")) {
        insight::check_if_installed("glmmTMB", minimum_version = "1.1.12")
    }

    # re.form=NA
    if (!isTRUE(checkmate::check_flag(vcov)) && !identical(vcov, "HC0")) {
        msg <- 'For this model type, `vcov` must be `TRUE`, `FALSE`, `"HC0"`.'
        stop_sprintf(msg)
    }

    if (!isFALSE(vcov) && (missing(re.form) || (!isTRUE(is.na(re.form))))) {
        msg <- "For this model type, `marginaleffects` only takes into account the uncertainty in fixed-effect parameters. This is often appropriate when `re.form=NA`, but may be surprising to users who set `re.form=NULL` (default) or to some other value. Call `options(marginaleffects_safe = FALSE)` to silence this warning."
        warn_sprintf(msg)
    }

    return(model)
}

conformal_cv_plus <- function(x,
                              data_train,
                              data_test,
                              R,
                              score,
                              conf_level,
                              mfx = NULL,
                              ...) {
    # cross-validation
    idx <- sample.int(nrow(data_train), nrow(data_train))
    idx <- split(idx, ceiling(seq_along(idx) / (length(idx) / R)))
    scores <- NULL
    for (i in idx) {
        data_cv <- data_train[-i, ]
        # re-fit the original model on training sets withholding the CV fold
        model_cv <- tryCatch(stats::update(mfx@model, data = data_cv),
            error = function(e) NULL)
        if (is.null(model_cv)) {
            if (is.call(mfx@call_model) && "data" %in% names(mfx@call_model)) {
                # if the model call has a data argument, we can update it
                mfx@call_model$data <- data_cv
                model_cv <- eval(mfx@call_model)
            } else {
                stop_sprintf("Failed to re-fit the model on the cross-validation set.")
            }
        }
        # use the updated model to make out-of-fold predictions
        # call_cv is the `predictions()` call, which we re-evaluate in-fold: newdata=train[i,]
        call_cv <- mfx@call
        call_cv[["model"]] <- model_cv
        call_cv[["newdata"]] <- data_train[i, ]
        call_cv[["vcov"]] <- FALSE # faster
        pred_cv <- eval(call_cv)
        # save the scores form each fold
        scores <- c(scores, get_conformal_score(pred_cv, score = score, mfx = mfx))
    }

    # test
    out <- refit(x, newdata = data_test)

    # bounds
    out <- get_conformal_bounds(out, score = scores, conf_level = conf_level, mfx = mfx)

    return(out)
}

unpack_matrix_cols <- function(x) {
    data.table::setDF(x)

    # what data types are we working with
    dcs <- sapply(x, function(x) class(x)[1])

    # do something if we have matrix columns
    if (any(mcols <- dcs == "matrix")) {
        # assume that any matrix columns are of the same number of cols
        nc <- ncol(x[mcols][[1L]])
        out <- lapply(x[mcols], as.vector)
        other <- lapply(x[!mcols], function(y, nc) rep(y, times = nc), nc = nc)
        out <- as.data.frame(out)
        other <- as.data.frame(other)
        out <- cbind(out, other)
        # put order back as it was
        out <- out[names(x)]
    } else {
        return(x)
    }

    data.table::setDT(out)

    out
}


unpack_matrix_1col <- function(x) {
    drop_dr <- vapply(x, inherits, logical(1), "DirichletRegData")
    if (any(drop_dr)) {
        x <- x[!drop_dr]
    }
    for (i in seq_along(x)) {
        if (inherits(x[[i]], "matrix") && ncol(x[[i]]) == 1) {
            x[[i]] <- drop(x[[i]])
        }
    }
    return(x)
}

# The content of this file was adapted from the `gratia` package
# https://github.com/gavinsimpson/gratia

# The MIT License (MIT)
#
# Copyright (c) 2013-2020 Gavin L. Simpson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

add_numderiv <- function(mfx, numderiv) {
    checkmate::assert(
        checkmate::check_choice(numderiv, c("richardson", "fdforward", "fdcenter")),
        checkmate::check_list(numderiv, min.len = 1)
    )

    if (isTRUE(checkmate::check_string(numderiv))) {
        numderiv <- list(numderiv)
    }

    if (length(numderiv) > 1) {
        if (numderiv[[1L]] %in% c("fdforward", "fdcenter")) {
            if (!all(names(numderiv)[-1L] %in% "eps")) {
                stop(
                    "The only valid argument for this numeric differentiation method is `eps`.",
                    call. = FALSE
                )
            }
        } else if (numderiv[[1L]] == "richardson") {
            valid <- c("eps", "d", "zero_tol", "size", "r", "v")
            if (!all(names(numderiv)[-1L] %in% valid)) {
                stop(
                    sprintf(
                        "The only valid arguments for this numeric differentiation method are: %s. See `?numDeriv::grad` for details.",
                        toString(valid)
                    ),
                    call. = FALSE
                )
            }
        }
    }

    mfx@numderiv <- numderiv
    return(mfx)
}

#' @rdname get_predict
#' @export
get_predict.tobit1 <- function(
    model,
    newdata = insight::get_data(model),
    type = "response",
    ...) {
    out <- stats::predict(model, what = type, newdata = newdata)
    out <- data.table(estimate = out)
    out <- add_rowid(out, newdata)
    return(out)
}

joint_test <- function(
    object,
    joint_index = NULL,
    hypothesis = 0,
    joint_test = "f",
    df = NULL,
    vcov = TRUE) {
    checkmate::assert_choice(joint_test, c("f", "chisq"))

    # do not use components() because this may be a model object
    mfx <- attr(object, "marginaleffects")

    # Create mfx object if it doesn't exist (needed for joint tests on model objects)
    if (is.null(mfx) && !inherits(object, c("slopes", "comparisons", "predictions", "hypotheses"))) {
        call <- construct_call(object, "hypotheses")
        object_sanitized <- sanitize_model(object, call = call, vcov = vcov)
        mfx <- new_marginaleffects_internal(
            call = call,
            model = object_sanitized
        )
        mfx@conf_level <- sanitize_conf_level(NULL)
    }

    if (joint_test == "f") {
        checkmate::assert_numeric(df, len = 2, null.ok = TRUE)
    } else {
        checkmate::assert_numeric(df, len = 1, null.ok = TRUE)
    }

    # theta_hat: P x 1 vector of estimated parameters
    if (inherits(object, c("slopes", "comparisons"))) {
        nam <- object$term
        if ("contrast" %in% names(object)) {
            nam <- paste(nam, object$contrast)
        }
        theta_hat <- stats::setNames(object$estimate, nam)
    } else {
        theta_hat <- get_coef(object)
    }

    # index
    checkmate::assert(
        checkmate::check_integerish(
            joint_index,
            lower = 1,
            upper = length(theta_hat)
        ),
        checkmate::check_character(joint_index),
        checkmate::check_true(joint_index)
    )

    if (isTRUE(joint_index)) {
        joint_index <- seq_along(theta_hat)
    } else if (isTRUE(checkmate::check_string(joint_index))) {
        joint_index <- grep(joint_index, names(theta_hat), perl = TRUE)
    }

    # V_hat: estimated covariance matrix
    V_hat <- get_vcov(object, vcov = vcov)

    # R: Q x P matrix for testing Q hypotheses on P parameters
    # build R matrix based on joint_index
    R <- matrix(0, nrow = length(joint_index), ncol = length(theta_hat))
    for (i in seq_along(joint_index)) {
        if (is.numeric(joint_index)) {
            R[i, joint_index[i]] <- 1
        } else {
            R[i, which(names(theta_hat) == joint_index[i])] <- 1
        }
    }

    # null hypothesis
    checkmate::assert(
        checkmate::check_number(hypothesis),
        checkmate::check_numeric(hypothesis, len = nrow(R)),
        checkmate::check_null(hypothesis)
    )
    if (is.null(hypothesis)) {
        hypothesis <- 0
    }
    r <- matrix(hypothesis, nrow = nrow(R), ncol = 1)

    # Calculate the difference between R*theta_hat and r
    diff <- R %*% theta_hat - r

    # Calculate the inverse of R*(V_hat/n)*R'
    inv <- solve(R %*% V_hat %*% t(R))

    # Calculate the Wald test statistic
    if (joint_test == "f") {
        wald_statistic <- t(diff) %*% inv %*% diff / dim(R)[1] # Q is the number of rows in R
    } else if (joint_test == "chisq") {
        wald_statistic <- t(diff) %*% inv %*% diff # Not normalized for chi-squared joint_test
    }

    # Degrees of freedom
    if (is.null(df)) {
        df1 <- dim(R)[1] # Q

        if (joint_test == "f") {
            # Check for lme models and warn about df heuristics
            if (inherits(object, "lme") || (!is.null(mfx) && inherits(mfx@model, "lme"))) {
                model_class <- if (inherits(object, "lme")) class(object)[1] else class(mfx@model)[1]
                msg <- "The `hypotheses()` functions uses simple heuristics to select degrees of freedom for this test. See the relevant section in `?hypotheses`. These rules are likely to yield inappropriate results for models of class `%s`. Please supply degrees of freedom values explicitly via the `df` argument."
                warn_sprintf(msg, model_class)
            }

            df2 <- NULL
            if (!is.null(mfx)) {
                df2 <- tryCatch(
                    insight::get_df(mfx@model),
                    error = function(e) NULL
                )
            }
            if (is.null(df2)) {
                df2 <- tryCatch(insight::get_df(object), error = function(e) NULL)
            }
            if (is.null(df2)) {
                # n: sample size
                n <- tryCatch(stats::nobs(object), error = function(e) NULL)
                if (is.null(n) && !is.null(mfx)) {
                    n <- tryCatch(
                        stats::nobs(mfx@model),
                        error = function(e) NULL
                    )
                }
                if (is.null(n)) {
                    stop_sprintf(
                        "Could not extract sample size from model object."
                    )
                }

                df2 <- n - length(theta_hat) # n - P
            }
        }
    } else {
        df1 <- df[1]

        if (joint_test == "f") {
            df2 <- df[2]
        }
    }

    # Calculate the p-value
    if (joint_test == "f") {
        p_value <- 1 - stats::pf(wald_statistic, df1, df2)
    } else if (joint_test == "chisq") {
        p_value <- 1 - stats::pchisq(wald_statistic, df1)
        df2 <- NULL
    }

    # Return the Wald joint_test statistic and p-value
    out <- data.frame(statistic = drop(wald_statistic), p.value = drop(p_value))
    class(out) <- c("hypotheses", "data.frame")
    if (joint_test == "f") {
        attr(out, "hypotheses_joint_label") <- "F"
    } else if (joint_test == "chisq") {
        attr(out, "hypotheses_joint_label") <- "ChiSq"
    }

    # degrees of freedom print
    if (joint_test == "f") {
        out$df1 <- df1
        out$df2 <- df2
    } else {
        out$df <- df1
    }

    # Create the print_head string
    print_head <- "\nJoint hypothesis test:\n"
    if (is.character(joint_index)) {
        for (i in joint_index) {
            print_head <- paste0(print_head, i, sprintf(" = %s\n", hypothesis))
        }
    } else if (
        inherits(
            object,
            c("marginaleffects", "comparisons", "slopes", "marginal_means")
        )
    ) {
        tmp <- paste0(
            get_labels(object, idx = joint_index),
            sprintf(" = %s\n", hypothesis)
        )
        print_head <- c(print_head, tmp)
    } else {
        tmp <- paste0(
            get_labels(stats::coef(object), idx = joint_index),
            sprintf(" = %s\n", hypothesis)
        )
        print_head <- c(print_head, tmp)
    }
    attr(out, "print_head") <- print_head

    # Add marginaleffects attribute for printing
    out <- add_attributes(out, mfx)

    # class before prune
    out <- prune_attributes(out)

    return(out)
}

#' @include set_coef.R
#' @rdname set_coef
#' @export
set_coef.mlm <- function(model, coefs, ...) {
    model$coefficients[] <- coefs

    return(model)
}

#' @include get_coef.R
#' @rdname get_coef
#' @export
get_coef.mlm <- function(model, ...) {
    out <- insight::get_parameters(model, ...)
    out <- stats::setNames(
        out$Estimate,
        sprintf("%s:%s", out$Response, out$Parameter)
    )
    return(out)
}


#' @include get_group_names.R
#' @rdname get_group_names
#' @export
get_group_names.mlm <- function(model, ...) {
    resp <- insight::get_response(model)
    return(names(resp))
}

plot_preprocess <- function(
    dat,
    v_x,
    v_color = NULL,
    v_facet_1 = NULL,
    v_facet_2 = NULL,
    condition = NULL,
    mfx = NULL
) {
    for (v in names(condition$condition)) {
        fun <- function(x, lab) {
            idx <- match(x, sort(unique(x)))
            factor(lab[idx], levels = lab)
        }
        if (identical(condition$condition[[v]], "threenum")) {
            dat[[v]] <- fun(dat[[v]], c("-SD", "Mean", "+SD"))
        } else if (identical(condition$condition[[v]], "fivenum")) {
            labs <- stats::setNames(
                sort(unique(dat[[v]])),
                format(sort(unique(dat[[v]])), digits = 2)
            )
            dat[[v]] <- fun(dat[[v]], names(labs))
        } else if (identical(condition$condition[[v]], "minmax")) {
            dat[[v]] <- fun(dat[[v]], c("Min", "Max"))
        } else if (identical(condition$condition[[v]], "quartile")) {
            dat[[v]] <- fun(dat[[v]], c("Q1", "Q2", "Q3"))
        }
    }
    if (check_variable_class(mfx, v_x, "categorical")) {
        dat[[v_x]] <- factor(dat[[v_x]])
    }
    # colors, linetypes, and facets are categorical attributes
    if (isTRUE(v_color %in% colnames(dat))) {
        dat[[v_color]] <- factor(dat[[v_color]])
    }
    if (isTRUE(v_facet_1 %in% colnames(dat))) {
        dat[[v_facet_1]] <- factor(dat[[v_facet_1]])
    }
    if (isTRUE(v_facet_2 %in% colnames(dat))) {
        dat[[v_facet_2]] <- factor(dat[[v_facet_2]])
    }

    if (!is.null(mfx) && !is.null(mfx@modeldata)) {
        vars <- c(v_x, v_color, v_facet_1, v_facet_2)
        for (v in vars) {
            if (isTRUE(v %in% colnames(dat)) && isTRUE(v %in% colnames(mfx@modeldata))) {
                attr(dat[[v]], "label") <- attr(mfx@modeldata[[v]], "label")
            }
        }
    }
    return(dat)
}


plot_build <- function(
    dat,
    v_x,
    v_color = NULL,
    v_facet_1 = NULL,
    v_facet_2 = NULL,
    dv = NULL,
    points = 0,
    rug = FALSE,
    gray = FALSE,
    mfx = NULL
) {
    checkmate::assert_flag(rug)
    checkmate::assert_flag(gray)

    is_discrete <- function(df, var, mfx = NULL) {
        if (is.null(var) || !isTRUE(var %in% colnames(df))) {
            return(FALSE)
        }
        out <- is.factor(df[[var]]) ||
            is.character(df[[var]]) ||
            is.logical(df[[var]])
        if (!is.null(mfx)) {
            out <- out || check_variable_class(mfx, var, "categorical")
        }
        return(out)
    }

    # create index before building ggplot to make sure it is available
    dat$marginaleffects_term_index <- get_unique_index(dat, mfx)
    multi_variables <- isTRUE(length(unique(dat$marginaleffects_term_index)) > 1)
    x_is_discrete <- is_discrete(dat, v_x, mfx)

    # Handle case where dv is a vector (e.g., binomial models with trials())
    # Use the first element for plotting the raw data points
    dv_plot <- if (length(dv) > 0) dv[1] else dv

    p <- ggplot2::ggplot(data = dat)

    if (
        points > 0 &&
            !x_is_discrete &&
            !check_variable_class(mfx, dv, "categorical")
    ) {
        dat_points <- mfx@modeldata
        if (!is.null(dat_points)) {
            if (is_discrete(dat_points, v_x, mfx)) {
                dat_points[[v_x]] <- factor(dat_points[[v_x]])
            }
            if (is_discrete(dat_points, v_color, mfx)) {
                dat_points[[v_color]] <- factor(dat_points[[v_color]])
            }
        }
        if (is_discrete(dat_points, v_color, mfx)) {
            if (isTRUE(gray)) {
                p <- p +
                    ggplot2::geom_point(
                        data = dat_points,
                        alpha = points,
                        ggplot2::aes(
                            x = .data[[v_x]],
                            y = .data[[dv_plot]],
                            shape = .data[[v_color]]
                        )
                    )
            } else {
                p <- p +
                    ggplot2::geom_point(
                        data = dat_points,
                        alpha = points,
                        ggplot2::aes(
                            x = .data[[v_x]],
                            y = .data[[dv_plot]],
                            color = .data[[v_color]]
                        )
                    )
            }
        } else {
            p <- p +
                ggplot2::geom_point(
                    data = dat_points,
                    alpha = points,
                    ggplot2::aes(x = .data[[v_x]], y = .data[[dv_plot]])
                )
        }
    }

    if (isTRUE(rug)) {
        p <- p + ggplot2::geom_rug(data = mfx@modeldata, ggplot2::aes(x = .data[[v_x]]))
        if (!is.null(dv_plot)) {
            p <- p +
                ggplot2::geom_rug(data = mfx@modeldata, ggplot2::aes(y = .data[[dv_plot]]))
        }
    }

    aes_args <- list(
        x = substitute(.data[[v_x]]),
        y = substitute(estimate)
    )

    if ("conf.low" %in% colnames(dat)) {
        aes_args$ymin <- substitute(conf.low)
        aes_args$ymax <- substitute(conf.high)
    }

    aes_args_ribbon <- aes_args
    aes_args_ribbon$fill <- aes_args$color
    aes_args_ribbon$color <- NULL

    # discrete x-axis
    if (x_is_discrete) {
        if (!is.null(v_color)) {
            if (gray) {
                aes_args$shape <- substitute(.data[[v_color]])
            } else {
                aes_args$color <- substitute(.data[[v_color]])
            }
        }
        aes_obj <- do.call(ggplot2::aes, aes_args)
        if ("conf.low" %in% colnames(dat)) {
            p <- p +
                ggplot2::geom_pointrange(
                    mapping = aes_obj,
                    position = ggplot2::position_dodge(0.15)
                )
        } else {
            p <- p +
                ggplot2::geom_point(
                    mapping = aes_obj,
                    position = ggplot2::position_dodge(0.15)
                )
        }

        # continuous x-axis
    } else {
        if (!is.null(v_color)) {
            if (gray) {
                aes_args$linetype <- substitute(.data[[v_color]])
                aes_args_ribbon$linetype <- substitute(.data[[v_color]])
            } else {
                aes_args$color <- substitute(.data[[v_color]])
                aes_args_ribbon$fill <- substitute(.data[[v_color]])
            }
        }
        aes_obj_ribbon <- do.call(ggplot2::aes, aes_args_ribbon)
        aes_args$ymin <- aes_args$ymax <- NULL
        aes_obj <- do.call(ggplot2::aes, aes_args)
        if ("conf.low" %in% colnames(dat)) {
            p <- p + ggplot2::geom_ribbon(aes_obj_ribbon, alpha = 0.1)
            p <- p + ggplot2::geom_line(aes_obj)
        }
        p <- p + ggplot2::geom_line(aes_obj)
    }

    # facets: 3rd and 4th variable and/or multiple effects
    ## If pass two facets then make facet grid
    if (!is.null(v_facet_1) && !is.null(v_facet_2)) {
        fo <- stats::as.formula(paste(
            v_facet_2,
            "~",
            ifelse(multi_variables, "marginaleffects_term_index +", ""),
            v_facet_1
        ))
        p <- p +
            ggplot2::facet_grid(fo, scales = "free", labeller = function(x) {
                lapply(
                    ggplot2::label_both(x),
                    gsub,
                    pattern = "marginaleffects_term_index: ",
                    replacement = ""
                )
            })
        ## if pass only 1 facet then facet_wrap
    } else if (!is.null(v_facet_1) && is.null(v_facet_2)) {
        fo <- stats::as.formula(paste(
            "~",
            ifelse(multi_variables, "marginaleffects_term_index +", ""),
            v_facet_1
        ))
        p <- p + ggplot2::facet_wrap(fo, scales = "free")
    } else if (multi_variables) {
        fo <- stats::as.formula("~ marginaleffects_term_index")
        p <- p + ggplot2::facet_wrap(fo, scales = "free")
    }

    return(p)
}

run <- FALSE
pkg <- requireNamespace("tinytest", quietly = TRUE)
home <- grepl("(?i)vince", Sys.info()["nodename"]) || identical(Sys.getenv("R_NOT_CRAN"), "true")
if (pkg && home && run) {
    tinytest::test_package("marginaleffects")
}

#' Predictions
#'
#' @description
#' Outcome predicted by a fitted model on a specified scale for a given combination of values of the predictor variables, such as their observed values, their means, or factor levels (a.k.a. "reference grid").
#'
#' * `predictions()`: unit-level (conditional) estimates.
#' * `avg_predictions()`: average (marginal) estimates.
#'
#' The `newdata` argument and the `datagrid()` function can be used to control where statistics are evaluated in the predictor space: "at observed values", "at the mean", "at representative values", etc.
#'
#' See the predictions vignette and package website for worked examples and case studies:

#' * <https://marginaleffects.com/chapters/predictions.html>
#' * <https://marginaleffects.com/>
#'
#' @rdname predictions
#' @param model Model object
#' @param variables Counterfactual variables.
#' * Output:
#'   - `predictions()`: The entire dataset is replicated once for each unique combination of `variables`, and predictions are made.
#'   - `avg_predictions()`: The entire dataset is replicated, predictions are made, and they are marginalized by `variables` categories.
#'   - Warning: This can be expensive in large datasets.
#'   - Warning: Users who need "conditional" predictions should use the `newdata` argument instead of `variables`.
#' * Input:
#'   - `NULL`: computes one prediction per row of `newdata`
#'   - Character vector: the dataset is replicated once of every combination of unique values of the variables identified in `variables`.
#'   - Named list: names identify the subset of variables of interest and their values. For numeric variables, the `variables` argument supports functions and string shortcuts:
#'     + A function which returns a numeric value
#'     + Numeric vector: Contrast between the 2nd element and the 1st element of the `x` vector.
#'     + "iqr": Contrast across the interquartile range of the regressor.
#'     + "sd": Contrast across one standard deviation around the regressor mean.
#'     + "2sd": Contrast across two standard deviations around the regressor mean.
#'     + "minmax": Contrast between the maximum and the minimum values of the regressor.
#'     + "threenum": mean and 1 standard deviation on both sides
#'     + "fivenum": Tukey's five numbers
#' @param newdata Grid of predictor values at which we evaluate predictions.
#' + Warning: Please avoid modifying your dataset between fitting the model and calling a `marginaleffects` function. This can sometimes lead to unexpected results.
#' + `NULL` (default): Unit-level predictions for each observed value in the dataset (empirical distribution). The dataset is retrieved using [insight::get_data()], which tries to extract data from the environment. This may produce unexpected results if the original data frame has been altered since fitting the model.
#' + string:
#'   - "mean": Predictions evaluated when each predictor is held at its mean or mode.
#'   - "median": Predictions evaluated when each predictor is held at its median or mode.
#'   - "balanced": Predictions evaluated on a balanced grid with every combination of categories and numeric variables held at their means.
#'   - "tukey": Predictions evaluated at Tukey's 5 numbers.
#'   - "grid": Predictions evaluated on a grid of representative numbers (Tukey's 5 numbers and unique values of categorical predictors).
#' + [datagrid()] call to specify a custom grid of regressors. For example:
#'   - `newdata = datagrid(cyl = c(4, 6))`: `cyl` variable equal to 4 and 6 and other regressors fixed at their means or modes.
#'   - See the Examples section and the [datagrid()] documentation.
#' + [subset()] call with a single argument to select a subset of the dataset used to fit the model, ex: `newdata = subset(treatment == 1)`
#' + [dplyr::filter()] call with a single argument to select a subset of the dataset used to fit the model, ex: `newdata = filter(treatment == 1)`
#' @param byfun A function such as `mean()` or `sum()` used to aggregate
#' estimates within the subgroups defined by the `by` argument. `NULL` uses the
#' `mean()` function. Must accept a numeric vector and return a single numeric
#' value. This is sometimes used to take the sum or mean of predicted
#' probabilities across outcome or predictor
#' levels. See examples section.
#' @param type string indicates the type (scale) of the predictions used to
#' compute contrasts or slopes. This can differ based on the model
#' type, but will typically be a string such as: "response", "link", "probs",
#' or "zero". When an unsupported string is entered, the model-specific list of
#' acceptable values is returned in an error message. When `type` is `NULL`, the
#' first entry in the error message is used by default. See the Type section in the documentation below.
#' @param transform A function applied to unit-level adjusted predictions and confidence intervals just before the function returns results. For bayesian models, this function is applied to individual draws from the posterior distribution, before computing summaries.
#'
#' @template references
#' @template deltamethod
#' @template model_specific_arguments
#' @template bayesian
#' @template equivalence
#' @template order_of_operations
#' @template parallel
#' @template options
#' @template return
#' @template type
#' @examplesIf interactive() || isTRUE(Sys.getenv("R_DOC_BUILD") == "true")
#' library("marginaleffects")
#' # Adjusted Prediction for every row of the original dataset
#' mod <- lm(mpg ~ hp + factor(cyl), data = mtcars)
#' pred <- predictions(mod)
#' head(pred)
#'
#' # Adjusted Predictions at User-Specified Values of the Regressors
#' predictions(mod, newdata = datagrid(hp = c(100, 120), cyl = 4))
#'
#' m <- lm(mpg ~ hp + drat + factor(cyl) + factor(am), data = mtcars)
#' predictions(m, newdata = datagrid(FUN_factor = unique, FUN_numeric = median))
#'
#' # Average Adjusted Predictions (AAP)
#' library(dplyr)
#' mod <- lm(mpg ~ hp * am * vs, mtcars)
#'
#' avg_predictions(mod)
#'
#' predictions(mod, by = "am")
#'
#' # Conditional Adjusted Predictions
#' plot_predictions(mod, condition = "hp")
#'
#' # Counterfactual predictions with the `variables` argument
#' # the `mtcars` dataset has 32 rows
#'
#' mod <- lm(mpg ~ hp + am, data = mtcars)
#' p <- predictions(mod)
#' head(p)
#' nrow(p)
#'
#' # average counterfactual predictions
#' avg_predictions(mod, variables = "am")
#'
#' # counterfactual predictions obtained by replicating the entire for different
#' # values of the predictors
#' p <- predictions(mod, variables = list(hp = c(90, 110)))
#' nrow(p)
#'
#'
#' # hypothesis test: is the prediction in the 1st row equal to the prediction in the 2nd row
#' mod <- lm(mpg ~ wt + drat, data = mtcars)
#'
#' predictions(
#'   mod,
#'   newdata = datagrid(wt = 2:3),
#'   hypothesis = "b1 = b2")
#'
#' # same hypothesis test using row indices
#' predictions(
#'   mod,
#'   newdata = datagrid(wt = 2:3),
#'   hypothesis = "b1 - b2 = 0")
#'
#' # same hypothesis test using numeric vector of weights
#' predictions(
#'   mod,
#'   newdata = datagrid(wt = 2:3),
#'   hypothesis = c(1, -1))
#'
#' # two custom contrasts using a matrix of weights
#' lc <- matrix(
#'   c(
#'     1, -1,
#'     2, 3),
#'   ncol = 2)
#' predictions(
#'   mod,
#'   newdata = datagrid(wt = 2:3),
#'   hypothesis = lc)
#'
#'
#' # `by` argument
#' mod <- lm(mpg ~ hp * am * vs, data = mtcars)
#' predictions(mod, by = c("am", "vs"))
#'
#' library(nnet)
#' nom <- multinom(factor(gear) ~ mpg + am * vs, data = mtcars, trace = FALSE)
#'
#' # first 5 raw predictions
#' p <- predictions(nom, type = "probs")
#' head(p)
#'
#' # average predictions
#' avg_predictions(nom, type = "probs", by = "group")
#'
#' by <- data.frame(
#'   group = c("3", "4", "5"),
#'   by = c("3,4", "3,4", "5"))
#'
#' predictions(nom, type = "probs", by = by)
#'
#' # sum of predicted probabilities for combined response levels
#' mod <- multinom(factor(cyl) ~ mpg + am, data = mtcars, trace = FALSE)
#' by <- data.frame(
#'   by = c("4,6", "4,6", "8"),
#'   group = as.character(c(4, 6, 8)))
#' predictions(mod, newdata = "mean", byfun = sum, by = by)
#'
#' @inheritParams slopes
#' @inheritParams comparisons
#' @export
predictions <- function(
    model,
    newdata = NULL,
    variables = NULL,
    vcov = TRUE,
    conf_level = 0.95,
    type = NULL,
    by = FALSE,
    byfun = NULL,
    wts = FALSE,
    transform = NULL,
    hypothesis = NULL,
    equivalence = NULL,
    df = Inf,
    numderiv = "fdforward",
    ...
) {
    # init
    if (inherits(model, "marginaleffects_internal")) {
        mfx <- model
    } else {
        # pass through ... to avoid calling `get_modeldata()` in inferences()
        if ("modeldata" %in% ...names()) {
            modeldata <- ...get("modeldata")
        } else {
            modeldata <- NULL
        }
        call <- construct_call(model, "predictions")
        model <- sanitize_model(model, call = call, newdata = newdata, wts = wts, vcov = vcov, by = by, ...)
        mfx <- new_marginaleffects_internal(
            call = call,
            model = model,
            modeldata = modeldata,
            by = by,
            byfun = byfun
        )
    }

    # newdata
    scall <- rlang::enquo(newdata)
    mfx <- add_newdata(mfx, scall, newdata = newdata, by = by, wts = wts, byfun = byfun)

    # inferences() dispatch
    methods <- c("rsample", "boot", "fwb", "simulation")
    if (isTRUE(checkmate::check_choice(vcov, methods))) {
        inferences_method <- vcov
        vcov <- FALSE
    } else {
        inferences_method <- NULL
    }

    dots <- list(...)
    sanity_dots(model = model, ...)

    # multiple imputation
    if (inherits(model, c("mira", "amest"))) {
        out <- process_imputation(mfx)
        return(out)
    }


    # sanity checks
    mfx <- add_numderiv(mfx, numderiv)
    mfx <- add_by(mfx, by)
    sanity_reserved(mfx)

    # if type is NULL, we backtransform if relevant
    # before add_hypothesis()
    link_to_response <- FALSE
    mfx@type <- sanitize_type(
        model = model,
        type = type,
        by = by,
        hypothesis = hypothesis,
        calling_function = mfx@calling_function
    )
    if (identical(mfx@type, "invlink(link)")) {
        # backtransform: yes
        if (is.null(hypothesis)) {
            link_to_response <- TRUE
            # backtransform: no
        } else {
            mfx@type <- "response"
            warn_sprintf(
                'The `type="invlink"` argument is not available unless `hypothesis` is `NULL` or a single number. The value of the `type` argument was changed to "response" automatically. To suppress this warning, use `type="response"` explicitly in your function call.'
            )
        }
    }

    mfx <- add_hypothesis(mfx, hypothesis)

    transform <- sanitize_transform(transform)

    mfx@conf_level <- sanitize_conf_level(conf_level, ...)

    # analogous to comparisons(variables=list(...))
    if (!is.null(variables)) {
        mfx <- add_variables(
            variables = variables,
            mfx = mfx
        )
        args <- list(
            model = mfx@model,
            newdata = mfx@newdata,
            grid_type = "counterfactual",
            marginaleffects_internal = mfx
        )
        for (v in mfx@variables) {
            args[[v$name]] <- v$value
        }
        mfx@newdata <- do.call("datagrid", args)
    }

    # trust newdata$rowid
    if (!"rowid" %in% colnames(mfx@newdata)) {
        mfx@newdata[["rowid"]] <- seq_len(nrow(mfx@newdata))
    }

    # Store unpadded newdata for degrees of freedom calculation
    unpadded_newdata <- mfx@newdata

    mfx@newdata <- pad(model, mfx@newdata)

    ############### sanity checks are over

    # pre-building the model matrix can speed up repeated predictions
    mfx@newdata <- add_model_matrix_attribute(mfx)

    # Try autodiff early exit
    autodiff_result <- NULL
    if (isTRUE(settings_get("autodiff")) && !isFALSE(vcov)) {
        # Get vcov matrix first
        mfx@vcov_type <- get_vcov_label(vcov)
        mfx@vcov_model <- get_vcov(mfx@model, vcov = vcov, type = if (link_to_response) "link" else mfx@type, ...)

        if (isTRUE(checkmate::check_matrix(mfx@vcov_model))) {
            # Filter out padding rows before autodiff when using by=character
            # Padding is only needed for model.matrix construction, which is already done
            if (is.character(mfx@by) && "rowid" %in% colnames(mfx@newdata)) {
                idx <- mfx@newdata$rowid > 0
                if (!all(idx)) {
                    # Filter newdata and model matrix
                    mfx@newdata <- mfx@newdata[idx, , drop = FALSE]
                    # Also filter the model matrix attribute
                    MM <- attr(mfx@newdata, "marginaleffects_model_matrix")
                    if (!is.null(MM)) {
                        attr(mfx@newdata, "marginaleffects_model_matrix") <- MM[idx, , drop = FALSE]
                    }
                }
            }
            # Try autodiff
            autodiff_result <- jax_predictions(
                mfx = mfx,
                vcov_matrix = mfx@vcov_model,
                ...
            )
        }
    }

    if (!is.null(autodiff_result)) {
        # SUCCESS: Use autodiff results and skip computation pipeline

        # Build result data.frame
        if (isFALSE(mfx@by)) {
            # Row-level results
            tmp <- data.frame(
                rowid = mfx@newdata$rowid,
                type = mfx@type,
                estimate = autodiff_result$estimate,
                std.error = autodiff_result$std.error
            )
            if ("rowidcf" %in% colnames(mfx@newdata)) {
                tmp[["rowidcf"]] <- mfx@newdata[["rowidcf"]]
            }
            # Unpad: remove padding rows added for missing factor levels
            tmp <- unpad(tmp, draws = NULL)$out
        } else {
            # Aggregated results (by=TRUE or by=character)
            tmp <- data.frame(
                type = mfx@type,
                estimate = autodiff_result$estimate,
                std.error = autodiff_result$std.error
            )

            # Add grouping columns for by=character
            if (is.character(mfx@by)) {
                # Extract unique group combinations
                bycols <- mfx@by
                if (inherits(mfx@newdata, "data.table")) {
                    group_data <- unique(mfx@newdata[, ..bycols])
                    data.table::setorderv(group_data, bycols)
                    group_data <- as.data.frame(group_data)
                } else {
                    group_data <- unique(mfx@newdata[, bycols, drop = FALSE])
                    if (nrow(group_data) > 1) {
                        ord <- do.call(order, group_data)
                        group_data <- group_data[ord, , drop = FALSE]
                    }
                }
                rownames(group_data) <- NULL
                tmp <- cbind(group_data, tmp)
            }
        }

        # Store jacobian
        mfx@jacobian <- autodiff_result$jacobian

    } else {
        # FALLBACK: Use standard computation pipeline

        # main estimation
        args <- list(
            mfx = mfx,
            type = if (link_to_response) "link" else mfx@type,
            hypothesis = mfx@hypothesis,
            by = by,
            byfun = byfun
        )

        args <- utils::modifyList(args, dots)
        tmp <- do_call(get_predictions, args)

        # hypothesis formula names are attached in by() in get_predictions()
        mfx@variable_names_by <- unique(c(
            mfx@variable_names_by,
            attr(tmp, "hypothesis_function_by")))

        # two cases when tmp is a data.frame
        # get_predict gets us rowid with the original rows
        if (inherits(tmp, "data.frame")) {
            setnames(
                tmp,
                old = c("Predicted", "SE", "CI_low", "CI_high"),
                new = c("estimate", "std.error", "conf.low", "conf.high"),
                skip_absent = TRUE
            )
        } else {
            tmp <- data.frame(mfx@newdata$rowid, mfx@type, tmp)
            colnames(tmp) <- c("rowid", "estimate")
            if ("rowidcf" %in% colnames(mfx@newdata)) {
                tmp[["rowidcf"]] <- mfx@newdata[["rowidcf"]]
            }
        }

        # bayesian posterior draws
        mfx@draws <- attr(tmp, "posterior_draws")

        if (!isFALSE(vcov)) {
            mfx@vcov_type <- get_vcov_label(vcov)
            mfx@vcov_model <- get_vcov(mfx@model, vcov = vcov, type = mfx@type, ...)

            # Delta method for standard errors
            if (!"std.error" %in% colnames(tmp) && is.null(mfx@draws) && isTRUE(checkmate::check_matrix(mfx@vcov_model))) {
                fun <- function(...) {
                    get_predictions(..., verbose = FALSE)$estimate
                }
                args <- list(
                    mfx = mfx,
                    model_perturbed = mfx@model,
                    vcov = mfx@vcov_model,
                    type = if (link_to_response) "link" else mfx@type,
                    FUN = fun,
                    hypothesis = mfx@hypothesis
                )
                args <- utils::modifyList(args, dots)
                se <- do_call(get_se_delta, args)
                if (is.numeric(se) && length(se) == nrow(tmp)) {
                    mfx@jacobian <- attr(se, "jacobian")
                    tmp[["std.error"]] <- as.vector(se) # drop attribute
                }
            }
        }
    }

    # Common path for both autodiff and fallback

    # degrees of freedom - use unpadded newdata
    # Temporarily replace newdata with unpadded version for df calculation
    original_newdata <- mfx@newdata
    mfx@newdata <- unpadded_newdata
    mfx <- add_degrees_of_freedom(
        mfx = mfx,
        df = df,
        by = by,
        hypothesis = mfx@hypothesis,
        vcov = vcov
    )
    # Restore padded newdata
    mfx@newdata <- original_newdata
    if (!is.null(mfx@df) && is.numeric(mfx@df)) {
        tmp$df <- mfx@df
    }

    # Confidence intervals (autodiff already has std.error)
    if (!isFALSE(vcov) && ("std.error" %in% colnames(tmp) || !is.null(mfx@draws))) {
        tmp <- get_ci(tmp, mfx)
    }

    out <- data.table::data.table(tmp)
    data.table::setDT(mfx@newdata)

    # it does not make sense to merge original data into aggregated results
    if (isFALSE(by) && !inherits(mfx@model, "mclogit")) {
        out <- merge_by_rowid(out, mfx@newdata)
    }

    # remove weights column (now handled by add_attributes)
    out[["marginaleffects_wts_internal"]] <- NULL

    # WARNING: we cannot sort rows at the end because `get_hypothesis()` is
    # applied in the middle, and it must already be sorted in the final order,
    # otherwise, users cannot know for sure what is going to be the first and
    # second rows, etc.
    out <- sort_columns(out, mfx@newdata, by)

    # equivalence tests
    out <- equivalence(out, equivalence = equivalence, df = mfx@df, draws = mfx@draws, ...)

    # after rename to estimate
    if (isTRUE(link_to_response)) {
        linv <- tryCatch(insight::link_inverse(mfx@model), error = function(e) identity)
        out <- backtransform(out, transform = linv, draws = mfx@draws)
    }
    out <- backtransform(out, transform = transform, draws = mfx@draws)
    new_draws <- attr(out, "posterior_draws") # important!
    if (!is.null(new_draws)) mfx@draws <- new_draws

    data.table::setDF(out)
    class(out) <- c("predictions", class(out))

    # before add_attributes()
    if (inherits(mfx@model, "brmsfit")) {
        insight::check_if_installed("brms")
        mfx@draws_chains <- brms::nchains(mfx@model)
    }

    # Add common attributes from mfx S4 slots
    # class before prune
    out <- add_attributes(out, mfx)
    out <- prune_attributes(out)


    if ("group" %in% names(out) && all(out$group == "main_marginaleffect")) {
        out$group <- NULL
    }

    if (!is.null(inferences_method)) {
        out <- inferences(out, method = inferences_method)
    }

    return(out)
}


#' Average predictions
#' @describeIn predictions Average predictions
#' @export
#'
avg_predictions <- function(
    model,
    newdata = NULL,
    variables = NULL,
    vcov = TRUE,
    conf_level = 0.95,
    type = NULL,
    by = TRUE,
    byfun = NULL,
    wts = FALSE,
    transform = NULL,
    hypothesis = NULL,
    equivalence = NULL,
    df = Inf,
    numderiv = "fdforward",
    ...
) {
    # order of the first few paragraphs is important
    # if `newdata` is a call to `typical` or `counterfactual`, insert `model`

    # group by focal variable automatically unless otherwise stated
    if (isTRUE(by)) {
        if (isTRUE(checkmate::check_character(variables))) {
            by <- variables
        } else if (isTRUE(checkmate::check_list(variables, names = "named"))) {
            by <- names(variables)
        }
    }

    #Construct predictions() call
    call_attr <- construct_call(model, "predictions")
    call_attr[["by"]] <- by

    out <- eval.parent(call_attr)

    return(out)
}

align_jacobian_vcov <- function(J, V, object, ...) {
    if (!isTRUE(ncol(J) == ncol(V))) {
        beta <- get_coef(object, ...)
        # Issue #718: ordinal::clm in test-pkg-ordinal.R
        if (
            anyNA(beta) &&
                ncol(J) > ncol(V) &&
                ncol(J) == length(beta) &&
                length(stats::na.omit(beta)) == ncol(V)
        ) {
            J <- J[, !is.na(beta), drop = FALSE]
        } else {
            cols <- intersect(colnames(J), colnames(V))
            if (length(cols) == 0) {
                stop_sprintf(
                    "The jacobian does not match the variance-covariance matrix."
                )
            }
            V <- V[cols, cols, drop = FALSE]
            J <- J[, cols, drop = FALSE]
        }
    }
    return(list(J = J, V = V))
}


#' Compute standard errors using the delta method
#'
#' @inheritParams slopes
#' @param FUN a function which accepts a `model` and other inputs and returns a
#'   vector of estimates (marginal effects, marginal means, etc.)
#' @param index data.frame over which we aggregate J_mean (matches tidy() output)
#' @return vector of standard errors
#' @noRd
get_se_delta <- function(
    model_perturbed,
    vcov,
    FUN,
    mfx = NULL,
    type = NULL,
    newdata = NULL,
    index = NULL,
    eps = NULL,
    J = NULL,
    hypothesis = NULL,
    calling_function = NULL,
    comparison = NULL,
    by = NULL,
    byfun = NULL,
    hi = NULL,
    lo = NULL,
    original = NULL,
    estimates = NULL,
    ...) {
    # Use mfx slots when available
    if (!is.null(mfx)) {
        eps <- if (is.null(eps)) mfx@eps else eps
        calling_function <- if (is.null(calling_function)) mfx@calling_function else calling_function
        comparison <- if (is.null(comparison)) mfx@comparison else comparison
        by <- if (is.null(by)) mfx@by else by
        byfun <- if (is.null(byfun)) mfx@byfun else byfun
    }
    # delta method does not work for these models
    bad <- c("brmsfit", "stanreg", "bart")
    if (any(inherits(model_perturbed, bad))) {
        return(NULL)
    }

    coefs <- get_coef(model_perturbed, ...)

    # TODO: this is a terrible sanity check
    # some vcov methods return an unnamed matrix, some have duplicate names
    flag <- anyDuplicated(colnames(vcov)) == 0 ||
        anyDuplicated(names(coefs)) == 0
    if (
        flag &&
            !is.null(dimnames(vcov)) &&
            all(names(coefs) %in% colnames(vcov))
    ) {
        bnames <- intersect(names(coefs), colnames(vcov))
        vcov <- vcov[bnames, bnames, drop = FALSE]
        colnames(vcov) <- row.names(vcov) <- names(coefs)
        coefs <- coefs[bnames]
    }

    # user-supplied jacobian machine (e.g. JAX)
    if (is.null(J)) {
        fun <- settings_get("jacobian_function")
        if (is.null(fun)) {
            fun <- function(...) NULL
        }
        if (!"..." %in% names(formals(fun))) {
            msg <- "The `marginaleffects_jacobian_function` option must accept the ... argument."
            stop_sprintf(msg)
        }
        J <- fun(
            coefs = coefs,
            mfx = mfx,
            newdata = newdata,
            model_perturbed = model_perturbed,
            hypothesis = hypothesis,
            type = type,
            by = by,
            byfun = byfun,
            hi = hi,
            lo = lo,
            original = original,
            estimates = estimates,
            comparison = comparison,
            calling_function = calling_function
        )
        checkmate::assert_matrix(J, mode = "numeric", ncols = length(coefs), null.ok = TRUE)

        # Unpad Jacobian for autodiff jax_jacobian if newdata was padded
        # JAX computes jacobian on padded newdata, but final results are unpadded
        if (!is.null(J) && !is.null(mfx) && "rowid" %in% colnames(mfx@newdata)) {
            idx <- mfx@newdata$rowid > 0
            if (!all(idx) && nrow(J) == nrow(mfx@newdata)) {
                J <- J[idx, , drop = FALSE]
            }
        }
    }

    # input: named vector of coefficients
    # output: gradient
    inner <- function(x) {
        names(x) <- names(coefs)
        model_tmp <- set_coef(model_perturbed, x, ...)
        # do not pass NULL arguments. Important for `deltam` to allow users to supply FUN without ...
        args <- list(
            mfx = mfx,
            model_perturbed = model_tmp,
            hypothesis = hypothesis,
            type = type,
            hi = hi,
            lo = lo,
            original = original,
            by = by,
            byfun = byfun
        )
        args <- c(args, list(...))
        if (inherits(model_perturbed, "gamlss")) {
            args[["safe"]] <- FALSE
        }
        if (!is.null(eps)) {
            args[["eps"]] <- eps
        }
        if (!is.null(type)) {
            args[["type"]] <- type
        }
        if (!is.null(newdata)) {
            args[["newdata"]] <- newdata
        }
        if (!is.null(J)) {
            args[["J"]] <- J
        }
        if (!is.null(eps)) {
            args[["eps"]] <- eps
        }

        if (inherits(model_perturbed, "glmmTMB")) {
            args$newparams <- x
        }

        g <- do_call(FUN, args)

        return(g)
    }

    if (is.null(J) || !is.null(hypothesis)) {
        args <- list(
            func = inner,
            x = coefs,
            numderiv = mfx@numderiv
        )
        J <- do.call("get_jacobian", args)
        colnames(J) <- names(get_coef(model_perturbed, ...))
    }

    # align J and V: This might be a problematic hack, but I have not found examples yet.
    V <- vcov
    if (!isTRUE(ncol(J) == ncol(V))) {
        aligned <- align_jacobian_vcov(J, V, model_perturbed, ...)
        J <- aligned$J
        V <- aligned$V
    }

    # Var(dydx) = J Var(beta) J'
    # computing the full matrix is memory-expensive, and we only need the diagonal
    # algebra trick: https://stackoverflow.com/a/42569902/342331
    if (isTRUE(settings_get("autodiff"))) {
        mAD <- settings_get("mAD")
        se <- mAD$utils$standard_errors(J, V)
    } else {
        se <- sqrt(rowSums(tcrossprod(J, V) * J))
    }
    se[se == 0] <- NA_real_
    attr(se, "jacobian") <- J

    return(se)
}

get_comparisons_data_logical <- function(model, newdata, variable, ...) {
    # custom data frame or function
    if (
        isTRUE(checkmate::check_function(variable$value)) ||
            isTRUE(checkmate::check_data_frame(variable$value))
    ) {
        out <- contrast_categories_custom(variable, newdata)
        return(out)
    }

    lo <- hi <- newdata
    lo[[variable$name]] <- FALSE
    hi[[variable$name]] <- TRUE
    lab <- suppressWarnings(tryCatch(
        sprintf(variable$label, TRUE, FALSE),
        error = function(e) variable$label
    ))
    out <- list(
        rowid = seq_len(nrow(newdata)),
        lo = lo,
        hi = hi,
        original = newdata,
        ter = rep(variable$name, nrow(newdata)),
        lab = rep(lab, nrow(newdata))
    )
    return(out)
}

#' Get a named variance-covariance matrix from a model object
#'
#' Mostly for internal use, but can be useful because the output is consistent across model classes.
#' @inheritParams slopes
#' @return A named square matrix of variance and covariances. The names must match the coefficient names.
#' @rdname get_vcov
#' @export
get_vcov <- function(model, ...) {
    UseMethod("get_vcov", model)
}


#' @rdname get_vcov
#' @export
get_vcov.default <- function(model, vcov = NULL, ...) {
    if (isFALSE(vcov)) {
        return(NULL)
    }

    vcov <- sanitize_vcov(model = model, vcov = vcov)
    if (isTRUE(checkmate::check_matrix(vcov))) {
        return(vcov)
    }

    # {insight}
    args <- get_varcov_args(model, vcov)
    args[["x"]] <- model
    args[["component"]] <- "all"

    # 1st try: with arguments
    out <- myTryCatch(do_call(get("get_varcov", asNamespace("insight")), args))

    # 2nd try: without arguments
    if (!isTRUE(checkmate::check_matrix(out$value, min.rows = 1))) {
        out <- myTryCatch(insight::get_varcov(model))
        if (isTRUE(checkmate::check_matrix(out$value, min.rows = 1))) {
            msg <- "Unable to extract a variance-covariance matrix using this `vcov` argument. Standard errors are computed using the default variance instead. Perhaps the model or argument is not supported by the `sandwich` ('HC0', 'HC3', ~clusterid, etc.) or `clubSandwich` ('CR0', etc.) packages. If you believe that the model is supported by one of these two packages, you can open a feature request on Github."
            warn_sprintf(msg)
        }
    }

    if (!isTRUE(checkmate::check_matrix(out$value, min.rows = 1))) {
        msg <- "Unable to extract a variance-covariance matrix from this model."
        warning(msg, call. = FALSE)
        return(NULL)

        # valid matrix with warning
    } else if (!is.null(out$warning)) {
        warning(out$warning$message, call. = FALSE)
    }

    out <- out[["value"]]

    # problem: no row.names
    if (is.null(row.names(out))) {
        coefs <- get_coef(model)
        if (ncol(out) == length(coefs)) {
            termnames <- names(stats::coef(model))
            if (length(termnames) == ncol(out)) {
                colnames(out) <- termnames
                row.names(out) <- termnames
            }
        } else {
            return(NULL)
        }
    }

    # problem: duplicate colnames
    if (anyDuplicated(colnames(out)) == 0) {
        coefs <- get_coef(model, ...)
        # 1) Check above is needed for `AER::tobit` and others where `out`
        # includes Log(scale) but `coef` does not Dangerous for `oridinal::clm`
        # and others where there are important duplicate column names in
        # `out`, and selecting with [,] repeats the first instance.

        # 2) Sometimes out has more columns than coefs
        if (all(names(coefs) %in% colnames(out))) {
            out <- out[names(coefs), names(coefs), drop = FALSE]
        }
    }

    return(out)

    # NOTES:
    # survival::coxph with 1 regressor produces a vector
}


#' Take a `summary()` style `vcov` argument and convert it to
#' `insight::get_varcov()`
#'
#' @keywords internal
get_varcov_args <- function(model, vcov) {
    if (is.null(vcov) || isTRUE(checkmate::check_matrix(vcov))) {
        out <- list()
        return(out)
    }

    if (isTRUE(checkmate::check_formula(vcov))) {
        out <- list("vcov" = "vcovCL", "vcov_args" = list("cluster" = vcov))
        return(out)
    }

    if (
        isTRUE(checkmate::check_choice(vcov, "satterthwaite")) ||
            isTRUE(checkmate::check_choice(vcov, "kewnard-roger"))
    ) {
        if (
            !isTRUE(inherits(model, "lmerMod")) &&
                !isTRUE(inherits(model, "lmerModTest"))
        ) {
            msg <- "Satterthwaite and Kenward-Roger corrections are only available for linear mixed effects models from the `lme4` package, and objects of class `lmerMod` or `lmerModTest`."
            stop(msg, call. = FALSE)
        }
        if (isTRUE(vcov == "satterthwaite")) {
            return(list())
        } else {
            return(list(vcov = "kenward-roger"))
        }
    }

    out <- switch(
        vcov,
        "stata" = list(vcov = "HC2"),
        "robust" = list(vcov = "HC3"),
        "bootstrap" = list(vcov = "BS"),
        "outer-product" = list(vcov = "OPG"),
        list(vcov = vcov)
    )
    return(out)
}


get_vcov_label <- function(vcov) {
    if (is.null(vcov)) {
        vcov <- ""
    }
    if (!is.character(vcov)) {
        return(NULL)
    }

    out <- switch(
        vcov,
        "stata" = "Stata",
        "robust" = "Robust",
        "kenward-roger" = "Kenward-Roger",
        "satterthwaite" = "Satterthwaite",
        "HC" = ,
        "HC0" = ,
        "HC1" = ,
        "HC2" = ,
        "HC3" = ,
        "HC4" = ,
        "HC4m" = ,
        "HC5" = ,
        "HAC" = ,
        "OPG" = vcov,
        "NeweyWest" = "Newey-West",
        "kernHAC" = "Kernel HAC",
        vcov
    )
    return(out)
}


#' internal get_vcov
#'
#' @export
#' @noRd
get_vcov.comparisons <- function(model, ...) {
    stats::vcov(model)
}

#' internal get_vcov
#'
#' @export
#' @noRd
get_vcov.slopes <- get_vcov.comparisons

#' internal get_vcov
#'
#' @export
#' @noRd
get_vcov.hypotheses <- get_vcov.comparisons

#' internal get_vcov
#'
#' @export
#' @noRd
get_vcov.predictions <- get_vcov.comparisons

---
title: "Model to Meaning"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Model to Meaning}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
   collapse = TRUE,
   comment = "#>"
)
```

The [marginaleffects.com](https://marginaleffects.com) website includes detailed documentation for both the `R` and `Python` versions of the `marginaleffects` package.

It also includes a complete free version of this forthcoming book

> *Model to Meaning How to Interpret Statistical Models with R and Python.* Arel-Bundock, Vincent. 2026. CRC Chapman & Hall.

In addition to full book chapters, [marginaleffects.com](https://marginaleffects.com) includes a lot of bonus content. In total, there are over 40 detailed vignettes and tutorials, including:

1. [Home](https://marginaleffects.com/)
1. [Who is this website for?](https://marginaleffects.com/chapters/who.html)
1. [Quick start](https://marginaleffects.com/bonus/get_started.html)
1. Interpretation
   - [Models and meaning](https://marginaleffects.com/chapters/challenge.html)
   - [Conceptual framework](https://marginaleffects.com/chapters/framework.html)
1. Quantities and Tests
   - [Hypothesis and equivalence tests](https://marginaleffects.com/chapters/hypothesis.html)
   - [Predictions](https://marginaleffects.com/chapters/predictions.html)
   - [Counterfactual comparisons](https://marginaleffects.com/chapters/comparisons.html)
   - [Slopes](https://marginaleffects.com/chapters/slopes.html)
1. Case Studies
   - [G-Computation](https://marginaleffects.com/chapters/gcomputation.html)
   - [Experiments](https://marginaleffects.com/chapters/experiments.html)
   - [Interactions, polynomials, and splines](https://marginaleffects.com/chapters/interactions.html)
   - [Categorical and ordinal outcomes](https://marginaleffects.com/chapters/categorical.html)
   - [Multilevel regression with post-stratification](https://marginaleffects.com/chapters/mrp.html)
   - [Machine learning](https://marginaleffects.com/chapters/ml.html)
   - [Uncertainty](https://marginaleffects.com/chapters/uncertainty.html)
1. Tutorials
   - [Plots](https://marginaleffects.com/bonus/plot.html)
   - [Bayes](https://marginaleffects.com/bonus/brms.html)
   - [Bootstrap](https://marginaleffects.com/bonus/bootstrap.html)
   - [Conjoint Experiments](https://marginaleffects.com/bonus/conjoint.html)
   - [Data grids](https://marginaleffects.com/bonus/grids.html)
   - [Elasticity](https://marginaleffects.com/bonus/elasticity.html)
   - [GAM](https://marginaleffects.com/bonus/gam.html)
   - [Heterogeneity](https://marginaleffects.com/bonus/heterogeneity.html)
   - [Interrupted Times Series](https://marginaleffects.com/bonus/interrupted_time_series.html)
   - [Inverse Probability Weighting](https://marginaleffects.com/bonus/ipw.html)
   - [Links](https://marginaleffects.com/bonus/links.html)
   - [Logit](https://marginaleffects.com/bonus/logit.html)
   - [Matching](https://marginaleffects.com/bonus/matching.html)
   - [Missing Data](https://marginaleffects.com/bonus/multiple_imputation.html)
   - [Mixed Effects](https://marginaleffects.com/bonus/lme4.html)
   - [Multinomial Logit](https://marginaleffects.com/bonus/mlogit.html)
   - [Performance](https://marginaleffects.com/bonus/performance.html)
   - [Power Analysis](https://marginaleffects.com/bonus/power.html)
   - [S Values](https://marginaleffects.com/bonus/svalues.html)
   - [Scikit-learn](https://marginaleffects.com/bonus/sklearn.html)
   - [Survival Analysis](https://marginaleffects.com/bonus/survival.html)
   - [Total Marginal Effects](https://marginaleffects.com/bonus/total_marginal_effect.html)
1. Miscellaneous
   - [Help](https://marginaleffects.com/bonus/help.html)
   - [Alternative Software](https://marginaleffects.com/bonus/alternative_software.html)
   - [Extensions](https://marginaleffects.com/bonus/extensions.html)
   - [FAQ](https://marginaleffects.com/bonus/faq.html)
   - [Meme gallery](https://marginaleffects.com/bonus/meme.html)
   - [Supported Models](https://marginaleffects.com/bonus/supported_models.html)
   - [Tables](https://marginaleffects.com/bonus/tables.html)
1. [Citation and License](https://marginaleffects.com/bonus/citation.html)

---
title: "Model to Meaning"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Model to Meaning}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

The [marginaleffects.com](https://marginaleffects.com) website includes detailed documentation for both the `R` and `Python` versions of the `marginaleffects` package.

It also includes a complete free version of this forthcoming book

> *Model to Meaning: How to Interpret Statistical Results Using marginaleffects for R and Python.* Arel-Bundock, Vincent. CRC Chapman & Hall.

In addition to full book chapters, [marginaleffects.com](https://marginaleffects.com) includes a lot of bonus content. In total, there are over 40 detailed vignettes and tutorials, including:

1. [Home](https://marginaleffects.com/)
1. [Who is this website for?](https://marginaleffects.com/chapters/who.html)
1. [Quick start](https://marginaleffects.com/bonus/get_started.html)
1. Interpretation
   - [Models and meaning](https://marginaleffects.com/chapters/challenge.html)
   - [Tools and data](https://marginaleffects.com/chapters/tools.html)
   - [Conceptual framework](https://marginaleffects.com/chapters/framework.html)
1. Quantities and Tests
   - [Hypothesis and equivalence tests](https://marginaleffects.com/chapters/hypothesis.html)
   - [Predictions](https://marginaleffects.com/chapters/predictions.html)
   - [Counterfactual comparisons](https://marginaleffects.com/chapters/comparisons.html)
   - [Slopes](https://marginaleffects.com/chapters/slopes.html)
1. Case Studies
   - [Interactions, polynomials, and splines](https://marginaleffects.com/chapters/interactions.html)
   - [Categorical and ordinal outcomes](https://marginaleffects.com/chapters/categorical.html)
   - [Experiments](https://marginaleffects.com/chapters/experiments.html)
   - [Mixed effects regression and post-stratification](https://marginaleffects.com/chapters/mrp.html)
   - [Machine learning](https://marginaleffects.com/chapters/ml.html)
   - [Uncertainty](https://marginaleffects.com/chapters/uncertainty.html)
1. Bonus Content
   - [Help](https://marginaleffects.com/bonus/help.html)
   - [FAQ](https://marginaleffects.com/bonus/faq.html)
   - [Plots](https://marginaleffects.com/bonus/plot.html)
   - [Alternative Software](https://marginaleffects.com/bonus/alternative_software.html)
   - [Bayes](https://marginaleffects.com/bonus/brms.html)
   - [Elasticity](https://marginaleffects.com/bonus/elasticity.html)
   - [Extensions](https://marginaleffects.com/bonus/extensions.html)
   - [GAM](https://marginaleffects.com/bonus/gam.html)
   - [G-Computation](https://marginaleffects.com/bonus/gcomputation.html)
   - [Heterogeneity](https://marginaleffects.com/bonus/heterogeneity.html)
   - [Inverse Probability Weighting](https://marginaleffects.com/bonus/ipw.html)
   - [Links](https://marginaleffects.com/bonus/links.html)
   - [Logit](https://marginaleffects.com/bonus/logit.html)
   - [Marginal Means](https://marginaleffects.com/bonus/marginalmeans.html)
   - [Matching](https://marginaleffects.com/bonus/matching.html)
   - [Meme gallery](https://marginaleffects.com/bonus/meme.html)
   - [Multiple Comparisons](https://marginaleffects.com/bonus/multiple_comparisons.html)
   - [NumPyro](https://marginaleffects.com/bonus/numpyro.html)
   - [Performance](https://marginaleffects.com/bonus/performance.html)
   - [Supported Models](https://marginaleffects.com/bonus/supported_models.html)
   - [S Values](https://marginaleffects.com/bonus/svalues.html)
   - [Tables](https://marginaleffects.com/bonus/tables.html)
1. Back Matter
   - [Bibliography](https://marginaleffects.com/chapters/bibliography.html)
1. [Citation and License](https://marginaleffects.com/bonus/citation.html)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_dataset.R
\name{get_dataset}
\alias{get_dataset}
\title{Download and Read Datasets from \code{marginaleffects} or Rdatasets}
\usage{
get_dataset(
  dataset = "thornton",
  package = "marginaleffects",
  docs = FALSE,
  search = NULL
)
}
\arguments{
\item{dataset}{String. Name of the dataset to download.
\itemize{
\item \code{marginaleffects} archive: \code{affairs}, \code{airbnb}, \code{immigration}, \code{lottery}, \code{military}, \code{thornton}
\item Rdatasets archive: The name of a dataset listed on the Rdatasets index. See the website or the \code{search} argument.
}}

\item{package}{String. Package name that originally published the data.}

\item{docs}{Logical. If TRUE open the documentation using \code{getOption("viewer")} or the Rstudio viewer.}

\item{search}{Regular expression. Download the dataset index from Rdatasets; search the "Package", "Item", and "Title" columns; and return the matching rows.}
}
\value{
A data frame containing the dataset.
library(marginaleffects)
}
\description{
Downloads a dataset from the \code{marginaleffects} or the Rdatasets archives, and return it as a data frame. Opens the documentation as an HTML page. Search available datasets.

\url{https://vincentarelbundock.github.io/Rdatasets/}
}
\examples{
dat <- get_dataset("Titanic", "Stat2Data")
head(dat)

get_dataset(search = "(?i)titanic")

# View documentation in the browser
# get_dataset("Titanic", "Stat2Data", docs = TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class.R
\docType{class}
\name{marginaleffects_internal-class}
\alias{marginaleffects_internal-class}
\title{Internal S4 class for marginaleffects}
\description{
This S4 class is used internally to hold common arguments passed between
functions to simplify the function signatures and reduce redundant argument passing.
}
\section{Slots}{

\describe{
\item{\code{by}}{Aggregation/grouping specification}

\item{\code{byfun}}{Function for aggregation when using by}

\item{\code{call}}{The original function call}

\item{\code{calling_function}}{The name of the calling function (comparisons, predictions, hypotheses)}

\item{\code{comparison}}{Comparison function specification}

\item{\code{cross}}{Boolean flag for cross-contrasts}

\item{\code{df}}{The degrees of freedom}

\item{\code{eps}}{Epsilon value for numerical derivatives}

\item{\code{jacobian}}{The jacobian matrix or NULL}

\item{\code{model}}{The fitted model object}

\item{\code{modeldata}}{The model data frame}

\item{\code{newdata}}{The new data frame for predictions}

\item{\code{type}}{The sanitized type from sanitize_type()}

\item{\code{vcov_model}}{The variance-covariance matrix}

\item{\code{wts}}{The weights specification}
}}

\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_draws.R
\name{posterior_draws}
\alias{posterior_draws}
\title{alias to \code{get_draws()} keep forever for backward compatibility with JSS}
\usage{
posterior_draws(x, shape = "long")
}
\description{
alias to \code{get_draws()} keep forever for backward compatibility with JSS
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/inferences.R
\name{inferences}
\alias{inferences}
\title{(EXPERIMENTAL) Bootstrap, Conformal, and Simulation-Based Inference}
\usage{
inferences(
  x,
  method,
  R = 1000,
  conf_type = "perc",
  data_train = NULL,
  data_test = NULL,
  data_calib = NULL,
  conformal_score = "residual_abs",
  estimator = NULL,
  ...
)
}
\arguments{
\item{x}{Object produced by one of the core \code{marginaleffects} functions, or a data frame suitable for the function supplied to the \code{estimator} argument.}

\item{method}{String
\itemize{
\item "delta": delta method standard errors
\item "boot" package
\item "fwb": fractional weighted bootstrap
\item "rsample" package
\item "simulation" from a multivariate normal distribution (Krinsky & Robb, 1986)
\item "conformal_split": prediction intervals using split conformal prediction (see Angelopoulos & Bates, 2022)
\item "conformal_cv+": prediction intervals using cross-validation+ conformal prediction (see Barber et al., 2020)
\item "conformal_full": prediction intervals using full conformal prediction (see Lei et al., 2018). \strong{Warning}: This method is computationally expensive and typically much slower than split or CV+ methods.
\item "conformal_quantile": prediction intervals using full conformal prediction (see Romano et al., 2020).
}}

\item{R}{Number of resamples, simulations, or cross-validation folds.}

\item{conf_type}{String: type of bootstrap interval to construct.
\itemize{
\item \code{boot}: "perc", "norm", "basic", or "bca"
\item \code{fwb}: "perc", "norm", "wald", "basic", "bc", or "bca"
\item \code{rsample}: "perc" or "bca"
\item \code{simulation}: "perc" or "wald"
}}

\item{data_train}{Data frame used to train/fit the model. If \code{NULL}, \code{marginaleffects} tries to extract the data from the original model object. Test data are inferred directly from the \code{newdata} supplied to the originating \code{marginaleffects} call (e.g., \code{predictions()}).}

\item{data_test}{Data frame make out of sample prediction. Only used for conformal inference. If \code{NULL}, the data frame supplied to \code{newdata} in the original \code{marginaleffects} call is used.}

\item{data_calib}{Data frame used for calibration in split conformal prediction.}

\item{conformal_score}{String. Warning: The \code{type} argument in \code{predictions()} must generate predictions which are on the same scale as the outcome variable. Typically, this means that \code{type} must be "response" or "probs".
\itemize{
\item "residual_abs" or "residual_sq" for regression tasks (numeric outcome)
\item "softmax" for classification tasks (when \code{predictions()} returns a \code{group} columns, such as multinomial or ordinal logit models.
}}

\item{estimator}{Function that accepts a data frame, fits a model, applies a \code{marginaleffects} function, and returns the object. Only supported with \code{method = "rsample"} or \code{method = "boot"}. When \code{method = "rsample"}, the output must include a "term" column. This is not always the case for \code{predictions()}, in which case users may have to create the column manually.}

\item{...}{\itemize{
\item If \code{method = "boot"}, additional arguments are passed to \code{boot::boot()}.
\item If \code{method = "fwb"}, additional arguments are passed to \code{fwb::fwb()}.
\item If \code{method = "rsample"}, additional arguments are passed to \code{rsample::bootstraps()}, unless the user supplies a \code{group} argument, in which case all arguments are passed to \code{rsample::group_bootstraps()}.
\item If \code{method = "conformal_full"}, additional arguments control the optimization process:
\itemize{
\item \code{var_multiplier}: multiplier for initial search bounds (default: 10)
\item \code{max_iter}: maximum iterations for root finding (default: 100)
\item \code{tolerance}: tolerance for root finding convergence (default: \code{.Machine$double.eps^0.25})
}
\item If \code{method = "conformal_quantile"}, additional arguments are passed to \code{quantregForest::quantregForest()} for fitting the quantile regression forest (e.g., \code{ntree}, \code{mtry}, \code{nodesize}, \code{nthreads}).
\item Additional arguments are ignored for other conformal methods (\code{conformal_split}, \verb{conformal_cv+}).
}}
}
\value{
A \code{marginaleffects} object with simulation or bootstrap resamples and objects attached.
}
\description{
Warning: This function is experimental. It may be renamed, the user interface may change, or the functionality may migrate to arguments in other \code{marginaleffects} functions.

Apply this function to a \code{marginaleffects} object to change the inferential method used to compute uncertainty estimates.
}
\details{
When \code{method = "simulation"}, we conduct simulation-based inference following the method discussed in Krinsky & Robb (1986):
\enumerate{
\item Draw \code{R} sets of simulated coefficients from a multivariate normal distribution with mean equal to the original model's estimated coefficients and variance equal to the model's variance-covariance matrix (classical, "HC3", or other).
\item Use the \code{R} sets of coefficients to compute \code{R} sets of estimands: predictions, comparisons, slopes, or hypotheses.
\item Take quantiles of the resulting distribution of estimands to obtain a confidence interval (when \code{conf_type = "perc"}) and the standard deviation of simulated estimates to estimate the standard error (which is used for a Z-test and Wald confidence intervals when \code{conf_type = "wald"}).
}

When \code{method = "fwb"}, drawn weights are supplied to the model fitting function's \code{weights} argument; if the model doesn't accept non-integer weights, this method should not be used. If weights were included in the original model fit, they are extracted by \code{\link[=weights]{weights()}} and multiplied by the drawn weights. These weights are supplied to the \code{wts} argument of the estimation function (e.g., \code{comparisons()}).

Warning: custom model classes are not supported by \code{inferences()} because they are not guaranteed to come with an appropriate \code{update()} method.
}
\section{References}{


Krinsky, I., and A. L. Robb. 1986. "On Approximating the Statistical Properties of Elasticities." Review of Economics and Statistics 68 (4): 715–9.

King, Gary, Michael Tomz, and Jason Wittenberg. "Making the most of statistical analyses: Improving interpretation and presentation." American journal of political science (2000): 347-361

Dowd, Bryan E., William H. Greene, and Edward C. Norton. "Computation of standard errors." Health services research 49.2 (2014): 731-750.

Angelopoulos, Anastasios N., and Stephen Bates. 2022. "A Gentle Introduction to Conformal Prediction and Distribution-Free Uncertainty Quantification." arXiv. https://doi.org/10.48550/arXiv.2107.07511.

Barber, Rina Foygel, Emmanuel J. Candes, Aaditya Ramdas, and Ryan J. Tibshirani. 2020. "Predictive Inference with the Jackknife+." arXiv. http://arxiv.org/abs/1905.02928.

Lei, Jing, Max G'Sell, Alessandro Rinaldo, Ryan J. Tibshirani, and Larry Wasserman. 2018. "Distribution-Free Predictive Inference for Regression." Journal of the American Statistical Association 113 (523): 1094–1111.

Romano, Yaniv, Evan Patterson, and Emmanuel Candes. 2020. "Conformalized quantile regression." Advances in neural information processing systems 32.
}

\section{Parallel computation}{


The \code{slopes()} and \code{comparisons()} functions can use parallelism to
speed up computation. Operations are parallelized for the computation of
standard errors, at the model coefficient level. There is always
considerable overhead when using parallel computation, mainly involved
in passing the whole dataset to the different processes. Thus, parallel
computation is most likely to be useful when the model includes many parameters
and the dataset is relatively small.

Warning: In many cases, parallel processing will not be useful at all.

To activate parallel computation, users must load the \code{future.apply} package,
call \code{plan()} function, and set a global option.

\code{options(marginaleffects_parallel = TRUE)}: parallelize delta method computation of standard errors.
\code{options(marginaleffects_parallel_inferences = TRUE)}: parallelize \code{"rsample"} or \code{"fwb"} bootstrap computation in \code{inferences()}.
\code{options(marginaleffects_parallel_packages = TRUE)}: vector of strings with the names of modeling packages used to fit the model, ex: c("survival", "splines")

For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(future.apply)
plan("multisession", workers = 4)
options(marginaleffects_parallel = FALSE)
options(marginaleffects_parallel_inferences = TRUE)
options(marginaleffects_parallel_packages = c("survival", "splines"))

slopes(model)
}\if{html}{\out{</div>}}

To disable parallelism in \code{marginaleffects} altogether, you can set a global option:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(marginaleffects_parallel = FALSE)
}\if{html}{\out{</div>}}
}

\examples{
\dontrun{
library(magrittr)
set.seed(1024)
mod <- lm(Sepal.Length ~ Sepal.Width * Species, data = iris)

# bootstrap
avg_predictions(mod, by = "Species") \%>\%
    inferences(method = "boot")

avg_predictions(mod, by = "Species") \%>\%
    inferences(method = "rsample")

# Fractional (bayesian) bootstrap
avg_slopes(mod, by = "Species") \%>\%
    inferences(method = "fwb") \%>\%
    get_draws("rvar") \%>\%
    data.frame()

# Simulation-based inference
slopes(mod) \%>\%
    inferences(method = "simulation") \%>\%
    head()

# Two-step estimation procedure: Propensity score + G-Computation
lalonde <- get_dataset("lalonde")
estimator <- function(data) {
    # Step 1: Estimate propensity scores
    fit1 <- glm(treat ~ age + educ + race, family = binomial, data = data)
    ps <- predict(fit1, type = "response")
    # Step 2: Fit weighted outcome model
    m <- lm(re78 ~ treat * (re75 + age + educ + race),
        data = data, weight = ps
    )
    # Step 3: Compute average treatment effect by G-computation
    avg_comparisons(m, variables = "treat", wts = ps, vcov = FALSE)
}
inferences(lalonde, method = "rsample", estimator = estimator)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/comparisons.R
\name{comparisons}
\alias{comparisons}
\alias{avg_comparisons}
\title{Comparisons Between Predictions Made With Different Regressor Values}
\usage{
comparisons(
  model,
  newdata = NULL,
  variables = NULL,
  comparison = "difference",
  type = NULL,
  vcov = TRUE,
  by = FALSE,
  conf_level = 0.95,
  transform = NULL,
  cross = FALSE,
  wts = FALSE,
  hypothesis = NULL,
  equivalence = NULL,
  df = Inf,
  eps = NULL,
  numderiv = "fdforward",
  ...
)

avg_comparisons(
  model,
  newdata = NULL,
  variables = NULL,
  type = NULL,
  vcov = TRUE,
  by = TRUE,
  conf_level = 0.95,
  comparison = "difference",
  transform = NULL,
  cross = FALSE,
  wts = FALSE,
  hypothesis = NULL,
  equivalence = NULL,
  df = Inf,
  eps = NULL,
  numderiv = "fdforward",
  ...
)
}
\arguments{
\item{model}{Model object}

\item{newdata}{Grid of predictor values at which we evaluate the comparisons.
\itemize{
\item Warning: Avoid modifying your dataset between fitting the model and calling a \code{marginaleffects} function. This can sometimes lead to unexpected results.
\item \code{NULL} (default): Unit-level contrasts for each observed value in the dataset (empirical distribution). The dataset is retrieved using \code{\link[insight:get_data]{insight::get_data()}}, which tries to extract data from the environment. This may produce unexpected results if the original data frame has been altered since fitting the model.
\item data frame: Unit-level contrasts for each row of the \code{newdata} data frame.
\item string:
\itemize{
\item "mean": Contrasts at the Mean. Contrasts when each predictor is held at its mean or mode.
\item "median": Contrasts at the Median. Contrasts when each predictor is held at its median or mode.
\item "balanced": Contrasts evaluated on a balanced grid with every combination of categories and numeric variables held at their means.
\item "tukey": Contrasts at Tukey's 5 numbers.
\item "grid": Contrasts on a grid of representative numbers (Tukey's 5 numbers and unique values of categorical predictors).
}
\item \code{\link[=datagrid]{datagrid()}} call to specify a custom grid of regressors. For example:
\itemize{
\item \code{newdata = datagrid(cyl = c(4, 6))}: \code{cyl} variable equal to 4 and 6 and other regressors fixed at their means or modes.
\item \code{newdata = datagrid(mpg = fivenum)}: \code{mpg} variable held at Tukey's five numbers (using the \code{fivenum} function), and other regressors fixed at their means or modes.
\item See the Examples section and the \link{datagrid} documentation.
}
\item \code{\link[=subset]{subset()}} call with a single argument to select a subset of the dataset used to fit the model, ex: \code{newdata = subset(treatment == 1)}
\item \code{\link[dplyr:filter]{dplyr::filter()}} call with a single argument to select a subset of the dataset used to fit the model, ex: \code{newdata = filter(treatment == 1)}
}}

\item{variables}{Focal variables
\itemize{
\item \code{NULL}: compute comparisons for all the variables in the model object (can be slow).
\item Character vector: subset of variables (usually faster).
\item Named list: names identify the subset of variables of interest, and values define the type of contrast to compute. Acceptable values depend on the variable type:
\itemize{
\item Factor or character variables:
\itemize{
\item "reference": Each factor level is compared to the factor reference (base) level
\item "all": All combinations of observed levels
\item "sequential": Each factor level is compared to the previous factor level
\item "pairwise": Each factor level is compared to all other levels
\item "minmax": The highest and lowest levels of a factor.
\item "revpairwise", "revreference", "revsequential": inverse of the corresponding hypotheses.
\item Vector of length 2 with the two values to compare.
\item Data frame with the same number of rows as \code{newdata}, with two columns of "lo" and "hi" values to compare.
\item Function that accepts a vector and returns a data frame with two columns of "lo" and "hi" values to compare. See examples below.
}
\item Logical variables:
\itemize{
\item NULL: contrast between TRUE and FALSE
\item Data frame with the same number of rows as \code{newdata}, with two columns of "lo" and "hi" values to compare.
\item Function that accepts a vector and returns a data frame with two columns of "lo" and "hi" values to compare. See examples below.
}
\item Numeric variables:
\itemize{
\item Numeric of length 1: Forward contrast for a gap of \code{x}, computed between the observed value and the observed value plus \code{x}. Users can set a global option to get a "center" or "backward" contrast instead: \code{options(marginaleffects_contrast_direction="center")}
\item Numeric vector of length 2: Contrast between the largest and the smallest elements of the \code{x} vector.
\item Data frame with the same number of rows as \code{newdata}, with two columns of "lo" and "hi" values to compare.
\item Function that accepts a vector and returns a data frame with two columns of "lo" and "hi" values to compare. See examples below.
\item "iqr": Contrast across the interquartile range of the regressor.
\item "sd": Contrast across one standard deviation around the regressor mean.
\item "2sd": Contrast across two standard deviations around the regressor mean.
\item "minmax": Contrast between the maximum and the minimum values of the regressor.
}
\item Examples:
\itemize{
\item \code{variables = list(gear = "pairwise", hp = 10)}
\item \code{variables = list(gear = "sequential", hp = c(100, 120))}
\item \verb{variables = list(hp = \\(x) data.frame(low = x - 5, high = x + 10))}
\item See the Examples section below for more.
}
}
}}

\item{comparison}{How should pairs of predictions be compared? Difference, ratio, odds ratio, or user-defined functions.
\itemize{
\item string: shortcuts to common contrast functions.
\itemize{
\item Supported shortcuts strings: difference, differenceavg, differenceavgwts, dydx, eyex, eydx, dyex, dydxavg, eyexavg, eydxavg, dyexavg, dydxavgwts, eyexavgwts, eydxavgwts, dyexavgwts, ratio, ratioavg, ratioavgwts, lnratio, lnratioavg, lnratioavgwts, lnor, lnoravg, lnoravgwts, lift, liftavg, liftavgwts, expdydx, expdydxavg, expdydxavgwts
\item See the Comparisons section below for definitions of each transformation.
}
\item function: accept two equal-length numeric vectors of adjusted predictions (\code{hi} and \code{lo}) and returns a vector of contrasts of the same length, or a unique numeric value.
\itemize{
\item See the Transformations section below for examples of valid functions.
}
}}

\item{type}{string indicates the type (scale) of the predictions used to
compute contrasts or slopes. This can differ based on the model
type, but will typically be a string such as: "response", "link", "probs",
or "zero". When an unsupported string is entered, the model-specific list of
acceptable values is returned in an error message. When \code{type} is \code{NULL}, the
first entry in the error message is used by default.}

\item{vcov}{Type of uncertainty estimates to report (e.g., for robust standard errors). Acceptable values:
\itemize{
\item FALSE: Do not compute standard errors. This can speed up computation considerably.
\item TRUE: Unit-level standard errors using the default \code{vcov(model)} variance-covariance matrix.
\item String which indicates the kind of uncertainty estimates to return.
\itemize{
\item Heteroskedasticity-consistent: \code{"HC"}, \code{"HC0"}, \code{"HC1"}, \code{"HC2"}, \code{"HC3"}, \code{"HC4"}, \code{"HC4m"}, \code{"HC5"}. See \code{?sandwich::vcovHC}
\item Heteroskedasticity and autocorrelation consistent: \code{"HAC"}
\item Mixed-Models degrees of freedom: "satterthwaite", "kenward-roger"
\item Other: \code{"NeweyWest"}, \code{"KernHAC"}, \code{"OPG"}. See the \code{sandwich} package documentation.
}
\item One-sided formula which indicates the name of cluster variables (e.g., \code{~unit_id}). This formula is passed to the \code{cluster} argument of the \code{sandwich::vcovCL} function.
\item Square covariance matrix
\item Function which returns a covariance matrix (e.g., \code{stats::vcov(model)})
}}

\item{by}{Aggregate unit-level estimates (aka, marginalize, average over). Valid inputs:
\itemize{
\item \code{FALSE}: return the original unit-level estimates.
\item \code{TRUE}: aggregate estimates for each term.
\item Character vector of column names in \code{newdata} or in the data frame produced by calling the function without the \code{by} argument.
\item Data frame with a \code{by} column of group labels, and merging columns shared by \code{newdata} or the data frame produced by calling the same function without the \code{by} argument.
\item See examples below.
\item For more complex aggregations, you can use the \code{FUN} argument of the \code{hypotheses()} function. See that function's documentation and the Hypothesis Test vignettes on the \code{marginaleffects} website.
}}

\item{conf_level}{numeric value between 0 and 1. Confidence level to use to build a confidence interval.}

\item{transform}{string or function. Transformation applied to unit-level estimates and confidence intervals just before the function returns results. Functions must accept a vector and return a vector of the same length. Support string shortcuts: "exp", "ln"}

\item{cross}{\itemize{
\item \code{FALSE}: Contrasts represent the change in adjusted predictions when one predictor changes and all other variables are held constant.
\item \code{TRUE}: Contrasts represent the changes in adjusted predictions when all the predictors specified in the \code{variables} argument are manipulated simultaneously (a "cross-contrast").
}}

\item{wts}{logical, string or numeric: weights to use when computing average predictions, contrasts or slopes. These weights only affect the averaging in \verb{avg_*()} or with the \code{by} argument, and not unit-level estimates. See \code{?weighted.mean}
\itemize{
\item string: column name of the weights variable in \code{newdata}. When supplying a column name to \code{wts}, it is recommended to supply the original data (including the weights variable) explicitly to \code{newdata}.
\item numeric: vector of length equal to the number of rows in the original data or in \code{newdata} (if supplied).
\item FALSE: Equal weights.
\item TRUE: Extract weights from the fitted object with \code{insight::find_weights()} and use them when taking weighted averages of estimates. Warning: \code{newdata=datagrid()} returns a single average weight, which is equivalent to using \code{wts=FALSE}
}}

\item{hypothesis}{specify a hypothesis test or custom contrast using a number , formula, string equation, vector, matrix, or function.
\itemize{
\item Number: The null hypothesis used in the computation of Z and p (before applying \code{transform}).
\item String: Equation to specify linear or non-linear hypothesis tests. If the terms in \code{coef(object)} uniquely identify estimates, they can be used in the formula. Otherwise, use \code{b1}, \code{b2}, etc. to identify the position of each parameter. The \verb{b*} wildcard can be used to test hypotheses on all estimates. If a named vector is used, the names are used as labels in the output. Examples:
\itemize{
\item \code{hp = drat}
\item \code{hp + drat = 12}
\item \code{b1 + b2 + b3 = 0}
\item \verb{b* / b1 = 1}
}
\item Formula: \code{lhs ~ rhs | group}
\itemize{
\item \code{lhs}
\itemize{
\item \code{ratio}
\item \code{difference}
\item Leave empty for default value
}
\item \code{rhs}
\itemize{
\item \code{pairwise} and \code{revpairwise}: pairwise differences between estimates in each row.
\item \code{reference}: differences between the estimates in each row and the estimate in the first row.
\item \code{sequential}: difference between an estimate and the estimate in the next row.
\item \code{meandev}: difference between an estimate and the mean of all estimates.
\item `meanotherdev: difference between an estimate and the mean of all other estimates, excluding the current one.
\item \code{poly}: polynomial contrasts, as computed by the \code{stats::contr.poly()} function.
\item \code{helmert}: Helmert contrasts, as computed by the \code{stats::contr.helmert()} function. Contrast 2nd level to the first, 3rd to the average of the first two, and so on.
\item \code{trt_vs_ctrl}: difference between the mean of estimates (except the first) and the first estimate.
\item \code{I(fun(x))}: custom function to manipulate the vector of estimates \code{x}. The function \code{fun()} can return multiple (potentially named) estimates.
}
\item \code{group} (optional)
\itemize{
\item Column name of \code{newdata}. Conduct hypothesis tests withing subsets of the data.
}
\item Examples:
\itemize{
\item \code{~ poly}
\item \code{~ sequential | groupid}
\item \code{~ reference}
\item \code{ratio ~ pairwise}
\item \code{difference ~ pairwise | groupid}
\item \code{~ I(x - mean(x)) | groupid}
\item \verb{~ I(\\(x) c(a = x[1], b = mean(x[2:3]))) | groupid}
}
}
\item Matrix or Vector: Each column is a vector of weights. The the output is the dot product between these vectors of weights and the vector of estimates. The matrix can have column names to label the estimates.
\item Function:
\itemize{
\item Accepts an argument \code{x}: object produced by a \code{marginaleffects} function or a data frame with column \code{rowid} and \code{estimate}
\item Returns a data frame with columns \code{term} and \code{estimate} (mandatory) and \code{rowid} (optional).
\item The function can also accept optional input arguments: \code{newdata}, \code{by}, \code{draws}.
\item This function approach will not work for Bayesian models or with bootstrapping. In those cases, it is easy to use \code{get_draws()} to extract and manipulate the draws directly.
}
\item See the Examples section below and the vignette: \url{https://marginaleffects.com/chapters/hypothesis.html}
}}

\item{equivalence}{Numeric vector of length 2: bounds used for the two-one-sided test (TOST) of equivalence, and for the non-inferiority and non-superiority tests. See Details section below.}

\item{df}{Degrees of freedom used to compute p values and confidence intervals. A single numeric value between 1 and \code{Inf}. When \code{df} is \code{Inf}, the normal distribution is used. When \code{df} is finite, the \code{t} distribution is used. See \link[insight:get_df]{insight::get_df} for a convenient function to extract degrees of freedom. Ex: \code{slopes(model, df = insight::get_df(model))}}

\item{eps}{NULL or numeric value which determines the step size to use when
calculating numerical derivatives: (f(x+eps)-f(x))/eps. When \code{eps} is
\code{NULL}, the step size is 0.0001 multiplied by the difference between
the maximum and minimum values of the variable with respect to which we
are taking the derivative. Changing \code{eps} may be necessary to avoid
numerical problems in certain models.}

\item{numderiv}{string or list of strings indicating the method to use to for the numeric differentiation used in to compute delta method standard errors.
\itemize{
\item "fdforward": finite difference method with forward differences
\item "fdcenter": finite difference method with central differences (default)
\item "richardson": Richardson extrapolation method
\item Extra arguments can be specified by passing a list to the \code{numDeriv} argument, with the name of the method first and named arguments following, ex: \code{numderiv=list("fdcenter", eps = 1e-5)}. When an unknown argument is used, \code{marginaleffects} prints the list of valid arguments for each method.
}}

\item{...}{Additional arguments are passed to the \code{predict()} method
supplied by the modeling package.These arguments are particularly useful
for mixed-effects or bayesian models (see the online vignettes on the
\code{marginaleffects} website). Available arguments can vary from model to
model, depending on the range of supported arguments by each modeling
package. See the "Model-Specific Arguments" section of the
\code{?slopes} documentation for a non-exhaustive list of available
arguments.}
}
\value{
A \code{data.frame} with one row per estimate. This data frame is pretty-printed by default, but users can interact with it as a regular data frame, with functions like \code{nrow()}, \code{head()}, \code{colnames()}, etc. Values can be extracted using standard \verb{[,]} or \code{$} operators, and manipulated using external packages like \code{dplyr} or \code{data.table}.

Columns may include:
\itemize{
\item \code{rowid}: row number of the \code{newdata} data frame
\item \code{group}: (optional) value of the grouped outcome (e.g., categorical outcome models)
\item \code{term}: the focal variable.
\item \code{estimate}: an estimate of the prediction, counterfactual comparison, or slope.
\item \code{std.error}: standard errors computed via the delta method.
\item \code{p.value}: p value associated to the \code{estimate} column. The null is determined by the \code{hypothesis} argument (0 by default).
\item \code{s.value}: Shannon information transforms of p values. See the S values vignette at \url{https://marginaleffects.com} the marginaleffects website.
\item \code{conf.low}: lower bound of the confidence (or credible) interval defined by the \code{conf_level} argument.
\item \code{conf.high}: upper bound of the confidence (or credible) interval defined by the \code{conf_level} argument.
\item \code{predicted_lo}: predicted outcome for the "low" value of the focal predictor in a counterfactual comparison.
\item \code{predicted_hi}: predicted outcome for the "high" value of the focal predictor in a counterfactual comparison.
}

See \code{?print.marginaleffects} for printing options.
}
\description{
Predict the outcome variable at different regressor values (e.g., college
graduates vs. others), and compare those predictions by computing a difference,
ratio, or some other function. \code{comparisons()} can return many quantities of
interest, such as contrasts, differences, risk ratios, changes in log odds, lift,
slopes, elasticities, etc.
\itemize{
\item \code{comparisons()}: unit-level (conditional) estimates.
\item \code{avg_comparisons()}: average (marginal) estimates.
}

\code{variables} identifies the focal regressors whose "effect" we are interested in. \code{comparison} determines how predictions with different regressor values are compared (difference, ratio, odds, etc.). The \code{newdata} argument and the \code{datagrid()} function control where statistics are evaluated in the predictor space: "at observed values", "at the mean", "at representative values", etc.

See the comparisons vignette and package website for worked examples and case studies:
\itemize{
\item \url{https://marginaleffects.com/chapters/comparisons.html}
\item \url{https://marginaleffects.com/}
}
}
\section{Functions}{
\itemize{
\item \code{avg_comparisons()}: Average comparisons

}}
\section{Standard errors using the delta method}{


Standard errors for all quantities estimated by \code{marginaleffects} can be obtained via the delta method. This requires differentiating a function with respect to the coefficients in the model using a finite difference approach. In some models, the delta method standard errors can be sensitive to various aspects of the numeric differentiation strategy, including the step size. By default, the step size is set to \code{1e-8}, or to \code{1e-4} times the smallest absolute model coefficient, whichever is largest.

\code{marginaleffects} can delegate numeric differentiation to the \code{numDeriv} package, which allows more flexibility. To do this, users can pass arguments to the \code{numDeriv::jacobian} function through a global option. For example:
\itemize{
\item \code{options(marginaleffects_numDeriv = list(method = "simple", method.args = list(eps = 1e-6)))}
\item \code{options(marginaleffects_numDeriv = list(method = "Richardson", method.args = list(eps = 1e-5)))}
\item \code{options(marginaleffects_numDeriv = NULL)}
}

See the "Standard Errors and Confidence Intervals" vignette on the \code{marginaleffects} website for more details on the computation of standard errors:

https://marginaleffects.com/vignettes/uncertainty.html

Note that the \code{inferences()} function can be used to compute uncertainty estimates using a bootstrap or simulation-based inference. See the vignette:

https://marginaleffects.com/vignettes/bootstrap.html
}

\section{Model-Specific Arguments}{


Some model types allow model-specific arguments to modify the nature of
marginal effects, predictions, marginal means, and contrasts. Please report
other package-specific \code{predict()} arguments on Github so we can add them to
the table below.

https://github.com/vincentarelbundock/marginaleffects/issues\tabular{llll}{
   Package \tab Class \tab Argument \tab Documentation \cr
   \code{brms} \tab \code{brmsfit} \tab \code{ndraws} \tab \link[brms:posterior_predict.brmsfit]{brms::posterior_predict} \cr
    \tab  \tab \code{re_formula} \tab \link[brms:posterior_predict.brmsfit]{brms::posterior_predict} \cr
   \code{lme4} \tab \code{merMod} \tab \code{re.form} \tab \link[lme4:predict.merMod]{lme4::predict.merMod} \cr
    \tab  \tab \code{allow.new.levels} \tab \link[lme4:predict.merMod]{lme4::predict.merMod} \cr
   \code{glmmTMB} \tab \code{glmmTMB} \tab \code{re.form} \tab \link[glmmTMB:predict.glmmTMB]{glmmTMB::predict.glmmTMB} \cr
    \tab  \tab \code{allow.new.levels} \tab \link[glmmTMB:predict.glmmTMB]{glmmTMB::predict.glmmTMB} \cr
    \tab  \tab \code{zitype} \tab \link[glmmTMB:predict.glmmTMB]{glmmTMB::predict.glmmTMB} \cr
   \code{mgcv} \tab \code{bam} \tab \code{exclude} \tab \link[mgcv:predict.bam]{mgcv::predict.bam} \cr
    \tab \code{gam} \tab \code{exclude} \tab \link[mgcv:predict.gam]{mgcv::predict.gam} \cr
   \code{robustlmm} \tab \code{rlmerMod} \tab \code{re.form} \tab \link[robustlmm:rlmerMod-class]{robustlmm::predict.rlmerMod} \cr
    \tab  \tab \code{allow.new.levels} \tab \link[robustlmm:rlmerMod-class]{robustlmm::predict.rlmerMod} \cr
   \code{MCMCglmm} \tab \code{MCMCglmm} \tab \code{ndraws} \tab  \cr
   \code{sampleSelection} \tab \code{selection} \tab \code{part} \tab \link[sampleSelection:predict.selection]{sampleSelection::predict.selection} \cr
}
}

\section{comparison argument functions}{


The following transformations can be applied by supplying one of the shortcut strings to the
\code{comparison} argument.
\code{hi} is a vector of adjusted predictions for the "high" side of the
contrast. \code{lo} is a vector of adjusted predictions for the "low" side of the
contrast. \code{y} is a vector of adjusted predictions for the original data. \code{x}
is the predictor in the original data. \code{eps} is the step size to use to
compute derivatives and elasticities.\tabular{ll}{
   Shortcut \tab Function \cr
   difference \tab \(hi, lo) hi - lo \cr
   differenceavg \tab \(hi, lo) mean(hi - lo) \cr
   dydx \tab \(hi, lo, eps) (hi - lo)/eps \cr
   eyex \tab \(hi, lo, eps, y, x) (hi - lo)/eps * (x/y) \cr
   eydx \tab \(hi, lo, eps, y, x) ((hi - lo)/eps)/y \cr
   dyex \tab \(hi, lo, eps, x) ((hi - lo)/eps) * x \cr
   dydxavg \tab \(hi, lo, eps) mean((hi - lo)/eps) \cr
   eyexavg \tab \(hi, lo, eps, y, x) mean((hi - lo)/eps * (x/y)) \cr
   eydxavg \tab \(hi, lo, eps, y, x) mean(((hi - lo)/eps)/y) \cr
   dyexavg \tab \(hi, lo, eps, x) mean(((hi - lo)/eps) * x) \cr
   ratio \tab \(hi, lo) hi/lo \cr
   ratioavg \tab \(hi, lo) mean(hi)/mean(lo) \cr
   lnratio \tab \(hi, lo) log(hi/lo) \cr
   lnratioavg \tab \(hi, lo) log(mean(hi)/mean(lo)) \cr
   lnor \tab \(hi, lo) log((hi/(1 - hi))/(lo/(1 - lo))) \cr
   lnoravg \tab \(hi, lo) log((mean(hi)/(1 - mean(hi)))/(mean(lo)/(1 - mean(lo)))) \cr
   lift \tab \(hi, lo) (hi - lo)/lo \cr
   liftavg \tab \(hi, lo) (mean(hi - lo))/mean(lo) \cr
   expdydx \tab \(hi, lo, eps) ((exp(hi) - exp(lo))/exp(eps))/eps \cr
   expdydxavg \tab \(hi, lo, eps) mean(((exp(hi) - exp(lo))/exp(eps))/eps) \cr
}
}

\section{Bayesian posterior summaries}{


By default, credible intervals in bayesian models are built as equal-tailed
intervals. This can be changed to a highest density interval by setting a global
option:

\code{options("marginaleffects_posterior_interval" = "eti")}

\code{options("marginaleffects_posterior_interval" = "hdi")}

By default, the center of the posterior distribution in bayesian models is
identified by the median. Users can use a different summary function by setting a
global option:

\code{options("marginaleffects_posterior_center" = "mean")}

\code{options("marginaleffects_posterior_center" = "median")}

When estimates are averaged using the \code{by} argument, the \code{tidy()} function, or
the \code{summary()} function, the posterior distribution is marginalized twice over.
First, we take the average \emph{across} units but \emph{within} each iteration of the
MCMC chain, according to what the user requested in \code{by} argument or
\code{tidy()/summary()} functions. Then, we identify the center of the resulting
posterior using the function supplied to the
\code{"marginaleffects_posterior_center"} option (the median by default).
}

\section{Equivalence, Inferiority, Superiority}{


\eqn{\theta} is an estimate, \eqn{\sigma_\theta} its estimated standard error, and \eqn{[a, b]} are the bounds of the interval supplied to the \code{equivalence} argument.

Non-inferiority:
\itemize{
\item \eqn{H_0}{H0}: \eqn{\theta \leq a}{\theta <= a}
\item \eqn{H_1}{H1}: \eqn{\theta > a}
\item \eqn{t=(\theta - a)/\sigma_\theta}{t=(\theta - a)/\sigma_\theta}
\item p: Upper-tail probability
}

Non-superiority:
\itemize{
\item \eqn{H_0}{H0}: \eqn{\theta \geq b}{\theta >= b}
\item \eqn{H_1}{H1}: \eqn{\theta < b}
\item \eqn{t=(\theta - b)/\sigma_\theta}{t=(\theta - b)/\sigma_\theta}
\item p: Lower-tail probability
}

Equivalence: Two One-Sided Tests (TOST)
\itemize{
\item p: Maximum of the non-inferiority and non-superiority p values.
}

Thanks to Russell V. Lenth for the excellent \code{emmeans} package and documentation which inspired this feature.
}

\section{Prediction types}{


The \code{type} argument determines the scale of the predictions used to compute quantities of interest with functions from the \code{marginaleffects} package. Admissible values for \code{type} depend on the model object. When users specify an incorrect value for \code{type}, \code{marginaleffects} will raise an informative error with a list of valid \code{type} values for the specific model object. The first entry in the list in that error message is the default type.

The \code{invlink(link)} is a special type defined by \code{marginaleffects}. It is available for some (but not all) models, and only for the \code{predictions()} function. With this link type, we first compute predictions on the link scale, then we use the inverse link function to backtransform the predictions to the response scale. This is useful for models with non-linear link functions as it can ensure that confidence intervals stay within desirable bounds, ex: 0 to 1 for a logit model. Note that an average of estimates with \code{type="invlink(link)"} will not always be equivalent to the average of estimates with \code{type="response"}. This type is default when calling \code{predictions()}. It is available---but not default---when calling \code{avg_predictions()} or \code{predictions()} with the \code{by} argument.

Some of the most common \code{type} values are:

response, link, E, Ep, average, class, conditional, count, cum.prob, cumhaz, cumprob, density, detection, disp, ev, expected, expvalue, fitted, hazard, invlink(link), latent, latent_N, linear, linear.predictor, linpred, location, lp, mean, numeric, p, ppd, pr, precision, prediction, prob, probability, probs, quantile, risk, rmst, scale, survival, unconditional, utility, variance, xb, zero, zlink, zprob
}

\section{Order of operations}{


Behind the scenes, the arguments of \code{marginaleffects} functions are evaluated in this order:
\enumerate{
\item \code{newdata}
\item \code{variables}
\item \code{comparison} and \code{slope}
\item \code{by}
\item \code{vcov}
\item \code{hypothesis}
\item \code{transform}
}
}

\section{Parallel computation}{


The \code{slopes()} and \code{comparisons()} functions can use parallelism to
speed up computation. Operations are parallelized for the computation of
standard errors, at the model coefficient level. There is always
considerable overhead when using parallel computation, mainly involved
in passing the whole dataset to the different processes. Thus, parallel
computation is most likely to be useful when the model includes many parameters
and the dataset is relatively small.

Warning: In many cases, parallel processing will not be useful at all.

To activate parallel computation, users must load the \code{future.apply} package,
call \code{plan()} function, and set a global option. For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(future.apply)
plan("multicore", workers = 4)
options(marginaleffects_parallel = TRUE)

slopes(model)
}\if{html}{\out{</div>}}

To disable parallelism in \code{marginaleffects} altogether, you can set a global option:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(marginaleffects_parallel = FALSE)
}\if{html}{\out{</div>}}
}

\section{Global options}{


The behavior of \code{marginaleffects} functions can be modified by setting global options.

Disable some safety checks and warnings:
\itemize{
\item \code{options(marginaleffects_startup_message = FALSE)}
\itemize{
\item Disable the startup message printed on \code{library(marginaleffects)}.
}
\item \code{options(marginaleffects_safe = FALSE)}
\itemize{
\item Disable safety checks and warnings.
}
\item \code{options(marginaleffects_print_omit = c("p.value", "s.value"))}
\itemize{
\item Omit some columns from the printed output.
}
}

Enforce lean return objects, sans information about the original model and
data, and other ancillary attributes. Note that this will disable some
advanced post-processing features and functions like \link{hypotheses}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(marginaleffects_lean = TRUE)`
}\if{html}{\out{</div>}}

Other options:
\itemize{
\item \code{marginaleffects_plot_gray}: logical. If \code{TRUE}, the default color of the plot is gray. Default is \code{FALSE}.
}
}

\examples{
\dontshow{if (interactive() || isTRUE(Sys.getenv("R_DOC_BUILD") == "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
library(marginaleffects)

# Linear model
tmp <- mtcars
tmp$am <- as.logical(tmp$am)
mod <- lm(mpg ~ am + factor(cyl), tmp)
avg_comparisons(mod, variables = list(cyl = "reference"))
avg_comparisons(mod, variables = list(cyl = "sequential"))
avg_comparisons(mod, variables = list(cyl = "pairwise"))

# GLM with different scale types
mod <- glm(am ~ factor(gear), data = mtcars)
avg_comparisons(mod, type = "response")
avg_comparisons(mod, type = "link")

# Contrasts at the mean
comparisons(mod, newdata = "mean")

# Contrasts between marginal means
comparisons(mod, newdata = "balanced")

# Contrasts at user-specified values
comparisons(mod, newdata = datagrid(am = 0, gear = tmp$gear))
comparisons(mod, newdata = datagrid(am = unique, gear = max))

m <- lm(mpg ~ hp + drat + factor(cyl) + factor(am), data = mtcars)
comparisons(m, variables = "hp", newdata = datagrid(FUN_factor = unique, FUN_numeric = median))

# Numeric contrasts
mod <- lm(mpg ~ hp, data = mtcars)
avg_comparisons(mod, variables = list(hp = 1))
avg_comparisons(mod, variables = list(hp = 5))
avg_comparisons(mod, variables = list(hp = c(90, 100)))
avg_comparisons(mod, variables = list(hp = "iqr"))
avg_comparisons(mod, variables = list(hp = "sd"))
avg_comparisons(mod, variables = list(hp = "minmax"))

# using a function to specify a custom difference in one regressor
dat <- mtcars
dat$new_hp <- 49 * (dat$hp - min(dat$hp)) / (max(dat$hp) - min(dat$hp)) + 1
modlog <- lm(mpg ~ log(new_hp) + factor(cyl), data = dat)
fdiff <- \(x) data.frame(x, x + 10)
avg_comparisons(modlog, variables = list(new_hp = fdiff))

# Adjusted Risk Ratio: see the contrasts vignette
mod <- glm(vs ~ mpg, data = mtcars, family = binomial)
avg_comparisons(mod, comparison = "lnratioavg", transform = exp)

# Adjusted Risk Ratio: Manual specification of the `comparison`
avg_comparisons(
  mod,
  comparison = function(hi, lo) log(mean(hi) / mean(lo)),
  transform = exp)
# cross contrasts
mod <- lm(mpg ~ factor(cyl) * factor(gear) + hp, data = mtcars)
avg_comparisons(mod, variables = c("cyl", "gear"), cross = TRUE)

# variable-specific contrasts
avg_comparisons(mod, variables = list(gear = "sequential", hp = 10))

# hypothesis test: is the `hp` marginal effect at the mean equal to the `drat` marginal effect
mod <- lm(mpg ~ wt + drat, data = mtcars)

comparisons(
  mod,
  newdata = "mean",
  hypothesis = "wt = drat")

# same hypothesis test using row indices
comparisons(
  mod,
  newdata = "mean",
  hypothesis = "b1 - b2 = 0")

# same hypothesis test using numeric vector of weights
comparisons(
  mod,
  newdata = "mean",
  hypothesis = c(1, -1))

# two custom contrasts using a matrix of weights
lc <- matrix(
  c(
    1, -1,
    2, 3),
  ncol = 2)
comparisons(
  mod,
  newdata = "mean",
  hypothesis = lc)

# Effect of a 1 group-wise standard deviation change
# First we calculate the SD in each group of `cyl`
# Second, we use that SD as the treatment size in the `variables` argument
library(dplyr)
mod <- lm(mpg ~ hp + factor(cyl), mtcars)
tmp <- mtcars \%>\%
  group_by(cyl) \%>\%
  mutate(hp_sd = sd(hp))
avg_comparisons(mod,
  variables = list(hp = function(x) data.frame(x, x + tmp$hp_sd)),
  by = "cyl")

# `by` argument
mod <- lm(mpg ~ hp * am * vs, data = mtcars)
comparisons(mod, by = TRUE)

mod <- lm(mpg ~ hp * am * vs, data = mtcars)
avg_comparisons(mod, variables = "hp", by = c("vs", "am"))

library(nnet)
mod <- multinom(factor(gear) ~ mpg + am * vs, data = mtcars, trace = FALSE)
by <- data.frame(
  group = c("3", "4", "5"),
  by = c("3,4", "3,4", "5"))
comparisons(mod, type = "probs", by = by)
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item Arel-Bundock V, Greifer N, Heiss A (2024). “How to Interpret Statistical Models Using marginaleffects for R and Python.” \emph{Journal of Statistical Software}, \emph{111}(9), 1-32. doi:10.18637/jss.v111.i09 \doi{10.18637/jss.v111.i09}
\item Greenland S. 2019. "Valid P-Values Behave Exactly as They Should: Some Misleading Criticisms of P-Values and Their Resolution With S-Values." The American Statistician. 73(S1): 106–114.
\item Cole, Stephen R, Jessie K Edwards, and Sander Greenland. 2020. "Surprise!" American Journal of Epidemiology 190 (2): 191–93. \doi{10.1093/aje/kwaa136}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinytest.R
\name{expect_hypotheses}
\alias{expect_hypotheses}
\title{\code{tinytest} helper}
\usage{
expect_hypotheses(object, se = TRUE, n_row = NULL, n_col = NULL, ...)
}
\description{
\code{tinytest} helper
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.marginaleffects}
\alias{print.marginaleffects}
\title{Print \code{marginaleffects} objects}
\usage{
\method{print}{marginaleffects}(
  x,
  style = getOption("marginaleffects_print_style", default = "summary"),
  digits = getOption("marginaleffects_print_digits", default = 3),
  p_eps = getOption("marginaleffects_print_p_eps", default = 0.001),
  topn = getOption("marginaleffects_print_topn", default = 5),
  nrows = getOption("marginaleffects_print_nrows", default = 30),
  ncols = getOption("marginaleffects_print_ncols", default = 30),
  type = getOption("marginaleffects_print_type", default = TRUE),
  column_names = getOption("marginaleffects_print_column_names", default = FALSE),
  ...
)
}
\arguments{
\item{x}{An object produced by one of the \code{marginaleffects} package functions.}

\item{style}{"summary", "data.frame", or "tinytable"}

\item{digits}{The number of digits to display.}

\item{p_eps}{p values smaller than this number are printed in "<0.001" style.}

\item{topn}{The number of rows to be printed from the beginning and end of tables with more than \code{nrows} rows.}

\item{nrows}{The number of rows which will be printed before truncation.}

\item{ncols}{The maximum number of column names to display at the bottom of the printed output.}

\item{type}{boolean: should the type be printed?}

\item{column_names}{boolean: should the column names be printed?}

\item{...}{Other arguments are currently ignored.}
}
\description{
This function controls the text which is printed to the console when one of the core \code{marginalefffects} functions is called and the object is returned: \code{predictions()}, \code{comparisons()}, \code{slopes()}, \code{hypotheses()}, \code{avg_predictions()}, \code{avg_comparisons()}, \code{avg_slopes()}.

All of those functions return standard data frames. Columns can be extracted by name, \code{predictions(model)$estimate}, and all the usual data manipulation functions work out-of-the-box:  \code{colnames()}, \code{head()}, \code{subset()}, \code{dplyr::filter()}, \code{dplyr::arrange()}, etc.

Some of the data columns are not printed by default. You can disable pretty printing and print the full results as a standard data frame using the \code{style} argument or by applying \code{as.data.frame()} on the object. See examples below.
}
\examples{
library(marginaleffects)
mod <- lm(mpg ~ hp + am + factor(gear), data = mtcars)
p <- predictions(mod, by = c("am", "gear"))
p

subset(p, am == 1)

print(p, style = "data.frame")

data.frame(p)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/components.R
\name{components.marginaleffects}
\alias{components.marginaleffects}
\title{Extract components from marginaleffects objects}
\usage{
\method{components}{marginaleffects}(object, component = NULL, ...)
}
\arguments{
\item{object}{A marginaleffects object (predictions, comparisons, slopes, or hypotheses)}

\item{component}{Character string specifying which component to extract. Must be a valid
slot name from the internal S4 object. If \code{NULL} (the default), \code{components()} prints a message with
all available component names. Common components include: "model", "newdata",
"modeldata", "call", "jacobian", "vcov_model", "type", "by", "comparison", "variables", etc.}

\item{...}{Ignored.}
}
\value{
The requested component from the mfx object
}
\description{
Extract components from marginaleffects objects
}
\details{
This function provides access to the internal components stored in the \code{mfx}
attribute of marginaleffects objects. The \code{mfx} attribute contains an S4 object of
class "marginaleffects_internal" with various slots containing model information,
data, and computational details used by the marginaleffects functions.

Warning: the internal slot names are not considered part of the public API and may change
without warning in future versions of the marginaleffects package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_vcov.R
\name{get_varcov_args}
\alias{get_varcov_args}
\title{Take a \code{summary()} style \code{vcov} argument and convert it to
\code{insight::get_varcov()}}
\usage{
get_varcov_args(model, vcov)
}
\description{
Take a \code{summary()} style \code{vcov} argument and convert it to
\code{insight::get_varcov()}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hypotheses.R
\name{hypotheses}
\alias{hypotheses}
\title{(Non-)Linear Tests for Null Hypotheses, Joint Hypotheses, Equivalence, Non Superiority, and Non Inferiority}
\usage{
hypotheses(
  model,
  hypothesis = NULL,
  vcov = NULL,
  conf_level = NULL,
  df = NULL,
  equivalence = NULL,
  joint = FALSE,
  joint_test = "f",
  multcomp = FALSE,
  numderiv = "fdforward",
  ...
)
}
\arguments{
\item{model}{Model object or object generated by the \code{comparisons()}, \code{slopes()}, or \code{predictions()} functions.}

\item{hypothesis}{specify a hypothesis test or custom contrast using a number , formula, string equation, vector, matrix, or function.
\itemize{
\item Number: The null hypothesis used in the computation of Z and p (before applying \code{transform}).
\item String: Equation to specify linear or non-linear hypothesis tests. If the terms in \code{coef(object)} uniquely identify estimates, they can be used in the formula. Otherwise, use \code{b1}, \code{b2}, etc. to identify the position of each parameter. The \verb{b*} wildcard can be used to test hypotheses on all estimates. If a named vector is used, the names are used as labels in the output. Examples:
\itemize{
\item \code{hp = drat}
\item \code{hp + drat = 12}
\item \code{b1 + b2 + b3 = 0}
\item \verb{b* / b1 = 1}
}
\item Formula: \code{lhs ~ rhs | group}
\itemize{
\item \code{lhs}
\itemize{
\item \code{ratio}
\item \code{difference}
\item Leave empty for default value
}
\item \code{rhs}
\itemize{
\item \code{pairwise} and \code{revpairwise}: pairwise differences between estimates in each row.
\item \code{reference}: differences between the estimates in each row and the estimate in the first row.
\item \code{sequential}: difference between an estimate and the estimate in the next row.
\item \code{meandev}: difference between an estimate and the mean of all estimates.
\item `meanotherdev: difference between an estimate and the mean of all other estimates, excluding the current one.
\item \code{poly}: polynomial contrasts, as computed by the \code{stats::contr.poly()} function.
\item \code{helmert}: Helmert contrasts, as computed by the \code{stats::contr.helmert()} function. Contrast 2nd level to the first, 3rd to the average of the first two, and so on.
\item \code{trt_vs_ctrl}: difference between the mean of estimates (except the first) and the first estimate.
\item \code{I(fun(x))}: custom function to manipulate the vector of estimates \code{x}. The function \code{fun()} can return multiple (potentially named) estimates.
}
\item \code{group} (optional)
\itemize{
\item Column name of \code{newdata}. Conduct hypothesis tests withing subsets of the data.
}
\item Examples:
\itemize{
\item \code{~ poly}
\item \code{~ sequential | groupid}
\item \code{~ reference}
\item \code{ratio ~ pairwise}
\item \code{difference ~ pairwise | groupid}
\item \code{~ I(x - mean(x)) | groupid}
\item \verb{~ I(\\(x) c(a = x[1], b = mean(x[2:3]))) | groupid}
}
}
\item Matrix or Vector: Each column is a vector of weights. The the output is the dot product between these vectors of weights and the vector of estimates. The matrix can have column names to label the estimates.
\item Function:
\itemize{
\item Accepts an argument \code{x}: object produced by a \code{marginaleffects} function or a data frame with column \code{rowid} and \code{estimate}
\item Returns a data frame with columns \code{term} and \code{estimate} (mandatory) and \code{rowid} (optional).
\item The function can also accept optional input arguments: \code{newdata}, \code{by}, \code{draws}.
\item This function approach will not work for Bayesian models or with bootstrapping. In those cases, it is easy to use \code{get_draws()} to extract and manipulate the draws directly.
}
\item See the Examples section below and the vignette: \url{https://marginaleffects.com/chapters/hypothesis.html}
}}

\item{vcov}{Type of uncertainty estimates to report (e.g., for robust standard errors). Acceptable values:
\itemize{
\item FALSE: Do not compute standard errors. This can speed up computation considerably.
\item TRUE: Unit-level standard errors using the default \code{vcov(model)} variance-covariance matrix.
\item String which indicates the kind of uncertainty estimates to return.
\itemize{
\item Heteroskedasticity-consistent: \code{"HC"}, \code{"HC0"}, \code{"HC1"}, \code{"HC2"}, \code{"HC3"}, \code{"HC4"}, \code{"HC4m"}, \code{"HC5"}. See \code{?sandwich::vcovHC}
\item Heteroskedasticity and autocorrelation consistent: \code{"HAC"}
\item Mixed-Models degrees of freedom: "satterthwaite", "kenward-roger"
\item Other: \code{"NeweyWest"}, \code{"KernHAC"}, \code{"OPG"}. See the \code{sandwich} package documentation.
}
\item One-sided formula which indicates the name of cluster variables (e.g., \code{~unit_id}). This formula is passed to the \code{cluster} argument of the \code{sandwich::vcovCL} function.
\item Square covariance matrix
\item Function which returns a covariance matrix (e.g., \code{stats::vcov(model)})
}}

\item{conf_level}{NULL or numeric value between 0 and 1. Confidence level to use to build a confidence interval. When \code{NULL} and \code{model} was generated by \code{marginaleffects}, the confidence level is taken from the \code{conf_level} attribute of the model. Otherwise, the default value is 0.95.}

\item{df}{Degrees of freedom used to compute p values and confidence intervals. A single numeric value between 1 and \code{Inf}. When using \code{joint_test="f"}, the \code{df} argument should be a numeric vector of length 2.}

\item{equivalence}{Numeric vector of length 2: bounds used for the two-one-sided test (TOST) of equivalence, and for the non-inferiority and non-superiority tests. See Details section below.}

\item{joint}{Joint test of statistical significance. The null hypothesis value can be set using the \code{hypothesis} argument.
\itemize{
\item FALSE: Hypotheses are not tested jointly.
\item TRUE: All parameters are tested jointly.
\item String: A regular expression to match parameters to be tested jointly. \code{grep(joint, perl = TRUE)}
\item Character vector of parameter names to be tested. Characters refer to the names of the vector returned by \code{marginaleffects::get_coef(object)}.
\item Integer vector of indices. Which parameters positions to test jointly.
}}

\item{joint_test}{A character string specifying the type of test, either "f" or "chisq". The null hypothesis is set by the \code{hypothesis} argument, with default null equal to 0 for all parameters.}

\item{multcomp}{Logical or string. If \code{TRUE} or string, apply multiple comparison adjustment to the p values and report family-wise confidence intervals. Valid strings: "holm", "hochberg", "hommel", "bonferroni", "BH", "BY", "fdr", "single-step", "Shaffer", "Westfall", "free". When \code{multcomp} is \code{TRUE}, the "holm" method is used.}

\item{numderiv}{string or list of strings indicating the method to use to for the numeric differentiation used in to compute delta method standard errors.
\itemize{
\item "fdforward": finite difference method with forward differences
\item "fdcenter": finite difference method with central differences (default)
\item "richardson": Richardson extrapolation method
\item Extra arguments can be specified by passing a list to the \code{numDeriv} argument, with the name of the method first and named arguments following, ex: \code{numderiv=list("fdcenter", eps = 1e-5)}. When an unknown argument is used, \code{marginaleffects} prints the list of valid arguments for each method.
}}

\item{...}{Additional arguments are passed to the \code{predict()} method
supplied by the modeling package.These arguments are particularly useful
for mixed-effects or bayesian models (see the online vignettes on the
\code{marginaleffects} website). Available arguments can vary from model to
model, depending on the range of supported arguments by each modeling
package. See the "Model-Specific Arguments" section of the
\code{?slopes} documentation for a non-exhaustive list of available
arguments.}
}
\description{
Uncertainty estimates are calculated as first-order approximate standard errors for linear or non-linear functions of a vector of random variables with known or estimated covariance matrix. In that sense, \code{\link{hypotheses}} emulates the behavior of the excellent and well-established \link[car:deltaMethod]{car::deltaMethod} and \link[car:linearHypothesis]{car::linearHypothesis} functions, but it supports more models; requires fewer dependencies; expands the range of tests to equivalence and superiority/inferiority; and offers convenience features like robust standard errors.

To learn more, read the hypothesis tests vignette, visit the
package website, or scroll down this page for a full list of vignettes:
\itemize{
\item \url{https://marginaleffects.com/chapters/hypothesis.html}
\item \url{https://marginaleffects.com/}
}

Warning #1: Tests are conducted directly on the scale defined by the \code{type} argument. For some models, it can make sense to conduct hypothesis or equivalence tests on the \code{"link"} scale instead of the \code{"response"} scale which is often the default.

Warning #2: For hypothesis tests on objects produced by the \code{marginaleffects} package, it is safer to use the \code{hypothesis} argument of the original function.  Using \code{hypotheses()} may not work in certain environments, in lists, or when working programmatically with *apply style functions.

Warning #3: The tests assume that the \code{hypothesis} expression is (approximately) normally distributed, which for non-linear functions of the parameters may not be realistic. More reliable confidence intervals can be obtained using the \code{inferences()} function with \code{method = "boot"}.
}
\section{Joint hypothesis tests}{

The test statistic for the joint Wald test is calculated as (R * theta_hat - r)' * inv(R * V_hat * R') * (R * theta_hat - r) / Q,
where theta_hat is the vector of estimated parameters, V_hat is the estimated covariance matrix, R is a Q x P matrix for testing Q hypotheses on P parameters,
r is a Q x 1 vector for the null hypothesis, and Q is the number of rows in R. If the test is a Chi-squared test, the test statistic is not normalized.

The p-value is then calculated based on either the F-distribution (for F-test) or the Chi-squared distribution (for Chi-squared test).
For the F-test, the degrees of freedom are Q and (n - P), where n is the sample size and P is the number of parameters.
For the Chi-squared test, the degrees of freedom are Q.
}

\section{Equivalence, Inferiority, Superiority}{


\eqn{\theta} is an estimate, \eqn{\sigma_\theta} its estimated standard error, and \eqn{[a, b]} are the bounds of the interval supplied to the \code{equivalence} argument.

Non-inferiority:
\itemize{
\item \eqn{H_0}{H0}: \eqn{\theta \leq a}{\theta <= a}
\item \eqn{H_1}{H1}: \eqn{\theta > a}
\item \eqn{t=(\theta - a)/\sigma_\theta}{t=(\theta - a)/\sigma_\theta}
\item p: Upper-tail probability
}

Non-superiority:
\itemize{
\item \eqn{H_0}{H0}: \eqn{\theta \geq b}{\theta >= b}
\item \eqn{H_1}{H1}: \eqn{\theta < b}
\item \eqn{t=(\theta - b)/\sigma_\theta}{t=(\theta - b)/\sigma_\theta}
\item p: Lower-tail probability
}

Equivalence: Two One-Sided Tests (TOST)
\itemize{
\item p: Maximum of the non-inferiority and non-superiority p values.
}

Thanks to Russell V. Lenth for the excellent \code{emmeans} package and documentation which inspired this feature.
}

\examples{
library(marginaleffects)
mod <- lm(mpg ~ hp + wt + factor(cyl), data = mtcars)

hypotheses(mod)

# Test of equality between coefficients
hypotheses(mod, hypothesis = "hp = wt")

# Non-linear function
hypotheses(mod, hypothesis = "exp(hp + wt) = 0.1")

# Robust standard errors
hypotheses(mod, hypothesis = "hp = wt", vcov = "HC3")

# b1, b2, ... shortcuts can be used to identify the position of the
# parameters of interest in the output of
hypotheses(mod, hypothesis = "b2 = b3")

# wildcard
hypotheses(mod, hypothesis = "b* / b2 = 1")

# term names with special characters have to be enclosed in backticks
hypotheses(mod, hypothesis = "`factor(cyl)6` = `factor(cyl)8`")

mod2 <- lm(mpg ~ hp * drat, data = mtcars)
hypotheses(mod2, hypothesis = "`hp:drat` = drat")

# predictions(), comparisons(), and slopes()
mod <- glm(am ~ hp + mpg, data = mtcars, family = binomial)
cmp <- comparisons(mod, newdata = "mean")
hypotheses(cmp, hypothesis = "b1 = b2")

mfx <- slopes(mod, newdata = "mean")
hypotheses(cmp, hypothesis = "b2 = 0.2")

pre <- predictions(mod, newdata = datagrid(hp = 110, mpg = c(30, 35)))
hypotheses(pre, hypothesis = "b1 = b2")

# The `hypothesis` argument can be used to compute standard errors for fitted values
mod <- glm(am ~ hp + mpg, data = mtcars, family = binomial)

f <- function(x) predict(x, type = "link", newdata = mtcars)
p <- hypotheses(mod, hypothesis = f)
head(p)

f <- function(x) predict(x, type = "response", newdata = mtcars)
p <- hypotheses(mod, hypothesis = f)
head(p)

# Complex aggregation
# Step 1: Collapse predicted probabilities by outcome level, for each individual
# Step 2: Take the mean of the collapsed probabilities by group and `cyl`
library(dplyr)
library(MASS)
library(dplyr)

dat <- transform(mtcars, gear = factor(gear))
mod <- polr(gear ~ factor(cyl) + hp, dat)

aggregation_fun <- function(x) {
  predictions(x, vcov = FALSE) |>
    mutate(group = ifelse(group \%in\% c("3", "4"), "3 & 4", "5")) |>
    summarize(estimate = sum(estimate), .by = c("rowid", "cyl", "group")) |>
    summarize(estimate = mean(estimate), .by = c("cyl", "group")) |>
    rename(term = cyl)
}

hypotheses(mod, hypothesis = aggregation_fun)

# Equivalence, non-inferiority, and non-superiority tests
mod <- lm(mpg ~ hp + factor(gear), data = mtcars)
p <- predictions(mod, newdata = "median")
hypotheses(p, equivalence = c(17, 18))

mfx <- avg_slopes(mod, variables = "hp")
hypotheses(mfx, equivalence = c(-.1, .1))

cmp <- avg_comparisons(mod, variables = "gear", hypothesis = ~pairwise)
hypotheses(cmp, equivalence = c(0, 10))

# joint hypotheses: character vector
model <- lm(mpg ~ as.factor(cyl) * hp, data = mtcars)
hypotheses(model, joint = c("as.factor(cyl)6:hp", "as.factor(cyl)8:hp"))

# joint hypotheses: regular expression
hypotheses(model, joint = "cyl")

# joint hypotheses: integer indices
hypotheses(model, joint = 2:3)

# joint hypotheses: different null hypotheses
hypotheses(model, joint = 2:3, hypothesis = 1)
hypotheses(model, joint = 2:3, hypothesis = 1:2)

# joint hypotheses: marginaleffects object
cmp <- avg_comparisons(model)
hypotheses(cmp, joint = "cyl")

# Multiple comparison adjustment
# p values and family-wise confidence intervals
cmp <- avg_comparisons(model)
hypotheses(cmp, multcomp = "hochberg")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_coef.R, R/methods_AER.R, R/methods_MASS.R,
%   R/methods_Rchoice.R, R/methods_afex.R, R/methods_aod.R, R/methods_betareg.R,
%   R/methods_nnet.R, R/methods_crch.R, R/methods_dataframe.R,
%   R/methods_flexsurv.R, R/methods_gamlss.R, R/methods_glmmTMB.R,
%   R/methods_glmtoolbox.R, R/methods_glmx.R, R/methods_lme4.R,
%   R/methods_mlm.R, R/methods_nlme.R, R/methods_pscl.R, R/methods_robustlmm.R,
%   R/methods_rstpm2.R, R/methods_sampleSelection.R, R/methods_scam.R,
%   R/methods_stats.R, R/methods_survey.R, R/methods_survival.R,
%   R/methods_tidymodels.R
\name{set_coef}
\alias{set_coef}
\alias{set_coef.default}
\alias{set_coef.tobit}
\alias{set_coef.polr}
\alias{set_coef.glmmPQL}
\alias{set_coef.hetprob}
\alias{set_coef.ivpml}
\alias{set_coef.afex_aov}
\alias{set_coef.glimML}
\alias{set_coef.betareg}
\alias{set_coef.multinom}
\alias{set_coef.crch}
\alias{set_coef.hxlr}
\alias{set_coef.data.frame}
\alias{set_coef.flexsurvreg}
\alias{set_coef.gamlss}
\alias{set_coef.glmmTMB}
\alias{set_coef.glmgee}
\alias{set_coef.glmx}
\alias{set_coef.merMod}
\alias{set_coef.lmerModLmerTest}
\alias{set_coef.lmerMod}
\alias{set_coef.mlm}
\alias{set_coef.lme}
\alias{set_coef.hurdle}
\alias{set_coef.zeroinfl}
\alias{set_coef.rlmerMod}
\alias{set_coef.stpm2}
\alias{set_coef.pstpm2}
\alias{set_coef.gsm}
\alias{set_coef.aft}
\alias{set_coef.selection}
\alias{set_coef.scam}
\alias{set_coef.glm}
\alias{set_coef.lm}
\alias{set_coef.nls}
\alias{set_coef.svyolr}
\alias{set_coef.survreg}
\alias{set_coef.model_fit}
\alias{set_coef.workflow}
\title{Internal function to set coefficients}
\usage{
set_coef(model, coefs, ...)

\method{set_coef}{default}(model, coefs, ...)

\method{set_coef}{tobit}(model, coefs, ...)

\method{set_coef}{polr}(model, coefs, ...)

\method{set_coef}{glmmPQL}(model, coefs, ...)

\method{set_coef}{hetprob}(model, coefs, ...)

\method{set_coef}{ivpml}(model, coefs, ...)

\method{set_coef}{afex_aov}(model, coefs, ...)

\method{set_coef}{glimML}(model, coefs, ...)

\method{set_coef}{betareg}(model, coefs, ...)

\method{set_coef}{multinom}(model, coefs, ...)

\method{set_coef}{crch}(model, coefs, ...)

\method{set_coef}{hxlr}(model, coefs, ...)

\method{set_coef}{data.frame}(model, coefs, ...)

\method{set_coef}{flexsurvreg}(model, coefs, ...)

\method{set_coef}{gamlss}(model, coefs, ...)

\method{set_coef}{glmmTMB}(model, coefs, ...)

\method{set_coef}{glmgee}(model, coefs, ...)

\method{set_coef}{glmx}(model, coefs, ...)

\method{set_coef}{merMod}(model, coefs, ...)

\method{set_coef}{lmerModLmerTest}(model, coefs, ...)

\method{set_coef}{lmerMod}(model, coefs, ...)

\method{set_coef}{mlm}(model, coefs, ...)

\method{set_coef}{lme}(model, coefs, ...)

\method{set_coef}{hurdle}(model, coefs, ...)

\method{set_coef}{zeroinfl}(model, coefs, ...)

\method{set_coef}{rlmerMod}(model, coefs, ...)

\method{set_coef}{stpm2}(model, coefs, ...)

\method{set_coef}{pstpm2}(model, coefs, ...)

\method{set_coef}{gsm}(model, coefs, ...)

\method{set_coef}{aft}(model, coefs, ...)

\method{set_coef}{selection}(model, coefs, ...)

\method{set_coef}{scam}(model, coefs, ...)

\method{set_coef}{glm}(model, coefs, ...)

\method{set_coef}{lm}(model, coefs, ...)

\method{set_coef}{nls}(model, coefs, ...)

\method{set_coef}{svyolr}(model, coefs, ...)

\method{set_coef}{survreg}(model, coefs, ...)

\method{set_coef}{model_fit}(model, coefs, ...)

\method{set_coef}{workflow}(model, coefs, ...)
}
\arguments{
\item{model}{object to modify}

\item{coefs}{vector of coefficients to insert in the model object}
}
\value{
Model object of the same class as the \code{model} argument, but with
different stored coefficients.
}
\description{
Set the coefficients in a model to different values and return the modified object (internal function)
}
\details{
To compute the variance of marginal effects we need to take the
Jacobian with
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predictions.R
\name{predictions}
\alias{predictions}
\alias{avg_predictions}
\title{Predictions}
\usage{
predictions(
  model,
  newdata = NULL,
  variables = NULL,
  vcov = TRUE,
  conf_level = 0.95,
  type = NULL,
  by = FALSE,
  byfun = NULL,
  wts = FALSE,
  transform = NULL,
  hypothesis = NULL,
  equivalence = NULL,
  df = Inf,
  numderiv = "fdforward",
  ...
)

avg_predictions(
  model,
  newdata = NULL,
  variables = NULL,
  vcov = TRUE,
  conf_level = 0.95,
  type = NULL,
  by = TRUE,
  byfun = NULL,
  wts = FALSE,
  transform = NULL,
  hypothesis = NULL,
  equivalence = NULL,
  df = Inf,
  numderiv = "fdforward",
  ...
)
}
\arguments{
\item{model}{Model object}

\item{newdata}{Grid of predictor values at which we evaluate predictions.
\itemize{
\item Warning: Please avoid modifying your dataset between fitting the model and calling a \code{marginaleffects} function. This can sometimes lead to unexpected results.
\item \code{NULL} (default): Unit-level predictions for each observed value in the dataset (empirical distribution). The dataset is retrieved using \code{\link[insight:get_data]{insight::get_data()}}, which tries to extract data from the environment. This may produce unexpected results if the original data frame has been altered since fitting the model.
\item string:
\itemize{
\item "mean": Predictions evaluated when each predictor is held at its mean or mode.
\item "median": Predictions evaluated when each predictor is held at its median or mode.
\item "balanced": Predictions evaluated on a balanced grid with every combination of categories and numeric variables held at their means.
\item "tukey": Predictions evaluated at Tukey's 5 numbers.
\item "grid": Predictions evaluated on a grid of representative numbers (Tukey's 5 numbers and unique values of categorical predictors).
}
\item \code{\link[=datagrid]{datagrid()}} call to specify a custom grid of regressors. For example:
\itemize{
\item \code{newdata = datagrid(cyl = c(4, 6))}: \code{cyl} variable equal to 4 and 6 and other regressors fixed at their means or modes.
\item See the Examples section and the \code{\link[=datagrid]{datagrid()}} documentation.
}
\item \code{\link[=subset]{subset()}} call with a single argument to select a subset of the dataset used to fit the model, ex: \code{newdata = subset(treatment == 1)}
\item \code{\link[dplyr:filter]{dplyr::filter()}} call with a single argument to select a subset of the dataset used to fit the model, ex: \code{newdata = filter(treatment == 1)}
}}

\item{variables}{Counterfactual variables.
\itemize{
\item Output:
\itemize{
\item \code{predictions()}: The entire dataset is replicated once for each unique combination of \code{variables}, and predictions are made.
\item \code{avg_predictions()}: The entire dataset is replicated, predictions are made, and they are marginalized by \code{variables} categories.
\item Warning: This can be expensive in large datasets.
\item Warning: Users who need "conditional" predictions should use the \code{newdata} argument instead of \code{variables}.
}
\item Input:
\itemize{
\item \code{NULL}: computes one prediction per row of \code{newdata}
\item Character vector: the dataset is replicated once of every combination of unique values of the variables identified in \code{variables}.
\item Named list: names identify the subset of variables of interest and their values. For numeric variables, the \code{variables} argument supports functions and string shortcuts:
\itemize{
\item A function which returns a numeric value
\item Numeric vector: Contrast between the 2nd element and the 1st element of the \code{x} vector.
\item "iqr": Contrast across the interquartile range of the regressor.
\item "sd": Contrast across one standard deviation around the regressor mean.
\item "2sd": Contrast across two standard deviations around the regressor mean.
\item "minmax": Contrast between the maximum and the minimum values of the regressor.
\item "threenum": mean and 1 standard deviation on both sides
\item "fivenum": Tukey's five numbers
}
}
}}

\item{vcov}{Type of uncertainty estimates to report (e.g., for robust standard errors). Acceptable values:
\itemize{
\item FALSE: Do not compute standard errors. This can speed up computation considerably.
\item TRUE: Unit-level standard errors using the default \code{vcov(model)} variance-covariance matrix.
\item String which indicates the kind of uncertainty estimates to return.
\itemize{
\item Heteroskedasticity-consistent: \code{"HC"}, \code{"HC0"}, \code{"HC1"}, \code{"HC2"}, \code{"HC3"}, \code{"HC4"}, \code{"HC4m"}, \code{"HC5"}. See \code{?sandwich::vcovHC}
\item Heteroskedasticity and autocorrelation consistent: \code{"HAC"}
\item Mixed-Models degrees of freedom: "satterthwaite", "kenward-roger"
\item Other: \code{"NeweyWest"}, \code{"KernHAC"}, \code{"OPG"}. See the \code{sandwich} package documentation.
\item "rsample", "boot", "fwb", and "simulation" are passed to the \code{method} argument of the \code{inferences()} function. To customize the bootstrap or simulation process, call \code{inferences()} directly.
}
\item One-sided formula which indicates the name of cluster variables (e.g., \code{~unit_id}). This formula is passed to the \code{cluster} argument of the \code{sandwich::vcovCL} function.
\item Square covariance matrix
\item Function which returns a covariance matrix (e.g., \code{stats::vcov(model)})
}}

\item{conf_level}{numeric value between 0 and 1. Confidence level to use to build a confidence interval.}

\item{type}{string indicates the type (scale) of the predictions used to
compute contrasts or slopes. This can differ based on the model
type, but will typically be a string such as: "response", "link", "probs",
or "zero". When an unsupported string is entered, the model-specific list of
acceptable values is returned in an error message. When \code{type} is \code{NULL}, the
first entry in the error message is used by default. See the Type section in the documentation below.}

\item{by}{Aggregate unit-level estimates (aka, marginalize, average over). Valid inputs:
\itemize{
\item \code{FALSE}: return the original unit-level estimates.
\item \code{TRUE}: aggregate estimates for each term.
\item Character vector of column names in \code{newdata} or in the data frame produced by calling the function without the \code{by} argument.
\item Data frame with a \code{by} column of group labels, and merging columns shared by \code{newdata} or the data frame produced by calling the same function without the \code{by} argument.
\item See examples below.
\item For more complex aggregations, you can use the \code{FUN} argument of the \code{hypotheses()} function. See that function's documentation and the Hypothesis Test vignettes on the \code{marginaleffects} website.
}}

\item{byfun}{A function such as \code{mean()} or \code{sum()} used to aggregate
estimates within the subgroups defined by the \code{by} argument. \code{NULL} uses the
\code{mean()} function. Must accept a numeric vector and return a single numeric
value. This is sometimes used to take the sum or mean of predicted
probabilities across outcome or predictor
levels. See examples section.}

\item{wts}{logical, string or numeric: weights to use when computing average predictions, contrasts or slopes. These weights only affect the averaging in \verb{avg_*()} or with the \code{by} argument, and not unit-level estimates. See \code{?weighted.mean}
\itemize{
\item string: column name of the weights variable in \code{newdata}. When supplying a column name to \code{wts}, it is recommended to supply the original data (including the weights variable) explicitly to \code{newdata}.
\item numeric: vector of length equal to the number of rows in the original data or in \code{newdata} (if supplied).
\item FALSE: Equal weights.
\item TRUE: Extract weights from the fitted object with \code{insight::find_weights()} and use them when taking weighted averages of estimates. Warning: \code{newdata=datagrid()} returns a single average weight, which is equivalent to using \code{wts=FALSE}
}}

\item{transform}{A function applied to unit-level adjusted predictions and confidence intervals just before the function returns results. For bayesian models, this function is applied to individual draws from the posterior distribution, before computing summaries.}

\item{hypothesis}{specify a hypothesis test or custom contrast using a number , formula, string equation, vector, matrix, or function.
\itemize{
\item Number: The null hypothesis used in the computation of Z and p (before applying \code{transform}).
\item String: Equation to specify linear or non-linear hypothesis tests. Two-tailed tests must include an equal \code{=} sign. One-tailed tests must start with \code{<} or \code{>}. If the terms in \code{coef(object)} uniquely identify estimates, they can be used in the formula. Otherwise, use \code{b1}, \code{b2}, etc. to identify the position of each parameter. The \verb{b*} wildcard can be used to test hypotheses on all estimates. When the hypothesis string represents a two-sided equation, the \code{estimate} column holds the value of the left side minus the right side of the equation. If a named vector is used, the names are used as labels in the output. Examples:
\itemize{
\item \code{hp = drat}
\item \code{hp + drat = 12}
\item \code{b1 + b2 + b3 = 0}
\item \verb{b* / b1 = 1}
\item \verb{<= 0}
\item \verb{>= -3.5}
\item \code{b1 >= 10}
}
\item Formula: \code{lhs ~ rhs | group}
\itemize{
\item \code{lhs}
\itemize{
\item \code{ratio} (null = 1)
\item \code{difference} (null = 0)
\item Leave empty for default value
}
\item \code{rhs}
\itemize{
\item \code{pairwise} and \code{revpairwise}: pairwise differences between estimates in each row.
\item \code{reference}: differences between the estimates in each row and the estimate in the first row.
\item \code{sequential}: difference between an estimate and the estimate in the next row.
\item \code{meandev}: difference between an estimate and the mean of all estimates.
\item \code{meanotherdev}: difference between an estimate and the mean of all other estimates, excluding the current one.
\item \code{poly}: polynomial contrasts, as computed by the \code{stats::contr.poly()} function.
\item \code{helmert}: Helmert contrasts, as computed by the \code{stats::contr.helmert()} function. Contrast 2nd level to the first, 3rd to the average of the first two, and so on.
\item \code{trt_vs_ctrl}: difference between the mean of estimates (except the first) and the first estimate.
\item \code{I(fun(x))}: custom function to manipulate the vector of estimates \code{x}. The function \code{fun()} can return multiple (potentially named) estimates.
}
\item \code{group} (optional)
\itemize{
\item Column name of \code{newdata}. Conduct hypothesis tests withing subsets of the data.
}
\item Examples:
\itemize{
\item \code{~ poly}
\item \code{~ sequential | groupid}
\item \code{~ reference}
\item \code{ratio ~ pairwise}
\item \code{difference ~ pairwise | groupid}
\item \code{~ I(x - mean(x)) | groupid}
\item \verb{~ I(\\(x) c(a = x[1], b = mean(x[2:3]))) | groupid}
}
}
\item Matrix or Vector: Each column is a vector of weights. The the output is the dot product between these vectors of weights and the vector of estimates. The matrix can have column names to label the estimates.
\item Function:
\itemize{
\item Accepts an argument \code{x}: object produced by a \code{marginaleffects} function or a data frame with column \code{rowid} and \code{estimate}
\item Returns a data frame with columns \code{term} and \code{estimate} (mandatory) and \code{rowid} (optional).
\item The function can also accept optional input arguments: \code{newdata}, \code{by}, \code{draws}.
\item This function approach will not work for Bayesian models or with bootstrapping. In those cases, it is easy to use \code{get_draws()} to extract and manipulate the draws directly.
}
\item See the Examples section below and the vignette: \url{https://marginaleffects.com/chapters/hypothesis.html}
\item Warning: When calling \code{predictions()} with \code{type="invlink(link)"} (the default in some models), \code{hypothesis} is tested and p values are computed on the link scale.
}}

\item{equivalence}{Numeric vector of length 2: bounds used for the two-one-sided test (TOST) of equivalence, and for the non-inferiority and non-superiority tests. For bayesian models, this report the proportion of posterior draws in the interval and the ROPE. See Details section below.}

\item{df}{Degrees of freedom used to compute p values and confidence intervals.
\itemize{
\item A single numeric value between 1 and \code{Inf}, or a numeric vector with length equal to the number of rows in the output. When \code{df} is \code{Inf}, the normal distribution is used. When \code{df} is finite, the \code{t} distribution is used.
\item "residual": Calls \link[insight:get_df]{insight::get_df} to extract degrees of freedom from the model automatically.
\item "satterthwaite" or "kenward-roger": Use the Satterthwaite or Kenward-Roger approximation to compute degrees of freedom in mixed effects models.
}}

\item{numderiv}{string or list of strings indicating the method to use to for the numeric differentiation used in to compute delta method standard errors.
\itemize{
\item "fdforward": finite difference method with forward differences (default)
\item "fdcenter": finite difference method with central differences
\item "richardson": Richardson extrapolation method
\item Extra arguments can be specified by passing a list to the \code{numDeriv} argument, with the name of the method first and named arguments following, ex: \code{numderiv=list("fdcenter", eps = 1e-5)}. When an unknown argument is used, \code{marginaleffects} prints the list of valid arguments for each method.
}}

\item{...}{Additional arguments are passed to the \code{predict()} method
supplied by the modeling package.These arguments are particularly useful
for mixed-effects or bayesian models (see the online vignettes on the
\code{marginaleffects} website). Available arguments can vary from model to
model, depending on the range of supported arguments by each modeling
package. See the "Model-Specific Arguments" section of the
\code{?slopes} documentation for a non-exhaustive list of available
arguments.}
}
\value{
A \code{data.frame} with one row per estimate. This data frame is pretty-printed by default, but users can interact with it as a regular data frame, with functions like \code{nrow()}, \code{head()}, \code{colnames()}, etc. Values can be extracted using standard \verb{[,]} or \code{$} operators, and manipulated using external packages like \code{dplyr} or \code{data.table}.

Columns may include:
\itemize{
\item \code{rowid}: row number of the \code{newdata} data frame
\item \code{group}: (optional) value of the grouped outcome (e.g., categorical outcome models)
\item \code{term}: the focal variable.
\item \code{estimate}: an estimate of the prediction, counterfactual comparison, or slope.
\item \code{std.error}: standard errors computed via the delta method.
\item \code{p.value}: p value associated to the \code{estimate} column. The null is determined by the \code{hypothesis} argument (0 by default).
\item \code{s.value}: Shannon information transforms of p values. See the S values vignette at \url{https://marginaleffects.com} the marginaleffects website.
\item \code{conf.low}: lower bound of the confidence (or credible) interval defined by the \code{conf_level} argument.
\item \code{conf.high}: upper bound of the confidence (or credible) interval defined by the \code{conf_level} argument.
\item \code{predicted_lo}: predicted outcome for the "low" value of the focal predictor in a counterfactual comparison.
\item \code{predicted_hi}: predicted outcome for the "high" value of the focal predictor in a counterfactual comparison.
\item \code{p.rope.unconditional}: share of posterior draws in the interval specified by the \code{equivalence} argument. This is only available for Bayesian models.
\item \code{p.rope.conditional}: share of posterior draws in the interval specified by the \code{equivalence} argument, among draws in the confidence interval. This is only available for Bayesian models.
\item \code{rope}: share of the posterior draws between \code{conf.low} and \code{conf.high} that are covered by the interval specified by the \code{equivalence} argument.
\item \code{statistic.noninf}: test statistic for non-inferiority test (when \code{equivalence} argument is used).
\item \code{statistic.nonsup}: test statistic for non-superiority test (when \code{equivalence} argument is used).
\item \code{p.value.noninf}: p-value for non-inferiority test (when \code{equivalence} argument is used).
\item \code{p.value.nonsup}: p-value for non-superiority test (when \code{equivalence} argument is used).
\item \code{p.value.equiv}: p-value for equivalence test using Two One-Sided Tests (TOST) approach (when \code{equivalence} argument is used).
}

See \code{?print.marginaleffects} for printing options.

The \code{data.frame}s produced by \code{marginaleffects} stores an attribute that holds many internal objects, such as the original model, data, and much other information that can be used for post-processing. This information can be inspected using the \code{components()} function.

Warning: The internal attributes retrieved by \code{components()} are not considered part of the public API of the package. Their names and contents can change without warning or notice. Users should not rely on them.

Warning: In some cases, the internal attributes used by \code{marginaleffects()} can use up a substantial amount of memory. To clear this data, use the \code{prune()} function or set \code{options(marginaleffects_lean=TRUE)}.
}
\description{
Outcome predicted by a fitted model on a specified scale for a given combination of values of the predictor variables, such as their observed values, their means, or factor levels (a.k.a. "reference grid").
\itemize{
\item \code{predictions()}: unit-level (conditional) estimates.
\item \code{avg_predictions()}: average (marginal) estimates.
}

The \code{newdata} argument and the \code{datagrid()} function can be used to control where statistics are evaluated in the predictor space: "at observed values", "at the mean", "at representative values", etc.

See the predictions vignette and package website for worked examples and case studies:
\itemize{
\item \url{https://marginaleffects.com/chapters/predictions.html}
\item \url{https://marginaleffects.com/}
}
}
\section{Functions}{
\itemize{
\item \code{avg_predictions()}: Average predictions

}}
\section{Standard errors using the delta method}{


Standard errors for all quantities estimated by \code{marginaleffects} can be obtained via the delta method. This requires differentiating a function with respect to the coefficients in the model using a finite difference approach. In some models, the delta method standard errors can be sensitive to various aspects of the numeric differentiation strategy, including the step size. By default, the step size is set to \code{1e-8}, or to \code{1e-4} times the smallest absolute model coefficient, whichever is largest.

\code{marginaleffects} can delegate numeric differentiation to the \code{numDeriv} package, which allows more flexibility. To do this, users can pass arguments to the \code{numDeriv::jacobian} function through a global option. For example:
\itemize{
\item \code{options(marginaleffects_numDeriv = list(method = "simple", method.args = list(eps = 1e-6)))}
\item \code{options(marginaleffects_numDeriv = list(method = "Richardson", method.args = list(eps = 1e-5)))}
\item \code{options(marginaleffects_numDeriv = NULL)}
}

See the "Uncertainty" chapter on the \code{marginaleffects} website for more details on the computation of standard errors, bootstrapping, and more:

https://marginaleffects.com/chapters/uncertainty.html
}

\section{Model-Specific Arguments}{


Some model types allow model-specific arguments to modify the nature of
marginal effects, predictions, marginal means, and contrasts. Please report
other package-specific \code{predict()} arguments on Github so we can add them to
the table below.

https://github.com/vincentarelbundock/marginaleffects/issues\tabular{llll}{
   Package \tab Class \tab Argument \tab Documentation \cr
   \code{brms} \tab \code{brmsfit} \tab \code{ndraws} \tab \link[brms:posterior_predict.brmsfit]{brms::posterior_predict} \cr
    \tab  \tab \code{re_formula} \tab \link[brms:posterior_predict.brmsfit]{brms::posterior_predict} \cr
   \code{lme4} \tab \code{merMod} \tab \code{re.form} \tab \link[lme4:predict.merMod]{lme4::predict.merMod} \cr
    \tab  \tab \code{allow.new.levels} \tab \link[lme4:predict.merMod]{lme4::predict.merMod} \cr
   \code{glmmTMB} \tab \code{glmmTMB} \tab \code{re.form} \tab \link[glmmTMB:predict.glmmTMB]{glmmTMB::predict.glmmTMB} \cr
    \tab  \tab \code{allow.new.levels} \tab \link[glmmTMB:predict.glmmTMB]{glmmTMB::predict.glmmTMB} \cr
    \tab  \tab \code{zitype} \tab \link[glmmTMB:predict.glmmTMB]{glmmTMB::predict.glmmTMB} \cr
   \code{mgcv} \tab \code{bam} \tab \code{exclude} \tab \link[mgcv:predict.bam]{mgcv::predict.bam} \cr
    \tab \code{gam} \tab \code{exclude} \tab \link[mgcv:predict.gam]{mgcv::predict.gam} \cr
   \code{robustlmm} \tab \code{rlmerMod} \tab \code{re.form} \tab \link[robustlmm:rlmerMod-class]{robustlmm::predict.rlmerMod} \cr
    \tab  \tab \code{allow.new.levels} \tab \link[robustlmm:rlmerMod-class]{robustlmm::predict.rlmerMod} \cr
   \code{MCMCglmm} \tab \code{MCMCglmm} \tab \code{ndraws} \tab  \cr
   \code{sampleSelection} \tab \code{selection} \tab \code{part} \tab \link[sampleSelection:predict.selection]{sampleSelection::predict.selection} \cr
}
}

\section{Bayesian posterior summaries}{


By default, credible intervals in bayesian models are built as equal-tailed
intervals. This can be changed to a highest density interval by setting a global
option:

\code{options("marginaleffects_posterior_interval" = "eti")}

\code{options("marginaleffects_posterior_interval" = "hdi")}

By default, the center of the posterior distribution in bayesian models is
identified by the median. Users can use a different summary function by setting a
global option:

\code{options("marginaleffects_posterior_center" = "mean")}

\code{options("marginaleffects_posterior_center" = "median")}

When estimates are averaged using the \code{by} argument, the \code{tidy()} function, or
the \code{summary()} function, the posterior distribution is marginalized twice over.
First, we take the average \emph{across} units but \emph{within} each iteration of the
MCMC chain, according to what the user requested in \code{by} argument or
\code{tidy()/summary()} functions. Then, we identify the center of the resulting
posterior using the function supplied to the
\code{"marginaleffects_posterior_center"} option (the median by default).
}

\section{Equivalence, Inferiority, Superiority}{


\eqn{\theta} is an estimate, \eqn{\sigma_\theta} its estimated standard error, and \eqn{[a, b]} are the bounds of the interval supplied to the \code{equivalence} argument.

Non-inferiority:
\itemize{
\item \eqn{H_0}{H0}: \eqn{\theta \leq a}{\theta <= a}
\item \eqn{H_1}{H1}: \eqn{\theta > a}
\item \eqn{t=(\theta - a)/\sigma_\theta}{t=(\theta - a)/\sigma_\theta}
\item p: Upper-tail probability
}

Non-superiority:
\itemize{
\item \eqn{H_0}{H0}: \eqn{\theta \geq b}{\theta >= b}
\item \eqn{H_1}{H1}: \eqn{\theta < b}
\item \eqn{t=(\theta - b)/\sigma_\theta}{t=(\theta - b)/\sigma_\theta}
\item p: Lower-tail probability
}

Equivalence: Two One-Sided Tests (TOST)
\itemize{
\item p: Maximum of the non-inferiority and non-superiority p values.
}

Thanks to Russell V. Lenth for the excellent \code{emmeans} package and documentation which inspired this feature.
}

\section{Order of operations}{


Behind the scenes, the arguments of \code{marginaleffects} functions are evaluated in this order:
\enumerate{
\item \code{newdata}
\item \code{variables}
\item \code{comparison} and \code{slope}
\item \code{by}
\item \code{vcov}
\item \code{hypothesis}
\item \code{transform}
}
}

\section{Parallel computation}{


The \code{slopes()} and \code{comparisons()} functions can use parallelism to
speed up computation. Operations are parallelized for the computation of
standard errors, at the model coefficient level. There is always
considerable overhead when using parallel computation, mainly involved
in passing the whole dataset to the different processes. Thus, parallel
computation is most likely to be useful when the model includes many parameters
and the dataset is relatively small.

Warning: In many cases, parallel processing will not be useful at all.

To activate parallel computation, users must load the \code{future.apply} package,
call \code{plan()} function, and set a global option.

\code{options(marginaleffects_parallel = TRUE)}: parallelize delta method computation of standard errors.
\code{options(marginaleffects_parallel_inferences = TRUE)}: parallelize \code{"rsample"} or \code{"fwb"} bootstrap computation in \code{inferences()}.
\code{options(marginaleffects_parallel_packages = TRUE)}: vector of strings with the names of modeling packages used to fit the model, ex: c("survival", "splines")

For example:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(future.apply)
plan("multisession", workers = 4)
options(marginaleffects_parallel = FALSE)
options(marginaleffects_parallel_inferences = TRUE)
options(marginaleffects_parallel_packages = c("survival", "splines"))

slopes(model)
}\if{html}{\out{</div>}}

To disable parallelism in \code{marginaleffects} altogether, you can set a global option:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(marginaleffects_parallel = FALSE)
}\if{html}{\out{</div>}}
}

\section{Global options}{


The behavior of \code{marginaleffects} functions can be modified by setting global options.

Disable some safety checks and warnings:
\itemize{
\item \code{options(marginaleffects_startup_message = FALSE)}
\itemize{
\item Disable the startup message printed on \code{library(marginaleffects)}.
}
\item \code{options(marginaleffects_safe = FALSE)}
\itemize{
\item Disable safety checks and warnings.
}
\item \code{options(marginaleffects_print_omit = c("p.value", "s.value"))}
\itemize{
\item Omit some columns from the printed output.
}
}

Enforce lean return objects, sans information about the original model and
data, and other ancillary attributes. Note that this will disable some
advanced post-processing features and functions like \link{hypotheses}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{options(marginaleffects_lean = TRUE)
}\if{html}{\out{</div>}}

Other options:
\itemize{
\item \code{marginaleffects_plot_gray}: logical. If \code{TRUE}, the default color of the plot is gray. Default is \code{FALSE}.
}
}

\section{Types}{


The \code{type} argument determines the scale of the predictions used to compute quantities of interest with functions from the \code{marginaleffects} package. Admissible values for \code{type} depend on the model object. When users specify an incorrect value for \code{type}, \code{marginaleffects} will raise an informative error with a list of valid \code{type} values for the specific model object. The first entry in the list in that error message is the default type.

The \code{invlink(link)} is a special type defined by \code{marginaleffects}. It is available for some (but not all) models, and only for the \code{predictions()} function. With this link type, we first compute predictions on the link scale, then we use the inverse link function to backtransform the predictions to the response scale. This is useful for models with non-linear link functions as it can ensure that confidence intervals stay within desirable bounds, ex: 0 to 1 for a logit model. Note that an average of estimates with \code{type="invlink(link)"} will not always be equivalent to the average of estimates with \code{type="response"}. This type is default when calling \code{predictions()}. It is available---but not default---when calling \code{avg_predictions()} or \code{predictions()} with the \code{by} argument.

Some of the most common \code{type} values are:\tabular{ll}{
   class \tab type \cr
   DirichletRegModel \tab response \cr
   Gam \tab invlink(link), response, link \cr
   Gls \tab lp \cr
   MCMCglmm \tab response \cr
   bam \tab response, link \cr
   bart \tab ev, ppd \cr
   betareg \tab response, link, precision, quantile, variance \cr
   bife \tab response, link \cr
   bracl \tab probs \cr
   brglmFit \tab response, link \cr
   brmsfit \tab response, link, prediction, average \cr
   brmultinom \tab probs, class \cr
   clm \tab prob, cum.prob, linear.predictor \cr
   clogit \tab expected, lp, risk, survival \cr
   coxph \tab survival, expected, lp, risk \cr
   coxph_weightit \tab survival, expected, lp, risk \cr
   crch \tab response, location, scale, density \cr
   fixest \tab invlink(link), response, link \cr
   flexsurvreg \tab survival, response, mean, link, lp, linear, rmst, hazard, cumhaz \cr
   gam \tab response, link \cr
   geeglm \tab response, link \cr
   glimML \tab response, link \cr
   glm \tab invlink(link), response, link \cr
   glm_weightit \tab invlink(link), probs, response, lp, link \cr
   glmerMod \tab response, link \cr
   glmgee \tab response \cr
   glmmPQL \tab response, link \cr
   glmmTMB \tab response, link, conditional, zprob, zlink, disp \cr
   glmrob \tab response, link \cr
   glmx \tab response \cr
   hetprob \tab pr, xb \cr
   hurdle \tab response, prob, count, zero \cr
   hxlr \tab location, cumprob, scale, density \cr
   iv_robust \tab response \cr
   ivpml \tab pr, xb \cr
   ivreg \tab response \cr
   lda \tab class, posterior \cr
   lm \tab response \cr
   lm_robust \tab response \cr
   lmerMod \tab response \cr
   lmerModLmerTest \tab response \cr
   lmrob \tab response \cr
   lrm \tab fitted, lp, mean \cr
   mblogit \tab response, latent, link \cr
   mclogit \tab response, latent, link \cr
   mhurdle \tab E, Ep, p \cr
   model_fit \tab numeric, prob, class \cr
   multinom \tab probs, latent \cr
   multinom_weightit \tab probs, response, mean \cr
   mvgam \tab response, link, expected, detection, latent_N \cr
   negbin \tab invlink(link), response, link \cr
   ols \tab lp \cr
   oohbchoice \tab probability, utility \cr
   ordinal_weightit \tab probs, response, link, lp, mean \cr
   orm \tab fitted, mean, lp \cr
   polr \tab probs \cr
   rendo.base \tab response, link \cr
   rlm \tab response \cr
   selection \tab response, link, unconditional, conditional \cr
   speedglm \tab response, link \cr
   speedlm \tab response \cr
   stanreg \tab response, link \cr
   survreg \tab response, link, quantile \cr
   svyglm \tab response, link \cr
   svyolr \tab probs \cr
   tobit \tab response, link \cr
   tobit1 \tab expvalue, linpred, prob \cr
   workflow \tab numeric, prob, class \cr
   zeroinfl \tab response, prob, count, zero \cr
}
}

\examples{
\dontshow{if (interactive() || isTRUE(Sys.getenv("R_DOC_BUILD") == "true")) withAutoprint(\{ # examplesIf}
library("marginaleffects")
# Adjusted Prediction for every row of the original dataset
mod <- lm(mpg ~ hp + factor(cyl), data = mtcars)
pred <- predictions(mod)
head(pred)

# Adjusted Predictions at User-Specified Values of the Regressors
predictions(mod, newdata = datagrid(hp = c(100, 120), cyl = 4))

m <- lm(mpg ~ hp + drat + factor(cyl) + factor(am), data = mtcars)
predictions(m, newdata = datagrid(FUN_factor = unique, FUN_numeric = median))

# Average Adjusted Predictions (AAP)
library(dplyr)
mod <- lm(mpg ~ hp * am * vs, mtcars)

avg_predictions(mod)

predictions(mod, by = "am")

# Conditional Adjusted Predictions
plot_predictions(mod, condition = "hp")

# Counterfactual predictions with the `variables` argument
# the `mtcars` dataset has 32 rows

mod <- lm(mpg ~ hp + am, data = mtcars)
p <- predictions(mod)
head(p)
nrow(p)

# average counterfactual predictions
avg_predictions(mod, variables = "am")

# counterfactual predictions obtained by replicating the entire for different
# values of the predictors
p <- predictions(mod, variables = list(hp = c(90, 110)))
nrow(p)


# hypothesis test: is the prediction in the 1st row equal to the prediction in the 2nd row
mod <- lm(mpg ~ wt + drat, data = mtcars)

predictions(
  mod,
  newdata = datagrid(wt = 2:3),
  hypothesis = "b1 = b2")

# same hypothesis test using row indices
predictions(
  mod,
  newdata = datagrid(wt = 2:3),
  hypothesis = "b1 - b2 = 0")

# same hypothesis test using numeric vector of weights
predictions(
  mod,
  newdata = datagrid(wt = 2:3),
  hypothesis = c(1, -1))

# two custom contrasts using a matrix of weights
lc <- matrix(
  c(
    1, -1,
    2, 3),
  ncol = 2)
predictions(
  mod,
  newdata = datagrid(wt = 2:3),
  hypothesis = lc)


# `by` argument
mod <- lm(mpg ~ hp * am * vs, data = mtcars)
predictions(mod, by = c("am", "vs"))

library(nnet)
nom <- multinom(factor(gear) ~ mpg + am * vs, data = mtcars, trace = FALSE)

# first 5 raw predictions
p <- predictions(nom, type = "probs")
head(p)

# average predictions
avg_predictions(nom, type = "probs", by = "group")

by <- data.frame(
  group = c("3", "4", "5"),
  by = c("3,4", "3,4", "5"))

predictions(nom, type = "probs", by = by)

# sum of predicted probabilities for combined response levels
mod <- multinom(factor(cyl) ~ mpg + am, data = mtcars, trace = FALSE)
by <- data.frame(
  by = c("4,6", "4,6", "8"),
  group = as.character(c(4, 6, 8)))
predictions(mod, newdata = "mean", byfun = sum, by = by)
\dontshow{\}) # examplesIf}
}
\references{
\itemize{
\item Arel-Bundock V, Greifer N, Heiss A (2024). “How to Interpret Statistical Models Using marginaleffects for R and Python.” \emph{Journal of Statistical Software}, \emph{111}(9), 1-32. doi:10.18637/jss.v111.i09 \doi{10.18637/jss.v111.i09}
\item Arel-Bundock (2026). "Model to Meaning: How to interpret statistical models in R and Python." CRC Press. https://routledge.com/9781032908724
\item Greenland S. 2019. "Valid P-Values Behave Exactly as They Should: Some Misleading Criticisms of P-Values and Their Resolution With S-Values." The American Statistician. 73(S1): 106–114.
\item Cole, Stephen R, Jessie K Edwards, and Sander Greenland. 2020. "Surprise!" American Journal of Epidemiology 190 (2): 191–93. \doi{10.1093/aje/kwaa136}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinytest.R
\name{expect_predictions}
\alias{expect_predictions}
\title{\code{tinytest} helper}
\usage{
expect_predictions(object, se = TRUE, n_row = NULL, n_col = NULL)
}
\description{
\code{tinytest} helper
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/methods_aod.R, R/methods_betareg.R,
%   R/sanity_model.R, R/methods_brms.R, R/methods_dbarts.R, R/methods_glmmTMB.R,
%   R/methods_lme4.R, R/methods_mclogit.R, R/methods_mlogit.R,
%   R/methods_ordinal.R, R/methods_plm.R, R/methods_quantreg.R,
%   R/methods_survey.R
\name{sanitize_model_specific.glimML}
\alias{sanitize_model_specific.glimML}
\alias{sanitize_model_specific.betareg}
\alias{sanitize_model_specific}
\alias{sanitize_model_specific.default}
\alias{sanitize_model_specific.brmsfit}
\alias{sanitize_model_specific.bart}
\alias{sanitize_model_specific.glmmTMB}
\alias{sanitize_model_specific.merMod}
\alias{sanitize_model_specific.mblogit}
\alias{sanitize_model_specific.mlogit}
\alias{sanitize_model_specific.clm}
\alias{sanitize_model_specific.plm}
\alias{sanitize_model_specific.rqs}
\alias{sanitize_model_specific.svyolr}
\alias{sanitize_model_specific.svyglm}
\title{Method to raise model-specific warnings and errors}
\usage{
\method{sanitize_model_specific}{glimML}(model, ...)

\method{sanitize_model_specific}{betareg}(model, ...)

sanitize_model_specific(model, ...)

\method{sanitize_model_specific}{default}(
  model,
  vcov = NULL,
  calling_function = "marginaleffects",
  ...
)

\method{sanitize_model_specific}{brmsfit}(model, ...)

\method{sanitize_model_specific}{bart}(model, ...)

\method{sanitize_model_specific}{glmmTMB}(model, vcov = TRUE, re.form, ...)

\method{sanitize_model_specific}{merMod}(model, re.form, ...)

\method{sanitize_model_specific}{mblogit}(model, calling_function = "marginaleffects", ...)

\method{sanitize_model_specific}{mlogit}(model, newdata, ...)

\method{sanitize_model_specific}{clm}(model, ...)

\method{sanitize_model_specific}{plm}(model, ...)

\method{sanitize_model_specific}{plm}(model, ...)

\method{sanitize_model_specific}{rqs}(model, ...)

\method{sanitize_model_specific}{svyolr}(model, wts = FALSE, by = FALSE, ...)

\method{sanitize_model_specific}{svyglm}(model, wts = FALSE, by = FALSE, ...)
}
\arguments{
\item{model}{Model object}

\item{...}{Additional arguments are passed to the \code{predict()} method
supplied by the modeling package.These arguments are particularly useful
for mixed-effects or bayesian models (see the online vignettes on the
\code{marginaleffects} website). Available arguments can vary from model to
model, depending on the range of supported arguments by each modeling
package. See the "Model-Specific Arguments" section of the
\code{?slopes} documentation for a non-exhaustive list of available
arguments.}

\item{vcov}{Type of uncertainty estimates to report (e.g., for robust standard errors). Acceptable values:
\itemize{
\item FALSE: Do not compute standard errors. This can speed up computation considerably.
\item TRUE: Unit-level standard errors using the default \code{vcov(model)} variance-covariance matrix.
\item String which indicates the kind of uncertainty estimates to return.
\itemize{
\item Heteroskedasticity-consistent: \code{"HC"}, \code{"HC0"}, \code{"HC1"}, \code{"HC2"}, \code{"HC3"}, \code{"HC4"}, \code{"HC4m"}, \code{"HC5"}. See \code{?sandwich::vcovHC}
\item Heteroskedasticity and autocorrelation consistent: \code{"HAC"}
\item Mixed-Models degrees of freedom: "satterthwaite", "kenward-roger"
\item Other: \code{"NeweyWest"}, \code{"KernHAC"}, \code{"OPG"}. See the \code{sandwich} package documentation.
}
\item One-sided formula which indicates the name of cluster variables (e.g., \code{~unit_id}). This formula is passed to the \code{cluster} argument of the \code{sandwich::vcovCL} function.
\item Square covariance matrix
\item Function which returns a covariance matrix (e.g., \code{stats::vcov(model)})
}}

\item{newdata}{Grid of predictor values at which we evaluate the slopes.
\itemize{
\item Warning: Please avoid modifying your dataset between fitting the model and calling a \code{marginaleffects} function. This can sometimes lead to unexpected results.
\item \code{NULL} (default): Unit-level slopes for each observed value in the dataset (empirical distribution). The dataset is retrieved using \code{\link[insight:get_data]{insight::get_data()}}, which tries to extract data from the environment. This may produce unexpected results if the original data frame has been altered since fitting the model.
\item \code{\link[=datagrid]{datagrid()}} call to specify a custom grid of regressors. For example:
\itemize{
\item \code{newdata = datagrid(cyl = c(4, 6))}: \code{cyl} variable equal to 4 and 6 and other regressors fixed at their means or modes.
\item See the Examples section and the \code{\link[=datagrid]{datagrid()}} documentation.
}
\item \code{\link[=subset]{subset()}} call with a single argument to select a subset of the dataset used to fit the model, ex: \code{newdata = subset(treatment == 1)}
\item \code{\link[dplyr:filter]{dplyr::filter()}} call with a single argument to select a subset of the dataset used to fit the model, ex: \code{newdata = filter(treatment == 1)}
\item string:
\itemize{
\item "mean": Slopes evaluated when each predictor is held at its mean or mode.
\item "median": Slopes evaluated when each predictor is held at its median or mode.
\item "balanced": Slopes evaluated on a balanced grid with every combination of categories and numeric variables held at their means.
\item "tukey": Slopes evaluated at Tukey's 5 numbers.
\item "grid": Slopes evaluated on a grid of representative numbers (Tukey's 5 numbers and unique values of categorical predictors).
}
}}

\item{wts}{logical, string or numeric: weights to use when computing average predictions, contrasts or slopes. These weights only affect the averaging in \verb{avg_*()} or with the \code{by} argument, and not unit-level estimates. See \code{?weighted.mean}
\itemize{
\item string: column name of the weights variable in \code{newdata}. When supplying a column name to \code{wts}, it is recommended to supply the original data (including the weights variable) explicitly to \code{newdata}.
\item numeric: vector of length equal to the number of rows in the original data or in \code{newdata} (if supplied).
\item FALSE: Equal weights.
\item TRUE: Extract weights from the fitted object with \code{insight::find_weights()} and use them when taking weighted averages of estimates. Warning: \code{newdata=datagrid()} returns a single average weight, which is equivalent to using \code{wts=FALSE}
}}

\item{by}{Aggregate unit-level estimates (aka, marginalize, average over). Valid inputs:
\itemize{
\item \code{FALSE}: return the original unit-level estimates.
\item \code{TRUE}: aggregate estimates for each term.
\item Character vector of column names in \code{newdata} or in the data frame produced by calling the function without the \code{by} argument.
\item Data frame with a \code{by} column of group labels, and merging columns shared by \code{newdata} or the data frame produced by calling the same function without the \code{by} argument.
\item See examples below.
\item For more complex aggregations, you can use the \code{FUN} argument of the \code{hypotheses()} function. See that function's documentation and the Hypothesis Test vignettes on the \code{marginaleffects} website.
}}
}
\value{
A warning, an error, or nothing
}
\description{
Method to raise model-specific warnings and errors
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinytest.R
\name{expect_margins}
\alias{expect_margins}
\title{\code{tinytest} helper}
\usage{
expect_margins(
  results,
  margins_object,
  se = TRUE,
  tolerance = 1e-05,
  verbose = FALSE
)
}
\description{
\code{tinytest} helper
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_slopes.R
\name{plot_slopes}
\alias{plot_slopes}
\title{Plot Conditional or Marginal Slopes}
\usage{
plot_slopes(
  model,
  variables = NULL,
  condition = NULL,
  by = NULL,
  newdata = NULL,
  type = NULL,
  vcov = NULL,
  conf_level = 0.95,
  wts = FALSE,
  slope = "dydx",
  rug = FALSE,
  gray = getOption("marginaleffects_plot_gray", default = FALSE),
  draw = TRUE,
  ...
)
}
\arguments{
\item{model}{Model object}

\item{variables}{Name of the variable whose marginal effect (slope) we want to plot on the y-axis.}

\item{condition}{Conditional slopes
\itemize{
\item Character vector (max length 4): Names of the predictors to display.
\item Named list (max length 4): List names correspond to predictors. List elements can be:
\itemize{
\item Numeric vector
\item Function which returns a numeric vector or a set of unique categorical values
\item Shortcut strings for common reference values: "minmax", "quartile", "threenum"
}
\item 1: x-axis. 2: color/shape. 3: facet (wrap if no fourth variable, otherwise cols of grid). 4: facet (rows of grid).
\item Numeric variables in positions 2 and 3 are summarized by Tukey's five numbers \code{?stats::fivenum}.
}}

\item{by}{Aggregate unit-level estimates (aka, marginalize, average over). Valid inputs:
\itemize{
\item \code{FALSE}: return the original unit-level estimates.
\item \code{TRUE}: aggregate estimates for each term.
\item Character vector of column names in \code{newdata} or in the data frame produced by calling the function without the \code{by} argument.
\item Data frame with a \code{by} column of group labels, and merging columns shared by \code{newdata} or the data frame produced by calling the same function without the \code{by} argument.
\item See examples below.
\item For more complex aggregations, you can use the \code{FUN} argument of the \code{hypotheses()} function. See that function's documentation and the Hypothesis Test vignettes on the \code{marginaleffects} website.
}}

\item{newdata}{When \code{newdata} is \code{NULL}, the grid is determined by the \code{condition} argument. When \code{newdata} is not \code{NULL}, the argument behaves in the same way as in the \code{slopes()} function.}

\item{type}{string indicates the type (scale) of the predictions used to
compute contrasts or slopes. This can differ based on the model
type, but will typically be a string such as: "response", "link", "probs",
or "zero". When an unsupported string is entered, the model-specific list of
acceptable values is returned in an error message. When \code{type} is \code{NULL}, the
first entry in the error message is used by default.}

\item{vcov}{Type of uncertainty estimates to report (e.g., for robust standard errors). Acceptable values:
\itemize{
\item FALSE: Do not compute standard errors. This can speed up computation considerably.
\item TRUE: Unit-level standard errors using the default \code{vcov(model)} variance-covariance matrix.
\item String which indicates the kind of uncertainty estimates to return.
\itemize{
\item Heteroskedasticity-consistent: \code{"HC"}, \code{"HC0"}, \code{"HC1"}, \code{"HC2"}, \code{"HC3"}, \code{"HC4"}, \code{"HC4m"}, \code{"HC5"}. See \code{?sandwich::vcovHC}
\item Heteroskedasticity and autocorrelation consistent: \code{"HAC"}
\item Mixed-Models degrees of freedom: "satterthwaite", "kenward-roger"
\item Other: \code{"NeweyWest"}, \code{"KernHAC"}, \code{"OPG"}. See the \code{sandwich} package documentation.
}
\item One-sided formula which indicates the name of cluster variables (e.g., \code{~unit_id}). This formula is passed to the \code{cluster} argument of the \code{sandwich::vcovCL} function.
\item Square covariance matrix
\item Function which returns a covariance matrix (e.g., \code{stats::vcov(model)})
}}

\item{conf_level}{numeric value between 0 and 1. Confidence level to use to build a confidence interval.}

\item{wts}{logical, string or numeric: weights to use when computing average predictions, contrasts or slopes. These weights only affect the averaging in \verb{avg_*()} or with the \code{by} argument, and not unit-level estimates. See \code{?weighted.mean}
\itemize{
\item string: column name of the weights variable in \code{newdata}. When supplying a column name to \code{wts}, it is recommended to supply the original data (including the weights variable) explicitly to \code{newdata}.
\item numeric: vector of length equal to the number of rows in the original data or in \code{newdata} (if supplied).
\item FALSE: Equal weights.
\item TRUE: Extract weights from the fitted object with \code{insight::find_weights()} and use them when taking weighted averages of estimates. Warning: \code{newdata=datagrid()} returns a single average weight, which is equivalent to using \code{wts=FALSE}
}}

\item{slope}{string indicates the type of slope or (semi-)elasticity to compute:
\itemize{
\item "dydx": dY/dX
\item "eyex": dY/dX * Y / X
\item "eydx": dY/dX * Y
\item "dyex": dY/dX / X
\item Y is the predicted value of the outcome; X is the observed value of the predictor.
}}

\item{rug}{TRUE displays tick marks on the axes to mark the distribution of raw data.}

\item{gray}{FALSE grayscale or color plot}

\item{draw}{\code{TRUE} returns a \code{ggplot2} plot. \code{FALSE} returns a \code{data.frame} of the underlying data.}

\item{...}{Additional arguments are passed to the \code{predict()} method
supplied by the modeling package.These arguments are particularly useful
for mixed-effects or bayesian models (see the online vignettes on the
\code{marginaleffects} website). Available arguments can vary from model to
model, depending on the range of supported arguments by each modeling
package. See the "Model-Specific Arguments" section of the
\code{?slopes} documentation for a non-exhaustive list of available
arguments.}
}
\value{
A \code{ggplot2} object
}
\description{
Plot slopes on the y-axis against values of one or more predictors (x-axis, colors/shapes, and facets).

The \code{by} argument is used to plot marginal slopes, that is, slopes made on the original data, but averaged by subgroups. This is analogous to using the \code{by} argument in the \code{slopes()} function.

The \code{condition} argument is used to plot conditional slopes, that is, slopes computed on a user-specified grid. This is analogous to using the \code{newdata} argument and \code{datagrid()} function in a \code{slopes()} call. All variables whose values are not specified explicitly are treated as usual by \code{datagrid()}, that is, they are held at their mean or mode (or rounded mean for integers). This includes grouping variables in mixed-effects models, so analysts who fit such models may want to specify the groups of interest using the \code{condition} argument, or supply model-specific arguments to compute population-level estimates. See details below.
See the "Plots" vignette and website for tutorials and information on how to customize plots:
\itemize{
\item https://marginaleffects.com/bonus/plot.html
\item https://marginaleffects.com
}
}
\section{Model-Specific Arguments}{


Some model types allow model-specific arguments to modify the nature of
marginal effects, predictions, marginal means, and contrasts. Please report
other package-specific \code{predict()} arguments on Github so we can add them to
the table below.

https://github.com/vincentarelbundock/marginaleffects/issues\tabular{llll}{
   Package \tab Class \tab Argument \tab Documentation \cr
   \code{brms} \tab \code{brmsfit} \tab \code{ndraws} \tab \link[brms:posterior_predict.brmsfit]{brms::posterior_predict} \cr
    \tab  \tab \code{re_formula} \tab \link[brms:posterior_predict.brmsfit]{brms::posterior_predict} \cr
   \code{lme4} \tab \code{merMod} \tab \code{re.form} \tab \link[lme4:predict.merMod]{lme4::predict.merMod} \cr
    \tab  \tab \code{allow.new.levels} \tab \link[lme4:predict.merMod]{lme4::predict.merMod} \cr
   \code{glmmTMB} \tab \code{glmmTMB} \tab \code{re.form} \tab \link[glmmTMB:predict.glmmTMB]{glmmTMB::predict.glmmTMB} \cr
    \tab  \tab \code{allow.new.levels} \tab \link[glmmTMB:predict.glmmTMB]{glmmTMB::predict.glmmTMB} \cr
    \tab  \tab \code{zitype} \tab \link[glmmTMB:predict.glmmTMB]{glmmTMB::predict.glmmTMB} \cr
   \code{mgcv} \tab \code{bam} \tab \code{exclude} \tab \link[mgcv:predict.bam]{mgcv::predict.bam} \cr
    \tab \code{gam} \tab \code{exclude} \tab \link[mgcv:predict.gam]{mgcv::predict.gam} \cr
   \code{robustlmm} \tab \code{rlmerMod} \tab \code{re.form} \tab \link[robustlmm:rlmerMod-class]{robustlmm::predict.rlmerMod} \cr
    \tab  \tab \code{allow.new.levels} \tab \link[robustlmm:rlmerMod-class]{robustlmm::predict.rlmerMod} \cr
   \code{MCMCglmm} \tab \code{MCMCglmm} \tab \code{ndraws} \tab  \cr
   \code{sampleSelection} \tab \code{selection} \tab \code{part} \tab \link[sampleSelection:predict.selection]{sampleSelection::predict.selection} \cr
}
}

\examples{
\dontshow{if (interactive() || isTRUE(Sys.getenv("R_DOC_BUILD") == "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
library(marginaleffects)
mod <- lm(mpg ~ hp * drat * factor(am), data = mtcars)

plot_slopes(mod, variables = "hp", condition = "drat")

plot_slopes(mod, variables = "hp", condition = c("drat", "am"))

plot_slopes(mod, variables = "hp", condition = list("am", "drat" = 3:5))

plot_slopes(mod, variables = "am", condition = list("hp", "drat" = range))

plot_slopes(mod, variables = "am", condition = list("hp", "drat" = "threenum"))

# marginal slopes
plot_slopes(mod, variables = "hp", by = "am")

# marginal slopes on a counterfactual grid
plot_slopes(mod,
  variables = "hp",
  by = "am",
  newdata = datagrid(am = 0:1, grid_type = "counterfactual")
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_predict.R, R/methods_MASS.R,
%   R/methods_MCMCglmm.R, R/methods_afex.R, R/methods_aod.R,
%   R/methods_betareg.R, R/methods_bife.R, R/methods_biglm.R, R/methods_nnet.R,
%   R/methods_brglm2.R, R/methods_brms.R, R/methods_crch.R, R/methods_dbarts.R,
%   R/methods_fixest.R, R/methods_flexsurv.R, R/methods_gamlss.R,
%   R/methods_glmmTMB.R, R/methods_glmtoolbox.R, R/methods_lme4.R,
%   R/methods_mclogit.R, R/methods_mhurdle.R, R/methods_mlr3.R,
%   R/methods_ordinal.R, R/methods_quantreg.R, R/methods_rms.R,
%   R/methods_robustlmm.R, R/methods_rstanarm.R, R/methods_rstpm2.R,
%   R/methods_stats.R, R/methods_survey.R, R/methods_survival.R,
%   R/methods_tidymodels.R, R/methods_tobit1.R
\name{get_predict}
\alias{get_predict}
\alias{get_predict.default}
\alias{get_predict.polr}
\alias{get_predict.glmmPQL}
\alias{get_predict.MCMCglmm}
\alias{get_predict.afex_aov}
\alias{get_predict.glimML}
\alias{get_predict.betareg}
\alias{get_predict.bife}
\alias{get_predict.biglm}
\alias{get_predict.multinom}
\alias{get_predict.brmultinom}
\alias{get_predict.brmsfit}
\alias{get_predict.crch}
\alias{get_predict.bart}
\alias{get_predict.fixest}
\alias{get_predict.flexsurvreg}
\alias{get_predict.gamlss}
\alias{get_predict.glmmTMB}
\alias{get_predict.glmgee}
\alias{get_predict.merMod}
\alias{get_predict.lmerModLmerTest}
\alias{get_predict.lmerMod}
\alias{get_predict.mblogit}
\alias{get_predict.mhurdle}
\alias{get_predict.Learner}
\alias{get_predict.clm}
\alias{get_predict.rq}
\alias{get_predict.rms}
\alias{get_predict.orm}
\alias{get_predict.lrm}
\alias{get_predict.ols}
\alias{get_predict.rlmerMod}
\alias{get_predict.stanreg}
\alias{get_predict.stpm2}
\alias{get_predict.pstpm2}
\alias{get_predict.gsm}
\alias{get_predict.aft}
\alias{get_predict.lm}
\alias{get_predict.glm}
\alias{get_predict.svyolr}
\alias{get_predict.svyglm}
\alias{get_predict.coxph}
\alias{get_predict.model_fit}
\alias{get_predict.workflow}
\alias{get_predict.tobit1}
\title{Get predicted values from a model object (internal function)}
\usage{
get_predict(model, newdata, type, ...)

\method{get_predict}{default}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{polr}(model, newdata = insight::get_data(model), type = "probs", ...)

\method{get_predict}{glmmPQL}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{MCMCglmm}(model, newdata, type = "response", ndraws = 1000, ...)

\method{get_predict}{afex_aov}(model, newdata = NULL, ...)

\method{get_predict}{glimML}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{betareg}(model, newdata, ...)

\method{get_predict}{bife}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{biglm}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{multinom}(model, newdata = insight::get_data(model), type = "probs", ...)

\method{get_predict}{brmultinom}(model, newdata = insight::get_data(model), type = "probs", ...)

\method{get_predict}{brmsfit}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{crch}(model, newdata = NULL, type = "location", ...)

\method{get_predict}{bart}(model, newdata = NULL, ...)

\method{get_predict}{fixest}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{flexsurvreg}(model, newdata, type, ...)

\method{get_predict}{gamlss}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{glmmTMB}(
  model,
  newdata = insight::get_data(model),
  type = "response",
  newparams = NULL,
  ...
)

\method{get_predict}{glmgee}(model, newdata, ...)

\method{get_predict}{merMod}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{lmerModLmerTest}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{lmerMod}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{mblogit}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{mhurdle}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{Learner}(model, newdata, type = NULL, ...)

\method{get_predict}{clm}(model, newdata = insight::get_data(model), type = "prob", ...)

\method{get_predict}{rq}(model, newdata = insight::get_data(model), type = NULL, ...)

\method{get_predict}{rms}(model, newdata = insight::get_data(model), type = NULL, ...)

\method{get_predict}{orm}(model, newdata = insight::get_data(model), type = NULL, ...)

\method{get_predict}{lrm}(model, newdata = insight::get_data(model), type = NULL, ...)

\method{get_predict}{ols}(model, newdata = insight::get_data(model), type = NULL, ...)

\method{get_predict}{rlmerMod}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{stanreg}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{stpm2}(model, newdata = NULL, ...)

\method{get_predict}{pstpm2}(model, newdata = NULL, ...)

\method{get_predict}{gsm}(model, newdata = NULL, ...)

\method{get_predict}{aft}(model, newdata = NULL, ...)

\method{get_predict}{lm}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{glm}(model, newdata = insight::get_data(model), type = "response", ...)

\method{get_predict}{svyolr}(model, newdata = insight::get_data(model), type = "probs", ...)

\method{get_predict}{svyglm}(
  model,
  newdata = insight::get_data(model),
  type = "response",
  se.fit = FALSE,
  ...
)

\method{get_predict}{coxph}(model, newdata = insight::get_data(model), type = "lp", ...)

\method{get_predict}{model_fit}(model, newdata, type = NULL, ...)

\method{get_predict}{workflow}(model, newdata, type = NULL, ...)

\method{get_predict}{tobit1}(model, newdata = insight::get_data(model), type = "response", ...)
}
\arguments{
\item{model}{Model object}

\item{newdata}{Grid of predictor values at which we evaluate the slopes.
\itemize{
\item Warning: Please avoid modifying your dataset between fitting the model and calling a \code{marginaleffects} function. This can sometimes lead to unexpected results.
\item \code{NULL} (default): Unit-level slopes for each observed value in the dataset (empirical distribution). The dataset is retrieved using \code{\link[insight:get_data]{insight::get_data()}}, which tries to extract data from the environment. This may produce unexpected results if the original data frame has been altered since fitting the model.
\item \code{\link[=datagrid]{datagrid()}} call to specify a custom grid of regressors. For example:
\itemize{
\item \code{newdata = datagrid(cyl = c(4, 6))}: \code{cyl} variable equal to 4 and 6 and other regressors fixed at their means or modes.
\item See the Examples section and the \code{\link[=datagrid]{datagrid()}} documentation.
}
\item \code{\link[=subset]{subset()}} call with a single argument to select a subset of the dataset used to fit the model, ex: \code{newdata = subset(treatment == 1)}
\item \code{\link[dplyr:filter]{dplyr::filter()}} call with a single argument to select a subset of the dataset used to fit the model, ex: \code{newdata = filter(treatment == 1)}
\item string:
\itemize{
\item "mean": Slopes evaluated when each predictor is held at its mean or mode.
\item "median": Slopes evaluated when each predictor is held at its median or mode.
\item "balanced": Slopes evaluated on a balanced grid with every combination of categories and numeric variables held at their means.
\item "tukey": Slopes evaluated at Tukey's 5 numbers.
\item "grid": Slopes evaluated on a grid of representative numbers (Tukey's 5 numbers and unique values of categorical predictors).
}
}}

\item{type}{string indicates the type (scale) of the predictions used to
compute contrasts or slopes. This can differ based on the model
type, but will typically be a string such as: "response", "link", "probs",
or "zero". When an unsupported string is entered, the model-specific list of
acceptable values is returned in an error message. When \code{type} is \code{NULL}, the
first entry in the error message is used by default.}

\item{...}{Additional arguments are passed to the \code{predict()} method
supplied by the modeling package.These arguments are particularly useful
for mixed-effects or bayesian models (see the online vignettes on the
\code{marginaleffects} website). Available arguments can vary from model to
model, depending on the range of supported arguments by each modeling
package. See the "Model-Specific Arguments" section of the
\code{?slopes} documentation for a non-exhaustive list of available
arguments.}
}
\value{
A data.frame of predicted values with a number of rows equal to the
number of rows in \code{newdata} and columns "rowid" and "estimate". A "group"
column is added for multivariate models or models with categorical outcomes.
}
\description{
Get predicted values from a model object (internal function)
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinytest.R
\name{expect_slopes}
\alias{expect_slopes}
\title{\code{tinytest} helper}
\usage{
expect_slopes(object, n_unique = NULL, pct_na = 5, se = TRUE, ...)
}
\description{
\code{tinytest} helper
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/refit.R
\name{refit}
\alias{refit}
\alias{refit.marginaleffects}
\alias{refit.predictions}
\alias{refit.comparisons}
\alias{refit.slopes}
\alias{refit.hypotheses}
\title{Refit a marginaleffects object with new data}
\usage{
refit(object, ...)

\method{refit}{marginaleffects}(object, data = NULL, newdata = NULL, vcov = NULL, ...)

\method{refit}{predictions}(object, data = NULL, newdata = NULL, vcov = NULL, ...)

\method{refit}{comparisons}(object, data = NULL, newdata = NULL, vcov = NULL, ...)

\method{refit}{slopes}(object, data = NULL, newdata = NULL, vcov = NULL, ...)

\method{refit}{hypotheses}(object, data = NULL, newdata = NULL, vcov = NULL, ...)
}
\arguments{
\item{object}{A marginaleffects object (predictions, comparisons, or slopes)}

\item{...}{Additional arguments passed to methods}

\item{data}{Optional data frame to refit the underlying model}

\item{newdata}{Optional data frame to re-evaluate the marginaleffects call}

\item{vcov}{Optional logical or variance-covariance matrix specification to pass to the marginaleffects call}
}
\value{
A marginaleffects object
}
\description{
Refit a marginaleffects object with new data
}
\details{
If \code{data} is supplied, the underlying model is refitted using that data.
If \code{newdata} is supplied, the marginaleffects call is re-evaluated with the new data.
Both can be supplied together to refit the model and make predictions on new data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_vcov.R, R/methods_MCMCglmm.R,
%   R/methods_afex.R, R/methods_aod.R, R/methods_biglm.R, R/methods_brms.R,
%   R/methods_dbarts.R, R/methods_gamlss.R, R/methods_glmmTMB.R,
%   R/methods_mhurdle.R, R/methods_mlr3.R, R/methods_rms.R, R/methods_rstpm2.R,
%   R/methods_scam.R, R/methods_systemfit.R, R/methods_tidymodels.R
\name{get_vcov}
\alias{get_vcov}
\alias{get_vcov.default}
\alias{get_vcov.MCMCglmm}
\alias{get_vcov.afex_aov}
\alias{get_vcov.glimML}
\alias{get_vcov.biglm}
\alias{get_vcov.brmsfit}
\alias{get_vcov.bart}
\alias{get_vcov.gamlss}
\alias{get_vcov.glmmTMB}
\alias{get_vcov.mhurdle}
\alias{get_vcov.Learner}
\alias{get_vcov.orm}
\alias{get_vcov.stpm2}
\alias{get_vcov.pstpm2}
\alias{get_vcov.gsm}
\alias{get_vcov.aft}
\alias{get_vcov.scam}
\alias{get_vcov.systemfit}
\alias{get_vcov.model_fit}
\alias{get_vcov.workflow}
\title{Get a named variance-covariance matrix from a model object}
\usage{
get_vcov(model, ...)

\method{get_vcov}{default}(model, vcov = NULL, ...)

\method{get_vcov}{MCMCglmm}(model, vcov = NULL, ...)

\method{get_vcov}{afex_aov}(model, vcov = NULL, ...)

\method{get_vcov}{glimML}(model, vcov = NULL, ...)

\method{get_vcov}{biglm}(model, vcov = NULL, ...)

\method{get_vcov}{brmsfit}(model, vcov = NULL, ...)

\method{get_vcov}{bart}(model, vcov = NULL, ...)

\method{get_vcov}{gamlss}(model, ...)

\method{get_vcov}{glmmTMB}(model, ...)

\method{get_vcov}{mhurdle}(model, vcov = NULL, ...)

\method{get_vcov}{Learner}(model, ...)

\method{get_vcov}{orm}(model, vcov = NULL, ...)

\method{get_vcov}{stpm2}(model, ...)

\method{get_vcov}{pstpm2}(model, ...)

\method{get_vcov}{gsm}(model, ...)

\method{get_vcov}{aft}(model, ...)

\method{get_vcov}{scam}(model, vcov = NULL, ...)

\method{get_vcov}{systemfit}(model, ...)

\method{get_vcov}{model_fit}(model, type = NULL, ...)

\method{get_vcov}{workflow}(model, type = NULL, ...)
}
\arguments{
\item{model}{Model object}

\item{...}{Additional arguments are passed to the \code{predict()} method
supplied by the modeling package.These arguments are particularly useful
for mixed-effects or bayesian models (see the online vignettes on the
\code{marginaleffects} website). Available arguments can vary from model to
model, depending on the range of supported arguments by each modeling
package. See the "Model-Specific Arguments" section of the
\code{?slopes} documentation for a non-exhaustive list of available
arguments.}

\item{vcov}{Type of uncertainty estimates to report (e.g., for robust standard errors). Acceptable values:
\itemize{
\item FALSE: Do not compute standard errors. This can speed up computation considerably.
\item TRUE: Unit-level standard errors using the default \code{vcov(model)} variance-covariance matrix.
\item String which indicates the kind of uncertainty estimates to return.
\itemize{
\item Heteroskedasticity-consistent: \code{"HC"}, \code{"HC0"}, \code{"HC1"}, \code{"HC2"}, \code{"HC3"}, \code{"HC4"}, \code{"HC4m"}, \code{"HC5"}. See \code{?sandwich::vcovHC}
\item Heteroskedasticity and autocorrelation consistent: \code{"HAC"}
\item Mixed-Models degrees of freedom: "satterthwaite", "kenward-roger"
\item Other: \code{"NeweyWest"}, \code{"KernHAC"}, \code{"OPG"}. See the \code{sandwich} package documentation.
}
\item One-sided formula which indicates the name of cluster variables (e.g., \code{~unit_id}). This formula is passed to the \code{cluster} argument of the \code{sandwich::vcovCL} function.
\item Square covariance matrix
\item Function which returns a covariance matrix (e.g., \code{stats::vcov(model)})
}}

\item{type}{string indicates the type (scale) of the predictions used to
compute contrasts or slopes. This can differ based on the model
type, but will typically be a string such as: "response", "link", "probs",
or "zero". When an unsupported string is entered, the model-specific list of
acceptable values is returned in an error message. When \code{type} is \code{NULL}, the
first entry in the error message is used by default.}
}
\value{
A named square matrix of variance and covariances. The names must match the coefficient names.
}
\description{
Mostly for internal use, but can be useful because the output is consistent across model classes.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/complete_levels.R
\name{complete_levels}
\alias{complete_levels}
\title{Create a data.frame with all factor or character levels}
\usage{
complete_levels(x, character_levels = NULL)
}
\description{
\code{model.matrix} breaks when \code{newdata} includes a factor
variable, but not all levels are present in the data. This is bad for us
because we often want to get predictions with one (or few) rows, where some
factor levels are inevitably missing.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_group_names.R, R/methods_MASS.R,
%   R/methods_nnet.R, R/methods_brglm2.R, R/methods_brms.R, R/methods_mclogit.R,
%   R/methods_mlm.R, R/methods_ordinal.R, R/methods_pscl.R, R/methods_survey.R
\name{get_group_names}
\alias{get_group_names}
\alias{get_group_names.default}
\alias{get_group_names.polr}
\alias{get_group_names.multinom}
\alias{get_group_names.bracl}
\alias{get_group_names.brmsfit}
\alias{get_group_names.mblogit}
\alias{get_group_names.mlm}
\alias{get_group_names.clm}
\alias{get_group_names.hurdle}
\alias{get_group_names.svyolr}
\title{Get levels of the outcome variable in grouped or multivariate models}
\usage{
get_group_names(model, ...)

\method{get_group_names}{default}(model, ...)

\method{get_group_names}{polr}(model, ...)

\method{get_group_names}{multinom}(model, ...)

\method{get_group_names}{bracl}(model, ...)

\method{get_group_names}{brmsfit}(model, ...)

\method{get_group_names}{mblogit}(model, type, ...)

\method{get_group_names}{mlm}(model, ...)

\method{get_group_names}{clm}(model, ...)

\method{get_group_names}{hurdle}(model, type = "count", ...)

\method{get_group_names}{svyolr}(model, ...)
}
\arguments{
\item{model}{Model object}

\item{...}{Additional arguments are passed to the \code{predict()} method
supplied by the modeling package.These arguments are particularly useful
for mixed-effects or bayesian models (see the online vignettes on the
\code{marginaleffects} website). Available arguments can vary from model to
model, depending on the range of supported arguments by each modeling
package. See the "Model-Specific Arguments" section of the
\code{?slopes} documentation for a non-exhaustive list of available
arguments.}

\item{type}{string indicates the type (scale) of the predictions used to
compute contrasts or slopes. This can differ based on the model
type, but will typically be a string such as: "response", "link", "probs",
or "zero". When an unsupported string is entered, the model-specific list of
acceptable values is returned in an error message. When \code{type} is \code{NULL}, the
first entry in the error message is used by default.}
}
\value{
A character vector
}
\description{
Get levels of the outcome variable in grouped or multivariate models
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/autodiff.R
\name{autodiff}
\alias{autodiff}
\title{EXPERIMENTAL -- Enable Automatic Differentiation with JAX}
\usage{
autodiff(autodiff = NULL, install = FALSE)
}
\arguments{
\item{autodiff}{Logical flag. If \code{TRUE}, enables automatic differentiation
with JAX. If \code{FALSE}, disables automatic differentiation and
reverts to finite difference methods. If \code{NULL} (default), the function
simply returns the current autodiff setting without changing it.}

\item{install}{Logical flag. If \code{TRUE}, installs the \code{marginaleffects}
Python package via \code{reticulate::py_install()}. Default is \code{FALSE}. This is
only necessary if you are self-managing a Python installation.}
}
\value{
When \code{autodiff} is \code{NULL}, returns \code{TRUE} if autodiff is enabled and
\code{FALSE} otherwise. Otherwise called for side effects of enabling or
disabling automatic differentiation or installing the Python package.
}
\description{
This function enables or disables automatic differentiation using the JAX
package in Python, which can considerably speed up and increase the accuracy
of standard errors when a model includes many parameters.
}
\details{
Automatic differentiation needs to be enabled once per session.

When \code{autodiff = TRUE}, this function:
\itemize{
\item Imports the \code{marginaleffects.autodiff} Python module via \code{\link[reticulate:import]{reticulate::import()}}
\item Sets the internal jacobian function to use JAX-based automatic differentiation
\item Provides faster and more accurate gradient computation for supported models
\item Falls back on the default finite difference method for unsupported models and calls.
}

Currently supports:
\itemize{
\item Model types: \code{lm}, \code{glm}, \code{ols}
\item Functions: \code{\link[=predictions]{predictions()}} and \code{\link[=comparisons]{comparisons()}}, along with \code{avg_} and \code{plot_} variants.
\item \code{type}: "response" or "link"
\item \code{by}: \code{TRUE}, \code{FALSE}, or character vector.
\item \code{comparison}: "difference" and "ratio"
}

For unsupported models or options, the function automatically falls back to
the default finite difference method.
}
\section{Python Configuration}{
By default, no manual configuration of Python should be necessary. On most
machines, unless you have explicitly configured \code{reticulate}, reticulate
defaults to an automatically managed ephemeral virtual environment with all
Python requirements declared via \code{reticulate::py_require()}.

If you prefer to use a manually managed Python installation, you can direct
\code{reticulate} and specify which Python executable or environment to use.
\code{reticulate} selects a Python installation using its \href{https://rstudio.github.io/reticulate/articles/versions.html#order-of-discovery}{Order of Discovery}.
As a convenience \code{autodiff(install=TRUE)} will install the \code{marginaleffects} Python
package in a self-managed virtual environment.

To specify an alternate Python version:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(reticulate)
use_python("/usr/local/bin/python")
}\if{html}{\out{</div>}}

To use a virtual environment:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{use_virtualenv("myenv")
}\if{html}{\out{</div>}}

These configuration commands should be called before calling \code{autodiff()}.
}

\examples{
\dontrun{
# Install the Python package (only needed once)
autodiff(install = TRUE)

# Enable automatic differentiation
autodiff(TRUE)

# Fit a model and compute marginal effects
mod <- glm(am ~ hp + wt, data = mtcars, family = binomial)
avg_comparisons(mod) # Will use JAX for faster computation

# Disable automatic differentiation
autodiff(FALSE)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class.R
\name{new_marginaleffects_internal}
\alias{new_marginaleffects_internal}
\title{Constructor for marginaleffects_internal class}
\usage{
new_marginaleffects_internal(
  model,
  call,
  by = FALSE,
  byfun = NULL,
  comparison = NULL,
  conf_level = 0.95,
  cross = FALSE,
  df = NULL,
  draws = NULL,
  draws_chains = 0,
  eps = NULL,
  hypothesis = NULL,
  hypothesis_null = NULL,
  hypothesis_direction = NULL,
  jacobian = NULL,
  modeldata = NULL,
  numderiv = list("fdforward"),
  type = NULL,
  variables = list(),
  variable_names_by = character(),
  variable_names_by_hypothesis = character(),
  vcov_model = NULL,
  vcov_type = NULL,
  wts = NULL
)
}
\arguments{
\item{model}{The fitted model object (required)}

\item{call}{The original function call (required)}

\item{df}{The degrees of freedom}

\item{type}{The sanitized type from sanitize_type()}

\item{vcov_model}{The variance-covariance matrix}

\item{wts}{The weights specification}
}
\value{
An object of class marginaleffects_internal
}
\description{
Constructor for marginaleffects_internal class
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_model_matrix.R
\name{get_model_matrix}
\alias{get_model_matrix}
\alias{get_model_matrix.default}
\title{Get a named model matrix}
\usage{
get_model_matrix(model, newdata)

\method{get_model_matrix}{default}(model, newdata)
}
\arguments{
\item{model}{Model object}

\item{newdata}{Grid of predictor values at which we evaluate the slopes.
\itemize{
\item Warning: Please avoid modifying your dataset between fitting the model and calling a \code{marginaleffects} function. This can sometimes lead to unexpected results.
\item \code{NULL} (default): Unit-level slopes for each observed value in the dataset (empirical distribution). The dataset is retrieved using \code{\link[insight:get_data]{insight::get_data()}}, which tries to extract data from the environment. This may produce unexpected results if the original data frame has been altered since fitting the model.
\item \code{\link[=datagrid]{datagrid()}} call to specify a custom grid of regressors. For example:
\itemize{
\item \code{newdata = datagrid(cyl = c(4, 6))}: \code{cyl} variable equal to 4 and 6 and other regressors fixed at their means or modes.
\item See the Examples section and the \code{\link[=datagrid]{datagrid()}} documentation.
}
\item \code{\link[=subset]{subset()}} call with a single argument to select a subset of the dataset used to fit the model, ex: \code{newdata = subset(treatment == 1)}
\item \code{\link[dplyr:filter]{dplyr::filter()}} call with a single argument to select a subset of the dataset used to fit the model, ex: \code{newdata = filter(treatment == 1)}
\item string:
\itemize{
\item "mean": Slopes evaluated when each predictor is held at its mean or mode.
\item "median": Slopes evaluated when each predictor is held at its median or mode.
\item "balanced": Slopes evaluated on a balanced grid with every combination of categories and numeric variables held at their means.
\item "tukey": Slopes evaluated at Tukey's 5 numbers.
\item "grid": Slopes evaluated on a grid of representative numbers (Tukey's 5 numbers and unique values of categorical predictors).
}
}}
}
\description{
Get a named model matrix
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/class.R
\name{extract_calling_function}
\alias{extract_calling_function}
\title{Extract calling function name from call object}
\usage{
extract_calling_function(call)
}
\arguments{
\item{call}{The call object from which to extract the function name}
}
\value{
Character string of the calling function ("comparisons", "predictions", or "hypotheses")
}
\description{
Extract calling function name from call object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_draws.R
\name{get_draws}
\alias{get_draws}
\title{Extract Posterior Draws or Bootstrap Resamples from \code{marginaleffects} Objects}
\usage{
get_draws(x, shape = "long")
}
\arguments{
\item{x}{An object produced by a \code{marginaleffects} package function, such as \code{predictions()}, \code{avg_slopes()}, \code{hypotheses()}, etc.}

\item{shape}{string indicating the shape of the output format:
\itemize{
\item "long": long format data frame
\item "DxP": Matrix with draws as rows and parameters as columns
\item "PxD": Matrix with draws as rows and parameters as columns
\item "rvar": Random variable datatype (see \code{posterior} package documentation).
}}
}
\value{
A data.frame with \code{drawid} and \code{draw} columns.
}
\description{
Extract Posterior Draws or Bootstrap Resamples from \code{marginaleffects} Objects
}
\details{
If DxP and PxD and the names returned by \code{coef(x)} are unique, \code{marginaleffects} sets parameter names to those names. Otherwise, it sets them to \code{b1}, \code{b2}, etc.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_coef.R, R/methods_MASS.R,
%   R/methods_WeightIt.R, R/methods_afex.R, R/methods_betareg.R,
%   R/methods_nnet.R, R/methods_brglm2.R, R/methods_brms.R,
%   R/methods_dataframe.R, R/methods_gamlss.R, R/methods_glmmTMB.R,
%   R/methods_glmtoolbox.R, R/methods_lme4.R, R/methods_mclogit.R,
%   R/methods_mgcv.R, R/methods_mlm.R, R/methods_sampleSelection.R,
%   R/methods_scam.R, R/methods_stats.R, R/methods_survey.R,
%   R/methods_systemfit.R, R/methods_tidymodels.R
\name{get_coef}
\alias{get_coef}
\alias{get_coef.default}
\alias{get_coef.polr}
\alias{get_coef.multinom_weightit}
\alias{get_coef.afex_aov}
\alias{get_coef.betareg}
\alias{get_coef.multinom}
\alias{get_coef.brmultinom}
\alias{get_coef.bracl}
\alias{get_coef.brmsfit}
\alias{get_coef.data.frame}
\alias{get_coef.gamlss}
\alias{get_coef.glmmTMB}
\alias{get_coef.glmgee}
\alias{get_coef.merMod}
\alias{get_coef.lmerModLmerTest}
\alias{get_coef.lmerMod}
\alias{get_coef.mblogit}
\alias{get_coef.gam}
\alias{get_coef.mlm}
\alias{get_coef.selection}
\alias{get_coef.scam}
\alias{get_coef.nls}
\alias{get_coef.svyolr}
\alias{get_coef.systemfit}
\alias{get_coef.workflow}
\title{Get a named vector of coefficients from a model object}
\usage{
get_coef(model, ...)

\method{get_coef}{default}(model, ...)

\method{get_coef}{polr}(model, ...)

\method{get_coef}{multinom_weightit}(model, ...)

\method{get_coef}{afex_aov}(model, ...)

\method{get_coef}{betareg}(model, ...)

\method{get_coef}{multinom}(model, ...)

\method{get_coef}{brmultinom}(model, ...)

\method{get_coef}{bracl}(model, ...)

\method{get_coef}{brmsfit}(model, ...)

\method{get_coef}{data.frame}(model, ...)

\method{get_coef}{gamlss}(model, ...)

\method{get_coef}{glmmTMB}(model, ...)

\method{get_coef}{glmgee}(model, ...)

\method{get_coef}{merMod}(model, ...)

\method{get_coef}{lmerModLmerTest}(model, ...)

\method{get_coef}{lmerMod}(model, ...)

\method{get_coef}{mblogit}(model, ...)

\method{get_coef}{gam}(model, ...)

\method{get_coef}{mlm}(model, ...)

\method{get_coef}{selection}(model, ...)

\method{get_coef}{scam}(model, ...)

\method{get_coef}{nls}(model, ...)

\method{get_coef}{svyolr}(model, ...)

\method{get_coef}{systemfit}(model, ...)

\method{get_coef}{workflow}(model, ...)
}
\arguments{
\item{model}{Model object}

\item{...}{Additional arguments are passed to the \code{predict()} method
supplied by the modeling package.These arguments are particularly useful
for mixed-effects or bayesian models (see the online vignettes on the
\code{marginaleffects} website). Available arguments can vary from model to
model, depending on the range of supported arguments by each modeling
package. See the "Model-Specific Arguments" section of the
\code{?slopes} documentation for a non-exhaustive list of available
arguments.}
}
\value{
A named vector of coefficients. The names must match those of the variance matrix.
}
\description{
Mostly for internal use, but can be useful because the output is consistent across model classes.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot_predictions.R
\name{plot_predictions}
\alias{plot_predictions}
\title{Plot Conditional or Marginal Predictions}
\usage{
plot_predictions(
  model,
  condition = NULL,
  by = NULL,
  newdata = NULL,
  type = NULL,
  vcov = NULL,
  conf_level = 0.95,
  wts = FALSE,
  transform = NULL,
  points = 0,
  rug = FALSE,
  gray = getOption("marginaleffects_plot_gray", default = FALSE),
  draw = TRUE,
  ...
)
}
\arguments{
\item{model}{Model object}

\item{condition}{Conditional predictions
\itemize{
\item Character vector (max length 4): Names of the predictors to display.
\item Named list (max length 4): List names correspond to predictors. List elements can be:
\itemize{
\item Numeric vector
\item Function which returns a numeric vector or a set of unique categorical values
\item Shortcut strings for common reference values: "minmax", "quartile", "threenum"
}
\item 1: x-axis. 2: color/shape. 3: facet (wrap if no fourth variable, otherwise cols of grid). 4: facet (rows of grid).
\item Numeric variables in positions 2 and 3 are summarized by Tukey's five numbers \code{?stats::fivenum}
}}

\item{by}{Marginal predictions
\itemize{
\item Character vector (max length 3): Names of the categorical predictors to marginalize across.
\item 1: x-axis. 2: color. 3: facets.
}}

\item{newdata}{When \code{newdata} is \code{NULL}, the grid is determined by the \code{condition} argument. When \code{newdata} is not \code{NULL}, the argument behaves in the same way as in the \code{predictions()} function.}

\item{type}{string indicates the type (scale) of the predictions used to
compute contrasts or slopes. This can differ based on the model
type, but will typically be a string such as: "response", "link", "probs",
or "zero". When an unsupported string is entered, the model-specific list of
acceptable values is returned in an error message. When \code{type} is \code{NULL}, the
first entry in the error message is used by default.}

\item{vcov}{Type of uncertainty estimates to report (e.g., for robust standard errors). Acceptable values:
\itemize{
\item FALSE: Do not compute standard errors. This can speed up computation considerably.
\item TRUE: Unit-level standard errors using the default \code{vcov(model)} variance-covariance matrix.
\item String which indicates the kind of uncertainty estimates to return.
\itemize{
\item Heteroskedasticity-consistent: \code{"HC"}, \code{"HC0"}, \code{"HC1"}, \code{"HC2"}, \code{"HC3"}, \code{"HC4"}, \code{"HC4m"}, \code{"HC5"}. See \code{?sandwich::vcovHC}
\item Heteroskedasticity and autocorrelation consistent: \code{"HAC"}
\item Mixed-Models degrees of freedom: "satterthwaite", "kenward-roger"
\item Other: \code{"NeweyWest"}, \code{"KernHAC"}, \code{"OPG"}. See the \code{sandwich} package documentation.
}
\item One-sided formula which indicates the name of cluster variables (e.g., \code{~unit_id}). This formula is passed to the \code{cluster} argument of the \code{sandwich::vcovCL} function.
\item Square covariance matrix
\item Function which returns a covariance matrix (e.g., \code{stats::vcov(model)})
}}

\item{conf_level}{numeric value between 0 and 1. Confidence level to use to build a confidence interval.}

\item{wts}{logical, string or numeric: weights to use when computing average predictions, contrasts or slopes. These weights only affect the averaging in \verb{avg_*()} or with the \code{by} argument, and not unit-level estimates. See \code{?weighted.mean}
\itemize{
\item string: column name of the weights variable in \code{newdata}. When supplying a column name to \code{wts}, it is recommended to supply the original data (including the weights variable) explicitly to \code{newdata}.
\item numeric: vector of length equal to the number of rows in the original data or in \code{newdata} (if supplied).
\item FALSE: Equal weights.
\item TRUE: Extract weights from the fitted object with \code{insight::find_weights()} and use them when taking weighted averages of estimates. Warning: \code{newdata=datagrid()} returns a single average weight, which is equivalent to using \code{wts=FALSE}
}}

\item{transform}{A function applied to unit-level adjusted predictions and confidence intervals just before the function returns results. For bayesian models, this function is applied to individual draws from the posterior distribution, before computing summaries.}

\item{points}{Number between 0 and 1 which controls the transparency of raw data points. 0 (default) does not display any points. Warning: The points displayed are raw data, so the resulting plot is not a "partial residual plot."}

\item{rug}{TRUE displays tick marks on the axes to mark the distribution of raw data.}

\item{gray}{FALSE grayscale or color plot}

\item{draw}{\code{TRUE} returns a \code{ggplot2} plot. \code{FALSE} returns a \code{data.frame} of the underlying data.}

\item{...}{Additional arguments are passed to the \code{predict()} method
supplied by the modeling package.These arguments are particularly useful
for mixed-effects or bayesian models (see the online vignettes on the
\code{marginaleffects} website). Available arguments can vary from model to
model, depending on the range of supported arguments by each modeling
package. See the "Model-Specific Arguments" section of the
\code{?slopes} documentation for a non-exhaustive list of available
arguments.}
}
\value{
A \code{ggplot2} object or data frame (if \code{draw=FALSE})
}
\description{
Plot predictions on the y-axis against values of one or more predictors (x-axis, colors/shapes, and facets).

The \code{by} argument is used to plot marginal predictions, that is, predictions made on the original data, but averaged by subgroups. This is analogous to using the \code{by} argument in the \code{predictions()} function.

The \code{condition} argument is used to plot conditional predictions, that is, predictions made on a user-specified grid. This is analogous to using the \code{newdata} argument and \code{datagrid()} function in a \code{predictions()} call. All variables whose values are not specified explicitly are treated as usual by \code{datagrid()}, that is, they are held at their mean or mode (or rounded mean for integers). This includes grouping variables in mixed-effects models, so analysts who fit such models may want to specify the groups of interest using the \code{condition} argument, or supply model-specific arguments to compute population-level estimates. See details below.

See the "Plots" vignette and website for tutorials and information on how to customize plots:
\itemize{
\item https://marginaleffects.com/bonus/plot.html
\item https://marginaleffects.com
}
}
\section{Model-Specific Arguments}{


Some model types allow model-specific arguments to modify the nature of
marginal effects, predictions, marginal means, and contrasts. Please report
other package-specific \code{predict()} arguments on Github so we can add them to
the table below.

https://github.com/vincentarelbundock/marginaleffects/issues\tabular{llll}{
   Package \tab Class \tab Argument \tab Documentation \cr
   \code{brms} \tab \code{brmsfit} \tab \code{ndraws} \tab \link[brms:posterior_predict.brmsfit]{brms::posterior_predict} \cr
    \tab  \tab \code{re_formula} \tab \link[brms:posterior_predict.brmsfit]{brms::posterior_predict} \cr
   \code{lme4} \tab \code{merMod} \tab \code{re.form} \tab \link[lme4:predict.merMod]{lme4::predict.merMod} \cr
    \tab  \tab \code{allow.new.levels} \tab \link[lme4:predict.merMod]{lme4::predict.merMod} \cr
   \code{glmmTMB} \tab \code{glmmTMB} \tab \code{re.form} \tab \link[glmmTMB:predict.glmmTMB]{glmmTMB::predict.glmmTMB} \cr
    \tab  \tab \code{allow.new.levels} \tab \link[glmmTMB:predict.glmmTMB]{glmmTMB::predict.glmmTMB} \cr
    \tab  \tab \code{zitype} \tab \link[glmmTMB:predict.glmmTMB]{glmmTMB::predict.glmmTMB} \cr
   \code{mgcv} \tab \code{bam} \tab \code{exclude} \tab \link[mgcv:predict.bam]{mgcv::predict.bam} \cr
    \tab \code{gam} \tab \code{exclude} \tab \link[mgcv:predict.gam]{mgcv::predict.gam} \cr
   \code{robustlmm} \tab \code{rlmerMod} \tab \code{re.form} \tab \link[robustlmm:rlmerMod-class]{robustlmm::predict.rlmerMod} \cr
    \tab  \tab \code{allow.new.levels} \tab \link[robustlmm:rlmerMod-class]{robustlmm::predict.rlmerMod} \cr
   \code{MCMCglmm} \tab \code{MCMCglmm} \tab \code{ndraws} \tab  \cr
   \code{sampleSelection} \tab \code{selection} \tab \code{part} \tab \link[sampleSelection:predict.selection]{sampleSelection::predict.selection} \cr
}
}

\section{Prediction types}{


The \code{type} argument determines the scale of the predictions used to compute quantities of interest with functions from the \code{marginaleffects} package. Admissible values for \code{type} depend on the model object. When users specify an incorrect value for \code{type}, \code{marginaleffects} will raise an informative error with a list of valid \code{type} values for the specific model object. The first entry in the list in that error message is the default type.

The \code{invlink(link)} is a special type defined by \code{marginaleffects}. It is available for some (but not all) models, and only for the \code{predictions()} function. With this link type, we first compute predictions on the link scale, then we use the inverse link function to backtransform the predictions to the response scale. This is useful for models with non-linear link functions as it can ensure that confidence intervals stay within desirable bounds, ex: 0 to 1 for a logit model. Note that an average of estimates with \code{type="invlink(link)"} will not always be equivalent to the average of estimates with \code{type="response"}. This type is default when calling \code{predictions()}. It is available---but not default---when calling \code{avg_predictions()} or \code{predictions()} with the \code{by} argument.

Some of the most common \code{type} values are:

response, link, E, Ep, average, class, conditional, count, cum.prob, cumhaz, cumprob, density, detection, disp, ev, expected, expvalue, fitted, hazard, invlink(link), latent, latent_N, linear, linear.predictor, linpred, location, lp, mean, numeric, p, ppd, pr, precision, prediction, prob, probability, probs, quantile, risk, rmst, scale, survival, unconditional, utility, variance, xb, zero, zlink, zprob
}

\examples{
\dontshow{if (interactive() || isTRUE(Sys.getenv("R_DOC_BUILD") == "true")) (if (getRversion() >= "3.4") withAutoprint else force)(\{ # examplesIf}
mod <- lm(mpg ~ hp + wt, data = mtcars)
plot_predictions(mod, condition = "wt")

mod <- lm(mpg ~ hp * wt * am, data = mtcars)
plot_predictions(mod, condition = c("hp", "wt"))

plot_predictions(mod, condition = list("hp", wt = "threenum"))

plot_predictions(mod, condition = list("hp", wt = range))

# marginal predictions
mod <- lm(mpg ~ hp * am, data = mtcars)
plot_predictions(mod, by = "am")

# marginal predictions on a counterfactual grid
plot_predictions(mod,
  by = "am",
  newdata = datagrid(am = 0:1, grid_type = "counterfactual")
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/datagrid.R
\name{datagrid}
\alias{datagrid}
\title{Data grids}
\usage{
datagrid(
  ...,
  model = NULL,
  newdata = NULL,
  by = NULL,
  grid_type = "mean_or_mode",
  response = FALSE,
  FUN = NULL,
  FUN_character = NULL,
  FUN_factor = NULL,
  FUN_logical = NULL,
  FUN_numeric = NULL,
  FUN_integer = NULL,
  FUN_binary = NULL,
  FUN_other = NULL
)
}
\arguments{
\item{...}{named arguments with vectors of values or functions for user-specified variables.
\itemize{
\item Functions are applied to the variable in the \code{model} dataset or \code{newdata}, and must return a vector of the appropriate type.
\item Character vectors are automatically transformed to factors if necessary.
\item The output will include all combinations of these variables (see Examples below.)
}}

\item{model}{Model object}

\item{newdata}{data.frame (one and only one of the \code{model} and \code{newdata} arguments can be used.)}

\item{by}{character vector with grouping variables within which \verb{FUN_*} functions are applied to create "sub-grids" with unspecified variables.}

\item{grid_type}{character. Determines the functions to apply to each variable. The defaults can be overridden by defining individual variables explicitly in \code{...}, or by supplying a function to one of the \verb{FUN_*} arguments.
\itemize{
\item "mean_or_mode": Character, factor, logical, and binary variables are set to their modes. Numeric, integer, and other variables are set to their means.
\item "balanced": Each unique level of character, factor, logical, and binary variables are preserved. Numeric, integer, and other variables are set to their means. Warning: When there are many variables and many levels per variable, a balanced grid can be very large. In those cases, it is better to use \code{grid_type="mean_or_mode"} and to specify the unique levels of a subset of named variables explicitly.
\item "dataframe": Similar to "mean_or_mode" but creates a data frame by binding columns element-wise rather than taking the cross-product. All explicitly specified vectors must have the same length (or length 1), and the result has as many rows as the longest vector. This differs from other grid types which use \code{expand.grid()} or \code{data.table::CJ()} to create all combinations.
\item "counterfactual": the entire dataset is duplicated for each combination of the variable values specified in \code{...}. Variables not explicitly supplied to \code{datagrid()} are set to their observed values in the original dataset.
}}

\item{response}{Logical should the response variable be included in the grid, even if it is not specified explicitly.}

\item{FUN}{a function to be applied to all variables in the grid. This is useful when you want to apply the same function to all variables, such as \code{mean} or \code{median}. If you specify \code{FUN}, it will override the \code{grid_type} defaults, but not other \verb{FUN_*} arguments below.}

\item{FUN_character}{the function to be applied to character variables.}

\item{FUN_factor}{the function to be applied to factor variables. This only applies if the variable in the original data is a factor. For variables converted to factor in a model-fitting formula, for example, \code{FUN_character} is used.}

\item{FUN_logical}{the function to be applied to logical variables.}

\item{FUN_numeric}{the function to be applied to numeric variables.}

\item{FUN_integer}{the function to be applied to integer-ish variables (including columns without decimal places).}

\item{FUN_binary}{the function to be applied to binary variables.}

\item{FUN_other}{the function to be applied to other variable types.}
}
\value{
A \code{data.frame} in which each row corresponds to one combination of the named
predictors supplied by the user via the \code{...} dots. Variables which are not
explicitly defined are held at their mean or mode.
}
\description{
Generate a data grid of user-specified values for use in the \code{newdata} argument of the \code{predictions()}, \code{comparisons()}, and \code{slopes()} functions. This is useful to define where in the predictor space we want to evaluate the quantities of interest. Ex: the predicted outcome or slope for a 37 year old college graduate.
}
\details{
If \code{datagrid} is used in a \code{predictions()}, \code{comparisons()}, or \code{slopes()} call as the
\code{newdata} argument, the model is automatically inserted in the \code{model} argument of \code{datagrid()}
call, and users do not need to specify either the \code{model} or \code{newdata} arguments. The same behavior will occur when the value supplied to \verb{newdata=} is a function call which starts with "datagrid". This is intended to allow users to create convenience shortcuts like:

\strong{Warning about hierarchical grouping variables:} When using the default \code{grid_type = "mean_or_mode"} with hierarchical models (such as mixed models with nested grouping factors), \code{datagrid()} may create invalid combinations of grouping variables. For example, if you have students nested within schools, or countries nested within regions, the modal values of each grouping variable may not correspond to valid nested relationships in the data. This can cause prediction errors. To avoid this issue, explicitly specify valid combinations of hierarchical grouping variables in the \code{datagrid()} call, or use \code{grid_type = "counterfactual"} to preserve the original data structure.

\preformatted{
mod <- lm(mpg ~ am + vs + factor(cyl) + hp, mtcars)
datagrid_bal <- function(...) datagrid(..., grid_type = "balanced")
predictions(model, newdata = datagrid_bal(cyl = 4))
}

If users supply a model, the data used to fit that model is retrieved using
the \code{insight::get_data} function.
}
\examples{
# The output only has 2 rows, and all the variables except `hp` are at their
# mean or mode.
datagrid(newdata = mtcars, hp = c(100, 110))

# We get the same result by feeding a model instead of a data.frame
mod <- lm(mpg ~ hp, mtcars)
datagrid(model = mod, hp = c(100, 110))

# Use in `marginaleffects` to compute "Typical Marginal Effects". When used
# in `slopes()` or `predictions()` we do not need to specify the
# `model` or `newdata` arguments.
slopes(mod, newdata = datagrid(hp = c(100, 110)))

# datagrid accepts functions
datagrid(hp = range, cyl = unique, newdata = mtcars)
comparisons(mod, newdata = datagrid(hp = fivenum))

# The full dataset is duplicated with each observation given counterfactual
# values of 100 and 110 for the `hp` variable. The original `mtcars` includes
# 32 rows, so the resulting dataset includes 64 rows.
dg <- datagrid(newdata = mtcars, hp = c(100, 110), grid_type = "counterfactual")
nrow(dg)

# We get the same result by feeding a model instead of a data.frame
mod <- lm(mpg ~ hp, mtcars)
dg <- datagrid(model = mod, hp = c(100, 110), grid_type = "counterfactual")
nrow(dg)

# Use `by` to hold variables at group-specific values
mod2 <- lm(mpg ~ hp + cyl, mtcars)
datagrid(model = mod2, hp = mean, by = "cyl")

# Use `FUN` to apply function to all variables
datagrid(model = mod2, FUN = median)

# Use `grid_type="dataframe"` for column-wise binding instead of cross-product
datagrid(model = mod2, hp = c(100, 200), cyl = c(4, 6), grid_type = "dataframe")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/prune.R
\name{prune.marginaleffects}
\alias{prune.marginaleffects}
\title{Prune marginaleffects objects to reduce memory usage}
\usage{
\method{prune}{marginaleffects}(tree, component = "all", ...)
}
\arguments{
\item{tree}{A marginaleffects object (predictions, comparisons, slopes, or hypotheses)}

\item{component}{A character string indicating which component to prune: "all" or "modeldata".}

\item{...}{Unused}
}
\value{
A pruned marginaleffects object
}
\description{
Remove large attributes from marginaleffects objects to reduce memory usage.
Warning: This will disable many useful post-processing features of \code{marginaleffects}
}
\details{
...
}
